@@ -3,7 +3,7 @@
 %%
 %%     Functions for reading and writing Wavefront ASCII files (.obj).
 %%
-%%  Copyright (c) 2001-2008 Bjorn Gustavsson
+%%  Copyright (c) 2001-2009 Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -94,9 +94,7 @@ hard_edges_1([#e3d_face{vs=Vs}|Fs], [Sg|Sgs], Acc0) ->
     Acc = add_edges(Vs, hd(Vs), Sg, Acc0),
     hard_edges_1(Fs, Sgs, Acc);
 hard_edges_1([], [], Acc) ->
-    R = sofs:relation(Acc),
-    F0 = sofs:relation_to_family(R),
-    F = sofs:to_external(F0),
+    F = rel2fam(Acc),
     foldl(fun({Edge,Sgs0}, He) ->
 		  %% Group 0 is special. Edges surrounded by faces with
 		  %% group zero will always be hard. Otherwise the rule
@@ -465,11 +463,12 @@ export(File, #e3d_file{objs=Objs,mat=Mat,creator=Creator}, Flags) ->
 
 export_object(F, #e3d_object{name=Name,obj=Mesh0}, Flags,
 	      Vbase, UVbase, Nbase) ->
+    IncludeNormals = proplists:get_bool(include_normals, Flags),
     Mesh = case proplists:get_bool(include_normals, Flags) of
 	       false -> Mesh0;
 	       true -> e3d_mesh:vertex_normals(Mesh0)
 	   end,
-    #e3d_mesh{fs=Fs0,vs=Vs,tx=Tx,ns=Ns} = Mesh,
+    #e3d_mesh{vs=Vs,tx=Tx,ns=Ns} = Mesh,
     mesh_info(F, Mesh),
     foreach(fun({X,Y,Z}) ->
 		    io:format(F, "v ~s ~s ~s\r\n",
@@ -484,10 +483,11 @@ export_object(F, #e3d_object{name=Name,obj=Mesh0}, Flags,
 			      [fmtf(X),fmtf(Y),fmtf(Z)])
 	    end, Ns),
     object_group(F, Name, Flags),
-    Fs1 = [{Mat,FaceRec} || #e3d_face{mat=Mat}=FaceRec <- Fs0],
-    Fs = sofs:to_external(sofs:relation_to_family(sofs:relation(Fs1))),
-    foreach(fun(Face) ->
-		    face_mat(F, Name, Face, Flags, Vbase, UVbase, Nbase)
+    GroupedFaces = group_smooth_groups(Mesh, IncludeNormals),
+    Fs1 = [{Mat,Pair} || {_SG,#e3d_face{mat=Mat}}=Pair <- GroupedFaces],
+    Fs = rel2fam(Fs1),
+    foreach(fun(MatFs) ->
+		    face_mat(F, Name, MatFs, Flags, Vbase, UVbase, Nbase)
 	    end, Fs),
     {Vbase+length(Vs),UVbase+length(Tx),Nbase+length(Ns)}.
 
@@ -502,14 +502,26 @@ object_group(F, Name, Flags) ->
 	false -> io:format(F, "g ~s\r\n", [Name])
     end.
 
-face_mat(F, Name, {Ms,Fs}, Flags, Vbase, UVbase, Nbase) ->
+group_smooth_groups(#e3d_mesh{fs=Fs}, false) ->
+    [{1,Face} || Face <- Fs];
+group_smooth_groups(#e3d_mesh{fs=Fs}, true) ->
+    [{SG,Face} || #e3d_face{sg=SG}=Face <- Fs].
+
+face_mat(F, Name, {Ms,SgFs0}, Flags, Vbase, UVbase, Nbase) ->
     mat_group(F, Name, Ms, Flags),
     io:put_chars(F, "usemtl"),
     foldl(fun(M, Prefix) ->
 		  io:format(F, "~c~s", [Prefix,atom_to_list(M)])
 	  end, $\s, Ms),
     eol(F),
-    foreach(fun(Vs) -> face(F, Vs, Vbase, UVbase, Nbase) end, Fs).
+    SgFs = rel2fam(SgFs0),
+    foreach(fun({SG,Faces}) ->
+		    io:format(F, "s ~w", [SG]),
+		    eol(F),
+		    foreach(fun(Face) ->
+				    face(F, Face, Vbase, UVbase, Nbase)
+			    end, Faces)
+	    end, SgFs).
 
 mat_group(F, Name, Ms, Flags) ->
     case proplists:get_bool(group_per_material, Flags) of
@@ -604,3 +616,10 @@ mesh_info(F, #e3d_mesh{vs=Vs,fs=Fs}) ->
 
 eol(F) ->
     io:put_chars(F, "\r\n").
+
+%%%
+%%% Common utilities.
+%%%
+
+rel2fam(R) ->
+    sofs:to_external(sofs:relation_to_family(sofs:relation(R))).
