@@ -105,7 +105,7 @@ volproject(Type,Chart,_Pinned,{_,BEdges},We) ->
 			e3d_mat:mul_vector(Rot,Vec)
 		end,
     Vs1 = lists:sort([{V,Transform(V)} || V <- Vs0]),
-    Tagged = leftOrRight(LoopInfo, Chart, We#we{vp=gb_trees:from_orddict(Vs1)}),
+    Tagged = leftOrRight(LoopInfo, Chart, We#we{vp=array:from_orddict(Vs1)}),
     %%io:format("Tagged ~w~n",[gb_sets:to_list(Tagged)]),
     [{V,fix_positions(V,Pos,CalcUV(Pos),Tagged)} || {V,Pos} <- Vs1].
 
@@ -353,7 +353,7 @@ projectFromCamera(Chart,{matrices,{MM,PM,VP}},We) ->
     Vs = wings_face:to_vertices(Chart, We),
     Proj = fun(V) ->
 		   {X,Y,Z} = wings_vertex:pos(V, We),
-		   {S,T, _} = glu:project(X,Y,Z,MM,PM,VP),
+		   {S,T, _} = wings_gl:project(X,Y,Z,MM,PM,VP),
 		   {V,{S,T,0.0}}
 	   end,
     lists:map(Proj, Vs).
@@ -368,7 +368,7 @@ chart_normal(Fs,We = #we{es=Etab}) ->
 	    [{_,BE}|_] = auv_placement:group_edge_loops(Fs,We),
 	    EdgeNormals = 
 		fun(#be{edge=Edge}, Sum0) ->
-			#edge{lf=LF,rf=RF} = gb_trees:get(Edge, Etab),
+			#edge{lf=LF,rf=RF} = array:get(Edge, Etab),
 			Sum1 = CalcNormal(LF,Sum0),
 			CalcNormal(RF,Sum1)
 		end,
@@ -441,7 +441,7 @@ lsqcm(Fs, Pinned, _Loop, We) ->
     %%?DBG("LSQ res ~p~n", [Vs2]),
     Patch = fun({Idt, {Ut,Vt}}) -> {Idt,{Ut,Vt,0.0}} end,
     Vs3 = lists:sort(lists:map(Patch, Vs2)),
-    TempVs = gb_trees:from_orddict(Vs3),
+    TempVs = array:from_orddict(Vs3),
     Area = fs_area(Fs, We, 0.0),
     MappedArea = fs_area(Fs, We#we{vp=TempVs}, 0.0),
     Scale = Area/MappedArea,
@@ -461,7 +461,7 @@ find_pinned_from_edges({Circumference, BorderEdges}, #we{vp=Vtab}) ->
 	    [#be{ve=TO},#be{vs=TO}|_] -> BorderEdges;
 	    _ -> reverse(BorderEdges)
 	end,
-    POS = fun(V) -> gb_trees:get(V,Vtab) end,
+    POS = fun(V) -> array:get(V,Vtab) end,
     Left = #link{no=1,pos=POS(Second#be.vs)},
     Right = foldl(fun(#be{vs=V}, #link{no=No,pos=Apos}) ->
 			  #link{no=No+1,pos=e3d_vec:add(Apos,POS(V))}
@@ -524,10 +524,10 @@ divide(What,0) -> What;
 divide(Vec,S) -> e3d_vec:divide(Vec,S).
      
 find_pinned({Circumference, BorderEdges}, We) ->
-    Vs = [gb_trees:get(V1, We#we.vp) || #be{vs=V1} <- BorderEdges],
+    Vs = [array:get(V1, We#we.vp) || #be{vs=V1} <- BorderEdges],
     Center = e3d_vec:average(Vs),
     AllC = lists:map(fun(#be{vs=Id}) ->
-			     Pos = gb_trees:get(Id, We#we.vp),
+			     Pos = array:get(Id, We#we.vp),
 			     Dist = e3d_vec:dist(Pos, Center),
 			     {Dist, Id, Pos}
 		     end, BorderEdges),
@@ -570,7 +570,7 @@ reorder_edge_loop(V1, [H|Tail], Acc) ->
 -record(lsq,{a,x0,ap,temp1,temp2,dr}).
 
 lsq_setup(Fs,We,Pinned) ->
-    {M,N,D,DR,L1,L2,_X0} = lsq_init(Fs,We,Pinned),
+    {M,N,D,DR,L1,L2} = lsq_init(Fs,We,Pinned),
     {Lquv0,{Np,Usum,Vsum}} =  
 	lists:mapfoldl(
 	  fun({P,{U,V} = UV}, {I,X,Y}) ->
@@ -601,15 +601,15 @@ lsq_init(Fs0,We0,Pinned0) ->
     We = wings_tesselation:triangulate(Fs0, We0),
     Fs = Fs0 ++ wings_we:new_items_as_ordset(face,We0,We),
     Pinned = gb_trees:from_orddict(lists:sort(Pinned0)),
-    lsq_init_fs(Fs,Pinned,We,{0,dict:new(),dict:new()},0,[],[],[]).
+    lsq_init_fs(Fs,Pinned,We,{0,dict:new(),dict:new()},0,[],[]).
 
-lsq_init_fs([F|Fs],P,We = #we{vp=Vtab},Ds0,N,Re0,Im0,X0) ->
-    Vs = [[A0|_],[B0|_],[C0|_]] = wings_face:vinfo_ccw(F,We),
+lsq_init_fs([F|Fs],P,We = #we{vp=Vtab},Ds0,N,Re0,Im0) ->
+    Vs = [[A0|_],[B0|_],[C0|_]] = wings_va:face_attr([vertex|uv], F, We),
     {[A,B,C],Ds} = update_dicts(Vs,Ds0),
 %%    {X1=Z0x,Y1=Z0y,X2=Z1x,Y2=Z1y,X3=Z2x,Y3=Z2y} = 
     {X1,Y1,X2,Y2,X3,Y3} = 
-	project_tri(gb_trees:get(A0,Vtab),gb_trees:get(B0,Vtab),
-		    gb_trees:get(C0,Vtab)), 
+	project_tri(array:get(A0,Vtab),array:get(B0,Vtab),
+		    array:get(C0,Vtab)), 
     %% Raimos old solution. 
     SqrtDT = try math:sqrt(abs((X2-X1)*(Y3-Y1)-(Y2-Y1)*(X3-X1))) 
 	     catch _:_ -> 0.000001
@@ -660,10 +660,9 @@ lsq_init_fs([F|Fs],P,We = #we{vp=Vtab},Ds0,N,Re0,Im0,X0) ->
     %% nlEnd(NL_ROW) ;
     %% }
 
-    lsq_init_fs(Fs,P,We,Ds,N+1,Re,Im,fixuv(Vs,P,Ds0)++X0);
-lsq_init_fs([],_,_We,{M,D,DR},N,Re0,Im0,X0) ->
-%%    io:format("X0 = ~p~n",[ X0]),
-    {M,N,D,DR,vecs(M,Re0,[]),vecs(M,Im0,[]),auv_matrix:vector(X0)}.
+    lsq_init_fs(Fs,P,We,Ds,N+1,Re,Im);
+lsq_init_fs([],_,_We,{M,D,DR},N,Re0,Im0) ->
+    {M,N,D,DR,vecs(M,Re0,[]),vecs(M,Im0,[])}.
 
 vecs(M,[R|Rs],Acc) ->
     vecs(M,Rs,[auv_matrix:vector(M,R)|Acc]);
@@ -682,20 +681,6 @@ update_dicts([[P|_]|Rest],N,D,DR,Acc) ->
 update_dicts([],N,D,DR,Acc) ->
     {lists:reverse(Acc),{N,D,DR}}.
 
-fixuv([[Id|UV]|R],Pinned,Ds={_,D,_}) ->
-    case dict:find(Id,D) of
-	{ok, _} -> fixuv(R,Pinned,Ds);
-	error -> 
-	    case {gb_trees:is_defined(Id, Pinned),UV} of
-		{true,_} ->   fixuv(R,Pinned,Ds);
-		{_,{U,V}}  -> [U,V|fixuv(R,Pinned,Ds)];
-		{_,_What} ->      
-%%		    io:format("~p ~p",[Id,_What]),
-		    [0.5,0.5|fixuv(R,Pinned,Ds)]
-	    end
-    end;
-fixuv([],_,_) -> [].
-
 project_tri(P0,P1,P2) ->
     L = e3d_vec:sub(P1,P0),
     X = e3d_vec:norm(L),
@@ -1058,12 +1043,12 @@ stretch_opt(We0, OVs) ->
     SUvs1 = gb_trees:to_list(SUvs0),
     
     Suvs = [{Id,{S0/Scale,T0/Scale,0.0}} || {Id,{S0,T0}} <- SUvs1],
-    We0#we{vp=gb_trees:from_orddict(Suvs)}.
+    We0#we{vp=array:from_orddict(Suvs)}.
 
 stretch_setup(Fs, We0, OVs) ->
     Be = wings_face:outer_edges(Fs, We0),
     Bv0 = foldl(fun(Edge, Acc) ->
-			#edge{vs=Vs,ve=Ve} = gb_trees:get(Edge, We0#we.es),
+			#edge{vs=Vs,ve=Ve} = array:get(Edge, We0#we.es),
 			[Vs,Ve|Acc]
 		end, [], Be),
     Bv = gb_sets:from_list(Bv0),
@@ -1211,9 +1196,9 @@ init_stretch([],_,F2S2,F2S8,F2Vs,V2Fs0,Uvs) ->
 
 calc_scale([{Face,[{Id1,P1},{Id2,P2},{Id3,P3}]}|R], Ovs, A2D, A3D,F2A,F2OVs) ->
     A2 = abs(area2d2(P1,P2,P3)/2),
-    Q1 = gb_trees:get(Id1,Ovs),
-    Q2 = gb_trees:get(Id2,Ovs),
-    Q3 = gb_trees:get(Id3,Ovs),    
+    Q1 = array:get(Id1,Ovs),
+    Q2 = array:get(Id2,Ovs),
+    Q3 = array:get(Id3,Ovs),    
     A3 = area3d(Q1,Q2,Q3),
     calc_scale(R,Ovs,A2+A2D,A3+A3D,[{Face,A3}|F2A],[{Face,{Q1,Q2,Q3}}|F2OVs]);
 calc_scale([],_Ovs,A2D,A3D,F2A,F2OVs) ->
