@@ -54,9 +54,9 @@ start(Op, #we{id=Id}=We0, OrigWe, St0) ->
 				   Es = wings_face:outer_edges(Chart,We1),
 				   gb_sets:union(gb_sets:from_list(Es),Acc)
 			   end, Cuts0, Charts),
-		 We1#we{he=Cuts, mode=material};
+		 We1#we{he=Cuts};
 	     delete_old ->
-		 We1#we{mode=material}
+		 We1
 	 end,
 
     St = St1#st{sel=[],selmode=face,shapes=gb_trees:from_orddict([{Id,We}])},
@@ -209,17 +209,7 @@ seg_event_6({action,{view,Cmd}}, #seg{st=St0}=Ss) ->
             wings_view:command(aim, St1),
             get_seg_event(Ss);
           false ->
-            HL0 = wings_pref:get_value(highlight_aim_at_unselected),
-            HL1 = wings_pref:get_value(highlight_aim_at_selected),
-            {_,X,Y} = wings_wm:local_mouse_state(),
-            {{_,Cmd0},St1} = case wings_pick:do_pick(X, Y, St0) of
-                  {add,_,St2} when HL0 =:= true ->
-                      {{view,highlight_aim},{add,St0,St2}};
-                  {delete,_,St2} when HL1 =:= true ->
-                      {{view,highlight_aim},{delete,St0,St2}};
-                  _Other -> 
-                      {{view,aim}, St0}
-            end,
+            {{_,Cmd0},St1} = wings:highlight_aim_setup(St0),
             wings_view:command(Cmd0,St1),
             get_seg_event(Ss)
         end;
@@ -323,12 +313,12 @@ seg_command(Cmd, #seg{st=#st{mat=Mat}=St0}=Ss) ->
 seg_command_debug(_, _) -> keep.
 -else.
 seg_command_debug({debug,select_features}, #seg{we=#we{id=Id}=We,st=St}=Ss) ->
-    Tot = gb_trees:size(We#we.es),
+    Tot = array:sparse_size(We#we.es),
     {Es,_,_} = auv_segment:find_features(We, 60, Tot div 50),
     Sel = [{Id,gb_sets:from_list(Es)}],
     get_seg_event(Ss#seg{st=St#st{selmode=edge,sel=Sel}});
 seg_command_debug({debug,select_seeds}, #seg{we=#we{id=Id}=We,st=St}=Ss) ->
-    Tot = gb_trees:size(We#we.es),
+    Tot = array:sparse_size(We#we.es),
     {Features,_,_} = auv_segment:find_features(We, 60, Tot div 50),
     {Seeds0,_} = auv_segment:build_seeds(Features, We),
     Seeds = [S || {_,S} <- Seeds0],
@@ -390,7 +380,7 @@ seg_map_charts_1([We0|Cs], Type, Id, N, Acc,Failed,Ss) ->
 	    seg_map_charts_1(Cs,Type,Id+1,N,Acc,[We1|Failed],Ss#seg{err=Message});
 	Vs0 ->
 	    Vs = auv_placement:rotate_area(Vs0,We1),
-	    We = We1#we{vp=gb_trees:from_orddict(sort(Vs))},
+	    We = We1#we{vp=array:from_orddict(sort(Vs))},
 	    seg_map_charts_1(Cs, Type, Id+1, N, [We|Acc], Failed, Ss)
     end;
 seg_map_charts_1([],_, _, _,Charts0,Failed,
