@@ -3,7 +3,7 @@
 %%
 %%     Render and capture a texture.
 %%
-%%  Copyright (c) 2002-2008 Dan Gudmundsson, Bjorn Gustavsson
+%%  Copyright (c) 2002-2009 Dan Gudmundsson, Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -12,9 +12,7 @@
 %%
 
 -module(auv_texture).
--export([get_texture/2, draw_options/0]).
-
--compile(export_all).
+-export([get_texture/2,draw_options/0]).
 
 -define(NEED_OPENGL, 1).
 -define(NEED_ESDL, 1).
@@ -24,13 +22,13 @@
 -include("auv.hrl").
 -include("e3d.hrl").
 
--import(lists, [foreach/2, reverse/1, sort/1, foldl/3, min/1,max/1]).
+-import(lists, [foreach/2,reverse/1,sort/1,foldl/3,member/2]).
+-import(erlang, [min/2,max/2]).
 -import(auv_segment, [map_vertex/2]).
 
--define(OPT_BG, [{type_sel,color},{undefined,ignore},{1.0,1.0,1.0}]).
--define(OPT_EDGES, [all_edges,{0.0,0.0,0.0}, 1.0, false]).
+-define(OPT_BG, [{type_sel,color},{undefined,ignore},{1.0,1.0,1.0,0.0}]).
+-define(OPT_EDGES, [all_edges,{0.0,0.0,0.0},1.0,false]).
 -define(OPT_FACES, [texture]).
--define(OPT_LIGHT, [64]).
 
 -record(opt, {texsz = {512,512},   %% Texture size
 	      no_renderers = 4,
@@ -63,8 +61,7 @@
  	 n,         % Normal            (binary) Optional
 	 bi,        % BiNormal          (binary) Optional
 	 bb,        % BoundingBox 3D pos
-	 uvc,       % Previous uv or vertex color (binary)
-	 uvc_mode   % material (uv) or vertex
+	 vc        % Vertex colors     (binary)
 	}).
 
 -record(chart, 
@@ -78,29 +75,28 @@
 -record(fs,
 	{vs,        % triangulated vertex id in uv window [[Id1,Id2,Id3]]
 	 vse,       % face vertex id's untriangulated for edge drawings
-	 mat,       % material 
 	 id}).      % Face Id
 
 %% Menu
 
 draw_options() ->
     [MaxTxs0|_] = gl:getIntegerv(?GL_MAX_TEXTURE_SIZE),
-    MaxTxs = max([min([4096,MaxTxs0]),256]),
+    MaxTxs = max(min(4096, MaxTxs0), 256),
     Prefs = get_pref(tx_prefs, pref_to_list(#opt{})),
     TexSz = proplists:get_value(texsz, Prefs, 512),
     Shaders = shaders(),
-    Qs = [{hframe,[{menu,gen_tx_sizes(MaxTxs,[]),TexSz,
+    Qs = [{hframe,[{menu,gen_tx_sizes(MaxTxs, []),TexSz,
 		    [{key,texsz}]}],
 	   [{title,?__(1,"Size")}]},
-	  {vframe, render_passes(Prefs,Shaders), [{title,?__(2,"Render")}]},
-	  {hframe, [{button,?__(4,"New Pass"),done,[{key,add_shader}]},
-		    {button,?__(5,"Delete Unused Pass"),done,[{key,del_shader}]}]}
+	  {vframe,render_passes(Prefs, Shaders), [{title,?__(2,"Render")}]},
+	  {hframe,[{button,?__(4,"New Pass"),done,[{key,add_shader}]},
+		   {button,?__(5,"Delete Unused Pass"),done,[{key,del_shader}]}]}
 	 ],
     
     wings_ask:dialog(?__(3,"Draw Options"), Qs,
 		     fun(Options) ->
 			     {{changed,New},Opt} = list_to_prefs(Options),
-			     set_pref([{tx_prefs, pref_to_list(Opt)}]),
+			     set_pref([{tx_prefs,pref_to_list(Opt)}]),
 			     if New -> 
 				     {auv,{draw_options,restart}};
 				true  ->
@@ -112,31 +108,30 @@ draw_options() ->
 %% Menu handling
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
-render_passes(Prefs,Shaders) ->
+render_passes(Prefs, Shaders) ->
     %% Ugly count of shaders
-    NoOfPasses = lists:max([length([ok||{{auv_pass,_},_}<- Prefs])-1,1]), 
+    NoOfPasses = max(length([ok || {{auv_pass,_},_} <- Prefs])-1, 1),
     Menu = renderers(Shaders),
     Background = 
 	{hframe, 
 	 [{label, integer_to_list(0) ++ ": "},
 	  {menu,[{?__(1,"Background"), auv_background}],auv_background,
 	   [{key,{auv_pass,0}},layout]}, 
-	  {value, get_def(Prefs,auv_opt,0), store_opt(0)},
+	  {value,get_def(Prefs,auv_opt,0), store_opt(0)},
 	  {button,?__(2,"Options"),keep,[option_hook(0,background(),[]),
 					 drop_flags(0)]}],[]},
     Other = [{hframe, 
 	      [{label, integer_to_list(Id) ++ ": "},
-	       {menu,Menu,default_menu(Id,Prefs),
+	       {menu,Menu,default_menu(Id, Prefs),
 		[{key,{auv_pass,Id}},layout,pass_hook(Id)]}, 
-	       {value, get_def(Prefs,auv_opt,Id), store_opt(Id)},
+	       {value,get_def(Prefs, auv_opt, Id),store_opt(Id)},
 	       {button,?__(3,"Options"),keep,
-		[option_hook(Id,Menu,Shaders),
+		[option_hook(Id, Menu, Shaders),
 		 drop_flags(Id)]}],
-	      []} 
-	     || Id <- lists:seq(1,NoOfPasses)],
+	      []} || Id <- lists:seq(1, NoOfPasses)],
     [Background|Other].
 
-default_menu(Pass,Prefs) -> 
+default_menu(Pass, Prefs) ->
     case get_def(Prefs, auv_pass, Pass) of
 	ignore -> default_menu(Pass);
 	Val -> Val
@@ -147,52 +142,38 @@ default_menu(_) -> ignore.
 
 get_def(List, What, Id) ->
     case proplists:get_value({What,Id}, List) of
-	undefined when What == auv_pass -> ignore;
-	undefined when What == auv_opt  -> [];
+	undefined when What =:= auv_pass -> ignore;
+	undefined when What =:= auv_opt  -> [];
 	Val -> Val
     end.
 
 background() ->
     [{?__(1,"Background"), auv_background}].
 renderers(Shaders) ->
-    Menu0 = [{"*"++Name++"*", {shader,Id}} || 
+    Menu0 = [{"*"++Name++"*",{shader,Id}} ||
 		#sh{name=Name,id=Id} <- Shaders],
-    Menu1 = Menu0,
-%%     Menu1 = case Shaders of 
-%% 		[] -> Menu0;
-%% 		_ -> [{?__(4, "Lights"), auv_light}|Menu0]
-%% 	    end,
     [{?__(1,"None"), ignore},
      {?__(2,"Draw Edges"),auv_edges},
-     {?__(3,"Draw Faces"),auv_faces}| Menu1
-    ].
+     {?__(3,"Draw Faces"),auv_faces}|Menu0].
 
 options(auv_background, [{type_sel,Type},{Image,_},Color],_) ->
     [{hradio,[{?__(1,"Image"),image},{?__(2,"Color"),color}],
       Type,[{key,type_sel},layout]},
      {hframe,[{label,?__(1,"Image")},image_selector(Image)],
       [is_enabled(image)]},
-     {hframe,[{label,?__(2,"Color")},{color,fix(Color,have_fbo())}],
+     {hframe,[{label,?__(2,"Color")},{color,fix(Color,wings_gl:have_fbo())}],
       [is_enabled(color)]}];
 options(auv_background, _Bad,Sh) ->  
     options(auv_background, ?OPT_BG,Sh);
-options(auv_edges,[Type,Color,Size,UseMat],_) ->
+options(auv_edges,[Type,Color,Size,UseVtxColors],_) ->
     [{vradio,[{?__(3,"Draw All Edges"),all_edges},
 	      {?__(4,"Draw Border Edges"), border_edges}], 
       Type, []},
-     {hframe,[{label,?__(5,"Edge Color:")},{color, Color}]},
+     {hframe,[{label,?__(5,"Edge Color:")},{color,Color}]},
      {hframe,[{label,?__(6,"Edge Width:")},{text,Size,[{range,{0.0,100.0}}]}]},
-     {?__(7,"Use face material (on border edges)"), UseMat}
+     {?__(8,"Use vertex colors (on border edges)"),UseVtxColors}
     ];
 options(auv_edges,_,Sh) -> options(auv_edges,?OPT_EDGES,Sh);
-options(auv_faces,[Type],_) ->
-    [{vradio,[{?__(8,"Use Material Colors"),materials},
-	      {?__(9,"Use (previous) Texture/Vertex colors"), texture}],
-      Type, []}];
-options(auv_faces,_,Sh) -> options(auv_faces,?OPT_FACES,Sh);
-options(auv_light,[Size],_Sh) -> 
-    [{hframe,[{label,?__(10,"No Lights:")},{text,Size,[{range,{4,100}}]}]}];
-options(auv_light,_,Sh) -> options(auv_light,?OPT_LIGHT,Sh);
 options({shader,Id},Vals,Sh) ->
     {value,Shader} = lists:keysearch(Id,#sh.id,Sh),
     shader_options(Shader,Vals);
@@ -202,7 +183,7 @@ options(Command,Vals,_) ->
 
 shader_options(#sh{args=Args,def=Defs,file=File}, Vals) ->
     case shader_menu(Args,reverse(Vals),[]) of
-	{failed,_} -> 
+	{failed,_} ->
 	    case shader_menu(Args,Defs,[]) of
 		{failed,What} -> 
 		    io:format("AUV: Bad default value ~p in ~p~n",
@@ -210,7 +191,8 @@ shader_options(#sh{args=Args,def=Defs,file=File}, Vals) ->
 		    [];
 		Menues -> Menues
 	    end;
-	Menues -> Menues
+	Menues ->
+	    Menues
     end.
 shader_menu([{uniform,color,_,_,Label}|As],[Col={_,_,_,_}|Vs],Acc) ->
     Menu = {hframe, [{label,Label},{color,Col}]},
@@ -236,10 +218,11 @@ shader_menu([{uniform,{image,_},_,_Def,Label}|As],[Def|Vs],Acc) ->
 shader_menu([{uniform,menu,_,_,Labels}|As],[Def|Vs],Acc) ->
     Menu = {menu,Labels,Def,[]},
     shader_menu(As,Vs,[Menu|Acc]);
-shader_menu([{auv,{auv_send_texture,Label,_}}|As],[Def0|Vs],Acc) ->
-    case Def0 of
-	{auv_send_texture, Def} -> Def;
-	Def -> Def
+shader_menu([{auv,{auv_send_texture,Label,Def0}}|As],Vs0,Acc) ->
+    case Vs0 of
+	[{auv_send_texture, Def}|Vs] -> ok;
+	[Def|Vs] -> ok;
+	[] -> Def = Def0, Vs = []
     end,
     Menu = {Label, Def, [{key,auv_send_texture}]},
     shader_menu(As,Vs,[Menu|Acc]);
@@ -275,7 +258,8 @@ is_enabled(Type) ->
 
 option_hook(Id,Renderers,Shaders) ->
     {hook, fun(is_disabled,{_Var,_I,Sto}) ->
-		   gb_trees:get({auv_pass,Id}, Sto) == ignore;
+		   Pass = gb_trees:get({auv_pass,Id}, Sto),
+		   Pass =:= ignore orelse Pass =:= auv_faces;
 	      (is_minimized, _) ->
 		   false;
 	      (update,{_Var,_I,_B,Sto}) ->
@@ -330,8 +314,8 @@ get_texture(St = #st{bb=#uvstate{}}, {Options,Shaders}) ->
 %% Texture Rendering
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
-render_image(Geom = #ts{uv=UVpos,pos=Pos,n=Ns,bi=BiNs,uvc=Uvc,uvc_mode=Mode}, 
-	     Passes,#opt{texsz={TexW,TexH}},Reqs) ->
+render_image(#ts{uv=UVpos,pos=Pos,n=Ns,bi=BiNs,vc=Vc}=Geom,
+	     Passes, #opt{texsz={TexW,TexH}}, Reqs) ->
     gl:pushAttrib(?GL_ALL_ATTRIB_BITS),
     
     Current = wings_wm:viewport(),
@@ -349,11 +333,8 @@ render_image(Geom = #ts{uv=UVpos,pos=Pos,n=Ns,bi=BiNs,uvc=Uvc,uvc_mode=Mode},
 	true -> gl:normalPointer(?GL_FLOAT, 0, Ns);
 	false -> ignore
     end,
-    case Mode of
-	vertex -> gl:colorPointer(3,?GL_FLOAT,0,Uvc);
-	_Other -> (catch gl:clientActiveTexture(?GL_TEXTURE0)),
-		  gl:texCoordPointer(2,?GL_FLOAT,0,Uvc)
-    end,
+
+    gl:colorPointer(3, ?GL_FLOAT, 0, Vc),
     case have_shaders() of
 	false -> ignore;
 	true  -> 
@@ -406,66 +387,14 @@ render_image(Geom = #ts{uv=UVpos,pos=Pos,n=Ns,bi=BiNs,uvc=Uvc,uvc_mode=Mode},
     end.
 
 %%%%%%%%% FBO stuff
-have_fbo() -> wings_gl:is_ext('GL_EXT_framebuffer_object').
 
 setup_fbo(W,H) ->
-    case have_fbo() of
-	false -> false;
-	true -> setup_fbo2(W,H) 
-    end.
-
-setup_fbo2(W,H) ->
-    [FB] = gl:genFramebuffersEXT(1),
-    [Col1,Col2] = gl:genTextures(2),
-    [Depth] = gl:genRenderbuffersEXT(1),
-    gl:bindFramebufferEXT(?GL_FRAMEBUFFER_EXT, FB),
-    %% Init color texture
-    gl:bindTexture(?GL_TEXTURE_2D, Col1),
-    gl:texParameterf(?GL_TEXTURE_2D,?GL_TEXTURE_MIN_FILTER,?GL_LINEAR),
-    gl:texImage2D(?GL_TEXTURE_2D, 0, ?GL_RGBA8, W, H, 0,
-		  ?GL_RGBA, ?GL_UNSIGNED_BYTE, 0),
-    gl:framebufferTexture2DEXT(?GL_FRAMEBUFFER_EXT,
-			       ?GL_COLOR_ATTACHMENT0_EXT,
-			       ?GL_TEXTURE_2D, Col1, 0),
-    
-    gl:bindTexture(?GL_TEXTURE_2D, Col2),
-    gl:texParameterf(?GL_TEXTURE_2D,?GL_TEXTURE_MIN_FILTER,?GL_LINEAR),
-    gl:texImage2D(?GL_TEXTURE_2D, 0, ?GL_RGBA8, W, H, 0,
-		  ?GL_RGBA, ?GL_UNSIGNED_BYTE, 0),
-    gl:framebufferTexture2DEXT(?GL_FRAMEBUFFER_EXT,
-			       ?GL_COLOR_ATTACHMENT1_EXT,
-			       ?GL_TEXTURE_2D, Col2, 0),
-    
-    %% Init depth texture
-    gl:bindRenderbufferEXT(?GL_RENDERBUFFER_EXT, Depth),
-    gl:renderbufferStorageEXT(?GL_RENDERBUFFER_EXT,
-			      ?GL_DEPTH_COMPONENT24, W, H),
-    gl:framebufferRenderbufferEXT(?GL_FRAMEBUFFER_EXT,
-				  ?GL_DEPTH_ATTACHMENT_EXT,
-				  ?GL_RENDERBUFFER_EXT, Depth),
-    Delete = 
-	fun() ->
-		gl:framebufferTexture2DEXT(?GL_FRAMEBUFFER_EXT,
-					   ?GL_COLOR_ATTACHMENT0_EXT,
-					   ?GL_TEXTURE_2D,0,0),
-		gl:deleteTextures(2,[Col1,Col2]),
-		gl:framebufferRenderbufferEXT(?GL_FRAMEBUFFER_EXT,
-					      ?GL_DEPTH_ATTACHMENT_EXT,
-					      ?GL_RENDERBUFFER_EXT, 0),
-		gl:deleteRenderbuffersEXT(1,[Depth]),
-		gl:bindFramebufferEXT(?GL_FRAMEBUFFER_EXT, 0),
-		gl:deleteFramebuffersEXT(1,[FB])
-	end,
-    case check_fbo_status(FB) of
-	false -> 		    
-	    gl:bindFramebufferEXT(?GL_FRAMEBUFFER_EXT, 0), 
-	    gl:deleteFramebuffersEXT(1,[FB]),
-	    false;
-	_ ->	    
-	    gl:bindFramebufferEXT(?GL_FRAMEBUFFER_EXT, FB),
-	    gl:drawBuffer(?GL_COLOR_ATTACHMENT0_EXT),
+    case wings_gl:setup_fbo({W,H}, [{color,[]}, {color,[]}, {depth,[]}]) of
+	not_supported -> false;
+	List ->
+	    [Col1,Col2] = [Col || {color, Col} <- List],
 	    #sh_conf{texsz={W,H},fbo_r=Col2,fbo_w=Col1,
-		     fbo_d=Delete}
+		     fbo_d=fun() -> wings_gl:delete_fbo(List) end}
     end.
 	
 error(Line) ->
@@ -512,9 +441,9 @@ get_texture(Wc, Wd, Hc, Hd, {W,H}=Info, DL, UsingFbo, ImageAcc)
 		gl:readBuffer(?GL_COLOR_ATTACHMENT0_EXT),
 		{4,?GL_RGBA}
 	end,
-    Mem = sdl_util:alloc(W*H*Sz, ?GL_UNSIGNED_BYTE),
+    Mem = wings_io:get_buffer(W*H*Sz, ?GL_UNSIGNED_BYTE),
     gl:readPixels(0,0,W,H,Type,?GL_UNSIGNED_BYTE,Mem),
-    ImageBin = sdl_util:getBin(Mem),
+    ImageBin = wings_io:get_bin(Mem),
     get_texture(Wc+1, Wd, Hc, Hd, Info, DL, UsingFbo, [ImageBin|ImageAcc]);
 get_texture(_Wc,Wd,Hc,Hd, Info, Dl, UsingFbo, ImageAcc) when Hc < Hd ->
     get_texture(0, Wd, Hc+1, Hd, Info, Dl, UsingFbo, ImageAcc);
@@ -557,33 +486,6 @@ calc_texsize(Vp, Tex, Orig) when Tex =< Vp ->
 calc_texsize(Vp, Tex, Orig) ->
     calc_texsize(Vp, Tex div 2, Orig).
 
-check_fbo_status(FB) ->
-    case gl:checkFramebufferStatusEXT(?GL_FRAMEBUFFER_EXT) of
-	?GL_FRAMEBUFFER_COMPLETE_EXT ->
-	    FB;
-	?GL_FRAMEBUFFER_UNSUPPORTED_EXT ->
-	    io:format("GL_FRAMEBUFFER_UNSUPPORTED_EXT~n",[]),
-	    false;
-	?GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT ->
-	    io:format("GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT~n",[]),
-	    false;
-        ?GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT ->
-	    io:format("GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT~n",[]),
-	    false;
-        ?GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT ->
-	    io:format("GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT~n",[]),
-	    false;
-        ?GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT    ->
-	    io:format("GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT~n",[]),
-	    false;
-        ?GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT ->
-	    io:format("GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT~n",[]),
-	    false;
-        ?GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT ->
-	    io:format("GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT~n",[]),
-	    false
-    end.
-
 get_pref(Key, Def) ->
     wpa:pref_get(autouv, Key, Def).
 set_pref(KeyVals) ->
@@ -646,29 +548,23 @@ set_viewport({X,Y,W,H}=Viewport) ->
 %% Data setup 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
-setup(St = #st{bb=#uvstate{id=RId, st=#st{mat=Mat,shapes=Sh0},
-			   orig_st=#st{mat=OrigMat,shapes=OrigSh}}},
-      Reqs) ->
-    We   = gb_trees:get(RId,Sh0),
-    Orig = gb_trees:get(RId,OrigSh),
-    Mats = merge_mats(gb_trees:to_list(OrigMat),Mat),
-    {Charts,{_Cnt,UVpos,Vpos,Ns,Ts,BB,Uvc}} = setup_charts(St,We,Orig,Mats,Reqs),
+setup(#st{bb=#uvstate{id=RId,st=#st{shapes=Sh0}}}=St, Reqs) ->
+    We = gb_trees:get(RId,Sh0),
+    {Charts,{_Cnt,UVpos,Vpos,Ns,Ts,BB,Vc}} = setup_charts(St, We, Reqs),
     #ts{charts=Charts,
 	uv =to_bin(UVpos,uv),
 	pos=to_bin(Vpos,pos),
 	n  =to_bin(Ns,pos),
 	bi =to_bin(Ts,pos),
 	bb = BB,
-	uvc=to_bin(Uvc,Orig#we.mode), 
-	uvc_mode=Orig#we.mode}.
+	vc=to_bin(Vc, vertex)}.
 
-setup_charts(#st{shapes=Cs0,selmode=Mode,sel=Sel},We,OrigWe,Mats,Reqs) ->
-    Ns = case lists:member(normal,Reqs) orelse lists:member(binormal,Reqs) of
+setup_charts(#st{shapes=Cs0,selmode=Mode,sel=Sel}, We, Reqs) ->
+    Ns = case member(normal, Reqs) orelse member(binormal, Reqs) of
 	     true -> setup_normals(We);
 	     false -> []
 	 end,
-    Start = {0,[],[],[],[],[],[]}, %% {UvPos,3dPos,Normal,Tangent,Uvc}
-    Mat = fun(Face) -> get_material(Face,We,OrigWe,Mats) end,
+    Start = {0,[],[],[],[],[],[]}, %% {UvPos,3dPos,Normal,Tangent,Vc}
     Shapes = if Sel =:= [] -> 
 		     gb_trees:values(Cs0);
 		Mode =:= body -> 
@@ -676,32 +572,36 @@ setup_charts(#st{shapes=Cs0,selmode=Mode,sel=Sel},We,OrigWe,Mats,Reqs) ->
 		true -> 
 		     gb_trees:values(Cs0)
 	     end,
-    Setup = fun(Ch,Acc) -> setup_chart(Ch,Mat,Ns,Reqs,We,OrigWe,Acc) end,
+    Setup = fun(Ch,Acc) ->
+		    setup_chart(Ch, Ns, Reqs, We, Acc)
+	    end,
     lists:mapfoldl(Setup, Start, Shapes).
 
-setup_chart(Uv = #we{id=Id},Mat,Ns,Reqs,WWe,OWe,State0) ->
-    OEs0 = outer_verts(Uv), 
+setup_chart(#we{id=Id}=Ch, Ns, Reqs, WWe, State0) ->
+    OEs0 = outer_verts(Ch),
     BB = [],
     {Fs,{OEs,UvBB,State}}  = 
-	create_faces(Uv,WWe,OWe,Ns,Mat,Reqs,{OEs0,BB,State0}),
+	create_faces(Ch, WWe, Ns, Reqs, {OEs0,BB,State0}),
     {#chart{id=Id,fs=Fs,oes=OEs,bb_uv=UvBB},State}.
 
-create_faces(We = #we{vp=Vtab,name=#ch{vmap=Vmap}},
-	     RWe = #we{vp=Vt3d},OWe=#we{mode=OldMode},
-	     NTab,GetMat,Reqs,State) ->
+create_faces(#we{vp=Vtab,name=#ch{vmap=Vmap}}=We,
+	     #we{vp=Vt3d}=RealWe, NTab, Reqs, State) ->
     Fs = wings_we:visible(We),
-    C=fun(Face,{OEs,UvBB,{Cnt,UVpos,Vpos,Ns,Ts,PosBB,Uvc}}) ->
+    C=fun(Face,{OEs,UvBB,{Cnt,UVpos,Vpos,Ns,Ts,PosBB,Vc}}) ->
 	      Vs0 = wings_face:vertices_ccw(Face,We),
-	      UVcoords = [gb_trees:get(V, Vtab) || V <- Vs0],
-	      Coords   = [gb_trees:get(map_vertex(V,Vmap),Vt3d) 
+	      UVcoords = [array:get(V, Vtab) || V <- Vs0],
+	      Coords   = [array:get(map_vertex(V,Vmap),Vt3d) 
 			  || V <- Vs0],
 	      Normals = if
-			    NTab=:= [] -> [];
-			    true -> fix_normals(Vs0,Vmap,
-						wings_face:vinfo_ccw(Face,RWe),
-						gb_trees:get(Face,NTab))
+			    NTab=:= [] ->
+				[];
+			    true ->
+				fix_normals(Vs0, Vmap,
+					    wings_va:face_attr([vertex|uv],
+							       Face, RealWe),
+					    gb_trees:get(Face,NTab))
 			end,
-	      OldUvc  = fix_uvc(Vs0,Face,OWe,Vmap,OldMode),
+	      OldVc  = fix_vc(Vs0, Face, RealWe, Vmap),
 	      Len = length(Vs0),
 	      FaceVs = lists:seq(0, Len-1),
 	      Vs = case Len of
@@ -711,9 +611,8 @@ create_faces(We = #we{vp=Vtab,name=#ch{vmap=Vmap}},
 %	      io:format("Face ~p Normals ~p~n", [Face,Normals]),
 	      Tangents = fix_tang_vecs(Vs,Coords,UVcoords,Normals,Reqs), 
 	      Indx = fun(I) -> [V+Cnt || V <- I] end,
-	      Mat  = GetMat(Face),
-	      {#fs{vs=Indx(Vs),vse=Indx(FaceVs),mat=Mat,id=Face},
-	       {map_oes(OEs,Vs0,Cnt+Len-1,Face,Mat),
+	      {#fs{vs=Indx(Vs),vse=Indx(FaceVs),id=Face},
+	       {map_oes(OEs, Vs0, Cnt+Len-1, Face),
 		e3d_vec:bounding_box(UvBB ++ UVcoords),
 		{Cnt+Len,
 		 UVcoords ++ UVpos,
@@ -721,7 +620,7 @@ create_faces(We = #we{vp=Vtab,name=#ch{vmap=Vmap}},
 		 Normals  ++ Ns,
 		 Tangents ++ Ts,
 		 e3d_vec:bounding_box(PosBB ++ Coords),
-		 OldUvc   ++ Uvc}}}
+		 OldVc   ++ Vc}}}
       end,
     lists:mapfoldl(C, State, Fs).
 
@@ -730,16 +629,16 @@ triangulate(FaceVs,Vcoords) ->
     T3dfaces = e3d_mesh:triangulate_face(E3dface, Vcoords),
     lists:append([FVs || #e3d_face{vs=FVs} <- T3dfaces]).
 
-map_oes([[A,B,Face]|OEs],Vs0,Cnt,Face,Mat) ->
-    MA = member(A,Vs0,Cnt),
-    MB = member(B,Vs0,Cnt),
-    [[MA,MB,Mat]|map_oes(OEs,Vs0,Cnt,Face,Mat)];
-map_oes([Other|OEs],Vs0,Cnt,Face,Mat) ->
-    [Other|map_oes(OEs,Vs0,Cnt,Face,Mat)];
-map_oes([],_,_,_,_) -> [].
+map_oes([[A,B,Face]|OEs], Vs0, Cnt, Face) ->
+    MA = find_index(A, Vs0, Cnt),
+    MB = find_index(B, Vs0, Cnt),
+    [[MA,MB,Face]|map_oes(OEs, Vs0, Cnt, Face)];
+map_oes([Other|OEs], Vs0, Cnt, Face) ->
+    [Other|map_oes(OEs, Vs0, Cnt, Face)];
+map_oes([], _, _, _) -> [].
 
-member(Val,[Val|_],Pos) -> Pos;
-member(Val,[_|R],Pos) -> member(Val,R,Pos-1).
+find_index(Val, [Val|_], Pos) -> Pos;
+find_index(Val, [_|R], Pos) -> find_index(Val, R, Pos-1).
 
 fix_normals(Vs,Vmap,VsI,Ns0) -> %% can be different order 
     fix_normals(Vs,n_zip(VsI,Ns0),Vmap).  
@@ -750,12 +649,9 @@ fix_normals([],_,_) -> [].
 fix_tang_vecs(_,_,_,[],_) -> [];
 fix_tang_vecs(_Vs,Coords,_UVCoords,Normals,Reqs) ->
     case lists:member(binormal,Reqs) of
-	true -> % Not tested enough, but I think it's better 
-% 	    All = fix_tang_vecs1(Vs,Coords,UVCoords,Normals,[]),
-% 	    R1 = fix_tang_vecs2(sort(All),[]),
-	    R2 = binormals(Normals,e3d_vec:normal(Coords)),
-%	    io:format("~p~n~p~n~n",[R1,R2]),
-	    R2;
+	true ->
+	    %% Not tested enough, but I think it's better
+	    binormals(Normals,e3d_vec:normal(Coords));
 	false ->
 	    []
     end.
@@ -771,83 +667,28 @@ binormals(Normals,FaceNormal) ->
 		e3d_vec:norm(e3d_vec:cross(Bi,N))
 	end,
     [R(N) || N <- Normals].
-
-%%% This code should be inside wings somewhere its copied 
-%%% from wpc_opengl.erl
-fix_tang_vecs1([A,B,C|Vs],Coords,UVCoords,Normals,Acc) ->
-    {UV1x,UV1y,_} = lists:nth(A+1,UVCoords),
-    {UV2x,UV2y,_} = lists:nth(B+1,UVCoords),
-    {UV3x,UV3y,_} = lists:nth(C+1,UVCoords),
-    V1 = lists:nth(A+1,Coords),N1 = lists:nth(A+1,Coords),
-    V2 = lists:nth(B+1,Coords),N2 = lists:nth(B+1,Coords),
-    V3 = lists:nth(C+1,Coords),N3 = lists:nth(C+1,Coords),
-    TBN1 = calcTS(V3,V1,V2,UV3x,UV3y,UV1x,UV1y,UV2x,UV2y,N1),
-    TBN2 = calcTS(V1,V2,V3,UV1x,UV1y,UV2x,UV2y,UV3x,UV3y,N2),
-    TBN3 = calcTS(V2,V3,V1,UV2x,UV2y,UV3x,UV3y,UV1x,UV1y,N3),
-    fix_tang_vecs1(Vs,Coords,UVCoords,Normals,
-		   [{A,TBN1},{B,TBN2},{C,TBN3}|Acc]);
-fix_tang_vecs1([],_Coords,_UVCoords,_Normals,Acc) -> 
-    Acc.
    
-calcTS(V1,V2,V3,S1,T1,S2,T2,S3,T3,N) ->
-    Side1 = e3d_vec:sub(V1,V2),
-    Side2 = e3d_vec:sub(V3,V2),
-    DT1 = T1-T2,
-    DT2 = T3-T2,
-    Stan1 = e3d_vec:norm(e3d_vec:sub(e3d_vec:mul(Side1,DT2),
-				     e3d_vec:mul(Side2,DT1))),	    
-    DS1 = S1-S2,
-    DS2 = S3-S2,
-    Ttan1 = e3d_vec:norm(e3d_vec:sub(e3d_vec:mul(Side1,DS2),
-				     e3d_vec:mul(Side2,DS1))),
-    %% OK we have the 2 tangents but cross them again with normal
-    %% so they are orthogonal
-    Stan  = e3d_vec:norm(e3d_vec:cross(Ttan1,N)),
-    Ttan  = e3d_vec:norm(e3d_vec:cross(Stan1,N)),
-    check_coordsys(Stan,Ttan,N).
-
-fix_tang_vecs2([{_,{S1,_,_}},{_,{S2,_,_}},{_,{S3,_,_}}],[]) ->
-    [S1,S2,S3];
-fix_tang_vecs2(List=[{Id,{_,_,N}}|_], Acc) ->
-    {Stans,R} = get_same(Id, List,[]),
-    T = e3d_vec:cross(e3d_vec:average(Stans),N),
-    fix_tang_vecs2(R,[e3d_vec:norm(T)|Acc]);
-fix_tang_vecs2([],Acc) ->
-    lists:reverse(Acc).
-
-get_same(Id, [{Id,{S,_,_}}|R],Acc) ->
-    get_same(Id,R,[S|Acc]);
-get_same(_,R,Acc) -> {Acc,R}.
-
-check_coordsys(Stan,Ttan,N) ->
-    Check = e3d_vec:dot(e3d_vec:cross(Stan,Ttan),N),
-    if 
-	Check < 0.0 -> {e3d_vec:mul(Stan,-1.0),e3d_vec:mul(Ttan,-1.0),N};
-	true ->        {Stan,Ttan,N}
-    end.
-
-fix_uvc(Vs,Face,OWe,Vmap,Mode) ->
+fix_vc(Vs, Face, We, Vmap) ->
     try 
-	Uvc = wings_face:vinfo_ccw(Face,OWe),
-	fix_uvc1(Vs,Uvc,Vmap,Mode,[])
+	Uvc = wings_va:face_attr([vertex|color], Face, We),
+	fix_vc1(Vs, Uvc, Vmap, [])
     catch error:_ ->
-	    fix_uvc1(Vs,[],Vmap,Mode,[])
+	    fix_vc1(Vs, [], Vmap, [])
     end.
-fix_uvc1([V|Vs],Uvc,Vmap,Mode, Acc) ->
-    Val = case find(map_vertex(V,Vmap),Uvc) of
-	      Color = {_,_,_} when Mode == vertex ->  Color;
-	      _ when Mode == vertex -> {1.0,1.0,1.0};
-	      Uv = {_,_} -> Uv;
-	      _ -> {0.0,0.0}
+
+fix_vc1([V|Vs], Uvc, Vmap, Acc) ->
+    Val = case find(map_vertex(V, Vmap), Uvc) of
+	      {_,_,_}=Color ->  Color;
+	      _ -> {1.0,1.0,1.0}
 	  end,
-    fix_uvc1(Vs,Uvc,Vmap,Mode,[Val|Acc]);
-fix_uvc1([],_,_,_,Acc) -> reverse(Acc).
+    fix_vc1(Vs, Uvc, Vmap, [Val|Acc]);
+fix_vc1([], _, _, Acc) -> reverse(Acc).
 
 find(V, [[V|Info]|_R]) -> Info;
 find(V, [_|R]) -> find(V,R);
 find(_, []) -> none.
 
-n_zip([[V|_]|R1],[[_|N]|R2]) -> 
+n_zip([[V|_]|R1],[N|R2]) ->
     [[V|N]|n_zip(R1,R2)];
 n_zip([],[]) -> [].
     
@@ -859,30 +700,14 @@ fix({R,G,B}, true) -> {R,G,B,1.0}.
     
 setup_normals(We = #we{fs=Ftab}) ->
     FN0	= [{Face,wings_face:normal(Face, We)} || Face <- gb_trees:keys(Ftab)],
-    Ns = wings_we:normals(FN0, We),  %% gb_tree of {Face, [VInfo|Normal]}    
+    Ns = wings_we:normals(FN0, We, none),
     gb_trees:from_orddict(sort(Ns)).
 
-get_material(Face, We, OrigWe, Materials) ->
-    Mat1 = wings_facemat:face(Face,We),
-    Mat = try
-	      case reverse(atom_to_list(Mat1)) of
-		  "vua_" ++ _ -> 
-		      try wings_facemat:face(Face,OrigWe)
-		      catch _:_ -> Mat1
-		      end;
-		  _ ->
-		      Mat1
-	      end
-	  catch _:_ ->
-		  Mat1
-	  end,
-    gb_trees:get(Mat, Materials).
-
 outer_verts(We = #we{es=Etab}) ->
     Fs = wings_we:visible(We),
     Outer = auv_util:outer_edges(Fs,We,false),    
     Verts = fun({Edge,Face}) -> 
-		    #edge{vs=Va,ve=Vb} = gb_trees:get(Edge, Etab),
+		    #edge{vs=Va,ve=Vb} = array:get(Edge, Etab),
 		    [Va,Vb,Face]
 	    end,
     lists:map(Verts, Outer).
@@ -891,37 +716,22 @@ outer_verts(We = #we{es=Etab}) ->
 %% and not on the process heap spent hours debugging this.. :-(
 to_bin(List, uv) -> to_bin3to2(List,[<<0:512>>]);
 to_bin(List, pos) -> to_bin3(List,[<<0:512>>]);
-to_bin(List, vertex) -> to_bin3(List,[<<0:512>>]);  %% Vertex colors
-to_bin(List, material) -> to_bin2(List,[<<0:512>>]).  %% UV coords.
+to_bin(List, vertex) -> to_bin3(List,[<<0:512>>]).  %% Vertex colors
 
 to_bin3([{A,B,C}|R],Acc) -> 
     to_bin3(R,[<<A:32/native-float,B:32/native-float,C:32/native-float>>|Acc]);
 to_bin3([],Acc) -> list_to_binary(Acc).
-to_bin2([{A,B}|R],Acc) -> 
-    to_bin2(R,[<<A:32/native-float,B:32/native-float>>|Acc]);
-to_bin2([],Acc) -> list_to_binary(Acc).
+
 to_bin3to2([{A,B,_}|R],Acc) -> 
     to_bin3to2(R,[<<A:32/native-float,B:32/native-float>>|Acc]);
 to_bin3to2([],Acc) -> list_to_binary(Acc).
 
-%
-% Workaround for ATI gl:'end' doesn't bite for line loop/strip..
+%% Workaround for ATI: gl:'end' doesn't bite for line loop/strip...
 vs_lines([A|R=[B|_]],Last) ->
     [A,B|vs_lines(R,Last)];
 vs_lines([B],Last) ->
     [B,Last].
 
-merge_mats([{MatName,Mat}|R], Mats) ->
-    case gb_trees:is_defined(MatName, Mats) of
-	true ->
-	    merge_mats(R, Mats);
-	false ->
-	    %% This can happen if the material has been deleted
-	    %% from the Outliner after entering the AutoUV window.
-	    merge_mats(R, gb_trees:insert(MatName, Mat, Mats))
-    end;
-merge_mats([], Mats) -> Mats.
-	     
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %% Builtin Shader Passes
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
@@ -965,9 +775,6 @@ pass({auv_edges, [all_edges,Color,Width,_UseMat]},_) ->
 			     Patched = vs_lines(Vs,hd(Vs)),
 			     gl:drawElements(?GL_LINES,length(Patched),
 					     ?GL_UNSIGNED_INT,Patched)
-%%                     Doesn't work
-%% 			     gl:drawElements(?GL_LINE_LOOP,length(Vs),
-%% 					     ?GL_UNSIGNED_INT,Vs) 
 			     end,
 	      foreach(Draw,Fs)
       end,
@@ -980,13 +787,7 @@ pass({auv_edges, [all_edges,Color,Width,_UseMat]},_) ->
 	    gl:disableClientState(?GL_VERTEX_ARRAY)
     end;
 pass({auv_edges, [border_edges,Color,Width,UseMat]},_) -> 
-    R= fun(#chart{oes=Es}) when UseMat ->
-	       Draw = fun([A,B,Mat]) ->
-			      gl:color4fv(get_diffuse(Mat)),
-			      gl:drawElements(?GL_LINES,2,?GL_UNSIGNED_INT,[A,B])
-		      end,
-	       foreach(Draw,Es);
-	  (#chart{oes=Es0}) ->
+    R= fun(#chart{oes=Es0}) ->
 	       Es = foldl(fun([A,B,_],Acc) -> [A,B|Acc] end, [], Es0),
 	       gl:drawElements(?GL_LINES,length(Es),?GL_UNSIGNED_INT,Es)
        end,
@@ -995,44 +796,34 @@ pass({auv_edges, [border_edges,Color,Width,UseMat]},_) ->
 	    gl:lineWidth(Width),
 	    gl:enableClientState(?GL_VERTEX_ARRAY),
 	    gl:disable(?GL_DEPTH_TEST),
-	    foreach(R, Charts),
+	    case UseMat of
+		true ->
+		    gl:enableClientState(?GL_COLOR_ARRAY),
+		    foreach(R, Charts),
+		    gl:disableClientState(?GL_COLOR_ARRAY);
+		_ ->
+		    foreach(R, Charts)
+	    end,
 	    gl:disableClientState(?GL_VERTEX_ARRAY)
     end;
 pass({auv_edges, _},Sh) ->
     pass({auv_edges, ?OPT_EDGES},Sh);
 
-pass({auv_faces, [Type]},_) ->
-    fun(#ts{charts=Charts,uvc_mode=Mode},_) ->  
+pass({auv_faces,[_]},_) ->
+    fun(#ts{charts=Charts}, _) ->
 	    gl:disable(?GL_DEPTH_TEST),
 	    gl:disable(?GL_ALPHA_TEST),
 	    gl:enable(?GL_BLEND),	   
 	    gl:blendFunc(?GL_SRC_ALPHA, ?GL_ONE_MINUS_SRC_ALPHA),
 	    gl:enableClientState(?GL_VERTEX_ARRAY),
 	    gl:clientActiveTexture(?GL_TEXTURE0),
-	    R = case Type of
-		    materials ->			
-			fun(#fs{vs=Vs,mat=Mat}) ->
-				gl:color4fv(get_diffuse(Mat)),
-				gl:drawElements(?GL_TRIANGLES,length(Vs),
-						?GL_UNSIGNED_INT,Vs)
-			end;
-		    _ when Mode == vertex -> 
-			gl:enableClientState(?GL_COLOR_ARRAY),
-			fun(#fs{vs=Vs}) ->
-				gl:drawElements(?GL_TRIANGLES,length(Vs),
-						?GL_UNSIGNED_INT,Vs)
-			end;
-		    _ -> 
-			gl:enableClientState(?GL_TEXTURE_COORD_ARRAY),
-			fun(#fs{vs=Vs,mat=Mat}) ->
-				set_diffuse_tx(Mat),
-				gl:drawElements(?GL_TRIANGLES,length(Vs),
-						?GL_UNSIGNED_INT,Vs)
-			end
+	    gl:enableClientState(?GL_COLOR_ARRAY),
+	    R = fun(#fs{vs=Vs}) ->
+			gl:drawElements(?GL_TRIANGLES,length(Vs),
+					?GL_UNSIGNED_INT,Vs)
 		end,
-	    erase({?MODULE,use_tx}),
 	    gl:disable(?GL_TEXTURE_2D),
-	    foreach(fun(#chart{fs=Fs}) -> foreach(R,Fs) end,Charts),
+	    foreach(fun(#chart{fs=Fs}) -> foreach(R, Fs) end, Charts),
 	    gl:disable(?GL_TEXTURE_2D),
 	    gl:disableClientState(?GL_VERTEX_ARRAY),
 	    gl:disableClientState(?GL_COLOR_ARRAY),
@@ -1040,33 +831,6 @@ pass({auv_faces, [Type]},_) ->
     end;
 pass({auv_faces, _},Sh) ->
     pass({auv_faces,?OPT_FACES},Sh);
-pass({auv_light, _},_Sh) ->
-%%     Circle = fun(N, Y, R) ->
-%% 		     Delta = math:pi()*2 / N,
-%% 		     [{R*math:cos(I*Delta), Y, R*math:sin(I*Delta)} 
-%% 		      || I <- lists:seq(0, N-1)]
-%% 	     end,
-%%     Sphere = fun(Ns,Nl) ->
-%% 		     Delta = math:pi() / Nl,
-%% 		     PosAndRads= [{math:cos(I*Delta), math:sin(I*Delta)} 
-%% 				  || I <- lists:seq(1, Nl-1)],
-%% 		     Circles = [Circle(Ns, Pos, Rad) || {Pos, Rad} <- PosAndRads],
-%% 		     lists:flatten(Circles)
-%% 	     end,
-%%  Lights = Sphere(8,4) ++ [{0.0,1.0,0.0}, {0.0,-1.0,0.0}],
-    Ligths = [{1.0,0.0,0.0}],
-    
-    VsS = wings_gl:compile(vertex, read_file("light1.vs")),
-    LP1 = wings_gl:link_prog([VsS]),
-    io:format("Light Prog ~p ~n",[LP1]),
-
-    fun(Ts = #ts{bb=BB=[Min,Max]},_) ->
-	    Center = e3d_vec:average(BB),
-	    Dist   = e3d_vec:dist(Min,Max),
-	    PM = get_lightProjM(Dist),
-	    render_lights(Ligths, PM, 10.0+Dist, Center,LP1,Ts)
-    end;
-
 pass({{shader,Id}, Opts},{Sh,Compiled}) ->
     shader_pass(lists:keysearch(Id,#sh.id,Sh),
 		lists:keysearch(Id,1,Compiled),Opts);
@@ -1083,7 +847,7 @@ shader_pass(_,false,_) ->
 shader_pass(false,_,_) ->    
     io:format("AUV: Not shader found skipped ~p~n", [?LINE]),
     ignore;
-shader_pass({value,#sh{args=Args,tex_units=TexUnits,reqs=Reqs}},
+shader_pass({value,#sh{name=_Name, args=Args,tex_units=TexUnits,reqs=Reqs}},
 	    {value,{_,Prog}},Opts) ->
     fun(Ts = #ts{charts=Charts},Config) ->
 	    gl:disable(?GL_DEPTH_TEST),
@@ -1094,8 +858,10 @@ shader_pass({value,#sh{args=Args,tex_units=TexUnits,reqs=Reqs}},
 	    try 
 		Conf = Config#sh_conf{prog=Prog,ts=Ts},
 		PerChartUniF = shader_uniforms(reverse(Args),Opts,Conf),
-		case lists:keysearch(auv_send_texture,1,Opts) of 
-		    {value,{auv_send_texture,true}} -> 
+		case send_texture(reverse(Args),Opts) of
+		    true ->
+			gl:enable(?GL_TEXTURE_2D),
+			gl:disable(?GL_BLEND),
 			draw_texture_square();
 		    _ -> 
 			gl:enableClientState(?GL_VERTEX_ARRAY),
@@ -1143,20 +909,30 @@ shader_pass({value,#sh{args=Args,tex_units=TexUnits,reqs=Reqs}},
 	    end
     end.
 
+send_texture([{auv,{auv_send_texture,_, _Def0}}|_],
+	     [{auv_send_texture,Def}|_]) ->
+    Def;
+send_texture([{auv,{auv_send_texture,_, _Def0}}|_],
+	     [Def|_]) ->
+    Def;
+send_texture([{auv,_}|Next], Opts) ->
+    send_texture(Next,Opts);
+send_texture([_|Next], [_|Opts]) ->
+    send_texture(Next,Opts);
+send_texture([],_) -> false.
+
+
 shader_uniforms([{uniform,color,Name,_,_}|As],[Val|Opts],Conf) ->
-    Loc = wings_gl:uloc(Conf#sh_conf.prog,Name),
-    gl:uniform4fv(Loc,1,[Val]),
+    wings_gl:set_uloc(Conf#sh_conf.prog,Name,Val),
     shader_uniforms(As,Opts,Conf);
 shader_uniforms([{uniform,float,Name,_,_}|As],[Val|Opts],Conf) ->
-    Loc = wings_gl:uloc(Conf#sh_conf.prog,Name),
-    gl:uniform1f(Loc,Val),
+    wings_gl:set_uloc(Conf#sh_conf.prog,Name, Val),
     shader_uniforms(As,Opts,Conf);
 shader_uniforms([{uniform,menu,Name,_,_}|As],[Vals|Opts],Conf) 
   when is_list(Vals) ->
     Loc = wings_gl:uloc(Conf#sh_conf.prog,Name),
     foldl(fun(Val,Cnt) -> gl:uniform1f(Loc+Cnt,Val),Cnt+1 end,0,Vals),
     shader_uniforms(As,Opts,Conf);
-
 shader_uniforms([{uniform,bool,Name,_,_}|As],[Val|Opts],Conf) ->
     Loc = wings_gl:uloc(Conf#sh_conf.prog,Name),
     BoolF = if Val -> 1.0; true -> 0.0 end,
@@ -1187,6 +963,8 @@ shader_uniforms([{auv,{auv_bg,Unit}}|Rest],Opts,Conf) ->
     Loc = wings_gl:uloc(Conf#sh_conf.prog, "auv_bg"),
     gl:activeTexture(?GL_TEXTURE0),
     gl:bindTexture(?GL_TEXTURE_2D, Conf#sh_conf.fbo_r),
+    gl:texParameteri(?GL_TEXTURE_2D, ?GL_TEXTURE_MAG_FILTER, ?GL_NEAREST),
+    gl:texParameteri(?GL_TEXTURE_2D, ?GL_TEXTURE_MIN_FILTER, ?GL_NEAREST),
     gl:uniform1i(Loc, Unit),
     shader_uniforms(Rest,Opts,Conf);
 shader_uniforms([{auv,auv_texsz}|As],Opts,Conf = #sh_conf{texsz={W,H}}) ->
@@ -1232,99 +1010,6 @@ get_requirements(Shaders) ->
 			List ++ Acc
 		end, [], Shaders).
 
-%%%%%%%%%%%% Ligth shader funcs
-
-get_lightProjM(Dist) ->
-    %% Calc ProjMatrix
-    gl:matrixMode(?GL_PROJECTION),
-    gl:pushMatrix(),
-    gl:loadIdentity(),
-    Sz     = Dist*math:tan(45*math:pi()/180/2),
-    gl:ortho(-Sz, Sz, -Sz, Sz, 1.0, Dist*2+10),
-    PM = gl:getDoublev(?GL_PROJECTION_MATRIX),
-    gl:popMatrix(),
-    PM.
-
-id(Id) ->
-    <<R:8,G:8,B:8,A:8>> = <<Id:32/big>>,
-    V = {A/255,B/255,G/255,(255-R)/255},
-    io:format("~p ~n",[V]),
-    V.
-
-render_lights([{Lx,Ly,Lz}| Ligths], PM, Dist, C={Cx,Cy,Cz}, LP1,Ts = #ts{charts=Charts}) ->
-    io:format("~p ~p ~p ~n", [C, Dist, {Lx,Ly,Lz}]),
-    gl:matrixMode(?GL_PROJECTION),  gl:pushMatrix(),  gl:loadMatrixd(PM),
-    gl:matrixMode(?GL_MODELVIEW),   gl:pushMatrix(),  gl:loadIdentity(),
-    glu:lookAt(Lx*Dist,Ly*Dist,Lz*Dist,Cx,Cy,Cz,0.0,1.0,0.0),
-    gl:clear(?GL_DEPTH_BUFFER_BIT bor ?GL_COLOR_BUFFER_BIT),
-%%    gl:enable(?GL_CULL_FACE),
-    gl:enable(?GL_DEPTH_TEST), 
-    gl:disable(?GL_BLEND), 
-    gl:disable(?GL_ALPHA_TEST), 
-    gl:disable(?GL_TEXTURE_2D), 
-    gl:shadeModel(?GL_FLAT),
-    gl:cullFace(?GL_BACK),
-    gl:enableClientState(?GL_VERTEX_ARRAY),
-    gl:disableClientState(?GL_COLOR_ARRAY),
-    gl:clientActiveTexture(?GL_TEXTURE1),
-    gl:enableClientState(?GL_TEXTURE_COORD_ARRAY),
-
-    gl:useProgram(LP1),    
-    IdPos = wings_gl:uloc(LP1,"id"),
-    
-    R = fun(#fs{vs=Vs,id=Id}) ->
-		%% gl:color4ubv(id(Id)),
-		%% XXX The following line is broken. id/1 returns a tuple,
-		%% but a list of tuples is needed for argument 3.
-		gl:uniform4fv(IdPos, 1, id(Id)),
-		gl:drawElements(?GL_TRIANGLES,length(Vs),
-				?GL_UNSIGNED_INT,Vs)
-	end,
-    foreach(fun(#chart{fs=Fs}) -> foreach(R,Fs) end,Charts),
-    gl:popMatrix(),
-    gl:matrixMode(?GL_PROJECTION),
-    gl:popMatrix(),
-    
-%%     %% Pass 2
-%%     gl:useProgram(Vert2d),
-%%     MVP = e3d_mat:mul(MM,PM),
-%%     foreach(fun(#chart{fs=Fs}) -> foreach(R,Fs) end,Charts),
-    
-%%     %% Hmm kan vi verkligen anvnda DEPTH_BUFFERN HR utan 
-%%     %% att swappa render buffers?
-    render_lights(Ligths, PM, Dist, C, LP1, Ts);
-render_lights([],_,_,_,LP1,_) -> 
-    gl:deleteProgram(LP1),
-    ok.
-
-%%%%%%%%%%%% Ligth shader funcs
-
-%%%%%%%%%%%%%%%% 
-%% Materials
-get_diffuse(Mat) ->
-    proplists:get_value(diffuse, proplists:get_value(opengl, Mat)).
-
-set_diffuse_tx(Mat) ->
-    case get({?MODULE,use_tx}) of
-	Mat -> ok;
-	_ ->
-	    Maps = proplists:get_value(maps, Mat),
-	    Ogl = proplists:get_value(opengl,Mat),
-	    gl:color4fv(proplists:get_value(diffuse,Ogl)),
-	    case proplists:get_value(diffuse, Maps, none) of
-		none ->
-		    gl:disable(?GL_TEXTURE_2D),
-		    put({?MODULE,use_tx},Mat),
-		    false;
-		Diff0 ->
-		    gl:enable(?GL_TEXTURE_2D),
-		    Diff = wings_image:txid(Diff0),
-		    gl:bindTexture(?GL_TEXTURE_2D,Diff),
-		    put({?MODULE,use_tx},Mat),
-		    true
-	    end
-    end.
-
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %% Shader loading/handling 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
