@@ -88,9 +88,9 @@ start_uvmap_1([{Id,_}|T], Action, St) ->
 	    wings_wm:send(SegWin, {add_faces,Fs,St}),
 	    wings_wm:raise(SegWin);
 	Op when element(1,Op) == edit ->	    
-	    create_window(Op, EditWin, Id, St,St);
+	    create_window(Op, EditWin, Id, St);
 	Op ->	    
-	    create_window(Op, SegWin, Id, St,St)
+	    create_window(Op, SegWin, Id, St)
     end,
     start_uvmap_1(T, Action, St);
 start_uvmap_1([], _, _) -> keep.
@@ -121,7 +121,7 @@ segment_or_edit(segment_old,Id,#st{selmode=face,sel=Sel}) ->
 segment_or_edit(segment_old,_Id,_) ->
     {seg_ui,object,keep_old}.
 	    
-create_window(Action, Name, Id, #st{shapes=Shs}=St, OrigSt) ->
+create_window(Action, Name, Id, #st{shapes=Shs}=St) ->
     #we{name=ObjName} = We = gb_trees:get(Id, Shs),
     Op = {replace,fun(Ev) -> auv_event(Ev, St) end},
     Segment = if element(1,Action) == edit -> ""; true -> ?__(1,"Segmenting") end,
@@ -133,12 +133,12 @@ create_window(Action, Name, Id, #st{shapes=Shs}=St, OrigSt) ->
     wings_wm:toplevel(Name, Title, {X,Y,highest}, {W,H},
 		      [resizable,closable,menubar,{properties,Props},
 		       {toolbar,CreateToolbar}], Op),
-    wings_wm:send(Name, {init,{Action,We},OrigSt}).
+    wings_wm:send(Name, {init,{Action,We}}).
 
-auv_event({init,Op,OrigSt}, St) ->
+auv_event({init,Op}, St) ->
     wings:init_opengl(St),
     case Op of
-	{{edit,What},We} -> start_edit(What, We, St, OrigSt);
+	{{edit,What},We} -> start_edit(What, We, St);
 	{{seg_ui,_,Oper},We} ->  auv_seg_ui:start(Oper, We, We, St)
     end;
 auv_event(redraw, _) ->
@@ -153,7 +153,7 @@ auv_event(_Ev, _) -> keep.
 %%% Start the UV editor.
 %%%
 
-start_edit(Mode, We, St, OrigSt) ->
+start_edit(Mode, We, St) ->
     MatNames0 = wings_facemat:all(We),
     MatNames1 = sofs:from_external(MatNames0, [{face,material}]),
     MatNames2 = sofs:converse(MatNames1),
@@ -162,13 +162,13 @@ start_edit(Mode, We, St, OrigSt) ->
     MatNames = [Mat || {Name,_}=Mat <- MatNames4, get_texture(Name, St) /= false],
     case MatNames of
 	[{MatName,_}] ->
-	    do_edit(MatName, Mode, We, St, OrigSt);
+	    do_edit(MatName, Mode, We, St);
 	_ ->
-	    do_edit(none, Mode, We, St,OrigSt)
+	    do_edit(none, Mode, We, St)
     end.
 
-do_edit(MatName, Mode, We, GeomSt, OrigSt) ->
-    AuvSt = create_uv_state(gb_trees:empty(), MatName, Mode, We, GeomSt,OrigSt),
+do_edit(MatName, Mode, We, GeomSt) ->
+    AuvSt = create_uv_state(gb_trees:empty(), MatName, Mode, We, GeomSt),
     new_geom_state(GeomSt, AuvSt).
 
 init_show_maps(Charts0, Fs, #we{name=WeName,id=Id}, GeomSt0) ->
@@ -176,7 +176,7 @@ init_show_maps(Charts0, Fs, #we{name=WeName,id=Id}, GeomSt0) ->
     Charts  = gb_trees:from_orddict(keysort(1, Charts1)),
     MatName0 = list_to_atom(WeName++"_auv"),
     {GeomSt1,MatName} = 
-	case gb_trees:is_defined(MatName0,GeomSt0#st.mat) of
+	case gb_trees:is_defined(MatName0, GeomSt0#st.mat) of
 	    true -> 
 		{GeomSt0,MatName0};
 	    false ->
@@ -190,16 +190,16 @@ init_show_maps(Charts0, Fs, #we{name=WeName,id=Id}, GeomSt0) ->
 	    wings_wm:send(EditWin, {add_faces,Fs,GeomSt}),
 	    wings_wm:send(geom, {new_state,GeomSt});
 	false ->
-	    create_window({edit,Fs},EditWin,Id,GeomSt,GeomSt0),
+	    create_window({edit,Fs}, EditWin, Id, GeomSt),
 	    wings_wm:send(geom, {new_state,GeomSt})
     end,
     GeomSt.
 
-create_uv_state(Charts, MatName, Fs, We, GeomSt = #st{shapes=Shs0},OrigSt) ->
+create_uv_state(Charts, MatName, Fs, We, #st{shapes=Shs0}=GeomSt) ->
     wings:mode_restriction([vertex,edge,face,body]),
     wings_wm:current_state(#st{selmode=body,sel=[]}),
 
-    Shs = gb_trees:update(We#we.id, We#we{fs=undefined,es=gb_trees:empty()}, Shs0),
+    Shs = gb_trees:update(We#we.id, We#we{fs=undefined,es=array:new()}, Shs0),
     FakeGeomSt = GeomSt#st{sel=[],shapes=Shs},
 
     Image = case get_texture(MatName,GeomSt) of
@@ -210,7 +210,6 @@ create_uv_state(Charts, MatName, Fs, We, GeomSt = #st{shapes=Shs0},OrigSt) ->
 		   id      = We#we.id,
 		   mode    = Fs,
 		   bg_img  = Image,
-		   orig_st = OrigSt,
 		   matname = MatName},
     St = FakeGeomSt#st{selmode=body,sel=[],shapes=Charts,bb=Uvs,
 		       repeatable=ignore,ask_args=none,drag_args=none},
@@ -313,11 +312,11 @@ redundant_separators([H|T]) ->
 insert_initial_uvcoords(Charts, Id, MatName, #st{shapes=Shs0}=St) ->
     We0 = gb_trees:get(Id, Shs0),
     We1 = update_uvs(gb_trees:values(Charts), We0),
-    We2 = insert_material(Charts, MatName, We1),
-    We = We2#we{mode=material},    
+    We2 = preserve_old_materials(We1, St),
+    We = insert_material(Charts, MatName, We2),
     Shs = gb_trees:update(Id, We, Shs0),
     St#st{shapes=Shs}.
-   
+
 update_selected_uvcoords(#st{bb=Uvs}=St) ->
     Charts = wpa:sel_fold(fun(_, We, Acc) -> [We|Acc] end, [], St),
     #uvstate{st=#st{shapes=Shs0}=GeomSt0,id=Id} = Uvs,
@@ -330,11 +329,7 @@ update_selected_uvcoords(#st{bb=Uvs}=St) ->
 
 %% update_uvs(Charts, We0) -> We
 %%  Update the UV coordinates for the original model.
-update_uvs(Cs, #we{es=Etab0}=GeomWe) ->
-    update_uvs_1(Cs, GeomWe, Etab0).
-
-update_uvs_1([#we{vp=Vpos0,name=#ch{vmap=Vmap}}=ChartWe|Cs],
-	     We, Etab0) ->
+update_uvs([#we{vp=Vpos0,name=#ch{vmap=Vmap}}=ChartWe|Cs], We0) ->
     VFace0 = wings_face:fold_faces(
 	       fun(Face, V, _, _, A) ->
 		       [{V,Face}|A]
@@ -343,39 +338,33 @@ update_uvs_1([#we{vp=Vpos0,name=#ch{vmap=Vmap}}=ChartWe|Cs],
     VFace1 = sofs:relation(VFace0),
     VFace2 = sofs:relation_to_family(VFace1),
     VFace = sofs:to_external(VFace2),
-    Vpos = gb_trees:to_list(Vpos0),
-    Etab = update_uvs_2(Vpos, VFace, Vmap, We, Etab0),
-    update_uvs_1(Cs, We, Etab);
-update_uvs_1([], We, Etab) -> We#we{es=Etab}.
+    Vpos = array:sparse_to_orddict(Vpos0),
+    We = update_uvs_1(Vpos, VFace, Vmap, We0),
+    update_uvs(Cs, We);
+update_uvs([], We) -> We.
 
-update_uvs_2([{V0,{X,Y,_}}|Vs], [{V0,Fs}|VFs], Vmap, We, Etab0) ->
+update_uvs_1([{V0,{X,Y,_}}|Vs], [{V0,Fs}|VFs], Vmap, We0) ->
     UV = {X,Y},
     V = auv_segment:map_vertex(V0, Vmap),
-    Etab = foldl(fun(Face, A) ->
-			 update_uvs_3(V, Face, UV, A, We)
-		 end, Etab0, Fs),
-    update_uvs_2(Vs, VFs, Vmap, We, Etab);
-update_uvs_2([{V0,none}|Vs], [{V0,Fs}|VFs], Vmap, We, Etab0) ->
+    We = wings_va:set_vtx_face_uvs(V, Fs, UV, We0),
+    update_uvs_1(Vs, VFs, Vmap, We);
+update_uvs_1([{V0,none}|Vs], [{V0,Fs}|VFs], Vmap, We0) ->
     V = auv_segment:map_vertex(V0, Vmap),
-    Etab = foldl(fun(Face, A) ->
-			 update_uvs_3(V, Face, none, A, We)
-		 end, Etab0, Fs),
-    update_uvs_2(Vs, VFs, Vmap, We, Etab);
-update_uvs_2([], [], _, _, Etab) -> Etab.
+    We = wings_va:set_vtx_face_uvs(V, Fs, none, We0),
+    update_uvs_1(Vs, VFs, Vmap, We);
+update_uvs_1([], [], _, We) -> We.
 
-update_uvs_3(V, Face, UV, Etab, We) ->
-    wings_vertex:fold(
-      fun(Edge, _, Rec0, Et) ->
-	      case Rec0 of
-		  #edge{vs=V,lf=Face,a=A} when A =/= UV ->
-		      Rec = gb_trees:get(Edge, Et),
-		      gb_trees:update(Edge, Rec#edge{a=UV}, Et);
-		  #edge{ve=V,rf=Face,b=B} when B =/= UV ->
-		      Rec = gb_trees:get(Edge, Et),
-		      gb_trees:update(Edge, Rec#edge{b=UV}, Et);
-		  _ -> Et
-	      end
-      end, Etab, V, We).
+%% preserve_old_materials(We0) -> We
+%%  If the object contains materials with colors and no
+%%  vertex colors, convert the materials to vertex colors.
+preserve_old_materials(We, St) ->
+    case not wings_va:any_colors(We) andalso
+	wings_facemat:any_interesting_materials(We) of
+	true ->
+	    wings_we:uv_to_color(We, St);
+	false ->
+	    We
+    end.
 
 insert_material(Cs, MatName, We) ->
     Faces = lists:append([wings_we:visible(W) || W <- gb_trees:values(Cs)]),
@@ -756,17 +745,7 @@ handle_event_3({action,Ev}, St) ->
             wings_view:command(aim, St1),
             get_event(St);
           false ->
-            HL0 = wings_pref:get_value(highlight_aim_at_unselected),
-            HL1 = wings_pref:get_value(highlight_aim_at_selected),
-            {_,X,Y} = wings_wm:local_mouse_state(),
-            {{_,Cmd},St1} = case wings_pick:do_pick(X, Y, St) of
-                  {add,_,St2} when HL0 =:= true ->
-                      {{view,highlight_aim},{add,St,St2}};
-                  {delete,_,St2} when HL1 =:= true ->
-                      {{view,highlight_aim},{delete,St,St2}};
-                  _Other -> 
-                      {{view,aim}, St}
-            end,
+            {{_,Cmd},St1} = wings:highlight_aim_setup(St),
             wings_view:command(Cmd,St1),
             get_event(St)
         end;
@@ -948,7 +927,7 @@ handle_command_1(stitch, St0 = #st{selmode=edge}) ->
 			      [auv_segment:map_edge(E,Emap)
 			       || E<-gb_sets:to_list(Es),
 				  begin 
-				      #edge{lf=LF,rf=RF}=gb_trees:get(E,Etab),
+				      #edge{lf=LF,rf=RF}=array:get(E, Etab),
 				      border(gb_sets:is_member(LF,Vis),
 					       gb_sets:is_member(RF,Vis))
 				  end] ++ A
@@ -964,7 +943,7 @@ handle_command_1(cut_edges, St0 = #st{selmode=edge,bb=#uvstate{id=Id,st=Geom}})
 			      A ++ [auv_segment:map_edge(E,Emap)
 				    || E <-gb_sets:to_list(Es),
 				       begin 
-					   #edge{lf=LF,rf=RF} = gb_trees:get(E,Etab),
+					   #edge{lf=LF,rf=RF} = array:get(E,Etab),
 					   gb_sets:is_member(LF,Vis) and 
 					       gb_sets:is_member(RF,Vis)
 				       end]
@@ -1068,7 +1047,7 @@ add_faces(NewFs,St0=#st{bb=ASt=#uvstate{id=Id,mode=Mode,st=GeomSt=#st{shapes=Shs
 	{_,object} -> St0;
 	{object,_} -> %% Force a chart rebuild, we are switching object mode
 	    We = gb_trees:get(Id,Shs0),
-	    Shs = gb_trees:update(Id, We#we{fs=undefined,es=gb_trees:empty()}, Shs0),
+	    Shs = gb_trees:update(Id, We#we{fs=undefined,es=array:new()}, Shs0),
 	    Fake = GeomSt#st{sel=[],shapes=Shs},
 	    St0#st{bb=ASt#uvstate{mode=object,st=Fake}};
 	{NewFs,Fs0} ->
@@ -1078,7 +1057,7 @@ add_faces(NewFs,St0=#st{bb=ASt=#uvstate{id=Id,mode=Mode,st=GeomSt=#st{shapes=Shs
 		    St0#st{bb=ASt#uvstate{mode=Fs}};
 		_ ->  %% Some new faces should be shown, force a chart rebuild
 		    We = gb_trees:get(Id,Shs0),
-		    Shs = gb_trees:update(Id, We#we{fs=undefined,es=gb_trees:empty()}, Shs0),
+		    Shs = gb_trees:update(Id, We#we{fs=undefined,es=array:new()}, Shs0),
 		    Fake = GeomSt#st{sel=[],shapes=Shs},
 		    St0#st{bb=ASt#uvstate{st=Fake,mode=Fs}}
 	    end
@@ -1107,7 +1086,7 @@ vertex_tighten(Vs0, We, A) ->
 
 body_tighten(#we{vp=Vtab}=We, A) ->
     Vis = gb_sets:from_ordset(wings_we:visible(We)),
-    Vs = [V || V <- gb_trees:keys(Vtab), not_bordering(V, Vis, We)],
+    Vs = [V || V <- wings_util:array_keys(Vtab), not_bordering(V, Vis, We)],
     wings_vertex_cmd:tighten(Vs, We, A).
 
 tighten(Magnet, St) ->
@@ -1138,17 +1117,17 @@ make_equal(Op,[Link0|R],We = #we{vp=Vtab}) ->
 		       [{_,_,A},{_,A,_}|_] -> reverse(Link0)
 		   end,
 	    D = foldl(fun({_E,Ve,Vs}, {X,Y,_}) ->
-			      {Xd,Yd,_} = e3d_vec:sub(gb_trees:get(Ve,Vtab),
-						      gb_trees:get(Vs,Vtab)),
+			      {Xd,Yd,_} = e3d_vec:sub(array:get(Ve,Vtab),
+						      array:get(Vs,Vtab)),
 			      {X+(Xd),Y+(Yd),0.0}
 		      end,
 		      {0.0,0.0,0.0},Link),
 	    Dist = element(E,D)/No,
 	    Vt = foldl(fun({_E,Ve,Vs}, Vt) ->
-			       Pos1 = gb_trees:get(Vs,Vt),
-			       Pos2 = gb_trees:get(Ve,Vt),
+			       Pos1 = array:get(Vs,Vt),
+			       Pos2 = array:get(Ve,Vt),
 			       Pos = setelement(E,Pos2,element(E,Pos1) + Dist),
-			       gb_trees:update(Ve,Pos,Vt)
+			       array:set(Ve,Pos,Vt)
 		       end,
 		       Vtab,Link),	    
 	    make_equal(Op,R,We#we{vp=Vt})
@@ -1187,9 +1166,9 @@ hide_charts(#st{shapes=Shs0,bb=UVs}=St) ->
 
 delete_charts(#st{shapes=Shs0}=St0) ->
     St1 = wpa:sel_map(fun(_, #we{vp=Vp0}=We) ->
-			      Vp1 = gb_trees:to_list(Vp0),
+			      Vp1 = array:sparse_to_orddict(Vp0),
 			      Vp = [{V,none} || {V,_} <- Vp1],
-			      We#we{vp=gb_trees:from_orddict(Vp)}
+			      We#we{vp=array:from_orddict(Vp)}
 		      end, St0),
     St = update_selected_uvcoords(St1),
     Shs = wpa:sel_fold(fun(_, #we{id=Id}, Shs) ->
@@ -1227,8 +1206,8 @@ stitch_charts([ChartStitches|Other],Moved,St0=#st{shapes=Sh0}) ->
     {Id1,Id2,{Vs1,Ve1,Vs2,Ve2}} = find_longest_dist(ChartStitches, St0),
     We1_0 = #we{vp=Vpos1}=gb_trees:get(Id1,Sh0),
     We2_0 = #we{vp=Vpos2}=gb_trees:get(Id2,Sh0),
-    Vs1P = gb_trees:get(Vs1,Vpos1),Ve1P = gb_trees:get(Ve1,Vpos1),
-    Vs2P = gb_trees:get(Vs2,Vpos2),Ve2P = gb_trees:get(Ve2,Vpos2),
+    Vs1P = array:get(Vs1,Vpos1),Ve1P = array:get(Ve1,Vpos1),
+    Vs2P = array:get(Vs2,Vpos2),Ve2P = array:get(Ve2,Vpos2),
     C1 = e3d_vec:average(Vs1P,Ve1P),
     C2 = e3d_vec:average(Vs2P,Ve2P),
     Sh = case {gb_sets:is_member(Id2,Moved),gb_sets:is_member(Id2,Moved)} of
@@ -1281,16 +1260,16 @@ find_longest_dist([{_,{Id1,_,{Vs1,Ve1}},{Id2,_,{Vs2,Ve2}}}|Rest],#st{shapes=Sh})
     %% Need to find a vector to use as basis to rotate the other chart to
     %% we grab the longest distance between to verts of chart1
     #we{vp=Vpos} = gb_trees:get(Id1,Sh),
-    Dist = e3d_vec:dist(gb_trees:get(Vs1,Vpos),gb_trees:get(Ve1,Vpos)),
+    Dist = e3d_vec:dist(array:get(Vs1,Vpos),array:get(Ve1,Vpos)),
     {Id1,Id2,find_longest_dist(Rest,Dist,Vs1,Ve1,Vs2,Ve2,Vpos)}.
 
 find_longest_dist([],_Dist,Bs1,Be1,Bs2,Be2,_Vpos) -> {Bs1,Be1,Bs2,Be2};
 find_longest_dist([This|Rest],Dist,Bs1,Be1,Bs2,Be2,Vpos) ->
     {_,{_,_,{Vs1,Ve1}},{_,_,{Vs2,Ve2}}} = This,
-    Dist1 = e3d_vec:dist(gb_trees:get(Vs1,Vpos),gb_trees:get(Be1,Vpos)),
-    Dist2 = e3d_vec:dist(gb_trees:get(Vs1,Vpos),gb_trees:get(Bs1,Vpos)),
-    Dist3 = e3d_vec:dist(gb_trees:get(Ve1,Vpos),gb_trees:get(Be1,Vpos)),
-    Dist4 = e3d_vec:dist(gb_trees:get(Ve1,Vpos),gb_trees:get(Bs1,Vpos)),
+    Dist1 = e3d_vec:dist(array:get(Vs1,Vpos),array:get(Be1,Vpos)),
+    Dist2 = e3d_vec:dist(array:get(Vs1,Vpos),array:get(Bs1,Vpos)),
+    Dist3 = e3d_vec:dist(array:get(Ve1,Vpos),array:get(Be1,Vpos)),
+    Dist4 = e3d_vec:dist(array:get(Ve1,Vpos),array:get(Bs1,Vpos)),
     if (Dist1>Dist),(Dist1>Dist2),(Dist1>Dist3),(Dist1>Dist4) ->
 	    find_longest_dist(Rest,Dist1,Vs1,Be1,Vs2,Be2,Vpos);
        (Dist2>Dist),(Dist2>Dist3),(Dist2>Dist4) ->
@@ -1316,16 +1295,16 @@ cluster_chart_moves2(_Id1,_Id2,R,Same,Other) ->
     {Same,Other++R}.
 
 average_pos([{V1,V2}|R], Vpos0) ->
-    Pos = e3d_vec:average(gb_trees:get(V1,Vpos0),gb_trees:get(V2,Vpos0)),
-    Vpos1 = gb_trees:update(V1,Pos,Vpos0),
-    Vpos  = gb_trees:update(V2,Pos,Vpos1),
+    Pos = e3d_vec:average(array:get(V1,Vpos0),array:get(V2,Vpos0)),
+    Vpos1 = array:set(V1,Pos,Vpos0),
+    Vpos  = array:set(V2,Pos,Vpos1),
     average_pos(R, Vpos);
 average_pos([],Vpos) -> Vpos.
 
 average_pos([{V1,V2}|R], Vpos1,Vpos2) ->
-    Pos = e3d_vec:average(gb_trees:get(V1,Vpos1),gb_trees:get(V2,Vpos2)),
-    Vp1 = gb_trees:update(V1,Pos,Vpos1),
-    Vp2 = gb_trees:update(V2,Pos,Vpos2),
+    Pos = e3d_vec:average(array:get(V1,Vpos1),array:get(V2,Vpos2)),
+    Vp1 = array:set(V1,Pos,Vpos1),
+    Vp2 = array:set(V2,Pos,Vpos2),
     average_pos(R, Vp1,Vp2);
 average_pos([],Vpos1,Vpos2) -> {Vpos1,Vpos2}.
 
@@ -1353,19 +1332,19 @@ displace_edges([{_,{Id,Edge1,{Vs1,Ve1}},{Id,_,{Vs2,Ve2}}}|Eds], Sh) ->
     Same = [{Vs1,Vs2},{Ve1,Ve2}],
     Vs = [{V1,V2} || {V1,V2} <- Same,
 		     V1 /= V2,
-		     gb_trees:get(V1,Vpos0) == gb_trees:get(V2,Vpos0)],
+		     array:get(V1,Vpos0) == array:get(V2,Vpos0)],
     case Vs of
 	[] ->  %% Already displaced
 	    displace_edges(Eds,Sh);
 	_ ->
 	    %% What Direction should we displace the verts?
-	    [Move1,Move2] = displace_dirs(0.005,Edge1,We),
+	    [Move1,Move2] = displace_dirs(0.0000001,Edge1,We),
 	    %% Make the move
 	    Vpos = foldl(fun({V1,V2},VpIn) ->
-				 Pos1 = e3d_vec:add(Move1,gb_trees:get(V1,Vpos0)),
-				 Vpos1 = gb_trees:update(V1,Pos1,VpIn),
-				 Pos2 = e3d_vec:add(Move2,gb_trees:get(V2,Vpos0)),
-				 gb_trees:update(V2,Pos2,Vpos1)
+				 Pos1 = e3d_vec:add(Move1,array:get(V1,Vpos0)),
+				 Vpos1 = array:set(V1,Pos1,VpIn),
+				 Pos2 = e3d_vec:add(Move2,array:get(V2,Vpos0)),
+				 array:set(V2,Pos2,Vpos1)
 			 end, Vpos0, Vs),	    
 	    displace_edges(Eds,gb_trees:update(Id, We#we{vp=Vpos},Sh))
     end;
@@ -1383,21 +1362,22 @@ displace_charts([{_,{Id1,_,_},{Id2,_,_}}|Eds], Moved, Sh) ->
 	    We0 = #we{vp=Vpos0} = gb_trees:get(Id1,Sh),
 	    C1 = wings_vertex:center(We0),
 	    C2 = wings_vertex:center(gb_trees:get(Id2,Sh)),
-	    Move = case e3d_vec:mul(e3d_vec:norm(e3d_vec:sub(C1,C2)),0.01) of
-		       {0.0,0.0,0.0} -> {0.0,0.01,0.0};
+	    Disp0 = e3d_vec:mul(e3d_vec:norm(e3d_vec:sub(C1,C2)),0.0000001),
+	    Move = case Disp0 of
+		       {0.0,0.0,0.0} -> {0.0,0.0000001,0.0};
 		       Disp -> Disp
 		   end,
 	    Vpos= [{V,e3d_vec:add(Pos,Move)} || 
-		      {V,Pos} <- gb_trees:to_list(Vpos0)],
-	    We = We0#we{vp=gb_trees:from_orddict(Vpos)},
+		      {V,Pos} <- array:sparse_to_orddict(Vpos0)],
+	    We = We0#we{vp=array:from_orddict(Vpos)},
 	    displace_charts(Eds,gb_sets:add(Id1,Moved),
 			    gb_trees:update(Id1,We,Sh))
     end.
 
 displace_dirs(Dist,Edge1,We = #we{es=Etab,vp=Vpos}) ->
-    #edge{vs=Vs1,ve=Ve1,lf=LF,rf=RF} = gb_trees:get(Edge1,Etab),
-    Vp1 = gb_trees:get(Vs1,Vpos),
-    Vp2 = gb_trees:get(Ve1,Vpos),
+    #edge{vs=Vs1,ve=Ve1,lf=LF,rf=RF} = array:get(Edge1,Etab),
+    Vp1 = array:get(Vs1,Vpos),
+    Vp2 = array:get(Ve1,Vpos),
     {Dx,Dy,_} = e3d_vec:norm(e3d_vec:sub(Vp1,Vp2)),
     Dir = {Dy,-Dx,0.0},
     EdgeFace1 = if LF < 0 -> RF; true -> LF end,
@@ -1425,8 +1405,8 @@ map_edges(WingsEs,#st{shapes=Sh}) ->
       fun(_L={E,[{Id1,E1},{Id2,E2}]},Acc) ->
 	      #we{name=#ch{vmap=Vmap1},es=Etab1}=gb_trees:get(Id1,Sh),
 	      #we{name=#ch{vmap=Vmap2},es=Etab2}=gb_trees:get(Id2,Sh),
-	      #edge{vs=Vs1,ve=Ve1} = gb_trees:get(E1,Etab1),
-	      #edge{vs=Vs2,ve=Ve2} = gb_trees:get(E2,Etab2),
+	      #edge{vs=Vs1,ve=Ve1} = array:get(E1,Etab1),
+	      #edge{vs=Vs2,ve=Ve2} = array:get(E2,Etab2),
 	      Vs1map = auv_segment:map_vertex(Vs1, Vmap1),
 	      R= case auv_segment:map_vertex(Vs2, Vmap2) of
 		     Vs1map -> {E,{Id1,E1,{Vs1,Ve1}},{Id2,E2,{Vs2,Ve2}}};
@@ -1484,7 +1464,11 @@ new_geom_state_1(Shs, #st{bb=#uvstate{id=Id,st=#st{shapes=Orig}}}=AuvSt) ->
     case {gb_trees:lookup(Id, Shs),gb_trees:lookup(Id, Orig)} of
 	{none,_} -> delete;
 	{{value,We},{value,We}} -> {AuvSt,false};
-	{{value,#we{es=Etab}},{value,#we{es=Etab}}} -> {AuvSt,false};
+	{{value,#we{es=Etab}=We},{value,#we{es=Etab}=OldWe}} ->
+	    case wings_va:any_update(We, OldWe) of
+		false -> {AuvSt,false};
+		true -> {rebuild_charts(We, AuvSt, []),true}
+	    end;
 	{{value,We},_} -> {rebuild_charts(We, AuvSt, []),true}
     end.
 
@@ -1527,7 +1511,7 @@ update_uv_tab_1([#we{id=Id,name=#ch{vmap=Vmap}}=We0|Cs], FvUvMap, Acc) ->
 	    %% all) in the chart. Throw away this chart.
 	    update_uv_tab_1(Cs, FvUvMap, Acc);
 	UVs1 ->
-	    UVs = gb_trees:from_orddict(UVs1),
+	    UVs = array:from_orddict(UVs1),
 	    We = We0#we{vp=UVs},
 	    update_uv_tab_1(Cs, FvUvMap, [{Id,We}|Acc])
     end;
@@ -1601,7 +1585,7 @@ face_sel_to_face([{K,We}|Cs], Faces, Sel) ->
 face_sel_to_face([], _, Sel) -> Sel.
 
 vertex_sel_to_vertex([{K,#we{vp=Vtab}=We}|Cs], Vs, Sel) ->
-    ChartVs = auv2geom_vs(gb_trees:keys(Vtab), We),
+    ChartVs = auv2geom_vs(wings_util:array_keys(Vtab), We),
     case ordsets:intersection(ChartVs, Vs) of
  	[] ->
 	    vertex_sel_to_vertex(Cs, Vs, Sel);
@@ -1614,7 +1598,7 @@ vertex_sel_to_vertex([], _, Sel) -> Sel.
 
 edge_sel_to_edge([{K,#we{es=Etab}=We}|Cs], Es, Sel) ->
     Vis = gb_sets:from_ordset(wings_we:visible(We)),
-    ChartEs0 = [E || {E,#edge{lf=Lf,rf=Rf}} <- gb_trees:to_list(Etab),
+    ChartEs0 = [E || {E,#edge{lf=Lf,rf=Rf}} <- array:sparse_to_orddict(Etab),
 		     gb_sets:is_member(Lf, Vis) orelse
 			 gb_sets:is_member(Rf, Vis)],
     ChartEs = auv2geom_edges(ChartEs0, We),
@@ -1711,13 +1695,13 @@ align_chart(Dir, St = #st{selmode=Mode}) ->
       fun(Sel, We = #we{vp=Vtab,es=Etab}) ->
 	      case gb_sets:to_list(Sel) of
 		  [V1,V2] when Mode == vertex -> 
-		      align_chart(Dir,gb_trees:get(V1,Vtab),
-				  gb_trees:get(V2,Vtab),
+		      align_chart(Dir,array:get(V1,Vtab),
+				  array:get(V2,Vtab),
 				  We);
 		  [E] when Mode == edge -> 
-		      #edge{vs=V1,ve=V2} = gb_trees:get(E, Etab),
-		      align_chart(Dir,gb_trees:get(V1,Vtab),
-				  gb_trees:get(V2,Vtab),
+		      #edge{vs=V1,ve=V2} = array:get(E, Etab),
+		      align_chart(Dir,array:get(V1,Vtab),
+				  array:get(V2,Vtab),
 				  We);
 		  _ -> align_error()
 	      end
@@ -1816,7 +1800,7 @@ reunfold(Method,#st{sel=Sel,selmode=vertex}=St0) ->
 		Msg = ?__(4,"chart")++" " ++ integer_to_list(I+1),
 		wings_pb:update(I/N, Msg),
 		Pinned = [begin
-			      {S,T,_} = gb_trees:get(V, Vtab),
+			      {S,T,_} = array:get(V, Vtab),
 			      {V,{S,T}}
 			  end || V <- gb_sets:to_list(Vs)],
 		{remap(Method, Pinned, Vs, We, St0),I+1}
@@ -1841,18 +1825,24 @@ remap(stretch_opt, _, _, We, St) ->
     ?SLOW(auv_mapping:stretch_opt(We, Vs3d));
 remap(proj_lsqcm, _, Sel, We0, St = #st{selmode=face}) ->
     Vs3d = orig_pos(We0, St),
-    try Vs0 = auv_mapping:projectFromChartNormal(gb_sets:to_list(Sel),We0#we{vp=Vs3d}),
-	Vtab = gb_trees:from_orddict(Vs0),
-	%%    We1 = #we{vs=Vtab} = update_and_scale_chart(Vs0,We0),
-	SelVs = wings_vertex:from_faces(Sel,We0),
-	Pinned = [begin
-		      {S,T,_} = gb_trees:get(V, Vtab),
-		      {V,{S,T}}
-		  end || V <- SelVs],
-	remap(lsqcm, Pinned, Sel, We0, St)
+    try
+	Fs0   = gb_sets:to_list(Sel),
+	Vs0   = auv_mapping:projectFromChartNormal(Fs0,We0#we{vp=Vs3d}),
+	case length(Vs0) == wings_util:array_entries(We0#we.vp) of
+	    true -> %% Everything projected nothing to unfold
+		update_and_scale_chart(Vs0,We0);
+	    false -> %% Unfold rest of faces
+		Vtab  = array:from_orddict(Vs0),
+		SelVs = wings_vertex:from_faces(Sel,We0),
+		Pinned = [begin
+			      {S,T,_} = array:get(V, Vtab),
+			      {V,{S,T}}
+			  end || V <- SelVs],
+		remap(lsqcm, Pinned, Sel, We0, St)
+	end
     catch throw:{_,What} ->
 	    wpa:error(What);
-	      throw:What ->
+	throw:What ->
 	    wpa:error(What)
     end;
 remap(Type, Pinned, _, We0, St) ->
@@ -1872,15 +1862,15 @@ orig_pos(We = #we{name=#ch{vmap=Vmap}},St) ->
     Vs3d = map(fun({V0,_Pos}) ->
 		       case gb_trees:lookup(V0, Vmap) of
 			   none -> 
-			       {V0, gb_trees:get(V0, Vs3d0)};
+			       {V0, array:get(V0, Vs3d0)};
 			   {value,V} ->
-			       {V0, gb_trees:get(V, Vs3d0)}
+			       {V0, array:get(V, Vs3d0)}
 		       end 
-	       end, gb_trees:to_list(We#we.vp)),
-    gb_trees:from_orddict(Vs3d).
+	       end, array:sparse_to_orddict(We#we.vp)),
+    array:from_orddict(Vs3d).
 
 update_and_scale_chart(Vs0,We0) ->
-    We1 = We0#we{vp=gb_trees:from_orddict(sort(Vs0))},
+    We1 = We0#we{vp=array:from_orddict(sort(Vs0))},
     Fs = wings_we:visible(We1),
     OldA = auv_mapping:fs_area(Fs,We0),
     NewA = auv_mapping:fs_area(Fs,We1),
@@ -2014,7 +2004,7 @@ auv2geom_vs(Vs, #we{name=#ch{vmap=Vmap}}) ->
     sort([auv_segment:map_vertex(V, Vmap) || V <- Vs]).
 
 geom2auv_vs(Vs, #we{name=#ch{vmap=Vmap},vp=Vtab}) ->
-    geom2auv_vs_1(gb_trees:keys(Vtab), gb_sets:from_list(Vs), Vmap, []).
+    geom2auv_vs_1(wings_util:array_keys(Vtab), gb_sets:from_list(Vs), Vmap, []).
 
 geom2auv_vs_1([V|Vs], VsSet, Vmap, Acc) ->
     case gb_sets:is_member(auv_segment:map_vertex(V, Vmap), VsSet) of
