@@ -3,7 +3,7 @@
 %%
 %%     Connect/Cut mode plugin.
 %%
-%%  Copyright (c) 2004-2008 Dan Gudmundsson
+%%  Copyright (c) 2004-2009 Dan Gudmundsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -98,10 +98,10 @@ handle_connect_event0(Ev, S) -> handle_connect_event1(Ev, S).
 handle_connect_event1(#mousebutton{button=1,x=X,y=Y,state=?SDL_PRESSED},
 		      #cs{st=St0}=C0) ->
     case wpa:pick(X,Y,St0) of
-	{add,MM,St = #st{selmode=edge,sel=[{Shape,Edge0}],shapes=Sh}} ->
+	{add,{_,_,MM},#st{selmode=edge,sel=[{Shape,Edge0}],shapes=Sh}=St} ->
 	    #we{es=Es} = gb_trees:get(Shape, Sh),
 	    [Edge] = gb_sets:to_list(Edge0),
-	    #edge{vs=V1,ve=V2} = gb_trees:get(Edge, Es),
+	    #edge{vs=V1,ve=V2} = array:get(Edge, Es),
 	    case cut_edge(X,Y,MM,St,C0) of
 		C0 -> 
 		    keep;
@@ -117,7 +117,7 @@ handle_connect_event1(#mousebutton{button=1,x=X,y=Y,state=?SDL_PRESSED},
 handle_connect_event1(#mousebutton{button=1,x=X,y=Y,state=?SDL_RELEASED},
 		      #cs{st=St0}=C0) ->
     case wpa:pick(X,Y,St0) of
-	{add,MM,St} ->
+	{add,{_,_,MM},St} ->
 	    C = do_connect(X,Y,MM,St,C0),
 	    update_hook(C),
 	    wings_draw:refresh_dlists(C#cs.st),
@@ -153,7 +153,7 @@ handle_connect_event1({slide_setup,Drag},_C) ->
 handle_connect_event1({new_state,St0=#st{shapes=Sh}},
 		      C0=#cs{mode=slide,we=Shape,v=[V|VR],backup=Old}) ->
     #we{vp=Vtab} = gb_trees:get(Shape, Sh),
-    Pos = gb_trees:get(V#vi.id, Vtab),
+    Pos = array:get(V#vi.id, Vtab),
     St1 = St0#st{sel=[],temp_sel=none, sh=true},    
     C1 = C0#cs{mode=normal,v=[V#vi{pos=Pos}|VR],st=St1},
     case connect_edge(C1) of
@@ -271,29 +271,13 @@ redraw(#cs{st=St}) ->
     wings:redraw("", St),
     keep.
 
-%%filter_hl(Hit,Cstate) -> true|false
 filter_hl(_, _) -> true.  %% Debug connection
-%% filter_hl(_, #cs{v=[]}) -> true;
-%% filter_hl({_,_,{Sh,_}},#cs{we=Prev}) when Sh /= Prev ->
-%%     false;
-%% filter_hl({edge,_,{Shape,Edge}}, #cs{last=Id,st=#st{shapes=Sh}}) ->
-%%     We = #we{es=Es} = gb_trees:get(Shape,Sh),
-%%     Ok = vertex_fs(Id, We),
-%%     #edge{lf=F1,rf=F2} = gb_trees:get(Edge, Es),
-%%     Fs = ordsets:from_list([F1,F2]),
-%%     length(ordsets:intersection(Fs,Ok)) == 1;
-%% filter_hl({vertex,_,{_,Id1}}, #cs{last=Id1}) -> true; %% Allow quitting
-%% filter_hl({vertex,_,{Shape,Id1}}, #cs{last=Id2,st=#st{shapes=Sh}}) ->
-%%     We = gb_trees:get(Shape,Sh),
-%%     Ok = vertex_fs(Id2, We),
-%%     Fs = vertex_fs(Id1, We),
-%%     length(ordsets:intersection(Fs,Ok)) == 1.
 
 do_connect(_X,_Y,MM,St0=#st{selmode=vertex,sel=[{Shape,Sel0}],shapes=Sh},
 	   C0=#cs{v=VL,loop_cut=LC,we=Prev}) ->
     [Id1] = gb_sets:to_list(Sel0),
     We0 = gb_trees:get(Shape, Sh),
-    Pos = gb_trees:get(Id1, We0#we.vp),
+    Pos = array:get(Id1, We0#we.vp),
     St1 = St0#st{sel=[],temp_sel=none, sh=true},
     Fs = vertex_fs(Id1, We0),
     VI = #vi{id=Id1,mm=MM,pos=Pos},
@@ -472,16 +456,16 @@ check_normal(Face,{Way,MM},We = #we{id=Id}) ->
     end.
 
 pos2Dto3D({IX,IY}, {V1Sp,V2Sp}, V1,V2,#we{vp=Vs}) ->
-    Pos1 = gb_trees:get(V1, Vs),
-    Pos2 = gb_trees:get(V2, Vs),
+    Pos1 = array:get(V1, Vs),
+    Pos2 = array:get(V2, Vs),
     TotDist = e3d_vec:dist(V1Sp,V2Sp),
     Dist = e3d_vec:dist(V1Sp,{float(IX),float(IY),0.0}) / TotDist,
     Vec = e3d_vec:mul(e3d_vec:sub(Pos2,Pos1),Dist),
     e3d_vec:add(Pos1, Vec).
 
 get_line(V1,V2,MM,#we{id=Id,vp=Vs}) ->
-    P1 = gb_trees:get(V1, Vs),
-    P2 = gb_trees:get(V2, Vs),
+    P1 = array:get(V1, Vs),
+    P2 = array:get(V2, Vs),
     Matrices = wings_u:get_matrices(Id, MM),
     V1Sp = setelement(3,obj_to_screen(Matrices, P1),0.0),
     V2Sp = setelement(3,obj_to_screen(Matrices, P2),0.0),
@@ -490,9 +474,9 @@ get_line(V1,V2,MM,#we{id=Id,vp=Vs}) ->
 calc_edgepos(X,Y0,Edge,MM,#we{id=Id,es=Es,vp=Vs},VL) ->
     {_,H} = wings_wm:win_size(),
     Y = H-Y0,
-    #edge{vs=V1,ve=V2,lf=F1,rf=F2} = gb_trees:get(Edge, Es),
-    Pos1 = gb_trees:get(V1, Vs),
-    Pos2 = gb_trees:get(V2, Vs),
+    #edge{vs=V1,ve=V2,lf=F1,rf=F2} = array:get(Edge, Es),
+    Pos1 = array:get(V1, Vs),
+    Pos2 = array:get(V2, Vs),
     Matrices = wings_u:get_matrices(Id, MM),
     V1Sp = setelement(3,obj_to_screen(Matrices, Pos1),0.0),
     V2Sp = setelement(3,obj_to_screen(Matrices, Pos2),0.0),
@@ -556,7 +540,7 @@ line_intersect2d({X1,Y1},{X2,Y2},{X3,Y3},{X4,Y4}) ->
     end.
 
 obj_to_screen({MVM,PM,VP}, {X,Y,Z}) ->
-    glu:project(X, Y, Z, MVM, PM, VP).
+    wings_gl:project(X, Y, Z, MVM, PM, VP).
 
 help(Cs = #cs{v=[]}) ->
     Msg1 = wings_msg:button_format(?__(1,"Select vertex or cut edge [press button to slide]")),
@@ -587,7 +571,7 @@ fake_sel_1(St0) ->
 	false -> St0;
 	true ->
 	    {_,X,Y} = wings_wm:local_mouse_state(),
-	    case wings_pick:do_pick(X, Y, St0) of
+	    case wpa:pick(X, Y, St0) of
 		{add,_,St} -> St;
 		_ -> St0
 	    end
@@ -633,9 +617,9 @@ gldraw_connect(Pos0, Pos1) ->
 
 slide(C=#cs{st=St=#st{shapes=Sh},we=Shape,v=[#vi{id=Id1,mm=MM}|_]},S,E) ->
     #we{vp=Vtab} = gb_trees:get(Shape, Sh),
-    Start0 = gb_trees:get(S, Vtab),
-    End0   = gb_trees:get(E, Vtab),
-    Curr   = gb_trees:get(Id1, Vtab),
+    Start0 = array:get(S, Vtab),
+    End0   = array:get(E, Vtab),
+    Curr   = array:get(Id1, Vtab),
     Matrices = wings_u:get_matrices(Shape, MM),
     P0 = {P0x,P0y,_} = obj_to_screen(Matrices, Start0),
     P1 = {P1x,P1y,_} = obj_to_screen(Matrices, End0),
