@@ -707,7 +707,7 @@ add_edges(Axis,St) ->
 add_edges(Axis,Es,We) ->
     lists:foldl(fun(Edge,A)->
             #we{es=Etab} = We,
-            #edge{vs=Va,ve=Vb} = gb_trees:get(Edge,Etab),
+            #edge{vs=Va,ve=Vb} = array:get(Edge,Etab),
             Pos1 = wings_vertex:pos(Va,We),
             Pos2 = wings_vertex:pos(Vb,We),
             get_distance(Axis,Pos1,Pos2) + A
@@ -782,7 +782,7 @@ to_axis(Axis,#st{selmode=edge,shapes=Shs,sel=[{Id,Sel}]}=St) ->
     Vec1 = axis_conversion(Axis),
     We = gb_trees:get(Id, Shs),
     [Edge] = gb_sets:to_list(Sel),
-    #edge{vs=V0s,ve=V0e} = gb_trees:get(Edge, We#we.es),
+    #edge{vs=V0s,ve=V0e} = array:get(Edge, We#we.es),
     Pos1 = wings_vertex:pos(V0s, We),
     Pos2 = wings_vertex:pos(V0e, We),
     Vec2 = e3d_vec:sub(Pos1,Pos2),
@@ -806,8 +806,8 @@ measure_angle(Axis,#st{selmode=edge,shapes=Shs,sel=[{Id0,Sel0},{Id1,Sel1}]}) ->
     We1 = gb_trees:get(Id1, Shs),
     [E0] = gb_sets:to_list(Sel0),
     [E1] = gb_sets:to_list(Sel1),
-    #edge{vs=V0s,ve=V0e} = gb_trees:get(E0, We0#we.es),
-    #edge{vs=V1s,ve=V1e} = gb_trees:get(E1, We1#we.es),
+    #edge{vs=V0s,ve=V0e} = array:get(E0, We0#we.es),
+    #edge{vs=V1s,ve=V1e} = array:get(E1, We1#we.es),
     Pos1 = wings_vertex:pos(V0s, We0),
     Pos2 = wings_vertex:pos(V0e, We0),
     Pos3 = wings_vertex:pos(V1s, We1),
@@ -818,8 +818,8 @@ measure_angle(Axis,#st{selmode=edge,shapes=Shs,sel=[{Id0,Sel0},{Id1,Sel1}]}) ->
 measure_angle(Axis,#st{selmode=edge,shapes=Shs,sel=[{Id,Sel}]}) ->
     We = gb_trees:get(Id, Shs),
     [E0,E1] = gb_sets:to_list(Sel),
-    #edge{vs=V0s,ve=V0e} = gb_trees:get(E0, We#we.es),
-    #edge{vs=V1s,ve=V1e} = gb_trees:get(E1, We#we.es),
+    #edge{vs=V0s,ve=V0e} = array:get(E0, We#we.es),
+    #edge{vs=V1s,ve=V1e} = array:get(E1, We#we.es),
     Pos1 = wings_vertex:pos(V0s, We),
     Pos2 = wings_vertex:pos(V0e, We),
     Pos3 = wings_vertex:pos(V1s, We),
@@ -949,7 +949,7 @@ add_areas(Axis,Fs,We) ->
 get_area(Axis,Face,We) ->
     #we{vp=Vtab} = We,
     Vs = wings_face:vertices_ccw(Face, We),
-    Vlist0 = [gb_trees:get(V, Vtab) || V <- Vs],
+    Vlist0 = [array:get(V, Vtab) || V <- Vs],
     Vlist1 = flatten_vpos_to_axis(Axis,Vlist0),
     FaceVs = lists:seq(0, length(Vs)-1),
     E3dFaces = [#e3d_face{vs=FaceVs}],
@@ -1014,10 +1014,9 @@ centers(Axis,Keys1,OrigSt,St) ->
 
 %%% Modifier keys
 mod_key_combo() ->
-    Mod = sdl_keyboard:getModState(),
-    Shift = (Mod band ?KMOD_SHIFT) =/= 0,
-    Ctrl = (Mod band ?KMOD_CTRL) =/= 0,
-    Alt = (Mod band ?KMOD_ALT) =/= 0,
+    Shift = wings_io:is_modkey_pressed(?KMOD_SHIFT),
+    Ctrl  = wings_io:is_modkey_pressed(?KMOD_CTRL),
+    Alt   = wings_io:is_modkey_pressed(?KMOD_ALT),
     {Shift,Ctrl,Alt}.
 
 %%% Utilities
