@@ -69,64 +69,17 @@ command(_,_) ->
     next.
 
 %%%% Setup
-contour_setup(inset_faces,St) ->
-    inset_faces_setup(inset_faces,extrude_faces(St));
+%% Start by Extruding the selection
+contour_setup(inset_faces,St0) ->
+    St = wings_face_cmd:extrude_faces(St0),
+    inset_faces_setup(inset_faces, St);
 contour_setup(inset_region, St0) ->
-    St = wings_sel:map(fun extrude_region_0/2, St0),
+    St = wings_face_cmd:extrude_region(St0),
     inset_regions_setup(inset_region, St);
 contour_setup(offset_region, St0) ->
-    St = wings_sel:map(fun extrude_region_0/2, St0),
+    St = wings_face_cmd:extrude_region(St0),
     offset_regions_setup(offset_region, St).
 
-%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-%%%% Extrude %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-                                                                              %%
-extrude_faces(St) ->                                                          %%
-    wings_sel:map(fun(Faces, We) ->                                           %%
-        wings_extrude_face:faces(Faces, We)                                   %%
-    end, St).                                                                 %%
-                                                                              %%
-%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-%%%% Extrude Region (from wings_face_cmd.erl) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-                                                                              %%
-extrude_region_0(Faces0, We0) ->                                              %%
-    %% We KNOW that a gb_set with fewer elements sorts before                 %%
-    %% a gb_set with more elements.                                           %%
-    Rs = lists:sort(wings_sel:face_regions(Faces0, We0)),                     %%
-    We = extrude_region_1(Rs, We0, []),                                       %%
-    extrude_region_vmirror(We0, We).                                          %%
-                                                                              %%
-extrude_region_1([Faces0|Rs0]=Rs, We0, Acc) ->                                %%
-    case gb_sets:size(Faces0) of                                              %%
-    1 ->                                                                      %%
-        [Face] = gb_sets:to_list(Faces0),                                     %%
-        extrude_region_1(Rs0, We0, [Face|Acc]);                               %%
-    _Other ->                                                                 %%
-        We = wings_extrude_face:faces(Acc, We0),                              %%
-        wings_extrude_face:regions(Rs, We)                                    %%
-    end;                                                                      %%
-extrude_region_1([], We, Faces) ->                                            %%
-    wings_extrude_face:faces(Faces, We).                                      %%
-                                                                              %%
-extrude_region_vmirror(_, #we{mirror=none}=We) -> We;                         %%
-extrude_region_vmirror(OldWe, #we{mirror=Face0}=We0) ->                       %%
-  %% Merge the mirror face and any newly created faces to one new mirror face %%
-  %% and flatten it.                                                          %%
-    FaceSet = gb_sets:singleton(Face0),                                       %%
-    Bordering = wings_face:extend_border(FaceSet, We0),                       %%
-    NewFaces = wings_we:new_items_as_gbset(face, OldWe, We0),                 %%
-    Dissolve0 = gb_sets:intersection(Bordering, NewFaces),                    %%
-    case gb_sets:is_empty(Dissolve0) of                                       %%
-    true -> We0;                                                              %%
-    false ->                                                                  %%
-        Dissolve = gb_sets:insert(Face0, Dissolve0),                          %%
-        We1 = wings_dissolve:faces(Dissolve, We0),                            %%
-        [Face] = NewFace = wings_we:new_items_as_ordset(face, We0, We1),      %%
-        We = wings_facemat:assign('_hole_', NewFace, We1),                    %%
-        wings_we:mirror_flatten(OldWe, We#we{mirror=Face})                    %%
-    end.                                                                      %%
-%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-
 offset_regions_setup(offset_region, St) ->
     State = drag_mode(offset_region),
     SetupSt = wings_sel_conv:more(St),
@@ -318,14 +271,14 @@ faces_data_0(Faces0,#we{es=Etab,vp=Vtab,fs=Ftab}=We,FaceNorms0,EAcc0,Vs0) ->
     end.
 
 faces_data_1(Edge,Face,Etab,Vtab,EAcc,Vs) ->
-    case gb_trees:get(Edge,Etab) of
+    case array:get(Edge,Etab) of
       #edge{vs=Va,ve=Vb,lf=Face,ltpr=NextEdge} ->
-        VposA = gb_trees:get(Va,Vtab),
-        VposB = gb_trees:get(Vb,Vtab),
+        VposA = array:get(Va,Vtab),
+        VposB = array:get(Vb,Vtab),
         faces_data_2(NextEdge,Face,Edge,Etab,Vtab,[VposB,VposA],[Edge|EAcc],[Vb|Vs]);
       #edge{vs=Va,ve=Vb,rf=Face,rtpr=NextEdge} ->
-        VposA = gb_trees:get(Va,Vtab),
-        VposB = gb_trees:get(Vb,Vtab),
+        VposA = array:get(Va,Vtab),
+        VposB = array:get(Vb,Vtab),
         faces_data_2(NextEdge,Face,Edge,Etab,Vtab,[VposA,VposB],[Edge|EAcc],[Va|Vs])
     end.
 
@@ -333,12 +286,12 @@ faces_data_2(LastEdge,Face,LastEdge,_,_,Vp,EAcc,Vs) ->
     {{Face,e3d_vec:normal(Vp)},EAcc,Vs};
 
 faces_data_2(Edge,Face,LastEdge,Etab,Vtab,Vp,EAcc,Vs) ->
-    case gb_trees:get(Edge,Etab) of
+    case array:get(Edge,Etab) of
       #edge{ve=V,lf=Face,ltpr=NextEdge} ->
-        Vpos = gb_trees:get(V,Vtab),
+        Vpos = array:get(V,Vtab),
         faces_data_2(NextEdge,Face,LastEdge,Etab,Vtab,[Vpos|Vp],[Edge|EAcc],[V|Vs]);
       #edge{vs=V,rf=Face,rtpr=NextEdge} ->
-        Vpos = gb_trees:get(V,Vtab),
+        Vpos = array:get(V,Vtab),
         faces_data_2(NextEdge,Face,LastEdge,Etab,Vtab,[Vpos|Vp],[Edge|EAcc],[V|Vs])
     end.
 
@@ -369,18 +322,18 @@ loop_vertices_data_0(EdgeSet0,FNtab,#we{es=Etab,vp=Vtab,mirror=M}=We,LNorms,VDat
     end.
 
 loop_vertices_data_1(Edge,EdgeSet,FNtab,Etab,Vtab,M,VData,Vs) ->
-    #edge{vs=Va,ve=Vb,rf=Rf,rtpr=NextEdge} = gb_trees:get(Edge, Etab),
-    VposA = gb_trees:get(Va,Vtab),
-    VposB = gb_trees:get(Vb,Vtab),
+    #edge{vs=Va,ve=Vb,rf=Rf,rtpr=NextEdge} = array:get(Edge, Etab),
+    VposA = array:get(Va,Vtab),
+    VposB = array:get(Vb,Vtab),
     FNorm = gb_trees:get(Rf,FNtab),
     VDir = e3d_vec:sub(VposB,VposA),
-    EdgeData = gb_trees:get(NextEdge,Etab),
+    EdgeData = array:get(NextEdge,Etab),
     loop_vertices_data_2(NextEdge,EdgeData,Va,VposA,Rf,Edge,FNtab,Etab,Vtab,M,
         EdgeSet,VDir,[],[FNorm],VData,[],Vs,0).
 
 loop_vertices_data_2(LastE,#edge{vs=Va,ve=Vb,lf=Face,rf=PrevFace},
       Vb,VposB,PrevFace,LastE,_,_,Vtab,M,EdgeSet,VDir,EDir0,VNorms,VData0,VPs,Vs0,Links) ->
-    VposA = gb_trees:get(Va,Vtab),
+    VposA = array:get(Va,Vtab),
     Dir = e3d_vec:sub(VposA,VposB),
     VNormal = e3d_vec:norm(e3d_vec:add(VNorms)),
     VData = case M of
@@ -406,7 +359,7 @@ loop_vertices_data_2(LastE,#edge{vs=Va,ve=Vb,lf=Face,rf=PrevFace},
 
 loop_vertices_data_2(CurE,#edge{vs=Va,ve=Vb,lf=Face,rf=PrevFace,ltpr=NextEdge,rtpr=IfNoFaceEdge},
       Vb,VposB,PrevFace,LastE,FNtab,Etab,Vtab,M,EdgeSet0,VDir,EDir0,VNorms0,VData0,VPs0,Vs0,Links) ->
-    VposA = gb_trees:get(Va,Vtab),
+    VposA = array:get(Va,Vtab),
     Dir = e3d_vec:sub(VposA,VposB),
     case  gb_trees:lookup(Face,FNtab) of
       none ->
@@ -419,7 +372,7 @@ loop_vertices_data_2(CurE,#edge{vs=Va,ve=Vb,lf=Face,rf=PrevFace,ltpr=NextEdge,rt
             {VData1,M};
           Face ->
             Vec = e3d_vec:norm(Dir),
-            FC = case gb_trees:get(NextEdge,Etab) of
+            FC = case array:get(NextEdge,Etab) of
               #edge{vs=Vb,rf=F} -> F;
               #edge{ve=Vb,lf=F} -> F
             end,
@@ -445,14 +398,14 @@ loop_vertices_data_2(CurE,#edge{vs=Va,ve=Vb,lf=Face,rf=PrevFace,ltpr=NextEdge,rt
             {VData1,M}
         end,
         NextVDir = e3d_vec:neg(Dir),
-        EdgeData = gb_trees:get(IfNoFaceEdge,Etab),
+        EdgeData = array:get(IfNoFaceEdge,Etab),
         [FNorm|_] = VNorms0,
         VPs = [VposB|VPs0],
         Vs = [Vb|Vs0],
         loop_vertices_data_2(IfNoFaceEdge,EdgeData,Va,VposA,PrevFace,LastE,FNtab,Etab,Vtab,Mir,
             EdgeSet,NextVDir,[],[FNorm],VData,VPs,Vs,Links+1);
       {value,FNorm} ->
-        EdgeData = gb_trees:get(NextEdge,Etab),
+        EdgeData = array:get(NextEdge,Etab),
         EDirs = [Dir|EDir0],
         VNorms = [FNorm|VNorms0],
         loop_vertices_data_2(NextEdge,EdgeData,Vb,VposB,Face,LastE,FNtab,Etab,Vtab,M,
@@ -461,10 +414,10 @@ loop_vertices_data_2(CurE,#edge{vs=Va,ve=Vb,lf=Face,rf=PrevFace,ltpr=NextEdge,rt
 
 loop_vertices_data_2(_CurE,#edge{vs=Va,ve=Vb,lf=PrevFace,rf=Face,rtpr=NextEdge},
       Va,VposA,PrevFace,LastE,FNtab,Etab,Vtab,M,EdgeSet0,VDir,EDir0,VNorms0,VData0,VPs0,Vs0,Links) ->
-    VposB = gb_trees:get(Vb,Vtab),
+    VposB = array:get(Vb,Vtab),
     Dir = e3d_vec:sub(VposB,VposA),
     FNorm = gb_trees:get(Face,FNtab),
-    EdgeData = gb_trees:get(NextEdge,Etab),
+    EdgeData = array:get(NextEdge,Etab),
     EDirs = [Dir|EDir0],
     VNorms = [FNorm|VNorms0],
     loop_vertices_data_2(NextEdge,EdgeData,Va,VposA,Face,LastE,FNtab,Etab,Vtab,M,
@@ -491,20 +444,20 @@ offset_loop_data_0(EdgeSet0,Faces,We,FNtab,LNorms,VData0,Vs0) ->
     end.
 
 offset_loop_data_1(Edge,EdgeSet,Faces,#we{es=Etab,vp=Vtab}=We,FNtab,VData,Vs) ->
-    #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf,ltsu=NextLeft,rtsu=NextRight} = gb_trees:get(Edge,Etab),
-    VposA = gb_trees:get(Va,Vtab),
-    VposB = gb_trees:get(Vb,Vtab),
+    #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf,ltsu=NextLeft,rtsu=NextRight} = array:get(Edge,Etab),
+    VposA = array:get(Va,Vtab),
+    VposB = array:get(Vb,Vtab),
     case gb_sets:is_member(Rf,Faces) of
       true ->
         VDir = e3d_vec:sub(VposB,VposA),
         FNorm = wings_face:normal(Lf,We),
-        EdgeData = gb_trees:get(NextLeft,Etab),
+        EdgeData = array:get(NextLeft,Etab),
         offset_loop_data_2(NextLeft,EdgeData,Va,VposA,Lf,Edge,We,FNtab,
             EdgeSet,VDir,[],[FNorm],VData,[],Vs,0);
       false ->
         VDir = e3d_vec:sub(VposA,VposB),
         FNorm = wings_face:normal(Rf,We),
-        EdgeData = gb_trees:get(NextRight,Etab),
+        EdgeData = array:get(NextRight,Etab),
         offset_loop_data_2(NextRight,EdgeData,Vb,VposB,Rf,Edge,We,FNtab,
             EdgeSet,VDir,[],[FNorm],VData,[],Vs,0)
     end.
@@ -513,7 +466,7 @@ offset_loop_data_2(LastE,#edge{vs=Va,ve=Vb,lf=PrevFace,rtsu=NextEdge},
         Vb,VposB,PrevFace,LastE,We,FNtab,EdgeSet,VDir,EDir0,VNorms,VData0,VPs,Vs0,Links) ->
     #we{vp=Vtab,mirror=M} = We,
     Mirror = M == PrevFace,
-    VposA = gb_trees:get(Va,Vtab),
+    VposA = array:get(Va,Vtab),
     Dir = e3d_vec:sub(VposA,VposB),
     VData = offset_loop_data_3(Mirror,Vb,VposB,VNorms,NextEdge,VDir,Dir,EDir0,FNtab,We,VData0),
     LoopNorm = e3d_vec:normal([VposB|VPs]),
@@ -524,7 +477,7 @@ offset_loop_data_2(LastE,#edge{vs=Va,ve=Vb,rf=PrevFace,ltsu=NextEdge},
         Va,VposA,PrevFace,LastE,We,FNtab,EdgeSet,VDir,EDir0,VNorms,VData0,VPs,Vs0,Links) ->
     #we{vp=Vtab,mirror=M} = We,
     Mirror = M == PrevFace,
-    VposB = gb_trees:get(Vb,Vtab),
+    VposB = array:get(Vb,Vtab),
     Dir = e3d_vec:sub(VposB,VposA),
     VData = offset_loop_data_3(Mirror,Va,VposA,VNorms,NextEdge,VDir,Dir,EDir0,FNtab,We,VData0),
     LoopNorm = e3d_vec:normal([VposA|VPs]),
@@ -540,8 +493,8 @@ offset_loop_data_2(CurE,#edge{vs=Va,ve=Vb,lf=Face,rf=PrevFace,ltsu=NextEdge,rtsu
             EdgeSet0,VDir,EDir0,FNtab,PrevFace,LastE,We,VData0,VPs0,Vs0,Links);
       false ->
         #we{es=Etab,vp=Vtab} = We,
-        EdgeData0 = gb_trees:get(NextEdge,Etab),
-        VposB = gb_trees:get(Vb,Vtab),
+        EdgeData0 = array:get(NextEdge,Etab),
+        VposB = array:get(Vb,Vtab),
         Dir = e3d_vec:sub(VposB,VposA),
         FNorm = wings_face:normal(Face,We),
         EDirs = [Dir|EDir0],
@@ -559,8 +512,8 @@ offset_loop_data_2(CurE,#edge{vs=Va,ve=Vb,lf=PrevFace,rf=Face,rtsu=NextEdge,ltsu
             EdgeSet0,VDir,EDir0,FNtab,PrevFace,LastE,We,VData0,VPs0,Vs0,Links);
       false ->
         #we{es=Etab,vp=Vtab} = We,
-        EdgeData0 = gb_trees:get(NextEdge,Etab),
-        VposA = gb_trees:get(Va,Vtab),
+        EdgeData0 = array:get(NextEdge,Etab),
+        VposA = array:get(Va,Vtab),
         Dir = e3d_vec:sub(VposA,VposB),
         FNorm = wings_face:normal(Face,We),
         EDirs = [Dir|EDir0],
@@ -572,7 +525,7 @@ offset_loop_data_2(CurE,#edge{vs=Va,ve=Vb,lf=PrevFace,rf=Face,rtsu=NextEdge,ltsu
 offset_loop_data_3(false,V,Vpos,VNorms0,NextEdge,VDir,Dir,EDir0,FNtab,We,VData0) ->
     #we{es=Etab} = We,
     VNorm = e3d_vec:norm(e3d_vec:add(VNorms0)),
-    NV = wings_vertex:other(V,gb_trees:get(NextEdge,Etab)),
+    NV = wings_vertex:other(V,array:get(NextEdge,Etab)),
     ANorm = vertex_normal(NV,FNtab,We),
     EDir = average_edge_dir(VNorm,VDir,Dir,EDir0),
     AvgDir = evaluate_vdata(VDir,Dir,VNorm),
@@ -580,7 +533,7 @@ offset_loop_data_3(false,V,Vpos,VNorms0,NextEdge,VDir,Dir,EDir0,FNtab,We,VData0)
     [{V,{Vpos,AvgDir,EDir,ScaledDir}}|VData0];
 
 offset_loop_data_3(true,V,Vpos,VNorms0,NE,VDir,Dir,EDir0,FNtab,#we{es=Etab},VData0) ->
-    Face = case gb_trees:get(NE,Etab) of
+    Face = case array:get(NE,Etab) of
       #edge{vs=V,lf=IsFaceInc} -> IsFaceInc;
       #edge{ve=V,rf=IsFaceInc} -> IsFaceInc
     end,
@@ -600,11 +553,11 @@ offset_loop_data_3(true,V,Vpos,VNorms0,NE,VDir,Dir,EDir0,FNtab,#we{es=Etab},VDat
 offset_loop_is_member(Mirror,V1,V2,Vpos1,CurE,NextE,VNorms0,NEdge,EdgeSet0,VDir,
         EDir0,FNtab,PFace,LastE,We,VData0,VPs0,Vs0,Links) ->
     #we{es=Etab,vp=Vtab} = We,
-    Vpos2 = gb_trees:get(V2,Vtab),
+    Vpos2 = array:get(V2,Vtab),
     Dir = e3d_vec:sub(Vpos2,Vpos1),
     NextVDir = e3d_vec:neg(Dir),
     EdgeSet = gb_sets:delete(CurE,EdgeSet0),
-    EdgeData = gb_trees:get(NextE,Etab),
+    EdgeData = array:get(NextE,Etab),
     [FNorm|_] = VNorms0,
     VData = offset_loop_data_3(Mirror,V1,Vpos1,VNorms0,NEdge,VDir,Dir,EDir0,
             FNtab,We,VData0),
@@ -682,7 +635,7 @@ vertex_normals([V|Vs],FaceNormTab,#we{vp=Vtab,mirror=M}=We,Acc) ->
         (_,Face,_,A) -> [gb_trees:get(Face,FaceNormTab)|A]
     end,[],V,We),
     VNorm = e3d_vec:norm(e3d_vec:add(FaceNorms)),
-    Vpos = gb_trees:get(V,Vtab),
+    Vpos = array:get(V,Vtab),
     vertex_normals(Vs,FaceNormTab,We,[{V,{Vpos,VNorm}}|Acc]);
 vertex_normals([],_,_,Acc) -> Acc.
 
@@ -735,15 +688,15 @@ collect_inset_face_data(Faces0,#we{es=Etab,vp=Vtab,fs=Ftab}=We,AllVs0,VData0,Sma
     end.
 
 traverse_1(Edge,Face,Etab,Vtab,AllVs) ->
-    case gb_trees:get(Edge,Etab) of
+    case array:get(Edge,Etab) of
       #edge{vs=Va,ve=Vb,lf=Face,ltpr=NextEdge} ->
-        VposA = gb_trees:get(Va,Vtab),
-        VposB = gb_trees:get(Vb,Vtab),
+        VposA = array:get(Va,Vtab),
+        VposB = array:get(Vb,Vtab),
         VDir = e3d_vec:norm_sub(VposB,VposA),
         traverse_2(NextEdge,Face,VDir,VposB,Edge,Etab,Vtab,{Va,{VposA,VDir}},[],AllVs,[]);
       #edge{vs=Va,ve=Vb,rf=Face,rtpr=NextEdge} ->
-        VposB = gb_trees:get(Vb,Vtab),
-        VposA = gb_trees:get(Va,Vtab),
+        VposB = array:get(Vb,Vtab),
+        VposA = array:get(Va,Vtab),
         VDir = e3d_vec:norm_sub(VposA,VposB),
         traverse_2(NextEdge,Face,VDir,VposA,Edge,Etab,Vtab,{Vb,{VposB,VDir}},[],AllVs,[])
     end.
@@ -758,9 +711,9 @@ traverse_2(LastEdge,_,PrevVDir,Vpos,LastEdge,_,_,{Vb,{Vpos,VDirA}},VPositions0,A
     {AllVs,{FNorm,FCntr,VData}};
 
 traverse_2(Edge,Face,PrevVDir,Vpos,LastEdge,Etab,Vtab,LastVert,VPositions0,AllVs0,Acc) ->
-    case gb_trees:get(Edge,Etab) of
+    case array:get(Edge,Etab) of
       #edge{vs=Va,ve=Vb,lf=Face,ltpr=NextEdge} ->
-        VposB = gb_trees:get(Vb, Vtab),
+        VposB = array:get(Vb, Vtab),
         VDirB = e3d_vec:norm_sub(Vpos,VposB),
         VDirA = e3d_vec:neg(PrevVDir),
         VData = [{Va,{Vpos,VDirB,VDirA}}|Acc],
@@ -768,7 +721,7 @@ traverse_2(Edge,Face,PrevVDir,Vpos,LastEdge,Etab,Vtab,LastVert,VPositions0,AllVs
         AllVs = [Va|AllVs0],
         traverse_2(NextEdge,Face,VDirB,VposB,LastEdge,Etab,Vtab,LastVert,VPositions,AllVs,VData);
       #edge{vs=Va,ve=Vb,rf=Face,rtpr=NextEdge} ->
-        VposA = gb_trees:get(Va, Vtab),
+        VposA = array:get(Va, Vtab),
         VDirA = e3d_vec:norm_sub(VposA,Vpos),
         VDirB = e3d_vec:neg(PrevVDir),
         VData = [{Vb,{Vpos,VDirA,VDirB}}|Acc],
