@@ -8,7 +8,7 @@
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 %%
-%%     $Id: wpc_region.erl 293 2008-03-21 09:57:09Z bjorng $
+%%     $Id$
 %%
 
 -module(wpc_region).
@@ -17,7 +17,7 @@
 
 -include_lib("wings.hrl").
 
--import(lists, [map/2,foldr/3,foldl/3,reverse/1,append/1]).
+-import(lists, [foldl/3]).
 -define(HUGE, 1.0E307).
 
 init() ->
