@@ -30,8 +30,8 @@ parse([], NewMenu, true) ->
     NewMenu;
 parse([], NewMenu, false) ->
     [sweep_menu_headings()|NewMenu];
-parse([A = {_,{extract_region,_}}|Rest], NewMenu, false) ->
-    parse(Rest, [sweep_menu_headings(),A|NewMenu], true);
+parse([S,A={_,{flatten,_}},S|Rest], NewMenu, false) when S==separator ->
+    parse(Rest, [S,A,S,sweep_menu_headings()|NewMenu], true);
 parse([Elem|Rest], NewMenu, Found) ->
     parse(Rest, [Elem|NewMenu], Found).
 
@@ -133,74 +133,18 @@ selection_ask([plane|Rest],Ask) ->
     Desc = ?__(1,"Choosing an axis perpendicular to the extrusion normal, will constrain all movement to its radial plane. Otherwise, it acts as an off axis component."),
     selection_ask(Rest,[{axis,Desc}|Ask]).
 
-%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-%%%% Extrude %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-sweep_extrude(Type, Axis,St) ->                                               %%
-    sweep_setup(Type, Axis,extrude_faces(St)).                                %%
-                                                                              %%
-extrude_faces(St) ->                                                          %%
-    wings_sel:map(fun(Faces, We) ->                                           %%
-        wings_extrude_face:faces(Faces, We)                                   %%
-    end, St).                                                                 %%
-                                                                              %%
-%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-%%%% Extrude Region (from wings_face_cmd.erl) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-sweep_region(Type, Axis, St0) ->                                              %%
-    St = wings_sel:map(fun extrude_region_0/2, St0),                          %%
-    sweep_setup(Type, Axis, St).                                              %%
-                                                                              %%
-extrude_region_0(Faces0, We0) ->                                              %%
-    %% We KNOW that a gb_set with fewer elements sorts before                 %%
-    %% a gb_set with more elements.                                           %%
-    Rs = lists:sort(wings_sel:face_regions(Faces0, We0)),                     %%
-    We = extrude_region_1(Rs, We0, []),                                       %%
-    extrude_region_vmirror(We0, We).                                          %%
-                                                                              %%
-extrude_region_1([Faces0|Rs0]=Rs, We0, Acc) ->                                %%
-    case gb_sets:size(Faces0) of                                              %%
-    1 ->                                                                      %%
-        [Face] = gb_sets:to_list(Faces0),                                     %%
-        extrude_region_1(Rs0, We0, [Face|Acc]);                               %%
-    _Other ->                                                                 %%
-        We = wings_extrude_face:faces(Acc, We0),                              %%
-        wings_extrude_face:regions(Rs, We)                                    %%
-    end;                                                                      %%
-extrude_region_1([], We, Faces) ->                                            %%
-    wings_extrude_face:faces(Faces, We).                                      %%
-                                                                              %%
-extrude_region_vmirror(_, #we{mirror=none}=We) -> We;                         %%
-extrude_region_vmirror(OldWe, #we{mirror=Face0}=We0) ->                       %%
-  %% Merge the mirror face and any newly created faces to one new mirror face %%
-  %% and flatten it.                                                          %%
-    FaceSet = gb_sets:singleton(Face0),                                       %%
-    Bordering = wings_face:extend_border(FaceSet, We0),                       %%
-    NewFaces = wings_we:new_items_as_gbset(face, OldWe, We0),                 %%
-    Dissolve0 = gb_sets:intersection(Bordering, NewFaces),                    %%
-    case gb_sets:is_empty(Dissolve0) of                                       %%
-    true -> We0;                                                              %%
-    false ->                                                                  %%
-        Dissolve = gb_sets:insert(Face0, Dissolve0),                          %%
-        We1 = wings_dissolve:faces(Dissolve, We0),                            %%
-        [Face] = NewFace = wings_we:new_items_as_ordset(face, We0, We1),      %%
-        We = wings_facemat:assign('_hole_', NewFace, We1),                    %%
-        wings_we:mirror_flatten(OldWe, We#we{mirror=Face})                    %%
-    end.                                                                      %%
-                                                                              %%
-%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-%%%% Extract (from wings_face_cmd.erl) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-sweep_extract(Type,Axis, St0) ->                                              %%
-    St1 = wings_sel:fold(                                                     %%
-        fun(Faces, We0, #st{sel=Sel0,onext=Oid}=S0) ->                        %%
-            We = wings_dissolve:complement(Faces, We0),                       %%
-            S = wings_shape:insert(We, extract, S0),                          %%
-            Sel = [{Oid,Faces}|Sel0],                                         %%
-            S#st{sel=Sel}                                                     %%
-        end, St0#st{sel=[]}, St0),                                            %%
-    Sel = St1#st.sel,                                                         %%
-    St = wings_sel:set(Sel, St1),                                             %%
-    sweep_region(Type, Axis, St).                                             %%
-                                                                              %%
-%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%%%% Start by Extruding Selection
+sweep_extrude(Type, Axis, St0) ->
+    St = wings_face_cmd:extrude_faces(St0),
+    sweep_setup(Type, Axis, St).
+
+sweep_region(Type, Axis, St0) ->
+    St = wings_face_cmd:extrude_region(St0),
+    sweep_setup(Type, Axis, St).
+
+sweep_extract(Type, Axis, St0) ->
+    St = wings_face_cmd:extract_region(St0),
+    sweep_region(Type, Axis, St).
 
 %%%% Setup
 sweep_setup(Type,Axis,St) ->
@@ -213,7 +157,7 @@ sweep_setup(Type,Axis,St) ->
     Tvs = wings_sel:fold(fun(Fs, #we{id=Id}=We, Acc) ->
 
             Regions = wings_sel:face_regions(Fs,We),
-            {AllVs,VsData} = collect_data(Regions, We, Axis, SelC, [], State, [], []),
+            {AllVs,VsData} = collect_data(Type, Regions, We, Axis, SelC, [], State, [], []),
             [{Id, {AllVs, sweep_fun(Type, VsData, State)}} | Acc]
          end, [], St),
     Units = units(Type),
@@ -225,10 +169,20 @@ units(relative) -> [angle,percent,percent,angle].
 
 
 %% LoopNorm is the extrude direction
-collect_data([Fs0|Rs], We, Axis0, SelC0, AllVs0, State, LVAcc0, ExData) ->
-    Fs = wings_face:extend_border(Fs0, We),
+collect_data(Type, [Fs0|Rs], #we{mirror=M}=We, Axis0, SelC0, AllVs0, State, LVAcc0, ExData) ->
+    Fs = gb_sets:delete_any(M, wings_face:extend_border(Fs0, We)),
     {OuterEs, RegVs} =  reg_data_0(Fs, We, [], []),
-    {LoopNorm, LoopVs} = loop_data_0(OuterEs, Fs, We),
+    {LoopNorm, LoopVs0} = loop_data_0(OuterEs, Fs, We),
+
+    LoopVs = case Type of
+      sweep_extrude -> LoopVs0;
+      _otherwise when M =/= none ->
+          MirEs = wings_face:to_edges([M],We),
+          LoopEs = OuterEs -- MirEs,
+          LoopVerts = wings_edge:to_vertices(LoopEs,We),
+          LoopVerts;
+      _otherwise -> LoopVs0
+    end,
     Axis = axis_conversion(Axis0,LoopNorm),
 
     ExVs = ordsets:subtract(RegVs, LoopVs),
@@ -245,9 +199,9 @@ collect_data([Fs0|Rs], We, Axis0, SelC0, AllVs0, State, LVAcc0, ExData) ->
     NW = non_warping_norm(Axis, LoopNorm),
     CN = specify_warp_and_center(Axis, NW, LoopC, SelC, State),
     Data = {{SelC, LoopC, LoopNorm, MaxR, NW, Axis}, CN},
-    collect_data(Rs, We, Axis0, SelC0, AllVs, State, LVAcc, [{Data,AllVpos}|ExData]);
+    collect_data(Type, Rs, We, Axis0, SelC0, AllVs, State, LVAcc, [{Data,AllVpos}|ExData]);
 
-collect_data([], _We, _Axis0, _SelC0, AllVs, _State, _LVs, VsData) ->
+collect_data(_Type, [], _We, _Axis0, _SelC0, AllVs, _State, _LVs, VsData) ->
     {AllVs, VsData}.
 
 
@@ -263,7 +217,7 @@ reg_data_0(Faces0, #we{es=Etab,fs=Ftab}=We, EAcc0, Vs0) ->
         {OuterEdges, ordsets:from_list(Vs0)}
     end.
 reg_data_1(Edge,Face,Etab,EAcc,Vs) ->
-    case gb_trees:get(Edge,Etab) of
+    case array:get(Edge,Etab) of
       #edge{ve=Vb,lf=Face,ltpr=NextEdge} ->
         reg_data_2(NextEdge,Face,Edge,Etab,[Edge|EAcc],[Vb|Vs]);
       #edge{vs=Va,rf=Face,rtpr=NextEdge} ->
@@ -273,7 +227,7 @@ reg_data_1(Edge,Face,Etab,EAcc,Vs) ->
 reg_data_2(LastEdge,_,LastEdge,_,EAcc,Vs) ->
     {EAcc,Vs};
 reg_data_2(Edge,Face,LastEdge,Etab,EAcc,Vs) ->
-    case gb_trees:get(Edge,Etab) of
+    case array:get(Edge,Etab) of
       #edge{ve=V,lf=Face,ltpr=NextEdge} ->
         reg_data_2(NextEdge,Face,LastEdge,Etab,[Edge|EAcc],[V|Vs]);
       #edge{vs=V,rf=Face,rtpr=NextEdge} ->
@@ -291,23 +245,23 @@ loop_data_1(Es0, Fs, Etab, Vtab, M, LNorms, Vs0) ->
       false ->
         {Edge, Es1} = gb_sets:take_smallest(Es0),
         {Es, LoopNorm, Links, Vs} = loop_data_2(Edge, Edge, Es1, Fs, Etab, Vtab, M, Vs0),
-        loop_data_1(Es, Fs, Etab, Vtab, M, [{Links,LoopNorm}|LNorms], Vs);
+        loop_data_1(Es, Fs, Etab, Vtab, M, [{Links,e3d_vec:neg(LoopNorm)}|LNorms], Vs);
       true ->
         AvgLoopNorm = e3d_vec:neg(average_loop_norm(LNorms)),
         {AvgLoopNorm, ordsets:from_list(Vs0)}
     end.
 
 loop_data_2(Edge, Edge, Es, Fs, Etab, Vtab, M, Vs) ->
-    E = gb_trees:get(Edge, Etab),
+    E = array:get(Edge, Etab),
     #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf,ltsu=NextLeft,rtsu=NextRight} = E,
     case gb_sets:is_member(Rf,Fs) of
-      true ->
-        VpA = gb_trees:get(Va,Vtab),
-        EData = gb_trees:get(NextLeft,Etab),
-        loop_data_3(NextLeft, EData, Edge, Es, Fs, Lf, Va, VpA, Etab, Vtab, M, [], Vs, 0);
       false ->
-        VpB = gb_trees:get(Vb,Vtab),
-        EData = gb_trees:get(NextRight,Etab),
+        VpA = array:get(Va,Vtab),
+        EData = array:get(NextLeft,Etab),
+        loop_data_3(NextLeft, EData, Edge, Es, Fs, Lf, Va, VpA, Etab, Vtab, M, [], Vs, 0);
+      true ->
+        VpB = array:get(Vb,Vtab),
+        EData = array:get(NextRight,Etab),
         loop_data_3(NextRight, EData, Edge, Es, Fs, Rf, Vb, VpB, Etab, Vtab, M, [], Vs, 0)
     end.
 
@@ -340,9 +294,9 @@ loop_data_3(CurE,#edge{vs=Va,ve=Vb,lf=PrevF,rf=Face,rtsu=NextEdge,ltsu=IfCurIsMe
         LastE, Es0, Fs, PrevF, Vb, VpB, Etab, Vtab, M, VPs0, Vs0, Links) ->
     case gb_sets:is_member(CurE,Es0) of
       true ->
-        EData = gb_trees:get(IfCurIsMember,Etab),
+        EData = array:get(IfCurIsMember,Etab),
         Es = gb_sets:delete(CurE,Es0),
-        VpA = gb_trees:get(Va,Vtab),
+        VpA = array:get(Va,Vtab),
         case M == PrevF of
           false ->
             VPs = [VpB|VPs0],
@@ -352,7 +306,7 @@ loop_data_3(CurE,#edge{vs=Va,ve=Vb,lf=PrevF,rf=Face,rtsu=NextEdge,ltsu=IfCurIsMe
             loop_data_3(IfCurIsMember,EData,LastE,Es,Fs,Face,Va,VpA,Etab,Vtab,M,VPs0,Vs0,Links)
         end;
       false ->
-        EData = gb_trees:get(NextEdge,Etab),
+        EData = array:get(NextEdge,Etab),
         loop_data_3(NextEdge, EData, LastE, Es0, Fs, Face, Vb, VpB, Etab, Vtab, M, VPs0, Vs0, Links)
     end;
 
@@ -360,9 +314,9 @@ loop_data_3(CurE,#edge{vs=Va,ve=Vb,lf=Face,rf=PrevF,ltsu=NextEdge,rtsu=IfCurIsMe
         LastE, Es0, Fs, PrevF, Va, VpA, Etab, Vtab, M, VPs0, Vs0, Links) ->
     case gb_sets:is_member(CurE,Es0) of
       true ->
-        EData = gb_trees:get(IfCurIsMember,Etab),
+        EData = array:get(IfCurIsMember,Etab),
         Es = gb_sets:delete(CurE,Es0),
-        VpB = gb_trees:get(Vb,Vtab),
+        VpB = array:get(Vb,Vtab),
         case M == PrevF of
           false ->
             VPs = [VpA|VPs0],
@@ -372,7 +326,7 @@ loop_data_3(CurE,#edge{vs=Va,ve=Vb,lf=Face,rf=PrevF,ltsu=NextEdge,rtsu=IfCurIsMe
             loop_data_3(IfCurIsMember,EData,LastE,Es,Fs,Face,Vb,VpB,Etab,Vtab,M,VPs0,Vs0,Links)
         end;
       false ->
-        EData = gb_trees:get(NextEdge,Etab),
+        EData = array:get(NextEdge,Etab),
         loop_data_3(NextEdge, EData, LastE, Es0, Fs, Face, Va, VpA, Etab, Vtab, M, VPs0, Vs0, Links)
     end.
 
@@ -380,7 +334,7 @@ loop_data_3(CurE,#edge{vs=Va,ve=Vb,lf=Face,rf=PrevF,ltsu=NextEdge,rtsu=IfCurIsMe
 %%%% Setup Utilities
 add_vpos_data(Type, Vs, #we{vp=Vtab}, Acc) ->
     lists:foldl(fun(V, A) ->
-          [{V, Type, gb_trees:get(V, Vtab)} | A]
+          [{V, Type, array:get(V, Vtab)} | A]
       end, Acc, Vs).
 
 non_warping_norm(Axis0,Norm) ->
@@ -411,7 +365,7 @@ sqr_length({X,Y,Z}) ->
 lowest_point_rel_to_norm(LoopVs, LoopNorm, LoopC, SelC, #we{vp=Vtab}) ->
     {L,LD,SD} = lists:foldl(fun
       (V, none) ->
-          VPos = gb_trees:get(V,Vtab),
+          VPos = array:get(V,Vtab),
           LVec = e3d_vec:sub(VPos, LoopC),
           SVec = e3d_vec:sub(VPos, SelC),
           Len = sqr_length(LVec),
@@ -419,7 +373,7 @@ lowest_point_rel_to_norm(LoopVs, LoopNorm, LoopC, SelC, #we{vp=Vtab}) ->
           SDot = e3d_vec:dot(LoopNorm, SVec),
           {Len,LDot,SDot};
       (V, {Len0,LDot0,SDot0}) ->
-          VPos = gb_trees:get(V,Vtab),
+          VPos = array:get(V,Vtab),
           LVec = e3d_vec:sub(VPos, LoopC),
           SVec = e3d_vec:sub(VPos, SelC),
           Len1 = sqr_length(LVec),
