@@ -64,53 +64,68 @@ command_help(turn_optimized) ->
 %%
 
 turn_edges(ModeFun, Opt, St0) ->
-    SelFun = fun(Es, #we{id=Id}=We0, Acc) ->
-		     {We,Sel} = turn_edges(Es, ModeFun, Opt, We0),
-		     {We,[{Id,gb_sets:from_list(Sel)}|Acc]}
+    SelFun = fun(Es0, #we{id=Id}=We0, Acc) ->
+		     Es1 = gb_sets:to_list(Es0),
+		     case wings_we:fully_visible_edges(Es1, We0) of
+			 [] -> {We0,Acc};
+			 Es ->
+			     {We,Sel} = turn_edges(Es, ModeFun, Opt, We0),
+			     {We,[{Id,gb_sets:from_list(Sel)}|Acc]}
+		     end
 	     end,
     {St,Sel} = wings_sel:mapfold(SelFun, [], St0),
     wings_sel:set(Sel, St).
 
-turn_edges(Edges0, ModeFun, Opt, #we{es=Etab}=We0) ->
-    Edges = gb_sets:to_list(Edges0),
+turn_edges(Edges, ModeFun, Opt, #we{es=Etab}=We0) ->
     validate_edges(Edges, Etab),
     foldl(fun(E0, {W0,Sel}) ->
 		  {E,W} = try_turn(E0, ModeFun, Opt, W0),
 		  {W,[E|Sel]}
 	  end, {We0,[]}, Edges).
 
-try_turn(Edge, ModeFun, Opt, #we{vp=Vtab}=We0) ->
+try_turn(Edge, ModeFun, Opt, #we{vp=Vtab,he=Htab0}=We0) ->
     {Vstart,Vend,V1,V2} = ModeFun(Edge, We0),
     case optimize(Opt, Vstart, Vend, V1, V2, Vtab) of
-	{Vert1,Vert2} ->
-	    We1 = wings_edge:dissolve_edge(Edge, We0),
-	    case gb_trees:is_defined(Edge, We1#we.es) of
-                true ->
-		    {Edge, We0};
-                false ->
-                    Vs0 = gb_sets:from_list([Vert1,Vert2]),
-                    We = wings_vertex_cmd:connect(Vs0, We1),
-                    if 
-                        We1#we.next_id =/= We#we.next_id ->
-			    {We1#we.next_id,We};
+    {Vert1,Vert2} ->
+        We1 = wings_edge:dissolve_edge(Edge, We0),
+        case array:get(Edge, We1#we.es) of
+                    #edge{} ->
+                      {Edge, We0};
+                    undefined ->
+                      Vs0 = gb_sets:from_list([Vert1,Vert2]),
+                      We = wings_vertex_cmd:connect(Vs0, We1),
+                      NewEdge = We1#we.next_id,
+                      if
+                        NewEdge =/= We#we.next_id ->
+                          Htab = check_hard_edges(Edge, NewEdge, Htab0),
+                          {NewEdge,We#we{he=Htab}};
                         true ->
-			    {Edge,We0}
-                    end
-	    end;
-	_ ->
-	    {Edge,We0}
+                          {Edge,We0}
+                      end
+        end;
+    _ ->
+        {Edge,We0}
+    end.
+
+check_hard_edges(Edge, NewEdge, Htab0) ->
+    case gb_sets:is_member(Edge, Htab0) of
+      true ->
+          Htab = gb_sets:delete(Edge, Htab0),
+          gb_sets:add(NewEdge, Htab);
+      false ->
+          Htab0
     end.
 
 cw_mode(Edge, #we{es=Etab}) ->
-    #edge{vs=Vstart,ve=Vend,rtsu=RNext,ltsu=LNext} = gb_trees:get(Edge, Etab),
-    RV = wings_vertex:other(Vend, gb_trees:get(RNext, Etab)),
-    LV = wings_vertex:other(Vstart, gb_trees:get(LNext, Etab)),
+    #edge{vs=Vstart,ve=Vend,rtsu=RNext,ltsu=LNext} = array:get(Edge, Etab),
+    RV = wings_vertex:other(Vend, array:get(RNext, Etab)),
+    LV = wings_vertex:other(Vstart, array:get(LNext, Etab)),
     {Vstart,Vend,RV,LV}.
 
 ccw_mode(Edge, #we{es=Etab}) ->
-    #edge{vs=Vstart,ve=Vend,rtpr=RNext,ltpr=LNext} = gb_trees:get(Edge, Etab),
-    RV = wings_vertex:other(Vstart, gb_trees:get(RNext, Etab)),
-    LV = wings_vertex:other(Vend, gb_trees:get(LNext, Etab)),
+    #edge{vs=Vstart,ve=Vend,rtpr=RNext,ltpr=LNext} = array:get(Edge, Etab),
+    RV = wings_vertex:other(Vstart, array:get(RNext, Etab)),
+    LV = wings_vertex:other(Vend, array:get(LNext, Etab)),
     {Vstart,Vend,RV,LV}.
 
 optimize(false, _, _, Va, Vb, _) -> {Va,Vb};
@@ -132,6 +147,6 @@ validate_edges(Edges, Etab) ->
     end.
     
 ve_collect_edge_faces([E|Es], Etab, Acc) ->
-    #edge{lf=Lf,rf=Rf} = gb_trees:get(E, Etab),
+    #edge{lf=Lf,rf=Rf} = array:get(E, Etab),
     ve_collect_edge_faces(Es, Etab, [Lf,Rf|Acc]);
 ve_collect_edge_faces([], _, Acc) -> Acc.
