@@ -22,7 +22,6 @@
 -define(NEED_OPENGL, 1).
 -define(NEED_ESDL, 1).
 -include_lib("wings.hrl").
--include_lib("e3d.hrl").
 
 -import(lists, [member/2,foldl/3]).
 
@@ -65,7 +64,7 @@ init() ->
     wings_pref:set_default(tweak_double_click_speed54,Val),
     wings_pref:set_default(tweak_ctrl,slide),
     wings_pref:set_default(tweak_mmb_select,false),
-	wings_pref:set_default(tweak_mag_step,0.05),
+    wings_pref:set_default(tweak_mag_step,0.05),
     case wings_pref:get_value(start_in_tweak) of
       true ->
         self() ! {external, launch_tweak},
@@ -112,10 +111,10 @@ tweak(Ask,maya,_St) when is_atom(Ask) ->
                          {Radio2,slide}],tweak_ctrl}],
        [{title,?__(7,"Button Options")},{hook,ClickHook}]}],
        [{title,?__(9,"Options Panel")},{hook,MmbHook}]},
-	   
-	   {hframe,[{slider,{text,MagStep,[{key,tweak_mag_step},{range,{0.001,0.1}}]}}],
-	   [{title,?__(10,"Magnet Radius Increment")}]}
-	   ]}],
+
+       {hframe,[{slider,{text,MagStep,[{key,tweak_mag_step},{range,{0.001,0.1}}]}}],
+       [{title,?__(10,"Magnet Radius Increment")}]}
+       ]}],
 
     PrefQs = [{Lbl,make_query(Ps)} || {Lbl,Ps} <- TweakPrefs],
     wings_ask:dialog(Ask, ?__(4,"Tweak Mode Preferences"),PrefQs,
@@ -129,7 +128,7 @@ tweak(Ask,_Cam,_St) when is_atom(Ask) ->
               (_, _) -> void
           end,
     DblClkSpd = wings_pref:get_value(tweak_double_click_speed54)/100000,
-	MagStep = wings_pref:get_value(tweak_mag_step),
+    MagStep = wings_pref:get_value(tweak_mag_step),
     TweakPrefs = [{vframe,
         [{?__(1,"Lmb single click Selects/Deselects"),tweak_single_click},
          {?__(2,"Lmb double click initiates Paint Select/Deselect"),tweak_double_click},
@@ -139,10 +138,10 @@ tweak(Ask,_Cam,_St) when is_atom(Ask) ->
        {vframe,[{vradio,[{Radio1,select},
                          {Radio2,slide}],tweak_ctrl}],
        [{title,?__(7,"Button Options")},{hook,ClickHook}]},
-	   
-	   {hframe,[{slider,{text,MagStep,[{key,tweak_mag_step},{range,{0.001,0.1}}]}}],
-	   [{title,?__(10,"Magnet Radius Increment")}]}
-	   ]}],
+
+       {hframe,[{slider,{text,MagStep,[{key,tweak_mag_step},{range,{0.001,0.1}}]}}],
+       [{title,?__(10,"Magnet Radius Increment")}]}
+       ]}],
 
     PrefQs = [{Lbl,make_query(Ps)} || {Lbl,Ps} <- TweakPrefs],
     wings_ask:dialog(Ask, ?__(4,"Tweak Mode Preferences"),PrefQs,
@@ -205,16 +204,14 @@ f2() -> ?SDLK_F2.
 f3() -> ?SDLK_F3.
 
 mod_key_combo() ->
-    Mod = sdl_keyboard:getModState(),
-    Shift = (Mod band shift()) =/= 0,
-    Ctrl = (Mod band ctrl()) =/= 0,
-    Alt = (Mod band alt()) =/= 0,
+    Shift = wings_io:is_modkey_pressed(?KMOD_SHIFT),
+    Ctrl  = wings_io:is_modkey_pressed(?KMOD_CTRL),
+    Alt   = wings_io:is_modkey_pressed(?KMOD_ALT),
     {Shift,Ctrl,Alt}.
 fkey_combo() ->
-    Keys = sdl_keyboard:getKeyState(),
-    F1 = element(f1()+1,Keys) =/= 0,
-    F2 = element(f2()+1,Keys) =/= 0,
-    F3 = element(f3()+1,Keys) =/= 0,
+    F1 = wings_io:is_key_pressed(f1()),
+    F2 = wings_io:is_key_pressed(f2()),
+    F3 = wings_io:is_key_pressed(f3()),
     [F1,F2,F3].
 
 %% Event handler for tweak mode
@@ -278,6 +275,12 @@ handle_tweak_event0(#keyboard{sym=$=}, #tweak{tmode=drag}=T) ->
 handle_tweak_event0(#keyboard{sym=$-}, #tweak{tmode=drag}=T) ->
     update_tweak_handler(magnet_radius(-1,T));
 
+handle_tweak_event0(#keyboard{sym=?SDLK_KP_PLUS}, #tweak{tmode=drag}=T) ->
+    update_tweak_handler(magnet_radius(1,T));
+
+handle_tweak_event0(#keyboard{sym=?SDLK_KP_MINUS}, #tweak{tmode=drag}=T) ->
+    update_tweak_handler(magnet_radius(-1,T));
+
 handle_tweak_event0(#keyboard{sym=C}, #tweak{tmode=drag}=T) ->
     case tweak_hotkey(C, T) of
       none -> keep;
@@ -292,7 +295,7 @@ handle_tweak_event0(#keyboard{unicode=C}=Ev, #tweak{tmode=wait,st=St0}=T) ->
               next ->
                 keep;
               Action ->
-                handle_tweak_event2({action,Action},T#tweak{st=St})
+                handle_tweak_event2({action,Action}, Ev, T#tweak{st=St})
             end;
       T1 -> update_tweak_handler(T1)
     end;
@@ -309,10 +312,10 @@ handle_tweak_event0(#mousemotion{x=X,y=Y,state=State,mod=Mod},
             #tweak{tmode=drag,cx=CX,cy=CY,ox=OX,oy=OY}=T0)
             when State =/= ?SDL_BUTTON_RMASK ->
     {GX,GY} = wings_wm:local2global(X, Y),
-    DX = float(GX-OX), %since last move X and Y
-    DY = float(GY-OY),
-    DxOrg = float(DX+CX), %total X
-    DyOrg = float(DY+CY), %total Y
+    DX = GX-OX, %since last move X and Y
+    DY = GY-OY,
+    DxOrg = DX+CX, %total X
+    DyOrg = DY+CY, %total Y
     FKeys = fkey_combo(),
     TKeys = wings_pref:get_value(tweak_xyz),
     C = tweak_constraints(FKeys,TKeys,[]),
@@ -493,19 +496,19 @@ handle_tweak_event1(#mousebutton{button=3,state=?SDL_RELEASED,x=X,y=Y},
     end;
 
 handle_tweak_event1(Ev,T) ->
-    handle_tweak_event2(Ev,T).
+    handle_tweak_event2(Ev, none, T).
 
-handle_tweak_event2(init_opengl, #tweak{st=St}) ->
+handle_tweak_event2(init_opengl, _, #tweak{st=St}) ->
     wings:init_opengl(St),
     keep;
-handle_tweak_event2(quit=Ev, T) ->
+handle_tweak_event2(quit=Ev, _, T) ->
     wings_wm:later(Ev),
     exit_tweak(T);
 
-handle_tweak_event2({current_state,St}, T) ->
+handle_tweak_event2({current_state,St}, _, T) ->
     update_tweak_handler(T#tweak{st=St});
 
-handle_tweak_event2({new_state,St1}, #tweak{st=St0}=T) ->
+handle_tweak_event2({new_state,St1}, _, #tweak{st=St0}=T) ->
     St2 = clear_temp_sel(St1),
     St3 = wings_undo:save(St0, St2),
     St = case St3 of
@@ -514,27 +517,13 @@ handle_tweak_event2({new_state,St1}, #tweak{st=St0}=T) ->
     end,
     update_tweak_handler(T#tweak{st=St});
 
-handle_tweak_event2({action,Action}, #tweak{tmode=wait,st=#st{}=St0}=T) ->
+handle_tweak_event2({action,Action}, Ev, #tweak{tmode=wait,st=#st{}=St0}=T) ->
     Hs = wings_pref:get_value(hilite_select),
     case Action of
-    {view,aim} ->
-        St = wings_view:command(aim, St0),
-        update_tweak_handler(T#tweak{st=St});
     {view,highlight_aim} ->
-        HL0 = wings_pref:get_value(highlight_aim_at_unselected),
-        HL1 = wings_pref:get_value(highlight_aim_at_selected),
-        {_,X,Y} = wings_wm:local_mouse_state(),
-        {{_,Cmd0},St1} = case wings_pick:do_pick(X, Y, St0) of
-              {add,_,St2} when HL0 =:= true ->
-                  {{view,highlight_aim},{add,St0,St2}};
-              {delete,_,St2} when HL1 =:= true ->
-                  {{view,highlight_aim},{delete,St0,St2}};
-              _Other ->
-                  {{view,aim}, St0}
-        end,
-        St = wings_view:command(Cmd0,St1),
-        update_tweak_handler(T#tweak{st=St});
-
+        {Cmd,_} = wings:highlight_aim_setup(St0),
+        St = wings_view:command(Cmd,St0),
+        do_cmd(Cmd, Ev, T#tweak{st=St});
     {edit,undo_toggle} ->
         St = wings_u:caption(wings_undo:undo_toggle(clear_temp_sel(St0))),
         update_tweak_handler(T#tweak{st=St});
@@ -553,24 +542,23 @@ handle_tweak_event2({action,Action}, #tweak{tmode=wait,st=#st{}=St0}=T) ->
     {select,all}=Cmd when Hs -> hotkey_select_setup(Cmd,T);
     {select, C}=Cmd when C==vertex;C==edge;C==face;C==body ->
         St = clear_temp_sel(St0),
-        do_cmd(Cmd,T#tweak{st=St});
+        do_cmd(Cmd, Ev, T#tweak{st=St});
     {select, {adjacent,_}}=Cmd ->
         St = clear_temp_sel(St0),
-        do_cmd(Cmd,T#tweak{st=St});
+        do_cmd(Cmd, Ev, T#tweak{st=St});
     {file,_}=Cmd ->
         St = clear_temp_sel(St0),
-        do_cmd(Cmd,T#tweak{st=St});
+        do_cmd(Cmd, Ev, T#tweak{st=St});
     keep -> keep;
-    Cmd ->
-        do_cmd(Cmd, T)
+    Cmd -> do_cmd(Cmd, Ev, T)
     end;
-handle_tweak_event2(close, _) ->
+handle_tweak_event2(close, _, _) ->
 % Close a second geometry window
     Active = wings_wm:this(),
     wings_wm:delete({object,Active}),
     delete;
 
-handle_tweak_event2(_, _) ->
+handle_tweak_event2(_, _, _) ->
     keep.
 
 %%%%
@@ -580,18 +568,18 @@ handle_tweak_event2(_, _) ->
 pick_event(X, Y, Cam, B, #tweak{st=#st{}=St0}=T0) ->
     {GX,GY} = wings_wm:local2global(X, Y),
     case wings_pick:do_pick(X, Y, St0) of
-      {add,MM,St1} ->
+      {add,What,St1} ->
         wings_wm:grab_focus(),
         wings_io:grab(),
-        begin_drag(MM, St1, T0),
+        begin_drag(What, St1, T0),
         do_tweak(0.0, 0.0, 0.0, 0.0, screen),
         Time = now(),
         T = T0#tweak{tmode=drag,ox=GX,oy=GY,cx=0,cy=0,dc={Time,0}},
         update_tweak_handler(T);
-      {delete,MM,_} ->
+      {delete,What,_} ->
         wings_wm:grab_focus(),
         wings_io:grab(),
-        begin_drag(MM, St0, T0),
+        begin_drag(What, St0, T0),
         do_tweak(0.0, 0.0, 0.0, 0.0, screen),
         Time = now(),
         T = T0#tweak{tmode=drag,ox=GX,oy=GY,cx=0,cy=0,dc={Time,0}},
@@ -626,21 +614,21 @@ remember_mode(#tweak{magnet=Mag,mag_type=MagType,mag_r=MagR,
              st=#st{selmode=Mode,sh=Sh}}) ->
     wpa:pref_set(?MODULE, sel_mode, {Mode,Sh,Mag,MagType,MagR}).
 
-do_cmd({tools, {tweak,false}}, #tweak{st=St}=T) ->
+do_cmd({tools, {tweak,false}}, _, #tweak{st=St}=T) ->
     exit_tweak(T#tweak{st=clear_temp_sel(St)});
 
-do_cmd({tools, {tweak,true}}, #tweak{st=St}=T) ->
+do_cmd({tools, {tweak,true}}, _, #tweak{st=St}=T) ->
     wings_plugin:command({tools, {tweak,true}}, St),
     exit_tweak(T#tweak{st=clear_temp_sel(St)});
 
-do_cmd(Cmd, #tweak{st=#st{}=St0}=T) ->
+do_cmd(Cmd, Ev, #tweak{st=#st{}=St0}=T) ->
     St1 = remember_command(Cmd, St0),
     case wings_plugin:command(Cmd,St1) of
-      next -> do_wings_cmd(Cmd,T);
+      next -> do_wings_cmd(Cmd, Ev, T);
       Result -> process_cmd_response(Result,T)
     end.
 
-do_wings_cmd({view,Cmd}, #tweak{st=#st{}=St0}=T) ->
+do_wings_cmd({view,Cmd}, _, #tweak{st=#st{}=St0}=T) ->
     case wings_view:command(Cmd,St0) of
         #st{}=St ->
             St1 = clear_temp_sel(St),
@@ -649,15 +637,15 @@ do_wings_cmd({view,Cmd}, #tweak{st=#st{}=St0}=T) ->
           Other
     end;
 
-do_wings_cmd(Cmd, #tweak{st=#st{}=St0}=T) ->
+do_wings_cmd(Cmd, Ev, #tweak{st=#st{}=St0}=T) ->
     St1 = remember_command(Cmd, St0),
-    Result = cmd_type(Cmd, St1),
+    Result = cmd_type(Cmd, Ev, St1),
     process_cmd_response(Result,T).
 
 process_cmd_response(Result,T) ->
     case Result of
       {save_state,St} ->
-          handle_tweak_event2({new_state,St}, T);
+          handle_tweak_event2({new_state,St}, none, T);
       #st{}=St ->
           update_tweak_handler(T#tweak{st=St});
       {drag,Drag} ->
@@ -673,19 +661,35 @@ process_cmd_response(Result,T) ->
           wings:save_windows(),
           exit(normal);
       {replace,Ev} ->
-          handle_tweak_event2(Ev, T);
+          handle_tweak_event2(Ev, none, T);
       Other ->
         Other
     end.
 
-cmd_type({select, Cmd}, St) -> wings_sel_cmd:command(Cmd, St#st{temp_sel=none});
-cmd_type(Cmd, St) -> wings:command(Cmd, St).
+cmd_type({select,Cmd}, _, St) ->
+    wings_sel_cmd:command(Cmd, St#st{temp_sel=none});
+cmd_type(Cmd, Ev, St) ->
+    case Ev of
+    none ->
+        wings:command(Cmd, St);
+    _ ->
+        %% The command was obtained through a hotkey, which for all
+        %% we know may be from an ancient version Wings or for a
+        %% plug-in that has been disabled.
+        try
+        wings:command(Cmd, St)
+        catch
+        error:_ ->
+            wings_hotkey:handle_error(Ev, Cmd),
+            St#st{repeatable=ignore}
+        end
+    end.
 
 hotkey_select_setup(Cmd,#tweak{st=St0}=T) ->
     {_,X,Y} = wings_wm:local_mouse_state(),
     case wings_pick:do_pick(X, Y, St0) of
-      {add,_,St} -> do_cmd(Cmd,T#tweak{st=St});
-      _Other     -> do_cmd(Cmd, T)
+      {add,_,St} -> do_cmd(Cmd, none, T#tweak{st=St});
+      _Other     -> do_cmd(Cmd, none, T)
     end.
 
 remember_command({C,_}=Cmd, St) when C =:= vertex; C =:= edge;
@@ -697,32 +701,35 @@ redraw(St) ->
     wings:redraw(St),
     keep.
 
-begin_drag(MM, St, T) ->
+begin_drag(SelItem, St, T) ->
     wings_draw:refresh_dlists(St),
     wings_dl:map(fun(D, _) ->
-             begin_drag_fun(D, MM, St, T)
+             begin_drag_fun(D, SelItem, St, T)
          end, []).
 
-begin_drag_fun(#dlo{src_sel={body,_},src_we=#we{vp=Vtab}=We}=D, _MM, _St, _T) ->
-    Vs = gb_trees:keys(Vtab),
+begin_drag_fun(#dlo{src_sel={body,_},src_we=#we{vp=Vtab}=We}=D, _, _, _) ->
+    Vs = wings_util:array_keys(Vtab),
     Center = wings_vertex:center(Vs, We),
     Id = e3d_mat:identity(),
     D#dlo{drag={matrix,Center,Id,e3d_mat:expand(Id)}};
-
-begin_drag_fun(#dlo{src_sel={Mode,Els},src_we=We}=D0, MM, St, T) ->
+begin_drag_fun(#dlo{src_sel={Mode,Els},src_we=#we{id=WeId}=We}=D0,
+	       {SelId,_,MM0}, St, T) ->
     Vs0 = sel_to_vs(Mode, gb_sets:to_list(Els), We),
     Center = wings_vertex:center(Vs0, We),
     {Vs,Magnet} = begin_magnet(T, Vs0, Center, We),
     D = wings_draw:split(D0, Vs, St),
+    MM = if
+	     WeId =:= SelId -> MM0;
+	     true -> original
+	 end,
     D#dlo{drag=#drag{vs=Vs0,pos0=Center,pos=Center,mag=Magnet,mm=MM}};
 begin_drag_fun(D, _, _, _) -> D.
 
-end_drag(#tweak{st=St0,ox=X,oy=Y}=T) ->
-    wings_wm:release_focus(),
-    wings_io:ungrab(X,Y),
+end_drag(#tweak{st=St0,ox=OX,oy=OY,cx=CX,cy=CY}=T) ->
+    show_cursor(OX,OY,CX,CY),
     St = wings_dl:map(fun end_drag/2, St0),
     help(T),
-    handle_tweak_event2({new_state,St},T#tweak{tmode=wait}).
+    handle_tweak_event2({new_state,St}, none, T#tweak{tmode=wait}).
 
 end_drag(#dlo{src_we=#we{id=Id},drag=#drag{}}=D0, #st{shapes=Shs0}=St0) ->
     #dlo{src_we=We} = D = wings_draw:join(D0),
@@ -770,9 +777,8 @@ end_drag(#dlo{src_we=#we{id=Id},drag={matrix,_,Matrix,_}}=D,
     {D2#dlo{vs=none,sel=none,drag=none},St};
 end_drag(D, St) -> {D, St}.
 
-end_pick(true, #tweak{st=#st{selmode=Selmode}=St0,ox=X,oy=Y}=T0) ->
-    wings_wm:release_focus(),
-    wings_io:ungrab(X,Y),
+end_pick(true, #tweak{st=#st{selmode=Selmode}=St0,ox=OX,oy=OY,cx=CX,cy=CY}=T0) ->
+    show_cursor(OX,OY,CX,CY),
     St1 = wings_dl:map(fun end_pick_1/2, St0),
     {_,X1,Y1} = wings_wm:local_mouse_state(),
     St = case wings_pick:do_pick(X1, Y1, St1#st{selmode=Selmode}) of
@@ -781,23 +787,17 @@ end_pick(true, #tweak{st=#st{selmode=Selmode}=St0,ox=X,oy=Y}=T0) ->
     end,
     T = T0#tweak{st=St0,tmode=wait},
     help(T),
-    handle_tweak_event2({new_state,St},T);
+    handle_tweak_event2({new_state,St}, none, T);
 
-end_pick(false, #tweak{st=St0,ox=X,oy=Y}=T) ->
-    wings_wm:release_focus(),
-    wings_io:ungrab(X,Y),
+end_pick(false, #tweak{st=St0,ox=OX,oy=OY,cx=CX,cy=CY}=T) ->
+    show_cursor(OX,OY,CX,CY),
     St = wings_dl:map(fun end_pick_1/2, St0),
     help(T),
-    handle_tweak_event2({new_state,St},T#tweak{tmode=wait}).
+    handle_tweak_event2({new_state,St}, none, T#tweak{tmode=wait}).
 
-end_pick_1(#dlo{mirror=M,ns=Ns,proxy_data=Pd,src_we=We},St0) ->
-    {#dlo{ns=Ns,mirror=M,proxy_data=Pd,src_we=We},St0}.
+end_pick_1(#dlo{mirror=M,ns=Ns,proxy=Proxy,proxy_data=Pd,src_we=We},St0) ->
+    {#dlo{ns=Ns,mirror=M,proxy=Proxy,proxy_data=Pd,src_we=We},St0}.
 
-sel_to_vs(edge, _, We) when ?IS_LIGHT(We) ->
-    Items = gb_sets:to_list(wings_sel:get_all_items(edge, We)),
-    wings_edge:to_vertices(Items,We);
-sel_to_vs(Mode, _, We) when ?IS_LIGHT(We) ->
-    gb_sets:to_list(wings_sel:get_all_items(Mode, We));
 sel_to_vs(vertex, Vs, _) -> Vs;
 sel_to_vs(edge, Es, We) -> wings_vertex:from_edges(Es, We);
 sel_to_vs(face, [Face], We) -> wings_face:vertices_ccw(Face, We);
@@ -815,87 +815,10 @@ do_tweak(DX, DY, DxOrg, DyOrg, Mode) ->
 %%  Additional functions for lookup around vertex
 %%
 
-collect_neib_faces(V,#we{mirror=MirrorFace}=We) ->
-    %% The We is not a complete one, but from the display lists. Therefore,
-    %% the face table is not complete. In particular, it does not contain
-    %% hidden faces (negative face numbers), so we must ignore any negative
-    %% face number.
-    wings_vertex:fold(fun(_, Face, _, A) when Face =/= MirrorFace,
-                          Face >= 0 ->
-                  [Face|A];
-             (_,_,_,A) ->
-                  A
-              end, [],V,We).
-
-collect_neib_verts(V,#we{es=Es}=We) ->
-    Facelist=collect_neib_faces(V,We),
-    foldl(fun(Face,D) ->
-          Edges = wings_face:to_edges([Face], We),
-          NearVerts=foldl(fun(E,B) ->
-                      #edge{vs=Vs,ve=Ve}=gb_trees:get(E,Es),
-                      if
-                          V==Vs -> [Ve|B];
-                          V==Ve -> [Vs|B];
-                          true -> B
-                      end
-                  end,[],Edges),
-          NearVerts ++ D
-      end, [],Facelist).
-
-check_if_face_contains_vs(Face,We,Vs)->
-    Verts=wings_face:to_vertices([Face],We),
-    foldl(fun({Vert,_,_,_,_},P)->
-          case member(Vert,Verts) of
-              true -> P;
-              _ -> none
-          end
-      end,all,Vs).
-
-check_if_Vs_have_V(V,Vs)->
-    foldl(fun({VinVs,_,_,_,_},Res)-> if VinVs==V -> true; true->Res end end,false,Vs).
-
-check_if_Vs_have_V12(V1,V2,Vs)->
-    case check_if_Vs_have_V(V1,Vs) of
-    true -> check_if_Vs_have_V(V2,Vs);
-    _ -> false
-    end.
-
-get_nverts(Vs)->
-    foldl(fun(_,S)->S+1 end,0,Vs).
-
-collect_neib_verts_vs(V,#we{es=Es}=We,Vs) ->
-    Facelist0=collect_neib_faces(V,We),
-    Facelist=case get_nverts(Vs) of
-         2-> foldl(fun(Face,FL)->
-                   Res=check_if_face_contains_vs(Face,We,Vs),
-                   case	Res of
-                       all -> [Face|FL];
-                       _ ->FL
-                   end
-               end,[],Facelist0);
-         _->Facelist0
-         end,
-    foldl(fun(Face,D) ->
-          Edges = wings_face:to_edges([Face], We),
-          NearVerts=foldl(fun(E,B) ->
-                      Edg=gb_trees:get(E,Es),
-                      #edge{vs=VS,ve=VE}=Edg,
-                      Have= case get_nverts(Vs) of
-                            2 -> check_if_Vs_have_V12(VE,VS,Vs);
-                            _ ->false
-                        end,
-                      if
-                          Have==true -> B;
-                          V==VS -> [VE|B];
-                          V==VE -> [VS|B];
-                          true -> B
-                      end
-                  end,[],Edges),
-          NearVerts ++ D
-      end, [],Facelist).
-
 collect_neib_verts_coor(V,We)->
-    VertList=collect_neib_verts(V,We),
+    VertList = wings_vertex:fold(fun(_,_,ERec,Acc) ->
+	               [wings_vertex:other(V,ERec)|Acc]
+				   end,[],V,We),
     foldl(fun(E,B) -> [wings_vertex:pos(E,We)|B] end,[],VertList).
 
 get_orig_pos(V,We,Vs)->
@@ -909,11 +832,14 @@ get_orig_pos(V,We,Vs)->
     end.
 
 collect_neib_verts_coor_vs(V,We,Vs)->
-    VertList=collect_neib_verts_vs(V,We,Vs),
+    VertList = wings_vertex:fold(fun(_,_,ERec,Acc) ->
+	               [wings_vertex:other(V,ERec)|Acc]
+				   end,[],V,We),
     foldl(fun(E,B) -> [get_orig_pos(E,We,Vs)|B] end,[],VertList).
 
 sub_pos_from_list(List,Pos) ->
-    foldl(fun(E,B) -> [e3d_vec:sub(E,Pos)|B] end,[],List).
+    foldl(fun
+	    (E,B) -> [e3d_vec:sub(E,Pos)|B] end,[],List).
 
 relax_vec(V, We) ->
     case collect_neib_verts_coor(V, We) of
@@ -927,21 +853,19 @@ relax_vec(V, We) ->
 
 slide_one_vec(Vpos, TweakPos, PosList) ->
     Dpos=e3d_vec:sub(TweakPos,Vpos),
-    {Dp,_} = foldl(fun(Vec, {VP,W}) ->
+    {Dp,_} = foldl(fun
+	    ({0.0,0.0,0.0},VPW) -> VPW;
+        (Vec, {VP,W}) ->
               Vn = e3d_vec:norm(Vec),
-              Dotp0 = e3d_vec:dot(Vn,Dpos),
-              {Dotp,Sign} = if
-                  Dotp0 < 0 -> {-Dotp0/1.5, -1.0};
-                  true -> {Dotp0, 1.0}
-              end,
+              Dotp = e3d_vec:dot(Vn,Dpos),
               if
-                  Dotp > W ->
+                  Dotp > W, Dotp > 0 ->
                       Len = e3d_vec:len(Vec),
                       Dotp2 = if
                           Dotp > Len -> Len;
                           true -> Dotp
                       end,
-                      {e3d_vec:mul(Vn, Dotp2 * Sign),Dotp};
+                      {e3d_vec:mul(Vn, Dotp2),Dotp};
                   true -> {VP,W}
               end
      end,{{0,0,0},0},PosList),
@@ -974,23 +898,21 @@ relax_vec_fn(V, #we{}=We,Pos0,Weight) ->
 %%
 
 collapse_short_edges(Tolerance, #we{es=Etab,vp=Vtab}=We) ->
-    Short = foldl(
-          fun({Edge,#edge{vs=Va,ve=Vb}}, A) ->
-              case gb_trees:is_defined(Va,Vtab) of
-              true->
-                  case gb_trees:is_defined(Vb,Vtab) of
-                  true->
-                      VaPos = wings_vertex:pos(Va, We),
-                      VbPos = wings_vertex:pos(Vb, We),
+    Short = array:sparse_foldl(
+          fun(Edge, #edge{vs=Va,ve=Vb}, A) ->
+              case array:get(Va,Vtab) of
+          undefined -> A;
+              VaPos ->
+                  case array:get(Vb,Vtab) of
+                  undefined -> A;
+                  VbPos ->
                       case abs(e3d_vec:dist(VaPos, VbPos)) of
                       Dist when Dist < Tolerance -> [Edge|A];
                       _Dist -> A
-                      end;
-                  false-> A
-                  end;
-              false -> A
+                      end
+                  end
               end
-          end, [], gb_trees:to_list(Etab)),
+          end, [], Etab),
     NothingCollapsed = Short == [],
     We1 = wings_collapse:collapse_edges(Short,We),
     {NothingCollapsed, We1}.
@@ -1062,26 +984,26 @@ do_tweak(#dlo{drag=#drag{vs=Vs,pos=Pos0,pos0=Orig,mag=Mag0,mm=MM}=Drag,
 do_tweak(D, _, _, _, _, _) -> D.
 
 obj_to_screen({MVM,PM,VP}, {X,Y,Z}) ->
-    glu:project(X, Y, Z, MVM, PM, VP).
+    wings_gl:project(X, Y, Z, MVM, PM, VP).
 
 screen_to_obj({MVM,PM,VP}, {Xs,Ys,Zs}) ->
-    glu:unProject(Xs, Ys, Zs, MVM, PM, VP).
+    wings_gl:unProject(Xs, Ys, Zs, MVM, PM, VP).
 
 tweak_pos(false, true, _, Pos0, TweakPos, #dlo{src_we=#we{}=We,src_sel={face,Sel0}}) ->
     Faces = gb_sets:to_list(Sel0),
     Normals = face_normals(Faces,We,[]),
     case Normals of
-    [[]] -> TweakPos;
-    _Otherwise ->
-        N = e3d_vec:average(Normals),
-    %% constraining by the plane
-    Dot = e3d_vec:dot(N, N),
-      if
-      Dot == 0.0 -> Pos0;
-      true ->
-        T = -e3d_vec:dot(N, e3d_vec:sub(TweakPos, Pos0)) / Dot,
-        e3d_vec:add_prod(TweakPos, N, T)
-      end
+	[[]] -> TweakPos;
+	_Otherwise ->
+	    N = e3d_vec:average(Normals),
+	    %% constraining by the plane
+	    Dot = e3d_vec:dot(N, N),
+	    if
+		Dot == 0.0 -> Pos0;
+		true ->
+		    T = -e3d_vec:dot(N, e3d_vec:sub(TweakPos, Pos0)) / Dot,
+		    e3d_vec:add_prod(TweakPos, N, T)
+	    end
     end;
 
 tweak_pos(false, true,Vs, Pos0, TweakPos, D) ->
@@ -1089,29 +1011,29 @@ tweak_pos(false, true,Vs, Pos0, TweakPos, D) ->
     N = e3d_vec:average(Normals),
     %% constraining by the plane
     Dot = e3d_vec:dot(N, N),
-      if
-      Dot == 0.0 -> Pos0;
-      true ->
-        T = -e3d_vec:dot(N, e3d_vec:sub(TweakPos, Pos0)) / Dot,
-        e3d_vec:add_prod(TweakPos, N, T)
-      end;
+    if
+	Dot == 0.0 -> Pos0;
+	true ->
+	    T = -e3d_vec:dot(N, e3d_vec:sub(TweakPos, Pos0)) / Dot,
+	    e3d_vec:add_prod(TweakPos, N, T)
+    end;
 
 %%%% Along Average Normal
 tweak_pos(true, false, _, Pos0, TweakPos, #dlo{src_we=#we{}=We,src_sel={face,Sel0}}) ->
     Faces = gb_sets:to_list(Sel0),
     Normals = face_normals(Faces,We,[]),
     case Normals of
-    [[]] -> TweakPos;
-    _Otherwise ->
-        N = e3d_vec:norm(e3d_vec:add(Normals)),
-        %% Return the point along the normal closest to TweakPos.
-        Dot = e3d_vec:dot(N, N),
-        if
-        Dot == 0.0 -> Pos0;
-        true ->
-          T = e3d_vec:dot(N, e3d_vec:sub(TweakPos, Pos0)) / Dot,
-          e3d_vec:add_prod(Pos0, N, T)
-        end
+	[[]] -> TweakPos;
+	_Otherwise ->
+	    N = e3d_vec:norm(e3d_vec:add(Normals)),
+	    %% Return the point along the normal closest to TweakPos.
+	    Dot = e3d_vec:dot(N, N),
+	    if
+		Dot == 0.0 -> Pos0;
+		true ->
+		    T = e3d_vec:dot(N, e3d_vec:sub(TweakPos, Pos0)) / Dot,
+		    e3d_vec:add_prod(Pos0, N, T)
+	    end
     end;
 
 tweak_pos(true, false, Vs, Pos0, TweakPos, D) ->
@@ -1119,12 +1041,12 @@ tweak_pos(true, false, Vs, Pos0, TweakPos, D) ->
     N = e3d_vec:norm(e3d_vec:add(Normals)),
     %% Return the point along the normal closest to TweakPos.
     Dot = e3d_vec:dot(N, N),
-      if
-      Dot == 0.0 -> Pos0;
-      true ->
-        T = e3d_vec:dot(N, e3d_vec:sub(TweakPos, Pos0)) / Dot,
-        e3d_vec:add_prod(Pos0, N, T)
-      end.
+    if
+	Dot == 0.0 -> Pos0;
+	true ->
+	    T = e3d_vec:dot(N, e3d_vec:sub(TweakPos, Pos0)) / Dot,
+	    e3d_vec:add_prod(Pos0, N, T)
+    end.
 
 %% vertex_normal(Vertex, DLO) -> UnormalizedNormal
 %%  Calculate the vertex normal. Will also work for vertices surrounded
@@ -1150,9 +1072,9 @@ face_normals([], _We, Normals) ->
     Normals.
 
 vertex_pos(V, Vtab, OrigVtab) ->
-    case gb_trees:lookup(V, Vtab) of
-    none -> gb_trees:get(V, OrigVtab);
-    {value,Pos} -> Pos
+    case array:get(V, Vtab) of
+    undefined -> array:get(V, OrigVtab);
+    Pos -> Pos
     end.
 
 help(#tweak{magnet=false}) ->
@@ -1295,16 +1217,16 @@ clear_temp_sel(#st{temp_sel={Mode,Sh}}=St) ->
 
 tweak_hotkey(C, #tweak{magnet=Mag,mag_type=MagType}=T) ->
     case magnet_hotkey(C, MagType) of
-    none -> constraint_hotkey();
+    none -> constraint_hotkey(T);
     toggle when Mag == true ->
         setup_magnet(T#tweak{magnet=false});
     toggle when Mag == false ->
         setup_magnet(T#tweak{magnet=true});
-    _ when Mag == false -> constraint_hotkey();
+    _ when Mag == false -> constraint_hotkey(T);
     Type -> setup_magnet(T#tweak{mag_type=Type})
     end.
 
-constraint_hotkey() ->
+constraint_hotkey(T) ->
 %% Alt + F1/2/3 toggles xyx constraints on/off
     Alt = mod_key_combo() == {false,false,true},
     Fkeys = fkey_combo(),
@@ -1314,7 +1236,7 @@ constraint_hotkey() ->
       true when Fkeys =/= [false,false,false] ->
         C = set_constraint_toggles(Fkeys,Constraints,[]),
         wings_pref:set_value(tweak_xyz,C),
-        none;
+        T;
       _other -> none
     end.
 set_constraint_toggles([true|Fkeys],[Pref|Constraints],C) ->
@@ -1348,7 +1270,6 @@ setup_magnet(#tweak{tmode=drag}=T) ->
              setup_magnet_fun(D, T)
          end, []),
     do_tweak(0.0, 0.0, 0.0, 0.0,screen),
-    wings_wm:dirty(),
     T;
 setup_magnet(T) -> T.
 
@@ -1367,7 +1288,7 @@ begin_magnet(#tweak{magnet=false}=T, Vs, Center, We) ->
     {[Va || {Va,_,_,_,_} <- Near],Mag};
 begin_magnet(#tweak{magnet=true}=T, Vs, Center, #we{vp=Vtab0}=We) ->
     Mirror = mirror_info(We),
-    Vtab1 = sofs:from_external(gb_trees:to_list(Vtab0), [{vertex,info}]),
+    Vtab1 = sofs:from_external(array:sparse_to_orddict(Vtab0), [{vertex,info}]),
     Vtab2 = sofs:drestriction(Vtab1, sofs:set(Vs, [vertex])),
     Vtab = sofs:to_external(Vtab2),
     Near = near(Center, Vs, Vtab, Mirror, T, We),
@@ -1505,26 +1426,40 @@ draw_magnet_1(#dlo{mirror=Mtx,drag=#drag{mm=Side,pos={X,Y,Z}}}, R,R2) ->
 
 draw_magnet_1(_, _,_) -> [].
 
-mirror_info(#we{mirror=none}) -> {[],none};
+mirror_info(#we{mirror=none}) -> {[],identity};
 mirror_info(#we{mirror=Face}=We) ->
-    PlaneNormal = wings_face:normal(Face, We),
     FaceVs = wpa:face_vertices(Face, We),
-    Origin = wings_vertex:center(FaceVs, We),
-    M0 = e3d_mat:translate(Origin),
-    M = e3d_mat:mul(M0, e3d_mat:project_to_plane(PlaneNormal)),
-    Flatten = e3d_mat:mul(M, e3d_mat:translate(e3d_vec:neg(Origin))),
+    Flatten = wings_we:mirror_projection(We),
     {FaceVs,Flatten}.
 
 mirror_matrix(V, {MirrorVs,Flatten}) ->
     case member(V, MirrorVs) of
-    false -> none;
-    true -> Flatten
+	false -> identity;
+	true -> Flatten
     end.
 
-mirror_constrain(none, Pos) -> Pos;
 mirror_constrain(Matrix, Pos) -> e3d_mat:mul_point(Matrix, Pos).
 
 geom_title(geom) ->
     ?__(1,"Geometry");
 geom_title({geom,N}) ->
     ?__(2,"Geometry #") ++ integer_to_list(N).
+
+%% After releasing lmb to conclude drag, unhide the cursor and make sure its
+%% inside the window.
+show_cursor(OX,OY,CX,CY) ->
+    X0 = OX+CX,
+    Y0 = OY+CY,
+    {{X1,Y1},{W,H}} = wings_wm:win_rect(),
+    Xw = X1+W,
+    Yh = Y1+H,
+    X = if X0 < X1 -> X1+20;
+           X0 > Xw -> Xw-20;
+           true -> X0
+        end,
+    Y = if Y0 < Y1 -> Y1+20;
+           Y0 > Yh -> Yh-20;
+           true -> Y0
+        end,
+    wings_wm:release_focus(),
+    wings_io:ungrab(X,Y).
