@@ -3,12 +3,12 @@
 %%
 %%     POV-Ray Plugin User Interface.
 %%
-%%  Copyright (c) 2007 Chris Hegarty
+%%  Copyright (c) 2007-2008 Chris Hegarty
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 %%
-%%     $Id: wpc_pov.erl 317 2008-04-13 22:35:25Z optigon $
+%%     $Id$
 %%
 
 -module(wpc_pov).
@@ -137,11 +137,23 @@ povray_menu(Menu) ->
 props(render, Attr) ->
     RenderFormat = proplists:get_value(render_format, Attr, ?DEF_RENDER_FORMAT),
     {value, {RenderFormat, Ext, Desc}} = lists:keysearch(RenderFormat, 1, wings_job:render_formats()),
-    [{title,?__(1,"Render")}, {ext,Ext}, {ext_desc,Desc}];
+    Title = case os:type() of
+        {win32,_} -> "Render";
+        _Other    -> ?__(1,"Render")
+	end,
+    [{title,Title}, {ext,Ext}, {ext_desc,Desc}];
 props(export, _Attr) ->
-    [{title,?__(2,"Export")},{ext,".pov"},{ext_desc,?__(3,"POV-Ray File")}];
+    {Title,File} = case os:type() of
+        {win32,_} -> {"Export","POV-Ray File"};
+        _Other    -> {?__(2,"Export"),?__(3,"POV-Ray File")}
+	end,
+    [{title,Title},{ext,".pov"},{ext_desc,File}];
 props(export_selected, _Attr) ->
-    [{title,?__(4,"Export Selected")},{ext,".pov"},{ext_desc,?__(3,"POV-Ray File")}].
+    {Title,File} = case os:type() of
+        {win32,_} -> {"Export","POV-Ray File"};
+        _Other    -> {?__(4,"Export Selected"),?__(3,"POV-Ray File")}
+	end,
+    [{title,Title},{ext,".pov"},{ext_desc,File}].
 
 %menu commands
 command({file,{export,{?TAG,A}}}, St) ->
@@ -1824,8 +1836,7 @@ hook(Props) when is_list(Props) ->
 		 none -> void
 	     end;
 	 (_, _) -> void
-     end};
-hook(Prop) -> hook([Prop]).
+     end}.
 
 hook(Op, Expr) -> hook([{Op,Expr}]).
     
