@@ -3,7 +3,7 @@
 %%
 %%     The main module of Wings 3D.
 %%
-%%  Copyright (c) 2001-2008 Bjorn Gustavsson
+%%  Copyright (c) 2001-2009 Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -19,7 +19,7 @@
 -export([save_windows/0]).
 -export([handle_drop/3]).
 -export([init_menubar/0]).
-
+-export([highlight_aim_setup/1]).
 -export([register_postdraw_hook/3,unregister_postdraw_hook/2]).
 
 -define(NEED_OPENGL, 1).
@@ -100,7 +100,7 @@ init(File) ->
     group_leader(wings_console:start(), self()),
     wings_init:init(),
     wings_text:init(),
-    wings_image:init(),
+    wings_image:init(wings_io:get_process_option()),
     wings_plugin:init(),
     wings_color:init(),
     wings_io:init(),
@@ -126,14 +126,14 @@ init(File) ->
     wings_view:init(),
     wings_file:init(),
     wings_u:caption(St),
-    put(wings_hitbuf, sdl_util:alloc(?HIT_BUF_SIZE, ?GL_INT)),
     wings_wm:init(),
     wings_file:init_autosave(),
     init_menubar(),
     wings_pb:init(),
     wings_ask:init(),
     wings_job:init(),
-    
+    wings_develop:init(),
+
     Op = main_loop_noredraw(St),		%Replace crash handler
                         %with this handler.
     
@@ -153,14 +153,12 @@ init(File) ->
     restore_windows(St),
     case catch wings_wm:enter_event_loop() of
     {'EXIT',normal} ->
-        wings_pref:finish(),
-        erase(wings_hitbuf),
-        sdl:quit();
+	    wings_pref:finish(),
+	    wings_io:quit();
     {'EXIT',Reason} ->
-        io:format("~P\n", [Reason,20]),
-        erase(wings_hitbuf),
-        sdl:quit(),
-        exit(Reason)
+	    io:format("~P\n", [Reason,20]),
+	    wings_io:quit(),
+	    exit(Reason)
     end.
 
 new_viewer(St) ->
@@ -177,32 +175,33 @@ new_viewer(Name, {X,Y}, Size, Props, ToolbarHidden, St) ->
     Op = main_loop_noredraw(St),
     Title = geom_title(Name),
     wings_wm:toplevel(Name, Title, {X,Y,highest}, Size,
-              [resizable,closable,{anchor,nw},
-               {toolbar,
-            fun(A, B, C) -> wings_toolbar:create(A, B, C) end},
-               menubar,
-               {properties,Props}],
-              Op),
+		      [resizable,closable,{anchor,nw},
+		       {toolbar,fun(A, B, C) ->
+					wings_toolbar:create(A, B, C)
+				end},
+		       menubar,
+		       {properties,Props}],
+		      Op),
     wings_wm:menubar(Name, get(wings_menu_template)),
     wings_wm:send({menubar,Name}, {current_state,St}),
+    wings_wm:send({toolbar,Name}, {current_state,St}),
     set_drag_filter(Name),
     if
-    ToolbarHidden -> wings_wm:hide({toolbar,Name});
-    true -> ok
+	ToolbarHidden -> wings_wm:hide({toolbar,Name});
+	true -> ok
     end,
     Name.
 
 free_viewer_num(N) ->
     case wings_wm:is_window({geom,N}) of
-    false -> N;
-    true -> free_viewer_num(N+1)
+	false -> N;
+	true -> free_viewer_num(N+1)
     end.
 
 open_file(none) -> ok;
 open_file(Name) -> wings_wm:send(geom, {open_file,Name}).
 
 init_opengl(St) ->
-    wings_draw_util:init(),
     wings_render:init(),
     wings_dl:init(),
     wings_draw:refresh_dlists(St),
@@ -212,13 +211,17 @@ redraw(St) ->
     redraw(info(St), St).
 
 redraw(Info, St) ->
-    wings_wm:clear_background(),
-    wings_render:render(St),
-    call_post_hook(St),
-    case Info =/= [] andalso wings_wm:get_prop(show_info_text) of
-    true -> wings_io:info(Info);
-    false -> ok
-    end.
+    Render =
+	fun() ->
+		wings_wm:clear_background(),
+		wings_render:render(St),
+		call_post_hook(St),
+		case Info =/= [] andalso wings_wm:get_prop(show_info_text) of
+		    true -> wings_io:info(Info);
+		    false -> ok
+		end
+	end,
+    wings_io:batch(Render).
 
 call_post_hook(St) ->
     case wings_wm:lookup_prop(postdraw_hook) of
@@ -244,9 +247,6 @@ unregister_postdraw_hook(Window, Id) ->
         ok
     end.
 
-save_state(St, St) ->
-    main_loop(St);
-
 save_state(St0, St1) ->
     St2 = wings_undo:save(St0, St1),
     St = case St2 of
@@ -282,8 +282,8 @@ handle_event({open_file,Name}, St0) ->
     end;
 handle_event(Ev, St) ->
     case wings_camera:event(Ev, St) of
-    next -> handle_event_0(Ev, St);
-    Other -> Other
+	next -> handle_event_0(Ev, St);
+	Other -> Other
     end.
 
 handle_event_0(#mousebutton{button=But,state=ButSt,mod=Mod}=Ev, St)
@@ -330,13 +330,13 @@ handle_event_2(Ev, St) -> handle_event_3(Ev, St).
         
 handle_event_3(#keyboard{}=Ev, St0) ->
     case do_hotkey(Ev, St0) of
-    next -> keep;
-    {Cmd,St} -> do_command(Cmd, St)
+	next -> keep;
+	{Cmd,St} -> do_command(Cmd, Ev, St)
     end;
 handle_event_3({action,Callback}, _) when is_function(Callback) ->
     Callback();
 handle_event_3({action,Cmd}, St) ->
-    do_command(Cmd, St);
+    do_command(Cmd, none, St);
 handle_event_3({vec_command,Command,St}, _) when is_function(Command) ->
     %% Use to execute command with vector arguments (see wings_vec.erl).
     command_response(Command(), none, St);
@@ -359,8 +359,8 @@ handle_event_3(redraw, St) ->
     main_loop_noredraw(St);
 handle_event_3(quit, St) ->
     case wings_wm:this() of
-    geom -> do_command({file,quit}, St);
-    _ -> keep
+	geom -> do_command({file,quit}, none, St);
+	_ -> keep
     end;
 handle_event_3({new_state,St}, St0) ->
     save_state(St0, St);
@@ -397,6 +397,8 @@ handle_event_3(language_changed, _) ->
     keep;
 handle_event_3({external,no_more_basic_menus}, _St) ->
     wings_help:no_more_basic_menus();
+handle_event_3({external,not_possible_to_save_prefs}, _St) ->
+    wings_help:not_possible_to_save_prefs();
 handle_event_3({external,launch_tweak}, St) ->
     wpc_tweak:command({tools,{tweak,false}},St);
 handle_event_3({external,Op}, St) ->
@@ -414,7 +416,7 @@ do_hotkey(Ev, #st{sel=[]}=St0) ->
         {add,_,St} ->
             case wings_hotkey:event(Ev, St) of
             next -> next;
-            {view,highlight_aim} -> {{view,aim},set_temp_sel(St0, St)};
+            {view,highlight_aim} -> highlight_aim_setup(St0);
             Cmd ->
                 case highlight_sel_style(Cmd) of
                 none -> {Cmd,St0};
@@ -476,6 +478,8 @@ highlight_sel_style({tools,save_bb}) -> temporary;
 highlight_sel_style({tools,{scale_to_bb,_}}) -> temporary;
 highlight_sel_style({tools,{scale_to_bb_prop,_}}) -> temporary;
 highlight_sel_style({tools,{move_to_bb,_}}) -> temporary;
+highlight_sel_style({tools,{move_bb_to_sel,_}}) -> temporary;
+highlight_sel_style({tools,{scale_bb_to_sel,_}}) -> temporary;
 highlight_sel_style({view,align_to_selection}) -> temporary;
 highlight_sel_style({view,aim}) -> temporary;
 highlight_sel_style({view,highlight_aim}) -> temporary;
@@ -483,14 +487,47 @@ highlight_sel_style({view,frame}) -> temporary;
 
 highlight_sel_style(_) -> none.
 
-do_command(Cmd, St0) ->
+do_command(Cmd, Event, St0) ->
     St = remember_command(Cmd, St0),
     {replace,
      fun(Ev) -> handle_event(Ev, St) end,
-     fun() -> raw_command(Cmd, none, St) end}.
+     fun() -> raw_command(Cmd, none, Event, St) end}.
 
 raw_command(Cmd, Args, St) ->
-    command_response(do_command_1(Cmd, St), Args, St).
+    raw_command(Cmd, Args, none, St).
+
+raw_command(Cmd, Args, Event, St) ->
+    command_response(raw_command_1(Cmd, Event, St), Args, St).
+
+raw_command_1(Cmd, Event, St0) ->
+    case wings_plugin:command(Cmd, St0#st{last_cmd=Cmd}) of
+	next ->
+	    %% Time the command if command timing is enabled,
+	    %% or just execute the command.
+	    Execute = fun() ->
+			      execute_command(Cmd, Event, St0)
+		      end,
+	    wings_develop:time_command(Execute, Cmd);
+	St0 -> St0;
+	#st{}=St -> {save_state,St};
+	Other -> Other
+    end.
+
+execute_command(Cmd, none, St) ->
+    %% The command was obtained directly from a menu.
+    %% It it not allowed to fail.
+    command(Cmd, St);
+execute_command(Cmd, Ev, St) ->
+    %% The command was obtained through a hotkey, which for all
+    %% we know may be from an ancient version Wings or for a
+    %% plug-in that has been disabled.
+    try
+	command(Cmd, St)
+    catch
+	error:_ ->
+	    wings_hotkey:handle_error(Ev, Cmd),
+	    St#st{repeatable=ignore}
+    end.
 
 command_response(#st{}=St, _, _) ->
     main_loop(clear_temp_sel(St));
@@ -518,14 +555,6 @@ command_response(quit, _, _) ->
     save_windows(),
     exit(normal).
 
-do_command_1(Cmd, St0) ->
-    case wings_plugin:command(Cmd, St0) of
-    next -> command(Cmd, St0);
-    St0 -> St0;
-    #st{}=St -> {save_state,St};
-    Other -> Other
-    end.
-
 remember_command({C,_}=Cmd, St) when C =:= vertex; C =:= edge;
                      C =:= face; C =:= body ->
     St#st{repeatable=Cmd,ask_args=none,drag_args=none};
@@ -571,155 +600,177 @@ repeatable(Mode, Cmd) ->
     _ -> no
     end.
 
-%% Vector and secondary-selection commands.
-command({shape,Shape}, St0) ->
-    case wings_shapes:command(Shape, St0) of
-        St0 -> St0;
-    #st{}=St -> {save_state,St};
-    Other -> Other
-    end;
-command({help,What}, St) ->
-    wings_help:command(What, St);
-
-%% Drag & drop.
-command({drop,What}, St) ->
-    drop_command(What, St);
-
-%% File menu.
-command({file,Command}, St) ->
-    wings_file:command(Command, St);
-
-%% Edit menu.
+%% Handle Undo/Redo first.
 command({edit,undo_toggle}, St) ->
     wings_u:caption(wings_undo:undo_toggle(St));
 command({edit,undo}, St) ->
     wings_u:caption(wings_undo:undo(St));
 command({edit,redo}, St) ->
     wings_u:caption(wings_undo:redo(St));
-command({edit,repeat}, #st{sel=[]}=St) -> St;
-command({edit,repeat}, #st{selmode=Mode,repeatable=Cmd0}=St) ->
-    case repeatable(Mode, Cmd0) of
-    no -> keep;
-    Cmd when is_tuple(Cmd) -> raw_command(Cmd, none, St)
+command(Cmd, St) ->
+    command_1(Cmd, St#st{last_cmd=Cmd}).
+
+%% Vector and secondary-selection commands.
+command_1({shape,Shape}, St0) ->
+    case wings_shapes:command(Shape, St0) of
+        St0 -> St0;
+    #st{}=St -> {save_state,St};
+    Other -> Other
     end;
-command({edit,repeat}, St) -> St;
-command({edit,repeat_args}, #st{sel=[]}=St) -> St;
-command({edit,repeat_args}, #st{selmode=Mode,repeatable=Cmd0,
-                ask_args=AskArgs}=St) ->
+command_1({help,What}, St) ->
+    wings_help:command(What, St);
+
+%% Drag & drop.
+command_1({drop,What}, St) ->
+    drop_command(What, St);
+
+%% File menu.
+command_1({file,Command}, St) ->
+    wings_file:command(Command, St);
+
+%% Edit menu.
+command_1({edit,repeat}, #st{sel=[]}=St) -> St;
+command_1({edit,repeat}, #st{selmode=Mode,repeatable=Cmd0}=St) ->
     case repeatable(Mode, Cmd0) of
-    no -> keep;
-    Cmd1 when is_tuple(Cmd1) ->
-        Cmd = replace_ask(Cmd1, AskArgs),
-        raw_command(Cmd, none, St)
+	no -> keep;
+	Cmd when is_tuple(Cmd) -> raw_command(Cmd, none, St)
     end;
-command({edit,repeat_args}, St) -> St;
-command({edit,repeat_drag}, #st{sel=[]}=St) -> St;
-command({edit,repeat_drag}, #st{selmode=Mode,repeatable=Cmd0,
-                ask_args=AskArgs,drag_args=DragArgs}=St) ->
+command_1({edit,repeat}, St) -> St;
+command_1({edit,repeat_args}, #st{sel=[]}=St) -> St;
+command_1({edit,repeat_args}, #st{selmode=Mode,repeatable=Cmd0,
+				  ask_args=AskArgs}=St) ->
     case repeatable(Mode, Cmd0) of
-    no -> keep;
-    Cmd1 when is_tuple(Cmd1) ->
-        Cmd = replace_ask(Cmd1, AskArgs),
-        raw_command(Cmd, DragArgs, St)
+	no -> keep;
+	Cmd1 when is_tuple(Cmd1) ->
+	    Cmd = replace_ask(Cmd1, AskArgs),
+	    raw_command(Cmd, none, St)
     end;
-command({edit,repeat_drag}, St) -> St;
-command({edit,purge_undo}, St) ->
+command_1({edit,repeat_args}, St) -> St;
+command_1({edit,repeat_drag}, #st{sel=[]}=St) -> St;
+command_1({edit,repeat_drag}, #st{selmode=Mode,repeatable=Cmd0,
+				  ask_args=AskArgs,drag_args=DragArgs}=St) ->
+    case repeatable(Mode, Cmd0) of
+	no -> keep;
+	Cmd1 when is_tuple(Cmd1) ->
+	    Cmd = replace_ask(Cmd1, AskArgs),
+	    raw_command(Cmd, DragArgs, St)
+    end;
+command_1({edit,repeat_drag}, St) -> St;
+command_1({edit,purge_undo}, St) ->
     purge_undo(St);
-command({edit,confirmed_purge_undo}, St) ->
+command_1({edit,confirmed_purge_undo}, St) ->
     wings_undo:init(St);
-command({edit,enable_patches}, St) ->
+command_1({edit,enable_patches}, St) ->
     wings_start:enable_patches(),
     St;
-command({edit,disable_patches}, St) ->
+command_1({edit,disable_patches}, St) ->
     wings_start:disable_patches(),
     St;
-command({edit,{preferences,Pref}}, St) ->
+command_1({edit,{preferences,Pref}}, St) ->
     wings_pref_dlg:command(Pref, St);
 
 %% Select menu.
-command({select,Command}, St) ->
+command_1({select,Command}, St) ->
     wings_sel_cmd:command(Command, St);
 
 %% View menu.
-command({view,Command}, St) ->
+command_1({view,Command}, St) ->
     wings_view:command(Command, St);
 
 %% Window menu.
-command({window,geom_viewer}, St) ->
+command_1({window,geom_viewer}, St) ->
     new_viewer(St),
     keep;
-command({window,outliner}, St) ->
+command_1({window,outliner}, St) ->
     wings_outliner:window(St);
-command({window,object}, St) ->
+command_1({window,object}, St) ->
     wings_shape:window(St);
-command({window,palette}, St) ->
+command_1({window,palette}, St) ->
     wings_palette:window(St);
-command({window,console}, _St) ->
+command_1({window,console}, _St) ->
     wings_console:window(),
     keep;
 
 %% Body menu.
-command({body,Cmd}, St) ->
+command_1({body,Cmd}, St) ->
     wings_body:command(Cmd, St);
 
 %% Face menu.
-command({face,Cmd}, St) ->
+command_1({face,Cmd}, St) ->
     wings_face_cmd:command(Cmd, St);
 
 %% Edge commands.
-command({edge,Cmd}, St) ->
+command_1({edge,Cmd}, St) ->
     wings_edge_cmd:command(Cmd, St);
 
 %% Vertex menu.
-command({vertex,Cmd}, St) ->
+command_1({vertex,Cmd}, St) ->
     wings_vertex_cmd:command(Cmd, St);
 
 %% Light menu.
-command({light,Cmd}, St) ->
+command_1({light,Cmd}, St) ->
     wings_light:command(Cmd, St);
 
 %% Material commands.
-command({material,Cmd}, St) ->
+command_1({material,Cmd}, St) ->
     wings_material:command(Cmd, St);
 
 %% Tools menu.
 
-command({tools,set_default_axis}, St) ->
-    wings:ask({[axis,point],[]}, St,
+command_1({tools,{set_default_axis, Type0}}, St) ->
+    Type = case Type0 of
+      axis_point -> [axis,point];
+      _ -> [Type0]
+    end,
+    wings:ask({Type,[]}, St,
           fun({Axis,Point}, _) ->
-              wings_pref:set_value(default_axis, {Point,Axis}),
-              keep
+               Default = {Point,Axis},
+               wings_pref:set_value(default_axis, Default),
+               keep;
+             (NewDef, _) ->
+               {Point,Axis} = wings_pref:get_value(default_axis),
+               Default = case Type0 of
+                 axis -> {Point, NewDef};
+                 point -> {NewDef, Axis}
+               end,
+               wings_pref:set_value(default_axis, Default),
+               keep
           end);
-command({tools,{align,Dir}}, St) ->
+command_1({tools,{align,Dir}}, St) ->
     {save_state,wings_align:align(Dir, St)};
-command({tools,{center,Dir}}, St) ->
+command_1({tools,{center,Dir}}, St) ->
     {save_state,wings_align:center(Dir, St)};
-command({tools,save_bb}, St) ->
+command_1({tools,save_bb}, St) ->
     wings_align:copy_bb(St);
-command({tools,{scale_to_bb,Dir}}, St) ->
+command_1({tools,{scale_to_bb,Dir}}, St) ->
     {save_state,wings_align:scale_to_bb(Dir, St)};
-command({tools,{scale_to_bb_prop,Dir}}, St) ->
+command_1({tools,{scale_to_bb_prop,Dir}}, St) ->
     {save_state,wings_align:scale_to_bb_prop(Dir, St)};
-command({tools,{move_to_bb,Dir}}, St) ->
+command_1({tools,{move_to_bb,Dir}}, St) ->
     {save_state,wings_align:move_to_bb(Dir, St)};
-command({tools,{virtual_mirror,Cmd}}, St) ->
+command_1({tools,{move_bb_to_sel,Dir}}, St) ->
+    {save_state,wings_align:move_bb_to_sel(Dir, St)};
+command_1({tools,{scale_bb_to_sel,Dir}}, St) ->
+    {save_state,wings_align:scale_bb_to_sel(Dir, St)};
+command_1({tools,{virtual_mirror,Cmd}}, St) ->
     wings_view:virtual_mirror(Cmd, St);
-command({tools, screenshot}, St) ->
+command_1({tools, screenshot}, St) ->
     wings_image:screenshot(),
     St;
-command({tools, area_volume_info}, St) ->
+command_1({tools, area_volume_info}, St) ->
     area_volume_info(St),
     St;
-command({tools, scene_size_info}, St) ->
-    scene_size_info(St),
-    St;
-command({tools, put_on_ground}, St) ->
+command_1({tools, put_on_ground}, St) ->
     {save_state,wings_align:put_on_ground(St)};
-command({tools, unitize}, St) ->
+command_1({tools, unitize}, St) ->
     {save_state,wings_align:unitize(St)};
 
+%% Develop menu.
+command_1({develop,Cmd}, St) ->
+    wings_develop:command(Cmd, St);
+
 %% wings_job action events.
-command({wings_job,Command}, St) ->
+command_1({wings_job,Command}, St) ->
     wings_job:command(Command, St).
 
 
@@ -738,13 +789,19 @@ popup_menu(X, Y, #st{selmode=Mode}=St) ->
     end.
 
 init_menubar() ->
-    Menus = [{?__(1,"File"),file,fun(St) -> wings_file:menu(St) end},
-          {?__(2,"Edit"),edit,fun edit_menu/1},
-          {?__(3,"View"),view,fun(St) -> wings_view:menu(St) end},
-          {?__(4,"Select"),select,fun(St) -> wings_sel_cmd:menu(St) end},
-          {?__(5,"Tools"),tools,fun tools_menu/1},
-          {?__(6,"Window"),window,fun window_menu/1},
-          {?__(7,"Help"),help,fun(St) -> wings_help:menu(St) end}],
+    Tail0 = [{?__(7,"Help"),help,fun wings_help:menu/1}],
+    Tail = case wings_pref:get_value(show_develop_menu) of
+	       true ->
+		   [{"Develop",develop,fun wings_develop:menu/1}|Tail0];
+	       false ->
+		   Tail0
+	   end,
+    Menus = [{?__(1,"File"),file,fun wings_file:menu/1},
+	     {?__(2,"Edit"),edit,fun edit_menu/1},
+	     {?__(3,"View"),view,fun wings_view:menu/1},
+	     {?__(4,"Select"),select,fun wings_sel_cmd:menu/1},
+	     {?__(5,"Tools"),tools,fun tools_menu/1},
+	     {?__(6,"Window"),window,fun window_menu/1}|Tail],
     put(wings_menu_template, Menus).
 
 edit_menu(St) ->
@@ -795,8 +852,13 @@ tools_menu(_) ->
      {?__(11,"Scale to Saved BB"),{scale_to_bb,Dirs}},
      {?__(12,"Scale to Saved BB Proportionally"),{scale_to_bb_prop,Dirs}},
      {?__(13,"Move to Saved BB"),{move_to_bb,wings_menu_util:all_xyz()}},
+     {?__(32,"Move BB to Selection"),{move_bb_to_sel,wings_menu_util:all_xyz()}},
+     {?__(33,"Scale BB to Selection"),{scale_bb_to_sel,Dirs}},
      separator,
-     {?__(14,"Set Default Axis"),set_default_axis,
+     {?__(14,"Set Default Axis"),{set_default_axis,
+       [{?__(34,"Set Axis and Point"),axis_point},
+        {?__(35,"Set Axis"),axis},
+        {?__(36,"Set Point"),point}]},
       ?__(15,"Define and store axis (with ref. point) for later use with any ")++
       ?__(16,"\"Default Axis\" command (e.g. Scale|Default Axis)")},
      separator,
@@ -814,8 +876,6 @@ tools_menu(_) ->
      separator,
      {?__(26,"Scene Info: Area & Volume"), area_volume_info,
       ?__(27,"Calculate area and volume for each object in the scene")},
-     {?__(32,"Memory Usage"), scene_size_info,
-      ?__(33,"Calculate the memory usage for the current scene from the saved state")},
      {?__(28,"Put on Ground"), put_on_ground,
       ?__(29,"Put selected objects on the ground plane")},
      {?__(30,"Unitize"), unitize,
@@ -998,7 +1058,7 @@ measure(Base, #st{selmode=edge,sel=[{Id,Es}],shapes=Shs}) ->
         1 ->
             We = gb_trees:get(Id, Shs),
             [Edge] = gb_sets:to_list(Es),
-            #edge{vs=Va,ve=Vb} = gb_trees:get(Edge, We#we.es),
+            #edge{vs=Va,ve=Vb} = array:get(Edge, We#we.es),
             {Xa,Ya,Za} = wings_vertex:pos(Va, We),
             {Xb,Yb,Zb} = wings_vertex:pos(Vb, We),
             Length = e3d_vec:dist({Xa,Ya,Za}, {Xb,Yb,Zb}),
@@ -1015,8 +1075,8 @@ measure(Base, #st{selmode=edge,sel=[{Id,Es}],shapes=Shs}) ->
         2 ->
             We = gb_trees:get(Id, Shs),
             [E0,E1] = gb_sets:to_list(Es),
-            #edge{vs=V0s,ve=V0e} = gb_trees:get(E0, We#we.es),
-            #edge{vs=V1s,ve=V1e} = gb_trees:get(E1, We#we.es),
+            #edge{vs=V0s,ve=V0e} = array:get(E0, We#we.es),
+            #edge{vs=V1s,ve=V1e} = array:get(E1, We#we.es),
             PosA = {X0s,Y0s,Z0s} = wings_vertex:pos(V0s, We),
             PosB = {X0e,Y0e,Z0e} = wings_vertex:pos(V0e, We),
             PosC = {X1s,Y1s,Z1s} = wings_vertex:pos(V1s, We),
@@ -1045,8 +1105,8 @@ measure(Base, #st{shapes=Shs, selmode=edge, sel=[{Id1,Sel1},{Id2,Sel2}]}) ->
       WeB = gb_trees:get(Id2, Shs),
       [Es0] = gb_sets:to_list(Sel1),
       [Es1] = gb_sets:to_list(Sel2),
-      #edge{vs=V0s,ve=V0e} = gb_trees:get(Es0, WeA#we.es),
-      #edge{vs=V1s,ve=V1e} = gb_trees:get(Es1, WeB#we.es),
+      #edge{vs=V0s,ve=V0e} = array:get(Es0, WeA#we.es),
+      #edge{vs=V1s,ve=V1e} = array:get(Es1, WeB#we.es),
       PosA = {X0s,Y0s,Z0s} = wings_vertex:pos(V0s, WeA),
       PosB = {X0e,Y0e,Z0e} = wings_vertex:pos(V0e, WeA),
       PosC = {X1s,Y1s,Z1s} = wings_vertex:pos(V1s, WeB),
@@ -1119,18 +1179,29 @@ item_list([], _Sep, Desc) -> [Desc|?__(1," selected")].
 
 shape_info(We) when ?IS_LIGHT(We) ->
     wings_light:info(We);
-shape_info(#we{id=Id,name=Name,fs=Ftab,es=Etab,vp=Vtab,mode=Mode}) ->
+shape_info(#we{id=Id,name=Name,fs=Ftab,es=Etab,vp=Vtab}=We) ->
     Faces = gb_trees:size(Ftab),
-    Edges = gb_trees:size(Etab),
-    Vertices = gb_trees:size(Vtab),
-    wings_util:format(?__(object_info,
-              "Object ~p \"~s\" has ~p polygons, "
-              "~p edges, ~p vertices.\n"
-              "Mode is ~s"),
-              [Id,Name,Faces,Edges,Vertices,object_mode(Mode)]).
+    Edges = wings_util:array_entries(Etab),
+    Vertices = wings_util:array_entries(Vtab),
+    wings_util:format(?__(new_object_info,
+			  "Object ~p \"~s\" has ~p polygons, "
+			  "~p edges, ~p vertices~s~s."),
+              [Id,Name,Faces,Edges,Vertices,
+	       vtx_attributes(We),hole_info(We)]).
 
-object_mode(vertex) -> ?__(vertex,"vertex color");
-object_mode(material) -> ?__(material,"material").
+vtx_attributes(We) ->
+    case wings_va:any_attributes(We) of
+	false -> "";
+	true -> ", " ++ ?__(1,"vertex attributes")
+    end.
+
+hole_info(#we{holes=[]}) ->
+    "";
+hole_info(#we{holes=Holes}) ->
+    case length(Holes) of
+	1 -> [", 1 ",?__(1,"hole")];
+	N -> [", ",integer_to_list(N)," ",?__(2,"holes")]
+    end.
 
 shape_info(Objs, Shs) ->
     shape_info(Objs, Shs, 0, 0, 0, 0).
@@ -1138,8 +1209,8 @@ shape_info(Objs, Shs) ->
 shape_info([{Id,_}|Objs], Shs, On, Vn, En, Fn) ->
     #we{fs=Ftab,es=Etab,vp=Vtab} = gb_trees:get(Id, Shs),
     Faces = gb_trees:size(Ftab),
-    Edges = gb_trees:size(Etab),
-    Vertices = gb_trees:size(Vtab),
+    Edges = wings_util:array_entries(Etab),
+    Vertices = wings_util:array_entries(Vtab),
     shape_info(Objs, Shs, On+1, Vn+Vertices, En+Edges, Fn+Faces);
 shape_info([], _Shs, N, Vertices, Edges, Faces) ->
     io_lib:format(?__(2,
@@ -1419,9 +1490,11 @@ save_windows_1([_|T]) -> save_windows_1(T);
 save_windows_1([]) -> [].
 
 save_window(Name, Ns) ->
-    Pos = wings_wm:win_ur({controller,Name}),
+    {MaxX,_} = wings_wm:win_size(desktop),
+    {PosX0,PosY} = wings_wm:win_ur({controller,Name}),
+    PosX = if PosX0 < 0 -> 20; PosX0 > MaxX -> 20; true -> PosX0 end,
     Size = wings_wm:win_size(Name),
-    W = {Name,Pos,Size},
+    W = {Name,{PosX,PosY},Size},
     [W|save_windows_1(Ns)].
 
 save_geom_window(Name, Ns) ->
@@ -1472,21 +1545,27 @@ restore_windows_1([{{geom,_}=Name,Pos0,Size,Ps0}|Ws], St) ->
     set_geom_props(Ps, Name),
     restore_windows_1(Ws, St);
 restore_windows_1([{{object,_}=Name,{_,_}=Pos,{_,_}=Size}|Ws], St) ->
-    wings_shape:window(Name, Pos, Size, St),
+    wings_shape:window(Name, validate_pos(Pos), Size, St),
     restore_windows_1(Ws, St);
 restore_windows_1([{outliner,{_,_}=Pos,{_,_}=Size}|Ws], St) ->
-    wings_outliner:window(Pos, Size, St),
+    wings_outliner:window(validate_pos(Pos), Size, St),
     restore_windows_1(Ws, St);
 restore_windows_1([{console,{_,_}=Pos,{_,_}=Size}|Ws], St) ->
-    wings_console:window(console, Pos, Size),
+    wings_console:window(console, validate_pos(Pos), Size),
     restore_windows_1(Ws, St);
 restore_windows_1([{palette,{_,_}=Pos,{_,_}=Size}|Ws], St) ->
-    wings_palette:window(Pos,Size,St),
+    wings_palette:window(validate_pos(Pos),Size,St),
     restore_windows_1(Ws, St);
 restore_windows_1([_|Ws], St) ->
     restore_windows_1(Ws, St);
 restore_windows_1([], _) -> ok.
 
+validate_pos({X,Y}=Pos) ->
+    case Y < 0 of
+      false -> Pos;
+      true -> {X,20}
+    end.
+
 geom_pos({X,Y}=Pos) ->
     {_,Upper0} = wings_wm:win_ul(desktop),
     Upper1 = case wings_wm:is_hidden({toolbar,geom}) of
@@ -1496,8 +1575,10 @@ geom_pos({X,Y}=Pos) ->
              Upper0+ToolbarH
          end,
     {_,TitleH} = wings_wm:win_size({controller,geom}),
-    case Upper1 + TitleH of
-    Upper when Y < Upper -> {X,Upper};
+    {_,MenuBarH} = wings_wm:win_size({menubar,geom}),
+    case Upper1 + TitleH + MenuBarH of
+    Upper when Y < Upper -> 
+      {X,Upper};
     _ -> Pos
     end.
 
@@ -1579,27 +1660,17 @@ area_volume(Face, We) ->
     Volume = e3d_vec:dot(V1, Bc)/6.0,
     {Area, Volume}.
 
-scene_size_info(St) ->
-    Bytes = byte_size(term_to_binary(St)),
-    MB = Bytes/1048576,
-	{Unit,Amount} = best_unit(Bytes,MB),
-    Usage = io_lib:format(?__(1,"Your scene is using approximately ~s ~s [~p bytes]"),
-            [wings_util:nice_float(Amount), Unit, Bytes]),
-    wings_help:help_window(?__(2,"Memory Usage"), [Usage]).
-
-best_unit(Bytes,MB) when  MB < 1.0 ->
-    {"KB",Bytes/1024};
-best_unit(_, MB) ->
-    {"MB",MB}.
-
 highlight_aim_setup(St0) ->
     HL0 = wings_pref:get_value(highlight_aim_at_unselected),
     HL1 = wings_pref:get_value(highlight_aim_at_selected),
     {_,X,Y} = wings_wm:local_mouse_state(),
     case wings_pick:do_pick(X, Y, St0) of
-      {add,_,St} when HL0 =:= true -> {{view,highlight_aim},{add,St0,St}};
-      {delete,_,St} when HL1 =:= true -> {{view,highlight_aim},{delete,St0,St}};
-      _Other           -> {{view,aim}, St0}
+	{add,MM,#st{selmode=Selmode,sel=Sel}} when HL0 =:= true ->
+	    {{view,{highlight_aim,{add,{Selmode,Sel,MM}}}},St0};
+	{delete,MM,#st{selmode=Selmode,sel=Sel}} when HL1 =:= true ->
+	    {{view,{highlight_aim,{delete,{Selmode,Sel,MM}}}},St0};
+	_Other ->
+	    {{view,aim},St0}
     end.
 
 hotkey_select_setup(Cmd,St0) ->
