@@ -77,7 +77,20 @@ menu(X, Y, St) ->
 	     ?__(30,"Rename selected objects")},
 	    separator,
 	    {?__(31,"Show All"),show_all,
-	     ?__(32,"Show all faces for this object")}|mode_dependent(St)],
+	     ?__(32,"Show all faces for this object")},
+	    {?__(33,"Vertex Attributes"),
+	     {vertex_attributes,
+	      [{?__(41,"Colors to Materials"),colors_to_materials,
+		?__(42,"Convert vertex colors to materials")},
+	       {?__(43,"Materials to Colors"),materials_to_colors,
+		?__(44,"Convert materials to vertex colors")},
+	       {?__(45,"Remove Colors"),remove_colors,
+		?__(46,"Remove all vertex colors")},
+	       {?__(47,"Remove UV Coordinates"),remove_uv_coordinates,
+		?__(48,"Remove all UV coordinates")},
+	       {?__(49,"Remove All Attributes"),remove_all_attributes,
+		?__(50,"Remove all vertex colors and UV coordinates")}]}}|
+	    mode_dependent(St)],
     wings_menu:popup_menu(X, Y, body, Menu).
 
 flip_fun(pick) ->
@@ -94,17 +107,16 @@ flip_fun(Axis) ->
 	end.
 dup(Type) ->
 %% Return {dup,Type} if Alt is pressed during Flip command, otherwise Type.
-    Mod = sdl_keyboard:getModState(),
-    case (Mod band ?KMOD_ALT) =/= 0 of
+    case wings_io:is_modkey_pressed(?KMOD_ALT) of
         true -> {dup,Type};
         false -> Type
     end.
 
 mode_dependent(St) ->
     SelObj = wings_sel:fold(fun(_, We, A) -> [We|A] end, [], St),
-    Area = foldl(fun(#we{has_shape=Shape}=We, A) ->
+    Kind = foldl(fun(We, A) ->
 			 Type = if 
-				    ?IS_ANY_LIGHT(We), Shape -> arealight;
+				    ?IS_AREA_LIGHT(We) -> arealight;
 				    true -> object
 				end,
 			 case A of
@@ -113,32 +125,14 @@ mode_dependent(St) ->
 			     _ -> mixed
 			 end
 		 end, none, SelObj),
-    ObjMode = foldl(fun(#we{mode=M}, none) -> M;
-		       (#we{mode=M}, M) -> M;
-		       (_, _) -> mixed
-		    end, none, SelObj),
-    Tail0 = vertex_color_item(Area),
-    Tail1 = mat_col_conv(ObjMode, Area, Tail0),
-    Tail2 = arealight_conv(Area, Tail1),
-    case mode_conv(ObjMode, Area, Tail2) of
-	[] -> [];
-	Tail -> [separator|Tail]
-    end.
+    Tail = vertex_color_item(Kind),
+    arealight_conv(Kind, Tail).
 
 vertex_color_item(object) ->
     [{?__(1,"Vertex Color"),vertex_color,
-      ?__(2,"Apply vertex colors to selected objects "
-	   "(first changing mode to vertex color mode if needed)")}];
+      ?__(3,"Apply vertex colors to selected objects")}];
 vertex_color_item(_) -> [].
     
-mat_col_conv(vertex, object, T) ->
-    [{?__(1,"Colors to Materials"),colors_to_materials,
-      ?__(2,"Convert vertex colors to materials")}|T];
-mat_col_conv(material, object, T) ->
-    [{?__(3,"Materials to Colors"),materials_to_colors,
-      ?__(4,"Convert materials to vertex colors")}|T];
-mat_col_conv(_, _, T) -> T.
-
 arealight_conv(arealight, T) ->
     [{?__(1,"Area Light to Object"),from_arealight,
       ?__(2,"Convert selected area lights to objects")}|T];
@@ -147,16 +141,6 @@ arealight_conv(object, T) ->
       ?__(4,"Convert selected objects to area lights")}|T];
 arealight_conv(mixed, T) -> T.
 
-mode_conv(material, object, T) ->
-    [{?__(1,"Vertex Color Mode"),vertex_color_mode,
-      ?__(2,"Change object mode to vertex color mode "
-	   "(materials will be kept, but not shown)")}|T];
-mode_conv(vertex, object, T) ->
-    [{?__(3,"Material Mode"),material_mode,
-      ?__(4,"Change object mode to material mode "
-	   "(vertex colors will be kept, but not shown)")}|T];
-mode_conv(_, _, T) -> T.
-
 command({move,Type}, St) ->
     wings_move:setup(Type, St);
 command({rotate,Type}, St) ->
@@ -215,14 +199,16 @@ command(from_arealight, St) ->
     from_arealight(St);
 command({from_arealight,Ids}, St) ->
     from_arealight(Ids, St);
-command(materials_to_colors, St) ->
+command({vertex_attributes,materials_to_colors}, St) ->
     {save_state,materials_to_colors(St)};
-command(colors_to_materials, St) ->
+command({vertex_attributes,colors_to_materials}, St) ->
     {save_state,colors_to_materials(St)};
-command(material_mode, St) ->
-    {save_state,set_mode(material, St)};
-command(vertex_color_mode, St) ->
-    {save_state,set_mode(vertex, St)};
+command({vertex_attributes,remove_colors}, St) ->
+    {save_state,va_remove(color, St)};
+command({vertex_attributes,remove_uv_coordinates}, St) ->
+    {save_state,va_remove(uv, St)};
+command({vertex_attributes,remove_all_attributes}, St) ->
+    {save_state,va_remove(all, St)};
 command({weld,Ask}, St) ->
     weld(Ask, St);
 command(vertex_color, St) ->
@@ -283,19 +269,19 @@ clean_isolated_vertices(We) ->
     end.
 		  
 clean_short_edges(Tolerance, #we{es=Etab,vp=Vtab}=We) ->
-    Short = foldl(
-	      fun({Edge,#edge{vs=Va,ve=Vb}}, A) ->
-		      VaPos = gb_trees:get(Va, Vtab),
-		      VbPos = gb_trees:get(Vb, Vtab),
+    Short = array:sparse_foldl(
+	      fun(Edge, #edge{vs=Va,ve=Vb}, A) ->
+		      VaPos = array:get(Va, Vtab),
+		      VbPos = array:get(Vb, Vtab),
 		      case abs(e3d_vec:dist(VaPos, VbPos)) of
 			  Dist when Dist < Tolerance -> [Edge|A];
 			  _Dist -> A
 		      end
-	      end, [], gb_trees:to_list(Etab)),
+	      end, [], Etab),
     foldl(fun(Edge, #we{es=Et}=W) ->
-		  case gb_trees:is_defined(Edge, Et) of
-		      true -> wings_collapse:collapse_edge(Edge, W);
-		      false -> W
+		  case array:get(Edge, Et) of
+		      undefined -> W;
+		      _ -> wings_collapse:collapse_edge(Edge, W)
 		  end
 	  end, We, Short).
 
@@ -348,11 +334,11 @@ cleanup_rep_3(Ves0, V, Face, Mat, Ftab0, Etab0, We0) ->
     cleanup_rep_3(NextVes, V, Face, Mat, Ftab, Etab, We).
 
 cleanup_rep_4([{Vtx,Edge}|T]=Ves, V0, Face, NewFace, Etab0) ->
-    Rec = case gb_trees:get(Edge, Etab0) of
+    Rec = case array:get(Edge, Etab0) of
 	      #edge{lf=Face}=Rec0 -> Rec0#edge{lf=NewFace};
 	      #edge{rf=Face}=Rec0 -> Rec0#edge{rf=NewFace}
 	  end,
-    Etab = gb_trees:update(Edge, Rec, Etab0),
+    Etab = array:set(Edge, Rec, Etab0),
     if
 	Vtx =:= V0 -> {Ves,Etab};
 	true -> cleanup_rep_4(T, V0, Face, NewFace, Etab)
@@ -367,24 +353,24 @@ repeated_vertex_1([_|T]) -> repeated_vertex_1(T);
 repeated_vertex_1([]) -> none.
 
 cleanup_patch_edge(V, Face, From, To, Etab) ->
-    R = case gb_trees:get(From, Etab) of
+    R = case array:get(From, Etab) of
 	    #edge{lf=Face,ve=V}=R0 -> R0#edge{ltpr=To};
 	    #edge{lf=Face,vs=V}=R0 -> R0#edge{ltsu=To};
 	    #edge{rf=Face,vs=V}=R0 -> R0#edge{rtpr=To};
 	    #edge{rf=Face,ve=V}=R0 -> R0#edge{rtsu=To}
 	end,
-    gb_trees:update(From, R, Etab).
+    array:set(From, R, Etab).
 
 %%
 %% A waist is a vertex shared by edges all of which cannot be
 %% reached from the incident edge of the vertex.
 %%
 cleanup_waists(#we{es=Etab,vp=Vtab}=We) ->
-    VsEs0 = foldl(fun({E,#edge{vs=Va,ve=Vb}}, A) ->
-			  [{Va,E},{Vb,E}|A]
-		  end, [], gb_trees:to_list(Etab)),
+    VsEs0 = array:sparse_foldl(fun(E, #edge{vs=Va,ve=Vb}, A) ->
+				       [{Va,E},{Vb,E}|A]
+			       end, [], Etab),
     VsEs = wings_util:rel2fam(VsEs0),
-    cleanup_waists_1(gb_trees:keys(Vtab), VsEs, We).
+    cleanup_waists_1(wings_util:array_keys(Vtab), VsEs, We).
 
 cleanup_waists_1([V|Vs], [{V,AllEs}|VsEs], #we{es=Etab0,vp=Vtab0,vc=Vct0}=We0) ->
     Es0 = wings_vertex:fold(fun(E, _, _, A) -> [E|A] end, [], V, We0),
@@ -396,8 +382,8 @@ cleanup_waists_1([V|Vs], [{V,AllEs}|VsEs], #we{es=Etab0,vp=Vtab0,vc=Vct0}=We0) -
 	    %% Repair by duplicating the original vertex.
 	    {NewV,We1} = wings_we:new_id(We0),
 	    Etab = patch_vtx_refs(Es, V, NewV, Etab0),
-	    Vtab = gb_trees:insert(NewV, gb_trees:get(V, Vtab0), Vtab0),
-	    Vct = gb_trees:insert(NewV, AnEdge, Vct0),
+	    Vtab = array:set(NewV, array:get(V, Vtab0), Vtab0),
+	    Vct = array:set(NewV, AnEdge, Vct0),
 	    We = We1#we{es=Etab,vp=Vtab,vc=Vct},
 	    io:format(?__(1,"Removed waist vertex: ~p\n"), [V]),
 
@@ -409,11 +395,11 @@ cleanup_waists_1([V|Vs], [{V,AllEs}|VsEs], #we{es=Etab0,vp=Vtab0,vc=Vct0}=We0) -
 cleanup_waists_1([], [], We) -> We.
 
 patch_vtx_refs([E|Es], OldV, NewV, Etab0) ->
-    Etab = case gb_trees:get(E, Etab0) of
+    Etab = case array:get(E, Etab0) of
 	       #edge{vs=OldV}=Rec ->
-		   gb_trees:update(E, Rec#edge{vs=NewV}, Etab0);
+		   array:set(E, Rec#edge{vs=NewV}, Etab0);
 	       #edge{ve=OldV}=Rec ->
-		   gb_trees:update(E, Rec#edge{ve=NewV}, Etab0)
+		   array:set(E, Rec#edge{ve=NewV}, Etab0)
 	   end,
     patch_vtx_refs(Es, OldV, NewV, Etab);
 patch_vtx_refs([], _, _, Etab) -> Etab.
@@ -436,7 +422,7 @@ delete_2edged_faces_1([], We) -> We.
 delete_if_bad(Face, #we{fs=Ftab,es=Etab}=We) ->
     case gb_trees:lookup(Face, Ftab) of
 	{value,Edge} ->
-	    case gb_trees:get(Edge, Etab) of
+	    case array:get(Edge, Etab) of
 		#edge{ltpr=Same,ltsu=Same,rtpr=Same,rtsu=Same} ->
 		    bad_edge;
 		#edge{ltpr=Same,ltsu=Same} ->
@@ -531,72 +517,48 @@ flip(Type, Plane, St0) ->
     flip_cmd(Type,Plane,St).
 
 flip_cmd(Type, Plane, St) ->
-    wings_sel:map(fun(_, We) -> 
-        case Plane of
-          {{_,_,_},{_,_,_}} -> flip_body_2(Plane, We);
-          {_,_,_} -> flip_body_3(Type, Plane, We);
-          _Otherwise -> flip_body_1(Type, Plane, We)
-        end
-    end, St).
+    wings_sel:map(fun(_, We) -> flip_body(Type, Plane, We) end, St).
 
-flip_body_1(Type, Plane0, We0) ->
-    Plane = flip_scale(Plane0),
-    Center = e3d_vec:average(wings_vertex:bounding_box(We0)),
-    {Cx,Cy,Cz} = flip_center(Type,Plane0,Center),
+flip_body(_Type, {{_,_,_}=Plane,{_,_,_}=Point}, We) ->
+    do_flip(Point, Plane, We);
+flip_body(Type, Plane0, We) ->
+    Plane = flip_normal(Plane0),
+    Point0 = e3d_vec:average(wings_vertex:bounding_box(We)),
+    Point = flip_center(Type, Plane0, Point0),
+    do_flip(Point, Plane, We).
+
+%% do_flip(Point, PlaneNormal, We0) -> We
+%%  Flip an object around the plane defined by the point Point
+%%  and the plane's normal PlaneNormal.
+%%
+do_flip({Cx,Cy,Cz}, PlaneNormal, We0) ->
+    ToZMat = e3d_mat:rotate_to_z(PlaneNormal),
+    FromZMat = e3d_mat:transpose(ToZMat),
     M0 = e3d_mat:translate(Cx, Cy, Cz),
-    M1 = e3d_mat:mul(M0, Plane),
-    M = e3d_mat:mul(M1, e3d_mat:translate(-Cx, -Cy, -Cz)),
+    M1 = e3d_mat:mul(M0, FromZMat),
+    M2 = e3d_mat:mul(M1, e3d_mat:scale(1.0, 1.0, -1.0)),
+    M3 = e3d_mat:mul(M2, ToZMat),
+    M = e3d_mat:mul(M3, e3d_mat:translate(-Cx, -Cy, -Cz)),
     We = wings_we:transform_vs(M, We0),
     wings_we:invert_normals(We).
 
-flip_body_2({Plane,Center}, #we{vp=Vtab0}=We) ->
-    Vtab1 = lists:foldl(fun(Vtx, A) ->
-            flip_vs(Vtx, Plane, Center, A)
-            end, [], gb_trees:to_list(Vtab0)),
-    Vtab = gb_trees:from_orddict(reverse(Vtab1)),
-    wings_we:invert_normals(We#we{vp=Vtab}).
+flip_normal(x) -> {1.0,0.0,0.0};
+flip_normal(y) -> {0.0,1.0,0.0};
+flip_normal(z) -> {0.0,0.0,1.0};
+flip_normal({Axis,_}) -> flip_normal(Axis);
+flip_normal({_,_,_}=Normal) -> Normal.
 
-flip_body_3(Type, Plane, #we{vp=Vtab0}=We) ->
-    Center0 = wings_vertex:center(We),
-    Center = flip_center(Type,Plane,Center0),
-    Vtab1 = lists:foldl(fun(Vtx, A) ->
-            flip_vs(Vtx, Plane, Center, A)
-            end, [], gb_trees:to_list(Vtab0)),
-    Vtab = gb_trees:from_orddict(reverse(Vtab1)),
-    wings_we:invert_normals(We#we{vp=Vtab}).
-
-flip_vs({V,Pos0}, Plane, Center, A) ->
-    ToCenter = e3d_vec:sub(Center, Pos0),
-    Dot = e3d_vec:dot(ToCenter, Plane),
-    Pos = wings_util:share(e3d_vec:add_prod(Pos0, Plane, 2.0*Dot)),
-    [{V,Pos}|A].
-
-flip_scale(x) -> e3d_mat:scale(-1.0, 1.0, 1.0);
-flip_scale(y) -> e3d_mat:scale(1.0, -1.0, 1.0);
-flip_scale(z) -> e3d_mat:scale(1.0, 1.0, -1.0);
-flip_scale({Axis,_}) -> flip_scale(Axis).
-
-flip_center(local,_,Center) -> Center;
-flip_center(global,x,{_,Cy,Cz}) -> {0.0,Cy,Cz};
-flip_center(global,y,{Cx,_,Cz}) -> {Cx,0.0,Cz};
-flip_center(global,z,{Cx,Cy,_}) -> {Cx,Cy,0.0};
-flip_center(global,{Xa,Ya,Za},{Cx,Cy,Cz}) ->
-    Plane = e3d_vec:norm({Xa,Ya,Za}),
-    intersect_vec_plane({Cx,Cy,Cz},{0.0,0.0,0.0},Plane,Plane);
-flip_center(point,{x,{X,_,_}},{_,Cy,Cz}) -> {X,Cy,Cz};
-flip_center(point,{y,{_,Y,_}},{Cx,_,Cz}) -> {Cx,Y,Cz};
-flip_center(point,{z,{_,_,Z}},{Cx,Cy,_}) -> {Cx,Cy,Z}.
-
-intersect_vec_plane(PosA,PosB,Vector,Plane) ->
-    %% Return point where Vector through PosA intersects with Plane at PosB
-    DotProduct = e3d_vec:dot(Vector,Plane),
-    case DotProduct of
-      0.0 ->
-        PosA;
-      _Otherwise ->
-        Intersection = e3d_vec:dot(e3d_vec:sub(PosB,PosA),Vector)/DotProduct,
-        e3d_vec:add(PosA, e3d_vec:mul(Plane, Intersection))
-    end.
+flip_center(local, _, Center) -> Center;
+flip_center(global, x, {_,Cy,Cz}) -> {0.0,Cy,Cz};
+flip_center(global, y, {Cx,_,Cz}) -> {Cx,0.0,Cz};
+flip_center(global, z, {Cx,Cy,_}) -> {Cx,Cy,0.0};
+flip_center(global, {_,_,_}=Plane, Point) ->
+    %% Project the Point down to the plane where the global axis
+    %% meets the origin.
+    e3d_vec:sub(Point, e3d_vec:mul(Plane, e3d_vec:dot(Point, Plane)));
+flip_center(point, {x,{X,_,_}}, {_,Cy,Cz}) -> {X,Cy,Cz};
+flip_center(point, {y,{_,Y,_}}, {Cx,_,Cz}) -> {Cx,Y,Cz};
+flip_center(point, {z,{_,_,Z}}, {Cx,Cy,_}) -> {Cx,Cy,Z}.
 
 %%%
 %%% The Tighten command.
@@ -607,7 +569,7 @@ tighten(St) ->
     wings_drag:setup(Tvs, [percent], St).
 
 tighten(_, #we{vp=Vtab}=We, A) ->
-    Vs = gb_trees:keys(Vtab),
+    Vs = wings_util:array_keys(Vtab),
     wings_vertex_cmd:tighten(Vs, We, A).
     
 %%%
@@ -630,22 +592,10 @@ combine(#st{shapes=Shs0,sel=[{Id,_}=S|_]=Sel0}=St) ->
     Sel2 = sofs:domain(Sel1),
     {Wes0,Shs2} = sofs:partition(1, Shs1, Sel2),
     Wes = sofs:to_external(sofs:range(Wes0)),
-    Mode = unify_modes(Wes),
-    We0 = wings_we:merge(Wes),
-    We = We0#we{id=Id,mode=Mode},
+    We1 = wings_we:merge(Wes),
+    We = We1#we{id=Id},
     Shs = gb_trees:from_orddict(sort([{Id,We}|sofs:to_external(Shs2)])),
     St#st{shapes=Shs,sel=[S]}.
-
-unify_modes([#we{mode=Mode}|Wes]) ->
-    unify_modes(Wes, Mode).
-
-unify_modes([#we{mode=Mode}|Wes], Mode) ->
-    unify_modes(Wes, Mode);
-unify_modes([_|_], _) ->
-    wings_u:error(?__(1,
-			 "Objects with vertex colors cannot be combined " 
-			 "with objects with materials."));
-unify_modes([], Mode) -> Mode.
 		    
 %%%
 %%% The Separate command.
@@ -683,9 +633,9 @@ do_auto_smooth(Angle, St) ->
     wings_sel:map(fun(_, We) -> auto_smooth_1(Cos, We) end, St).
 
 auto_smooth_1(Cos, #we{es=Etab,he=Htab0}=We) ->
-    Htab = foldl(fun({E,R}, A) ->
-			 auto_smooth(E, R, Cos, A, We)
-		 end, Htab0, gb_trees:to_list(Etab)),
+    Htab = array:sparse_foldl(fun(E, R, A) ->
+				      auto_smooth(E, R, Cos, A, We)
+			      end, Htab0, Etab),
     We#we{he=Htab}.
 
 auto_smooth(Edge, #edge{lf=Lf,rf=Rf}, Cos, H0, We) ->
@@ -756,9 +706,8 @@ to_arealight_1([], Shs, St) ->
 to_arealight_1([We0|Wes], Shs, St) when ?IS_ANY_LIGHT(We0) ->
     to_arealight_1(Wes, Shs, St);
 to_arealight_1([#we{id=Id}=We0|Wes], Shs, St) ->
-    #we{light=Light,has_shape=HasShape} = 
-	wings_light:import([{opengl,[{type,area}]}]),
-    We = We0#we{light=Light,has_shape=HasShape},
+    #we{light=Light} = wings_light:import([{opengl,[{type,area}]}]),
+    We = We0#we{light=Light},
     to_arealight_1(Wes, gb_trees:update(Id, We, Shs), St).
 
 
@@ -768,8 +717,7 @@ to_arealight_1([#we{id=Id}=We0|Wes], Shs, St) ->
 %%%
 from_arealight(#st{shapes=Shs}=St) ->
     Wes = wings_sel:fold(
-	    fun (_, #we{has_shape=HasShape}=We, A) 
-		when ?IS_ANY_LIGHT(We), HasShape -> 
+	    fun (_, We, A) when ?IS_AREA_LIGHT(We)-> 
 		    [We|A];
 		(_, _, A) ->
 		    A
@@ -782,22 +730,11 @@ from_arealight(Objects, #st{shapes=Shs}=St) ->
 
 from_arealight_1([], Shs, St) -> 
     St#st{shapes=Shs};
-from_arealight_1([#we{id=Id,has_shape=HasShape}=We|Wes], Shs, St)
-    when ?IS_ANY_LIGHT(We), HasShape ->
+from_arealight_1([#we{id=Id}=We|Wes], Shs, St) when ?IS_AREA_LIGHT(We) ->
     from_arealight_1(Wes, gb_trees:update(Id, We#we{light=none}, Shs), St);
-from_arealight_1([We|Wes], Shs, St) 
-  when ?IS_ANY_LIGHT(We) ->
+from_arealight_1([We|Wes], Shs, St) when ?IS_ANY_LIGHT(We) ->
     to_arealight_1(Wes, Shs, St).
 
-
-
-%%%
-%%% Set Mode.
-%%%
-
-set_mode(Mode, St) ->
-    wings_sel:map(fun(_, We) -> We#we{mode=Mode} end, St).
-
 %%%
 %%% Convert materials to vertex colors.
 %%%
@@ -818,12 +755,11 @@ colors_to_materials(St0) ->
 			  end, St0, St0),
     St#st{mat=Mat}.
 
-colors_to_materials_1(#we{mode=vertex,fs=Ftab}=We0, St) ->
-    colors_to_materials_2(gb_trees:keys(Ftab), We0#we{mode=material}, [], St);
-colors_to_materials_1(We, St) -> {We,St}.
+colors_to_materials_1(#we{fs=Ftab}=We0, St) ->
+    colors_to_materials_2(gb_trees:keys(Ftab), We0, [], St).
 
 colors_to_materials_2([F|Fs], We, Acc, St0) ->
-    Colors = [C || [_|C] <- wings_face:vinfo_ccw(F, We)],
+    Colors = [C || C <- wings_va:face_attr(color, F, We)],
     case wings_color:average(Colors) of
 	none ->
 	    colors_to_materials_2(Fs, We, Acc, St0);
@@ -855,6 +791,15 @@ fmt_int(I) ->
 fmt_int(3, L) -> L;
 fmt_int(N, L) -> fmt_int(N+1, [$0|L]).
 
+%%%
+%%% Removing vertex attributes.
+%%%
+
+va_remove(What, St) ->
+    wings_sel:map(fun(_, We) ->
+			  wings_va:remove(What, We)
+		  end, St).
+
 %%%
 %%% The Weld command.
 %%%
@@ -865,13 +810,41 @@ weld(Ask, _) when is_atom(Ask) ->
 	    {text,1.0E-3,[{range,{1.0E-5,10.0}}]}]}],
     wings_ask:dialog(Ask, ?__(2,"Weld"), Qs,
 		     fun(Res) -> {body,{weld,Res}} end);
-weld([Tolerance], St0) ->
-    St1 = combine(St0),
-    {St2,Sel} = wings_sel:mapfold(fun(_, We, Acc) ->
-					  weld_1(Tolerance, We, Acc)
-				  end, [], St1),
-    St = wings_sel:set(vertex, Sel, St2),
-    {save_state,wings_sel:valid_sel(St)}.
+weld([Tolerance], #st{shapes=Shs0,sel=Sel0}=St0) ->
+    Unselected = gb_trees:keys(Shs0) -- [ Id || {Id,_} <- Sel0 ],
+    #st{shapes=Shs}=St1 = separate(St0),
+    Selected = gb_trees:keys(Shs) -- Unselected,
+    Sel = [ {Id, gb_sets:singleton(0)} || Id <- Selected ],
+    St = combine(St1#st{sel=Sel}),
+    weld_objects(Tolerance, gb_sets:empty(), St).
+
+%% Cycle over St until it doesn't change. This is done because matching faces
+%% that are connected by a single vertex can't be processed (as far as I've
+%% tried). So the hope is that another face pair in the shape will be connect
+%% the next time through the St, which will make the unprocessed face pair share
+%% at least one edge. All the same, not all shapes will be processed. If there
+%% only single vert mathcing face pairs, then nothing can be done. If you know
+%% how to weld faces with only one common vertex, append the code at
+%% "single vertex").
+weld_objects(Tolerance, SelAcc0, St0) ->
+    Empty = gb_sets:empty(),
+    {St1,Sel0} = wings_sel:mapfold(fun(_, We0, Acc) ->
+					  case weld_1(Tolerance, We0, Acc) of
+					    {We0,_} when SelAcc0 =:= Empty ->
+					      wings_u:error(?__(1,"Found no faces to weld."));
+					    We -> We
+					  end
+				  end, [], St0),
+	case St0 =:= St1 of
+	  true ->
+	   [{Id,_}] = Sel0,
+       St = wings_sel:set(vertex, [{Id,SelAcc0}], St1),
+       {save_state,wings_sel:valid_sel(St)};
+      false ->
+        [{_,Vs}] = Sel0,
+        SelAcc = gb_sets:union(SelAcc0,Vs),
+        weld_objects(Tolerance, SelAcc, St1)
+    end.
 
 weld_1(Tol, #we{id=Id,fs=Fs0}=We0, Acc) ->
     Fs = weld_1_list(gb_trees:keys(Fs0), Tol, We0, []),
@@ -883,7 +856,7 @@ weld_1(Tol, #we{id=Id,fs=Fs0}=We0, Acc) ->
     Part = sofs:to_external(Part1),
     case weld_2(Part, Tol, We0) of
 	We0 ->
-	    wings_u:error(?__(1,"Found no faces to weld."));
+	    {We0,[{Id,gb_sets:singleton(0)}]}; % Nothing to weld or done
 	We ->
 	    {We,[{Id,weld_selection(lists:append(Part), We0, We)}|Acc]}
     end.
@@ -911,25 +884,174 @@ weld_part([], _, We) -> We.
 
 weld_part_1(Fa, [Fb|Fs], Tol, We0, Acc) ->
     case try_weld(Fa, Fb, Tol, We0) of
-	no -> weld_part_1(Fa, Fs, Tol, We0, [Fb|Acc]);	
+	We0 -> weld_part_1(Fa, Fs, Tol, We0, [Fb|Acc]);
 	We -> weld_part(Fs++Acc, Tol, We)
     end;
 weld_part_1(_, [], Tol, We, Acc) ->
     weld_part(Acc, Tol, We).
 
 try_weld(Fa, Fb, Tol, We) ->
-    case wings_face:are_neighbors(Fa, Fb, We) of
-	true -> no;
-	false ->
-	    Na = wings_face:normal(Fa, We),
-	    Nb = wings_face:normal(Fb, We),
-	    case e3d_vec:dot(Na, Nb) of
-		Dot when Dot < -0.99 ->
-		    try_weld_1(Fa, Fb, Tol, We);
-		_Dot -> no
-	    end
+    Na = wings_face:normal(Fa, We),
+    Nb = wings_face:normal(Fb, We),
+    case e3d_vec:dot(Na, Nb) of
+      Dot when Dot < -0.99 ->
+       case wings_face:are_neighbors(Fa, Fb, We) of
+         true ->
+           case shared_edges(Fa,Fb,We) of
+             [] ->  %io:format("~p\n",["single vertex"]),
+               We;
+             CommonEs ->
+               weld_neighbours(Fa, Fb, CommonEs, Tol, We)
+           end;
+         false ->
+           try_weld_1(Fa, Fb, Tol, We)
+       end;
+     _Dot -> We
     end.
 
+shared_edges(Fa, Fb, We) ->
+    FaEs = wings_face:to_edges([Fa], We),
+    FbEs = wings_face:to_edges([Fb], We),
+    AllEs = lists:sort(FaEs ++ FbEs),
+    get_shared_edges(AllEs).
+
+get_shared_edges([E,E|Es]) ->
+    [E|get_shared_edges(Es)];
+get_shared_edges([_|Es]) ->
+    get_shared_edges(Es);
+get_shared_edges([]) ->
+    [].
+
+weld_neighbours(Fa, Fb, CommonEs, Tol, We0) ->
+    Vs0 = wings_edge:to_vertices(CommonEs,We0),
+    #we{fs=Ftab}=We1 = dissolve_edges(CommonEs,We0),
+    case gb_trees:is_defined(Fa,Ftab) of
+      true ->
+        case check_weld_neighbours(Fa, Vs0, Tol, We1) of
+          error ->
+            wings_dissolve:faces([Fa,Fb],We0);
+          Other -> Other
+        end;
+      false ->
+          % check for other face id in case both faces were deleted
+          case gb_trees:is_defined(Fb,Ftab) of
+            true ->
+              case check_weld_neighbours(Fb, Vs0, Tol, We1) of
+                error ->
+                  wings_dissolve:faces([Fa,Fb],We0);
+                Other -> Other
+              end;
+            false -> wings_dissolve:faces([Fa,Fb],We0)
+          end
+    end.
+
+check_weld_neighbours(Face, Vs0, Tol, We0) ->
+    Vs1 = wings_face:to_vertices([Face], We0),
+    if
+      Vs0 =:= Vs1 -> error;
+      true ->
+        Vs =  Vs1 -- Vs0,
+        if
+          Vs =:= [] -> error;
+          length(Vs) rem 2 =:= 0 ->
+            NVs = Vs0 -- Vs,
+            case get_vs_pairs(Vs, Tol, Face, We0) of
+              error -> error;
+              CPList ->
+                connect_and_collapse(Face, CPList, NVs, [], We0)
+            end;
+          true -> We0
+        end
+    end.
+
+get_vs_pairs(Vs, Tol, Face, We) ->
+    [{V,P}|VposList]=All = wings_util:add_vpos(Vs, We),
+    closest_pair(VposList, {V,P}, {none,Tol}, Face, All, Tol, [], We).
+
+closest_pair([{Vb,VposB}|VposList], {Va,VposA}=V, {V0,D0}, Face, All, Tol, Acc, We) ->
+    D1 = e3d_vec:dist(VposA,VposB),
+    D = if
+      D1 =< D0 ->
+        case wings_vertex:edge_through(Va,Vb,Face,We) of
+          none -> {{Vb,VposB},D1};
+          _other -> {V0,D0}
+        end;
+      true -> {V0,D0}
+    end,
+    closest_pair(VposList, V, D, Face, All, Tol, Acc, We);
+closest_pair([], {V1,_}, {{V2,_},_}, _, [_,_], _Tol, Acc, _We) ->
+    [{V1,V2}|Acc];
+closest_pair([], {V1,_}=Vp1, {{V2,_}=Vp2,_}, Face, All0, Tol, Acc, We) ->
+    [{V,P}|VposList]=All = All0 -- [Vp1,Vp2],
+    closest_pair(VposList, {V,P}, {none,Tol}, Face, All, Tol, [{V1,V2}|Acc], We);
+closest_pair(_,_,_,_,_,_,_,_) -> error.
+
+dissolve_edges([E|CommonEs],We0) ->
+    case catch wings_edge:dissolve_edge(E,We0) of
+      #we{}=We ->
+         dissolve_edges(CommonEs,We);
+       _ -> We0
+    end;
+dissolve_edges([],We) ->
+    We.
+
+connect_and_collapse(Face, [{Va,Vb}|CPList], NVs, [], We0) ->
+    {We1, NewFace} = wings_vertex:force_connect(Va, Vb, Face, We0),
+    [E] = wings_we:new_items_as_ordset(edge, We0, We1),
+    #we{fs=Ftab}=We2 = wings_collapse:collapse_edge(E, We1),
+    case gb_trees:is_defined(Face,Ftab) of
+      true ->
+        case gb_trees:is_defined(NewFace,Ftab) of
+          true ->
+            NFVs = wings_face:to_vertices(NewFace, We2) -- NVs,
+            NFCPList = get_pairs_for_this_face(CPList,NFVs),
+            FCPList = CPList -- NFCPList,
+            connect_and_collapse(NewFace, NFCPList, NVs, [{Face,FCPList}], We2);
+          false ->
+            connect_and_collapse(Face, CPList, NVs, [], We2)
+        end;
+      false ->
+        case gb_trees:is_defined(NewFace,Ftab) of
+          false -> We2;
+          true ->
+            connect_and_collapse(NewFace, CPList, NVs, [], We2)
+        end
+    end;
+connect_and_collapse(Face, [{Va,Vb}|CPList], NVs, StoredFs, We0) ->
+    {We1, NewFace} = wings_vertex:force_connect(Va, Vb, Face, We0),
+    [E] = wings_we:new_items_as_ordset(edge, We0, We1),
+    #we{fs=Ftab}=We2 = wings_collapse:collapse_edge(E, We1),
+    case gb_trees:is_defined(Face,Ftab) of
+      true ->
+        case gb_trees:is_defined(NewFace,Ftab) of
+          true ->
+            NFVs = wings_face:to_vertices(NewFace, We2) -- NVs,
+            NFCPList = get_pairs_for_this_face(CPList,NFVs),
+            FCPList = CPList -- NFCPList,
+            F = {Face,FCPList},
+            connect_and_collapse(NewFace, NFCPList, NVs, [F|StoredFs], We2);
+          false ->
+            connect_and_collapse(Face, CPList, NVs, StoredFs, We2)
+        end;
+      false ->
+        case gb_trees:is_defined(NewFace,Ftab) of
+          false ->
+            [{OldFace,OldCPList}|Rest] = StoredFs,
+            connect_and_collapse(OldFace, OldCPList, NVs, Rest, We2);
+          true ->
+            connect_and_collapse(NewFace, CPList, NVs, StoredFs, We2)
+        end
+    end.
+
+% Assume if Va is in the face then Vb is as well
+get_pairs_for_this_face([{Va,Vb}|CPList],FVs) ->
+    case lists:member(Va,FVs) of
+      true -> [{Va,Vb}|get_pairs_for_this_face(CPList,FVs)];
+      false -> get_pairs_for_this_face(CPList,FVs)
+    end;
+get_pairs_for_this_face([],_FVs) ->
+    [].
+
 try_weld_1(Fa, Fb, Tol, We0) ->
     N = wings_face:vertices(Fa, We0),
     IterA = wings_face:iterator(Fa, We0),
@@ -937,7 +1059,8 @@ try_weld_1(Fa, Fb, Tol, We0) ->
     PosA = wings_vertex:pos(Va, We0),
     IterB0 = weld_synced_iterator(N, Fb, PosA, We0),
     case weld_same_positions(N, IterA, IterB0, Tol, We0) of
-	false -> no;
+	false ->
+	    We0;
 	true ->
 	    {Vb,_,_,_} = wings_face:next_ccw(IterB0),
 	    We = wings_face_cmd:force_bridge(Fa, Va, Fb, Vb, We0),
@@ -989,12 +1112,5 @@ weld_selection([], _, _, Acc) ->
 
 set_color(Color, St) ->
     wings_sel:map(fun(_, We) ->
-			  set_color_1(Color, We)
+			  wings_va:set_body_color(Color, We)
 		  end, St).
-
-set_color_1(Color, #we{es=Etab0}=We) ->
-    Etab1 = foldl(fun({E,Rec}, A) ->
-			  [{E,Rec#edge{a=Color,b=Color}}|A]
-		  end, [], gb_trees:to_list(Etab0)),
-    Etab = gb_trees:from_orddict(reverse(Etab1)),
-    We#we{es=Etab,mode=vertex}.
