@@ -3,7 +3,7 @@
 %%
 %%     This module handles interactive commands.
 %%
-%%  Copyright (c) 2001-2008 Bjorn Gustavsson
+%%  Copyright (c) 2001-2009 Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -48,7 +48,9 @@
 
 %% Drag per object.
 -record(do,
-	{funs					%List of transformation funs.
+	{funs,					%List of transformation funs.
+	 we_funs				%List of funs that operate on
+						%  the We.
 	}).
 
 setup(Tvs, Unit, St) ->
@@ -73,7 +75,7 @@ setup(Tvs, Units, Flags, St) ->
 	{general,General} ->
 	    wings_draw:invalidate_dlists(St),
 	    break_apart_general(General);
-	_ ->
+	_ = _BR ->
 	    wings_draw:invalidate_dlists(St),
 	    break_apart(Tvs, St)
     end,
@@ -188,26 +190,28 @@ break_apart(Tvs, St) ->
 
 break_apart_1(#dlo{src_we=#we{id=Id}=We}=D0, [{Id,TvList0}|Tvs], St) ->
     TvList = mirror_constrain(TvList0, We),
-    {Vs,FunList} = combine_tvs(TvList, We),
+    {Vs,FunList,WeFuns} = combine_tvs(TvList, We),
     D1 = if
 	     ?IS_LIGHT(We) -> D0#dlo{split=none};
 	     true -> D0
 	 end,
     D = wings_draw:split(D1, Vs, St),
-    Do = #do{funs=FunList},
+    Do = #do{funs=FunList,we_funs=WeFuns},
     {D#dlo{drag=Do},Tvs};
 break_apart_1(D, Tvs, _) -> {D,Tvs}.
 
 combine_tvs(TvList, #we{vp=Vtab}) ->
-    {FunList,VecVs0} = split_tv(TvList, [], []),
+    {FunList,WeFuns,VecVs0} = split_tv(TvList, [], [], []),
     SS = sofs:from_term(VecVs0, [{vec,[vertex]}]),
     FF = sofs:relation_to_family(SS),
     FU = sofs:family_union(FF),
     VecVs1 = sofs:to_external(FU),
     Affected = foldl(fun({_,Vs}, A) -> Vs++A end, [], VecVs1),
     case insert_vtx_data(VecVs1, Vtab, []) of
-	[] -> combine_tv_1(FunList, Affected, []);
-	VecVs -> combine_tv_1(FunList, Affected, [translate_fun(VecVs)])
+	[] ->
+	    combine_tv_1(FunList, WeFuns, Affected, []);
+	VecVs ->
+	    combine_tv_1(FunList, WeFuns, Affected, [translate_fun(VecVs)])
     end.
 
 translate_fun(VecVs) ->
@@ -219,15 +223,19 @@ translate_fun(VecVs) ->
 		  end, Acc, VecVs)
     end.
 
-combine_tv_1([{Aff,Fun}|T], Aff0, FunList) ->
-    combine_tv_1(T, Aff++Aff0, [Fun|FunList]);
-combine_tv_1([], Aff, FunList) -> {Aff,FunList}.
+combine_tv_1([{Aff,Fun}|T], WeFuns, Aff0, FunList) ->
+    combine_tv_1(T, WeFuns, Aff++Aff0, [Fun|FunList]);
+combine_tv_1([], WeFuns, Aff, FunList) ->
+    {Aff,FunList,WeFuns}.
 
-split_tv([{_,F}=Fun|T], Facc, Vacc) when is_function(F) ->
-    split_tv(T, [Fun|Facc], Vacc);
-split_tv([L|T], Facc, Vacc) when is_list(L) ->
-    split_tv(T, Facc, L++Vacc);
-split_tv([], Funs, VecVs) -> {Funs,VecVs}.
+split_tv([{we,F}|T], WeFacc, Facc, Vacc) when is_function(F, 2) ->
+    split_tv(T, [F|WeFacc], Facc, Vacc);
+split_tv([{_,F}=Fun|T], WeFacc, Facc, Vacc) when is_function(F, 2) ->
+    split_tv(T, WeFacc, [Fun|Facc], Vacc);
+split_tv([L|T], WeFacc, Facc, Vacc) when is_list(L) ->
+    split_tv(T, WeFacc, Facc, L++Vacc);
+split_tv([], WeFuns, Funs, VecVs) ->
+    {Funs,WeFuns,VecVs}.
 
 insert_vtx_data([{Vec,Vs0}|VecVs], Vtab, Acc) ->
     Vs = insert_vtx_data_1(Vs0, Vtab, []),
@@ -235,68 +243,62 @@ insert_vtx_data([{Vec,Vs0}|VecVs], Vtab, Acc) ->
 insert_vtx_data([], _Vtab, Acc) -> Acc.
 
 insert_vtx_data_1([V|Vs], Vtab, Acc) ->
-    insert_vtx_data_1(Vs, Vtab, [{V,gb_trees:get(V, Vtab)}|Acc]);
+    insert_vtx_data_1(Vs, Vtab, [{V,array:get(V, Vtab)}|Acc]);
 insert_vtx_data_1([], _Vtab, Acc) -> Acc.
 
-mirror_constrain(Tvs, #we{mirror=none}) -> Tvs;
-mirror_constrain(Tvs, #we{mirror=Face,fs=Ftab}=We) ->
-    case gb_trees:is_defined(Face, Ftab) of
-	false -> Tvs;
-	true ->
-	    [V|_] = Vs = wings_face:vertices_cw(Face, We),
-	    VsSet = ordsets:from_list(Vs),
-	    N = wings_face:face_normal_cw(Vs, We),
-	    Vpos = wings_vertex:pos(V, We),
-	    mirror_constrain_1(Tvs, VsSet, {N,Vpos}, [])
-    end.
+mirror_constrain(Tvs, #we{mirror=none}) ->
+    Tvs;
+mirror_constrain(Tvs, #we{mirror=Face}=We) ->
+    M = wings_we:mirror_projection(We),
+    Vs = wings_face:vertices_cw(Face, We),
+    VsSet = ordsets:from_list(Vs),
+    mirror_constrain_1(Tvs, VsSet, M, []).
 
-mirror_constrain_1([{Vs,Tr0}=Fun|Tvs], VsSet, N, Acc) when is_function(Tr0) ->
+mirror_constrain_1([{we,Tr}=Fun|Tvs], VsSet, M, Acc) when is_function(Tr) ->
+    mirror_constrain_1(Tvs, VsSet, M, [Fun|Acc]);
+mirror_constrain_1([{Vs,Tr0}=Fun|Tvs], VsSet, M, Acc) when is_function(Tr0) ->
     case ordsets:intersection(ordsets:from_list(Vs), VsSet) of
 	[] ->
-	    mirror_constrain_1(Tvs, VsSet, N, [Fun|Acc]);
+	    mirror_constrain_1(Tvs, VsSet, M, [Fun|Acc]);
 	[_|_]=Mvs ->
-	    Tr = constrain_fun(Tr0, N, Mvs),
-	    mirror_constrain_1(Tvs, VsSet, N, [{Vs,Tr}|Acc])
+	    Tr = constrain_fun(Tr0, M, Mvs),
+	    mirror_constrain_1(Tvs, VsSet, M, [{Vs,Tr}|Acc])
     end;
-mirror_constrain_1([VecVs0|Tvs], VsSet, N, Acc) ->
+mirror_constrain_1([VecVs0|Tvs], VsSet, M, Acc) ->
     VecVs1 = sofs:from_term(VecVs0, [{vec,[vertex]}]),
     VecVs2 = sofs:family_to_relation(VecVs1),
     VecVs3 = sofs:to_external(VecVs2),
-    VecVs = mirror_constrain_2(VecVs3, VsSet, N, []),
-    mirror_constrain_1(Tvs, VsSet, N, [VecVs|Acc]);
+    VecVs = mirror_constrain_2(VecVs3, VsSet, M, []),
+    mirror_constrain_1(Tvs, VsSet, M, [VecVs|Acc]);
 mirror_constrain_1([], _, _, Acc) -> Acc.
 
-mirror_constrain_2([{Vec0,V}|T], VsSet, {N,_}=Plane, Acc) ->
+mirror_constrain_2([{Vec0,V}|T], VsSet, M, Acc) ->
     case member(V, VsSet) of
 	false ->
-	    mirror_constrain_2(T, VsSet, Plane, [{Vec0,[V]}|Acc]);
+	    mirror_constrain_2(T, VsSet, M, [{Vec0,[V]}|Acc]);
 	true ->
-	    Vec = project_vector(Vec0, N),
-	    mirror_constrain_2(T, VsSet, Plane, [{Vec,[V]}|Acc])
+	    Vec = e3d_mat:mul_vector(M, Vec0),
+	    mirror_constrain_2(T, VsSet, M, [{Vec,[V]}|Acc])
     end;
 mirror_constrain_2([], _, _, Acc) -> Acc.
 
-project_vector(Vec, Plane) ->
-    e3d_vec:sub(Vec, e3d_vec:mul(Plane, e3d_vec:dot(Vec, Plane))).
-
-constrain_fun(Tr0, Plane, Vs) ->
+constrain_fun(Tr0, M, Vs) ->
     fun(Cmd, Arg) ->
 	    case Tr0(Cmd, Arg) of
-		Tr when is_function(Tr) ->
-		    constrain_fun(Tr, Plane, Vs);
+		Tr when is_function(Tr, 2) ->
+		    constrain_fun(Tr, M, Vs);
 		List ->
-		    constrain_vs(List, Vs, Plane, [])
+		    constrain_vs(List, Vs, M, [])
 	    end
     end.
 
-constrain_vs([{V,Pos0}=H|T], Vs, {N,Point}=Plane, Acc) ->
+constrain_vs([{V,Pos0}=H|T], Vs, M, Acc) ->
     case member(V, Vs) of
-	false -> constrain_vs(T, Vs, Plane, [H|Acc]);
+	false ->
+	    constrain_vs(T, Vs, M, [H|Acc]);
 	true ->
-	    ToPoint = e3d_vec:sub(Point, Pos0),
-	    Dot = e3d_vec:dot(ToPoint, N),
-	    Pos = e3d_vec:add_prod(Pos0, N, Dot),
-	    constrain_vs(T, Vs, Plane, [{V,Pos}|Acc])
+	    Pos = e3d_mat:mul_point(M, Pos0),
+	    constrain_vs(T, Vs, M, [{V,Pos}|Acc])
     end;
 constrain_vs([], _, _, Acc) -> Acc.
 
@@ -306,11 +308,13 @@ insert_matrix(Tvs) ->
 			 insert_matrix_fun(D, Data, Id)
 		 end, sort(Tvs)).
 
-insert_matrix_fun(#dlo{work=Work,edges=Edges,sel=Sel,src_sel=SrcSel,
-		       src_we=#we{id=Id}=We,mirror=M,proxy_data=Pd},
+insert_matrix_fun(#dlo{work=Work,edges=Edges,sel=Sel,src_sel=SrcSel,smooth=Smooth,
+		       src_we=#we{id=Id}=We,mirror=M,
+		       proxy=Proxy,proxy_data=Pd,open=Open},
 		  [{Id,Tr}|Tvs], Matrix) ->
     {#dlo{work=Work,edges=Edges,sel=Sel,drag={matrix,Tr,Matrix,Matrix},
-	  src_we=We,src_sel=SrcSel,mirror=M,proxy_data=Pd},Tvs};
+	  src_we=We,src_sel=SrcSel,mirror=M,smooth=Smooth,
+	  proxy=Proxy,proxy_data=Pd,open=Open},Tvs};
 insert_matrix_fun(D, Tvs, _) -> {D,Tvs}.
 
 break_apart_general(Tvs) ->
@@ -604,7 +608,7 @@ ungrab(#drag{x=Ox,y=Oy}) ->
 
 invalidate_fun(#dlo{drag=none}=D, _) -> D;
 invalidate_fun(#dlo{src_we=We}=D, _) ->
-    wings_draw:abort_split(D#dlo{src_we=We#we{es=none}}).
+    wings_draw:abort_split(D#dlo{src_we=We#we{es=array:new()}}).
 
 numeric_input(Drag0) ->
     {_,X,Y} = wings_wm:local_mouse_state(),
@@ -692,11 +696,11 @@ view_changed(#drag{flags=Flags}=Drag0) ->
 	    case member(keep_drag,Flags) of
 	      true ->
 	        wings_dl:map(fun view_changed_fun/2, []),
-	        {_,X,Y} = sdl_mouse:getMouseState(),
+	        {_,X,Y} = wings_io:get_mouse_state(),
 	        Drag0#drag{x=X,y=Y};
 	      false ->
 	    wings_dl:map(fun view_changed_fun/2, []),
-	    {_,X,Y} = sdl_mouse:getMouseState(),
+	    {_,X,Y} = wings_io:get_mouse_state(),
 	        Drag0#drag{x=X,y=Y,xs=0,ys=0,zs=0,fp=0,psum=[0,0,0,0]}
 	    end
     end.
@@ -1068,7 +1072,9 @@ motion_update_fun(#dlo{drag={matrix,Trans,Matrix0,_}}=D, Move) ->
     D#dlo{drag={matrix,Trans,Matrix0,Matrix}};
 motion_update_fun(#dlo{drag={general,Fun}}=D, Move) ->
     Fun(Move, D);
-motion_update_fun(#dlo{drag=#do{funs=Tv}}=D, Move) ->
+motion_update_fun(#dlo{drag=#do{funs=Tv,we_funs=WeFuns},src_we=We0}=D0, Move) ->
+    We = foldl(fun(WeFun, W) -> WeFun(W, Move) end, We0, WeFuns),
+    D = D0#dlo{src_we=We},
     Vtab = foldl(fun(F, A) -> F(Move, A) end, [], Tv),
     wings_draw:update_dynamic(D, Vtab);
 motion_update_fun(D, _) -> D.
@@ -1145,15 +1151,19 @@ normalize(Move, #drag{mode_fun=ModeFun,mode_data=ModeData,
     St#st{shapes=Shs}.
 
 normalize_fun(#dlo{drag=none}=D, _Move, Shs) -> {D,Shs};
-normalize_fun(#dlo{drag={matrix,_,_,_},transparent=#we{id=Id}=We}=D,
-	      _Move, Shs0) when ?IS_LIGHT(We) ->
+normalize_fun(#dlo{drag={matrix,_,_,_},transparent=#we{id=Id}=We,
+		   proxy_data=PD}=D0, _Move, Shs0) when ?IS_LIGHT(We) ->
     Shs = gb_trees:update(Id, We, Shs0),
-    {D#dlo{work=none,drag=none,src_we=We,transparent=false},Shs};
-normalize_fun(#dlo{drag={matrix,_,_,Matrix},src_we=#we{id=Id}=We0}=D0,
+    D = D0#dlo{work=none,smooth=none,drag=none,src_we=We,transparent=false,
+	   proxy_data=wings_proxy:invalidate(PD, dl)},
+    {wings_draw:changed_we(D, D),Shs};
+normalize_fun(#dlo{drag={matrix,_,_,Matrix},src_we=#we{id=Id}=We0,
+		   proxy_data=PD}=D0,
 	      _Move, Shs0) ->
     We = wings_we:transform_vs(Matrix, We0),
     Shs = gb_trees:update(Id, We, Shs0),
-    D = D0#dlo{work=none,edges=none,sel=none,drag=none,src_we=We,mirror=none},
+    D = D0#dlo{work=none,smooth=none,edges=none,sel=none,drag=none,src_we=We,
+	       mirror=none,proxy_data=wings_proxy:invalidate(PD, dl)},
     {wings_draw:changed_we(D, D),Shs};
 normalize_fun(#dlo{drag={general,Fun},src_we=#we{id=Id}=We}=D0, Move, Shs) ->
     D1 = Fun({finish,Move}, D0),
