@@ -3,7 +3,7 @@
 %%
 %%     This module contains most edge command and edge utility functions.
 %%
-%%  Copyright (c) 2001-2008 Bjorn Gustavsson.
+%%  Copyright (c) 2001-2009 Bjorn Gustavsson.
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -46,7 +46,7 @@ to_vertices(Edges, #we{es=Etab}) ->
     to_vertices(gb_sets:to_list(Edges), Etab, []).
 
 to_vertices([E|Es], Etab, Acc) ->
-    #edge{vs=Va,ve=Vb} = gb_trees:get(E, Etab),
+    #edge{vs=Va,ve=Vb} = array:get(E, Etab),
     to_vertices(Es, Etab, [Va,Vb|Acc]);
 to_vertices([], _Etab, Acc) -> ordsets:from_list(Acc).
 
@@ -60,7 +60,7 @@ from_faces(Faces, We) ->
 cut(Edge, 2, We) ->
     fast_cut(Edge, default, We);
 cut(Edge, N, #we{es=Etab}=We) ->
-    #edge{vs=Va,ve=Vb} = gb_trees:get(Edge, Etab),
+    #edge{vs=Va,ve=Vb} = array:get(Edge, Etab),
     PosA = wings_vertex:pos(Va, We),
     PosB = wings_vertex:pos(Vb, We),
     Vec = e3d_vec:mul(e3d_vec:sub(PosB, PosA), 1/N),
@@ -73,84 +73,81 @@ cut_1(N, Edge, Pos0, Vec, We0) ->
     {We,NewE} = fast_cut(Edge, Pos, We0),
     cut_1(N-1, NewE, Pos, Vec, We).
 
-%% fast_cut(Edge, Position, We0) -> {We,NewVertex,NewEdge}
+%% fast_cut(Edge, Position, We0) -> {We,NewElement}
+%%      NewElement = ID for the new vertex and the new Edge
 %%  Cut an edge in two parts. Position can be given as
 %%  the atom `default', in which case the position will
 %%  be set to the midpoint of the edge.
 
 fast_cut(Edge, Pos0, We0) ->
-    {NewEdge=NewV,We} = wings_we:new_ids(1, We0),
-    #we{es=Etab0,vc=Vct0,vp=Vtab0,he=Htab0} = We,
-    Template = gb_trees:get(Edge, Etab0),
-    #edge{vs=Vstart,ve=Vend,a=ACol,b=BCol,lf=Lf,rf=Rf,
-	  ltpr=EdgeA,rtsu=EdgeB,rtpr=NextBCol} = Template,
-    VendPos = gb_trees:get(Vend, Vtab0),
-    Vct1 = gb_trees:update(Vend, NewEdge, Vct0),
+    {NewEdge=NewV,We1} = wings_we:new_ids(1, We0),
+    #we{es=Etab0,vc=Vct0,vp=Vtab0,he=Htab0} = We1,
+    Template = array:get(Edge, Etab0),
+    #edge{vs=Vstart,ve=Vend,ltpr=EdgeA,rtsu=EdgeB} = Template,
+    VendPos = array:get(Vend, Vtab0),
+    Vct1 = array:set(Vend, NewEdge, Vct0),
     VstartPos = wings_vertex:pos(Vstart, Vtab0),
     if
 	Pos0 =:= default ->
-	    NewVPos0 = e3d_vec:average([VstartPos,VendPos]);
+	    NewVPos0 = e3d_vec:average(VstartPos, VendPos);
 	true ->
 	    NewVPos0 = Pos0
     end,
     NewVPos = wings_util:share(NewVPos0),
-    Vct = gb_trees:insert(NewV, NewEdge, Vct1),
-    Vtab = gb_trees:insert(NewV, NewVPos, Vtab0),
+    Vct = array:set(NewV, NewEdge, Vct1),
+    Vtab = array:set(NewV, NewVPos, Vtab0),
 
-    %% Here we handle vertex colors/UV coordinates.
-    AColOther = get_vtx_color(EdgeA, Lf, Etab0),
-    BColOther = get_vtx_color(NextBCol, Rf, Etab0),
+    NewEdgeRec = Template#edge{vs=NewV,ltsu=Edge,rtpr=Edge},
+    Etab1 = array:set(NewEdge, NewEdgeRec, Etab0),
+    Etab2 = patch_edge(EdgeA, NewEdge, Edge, Etab1),
+    Etab3 = patch_edge(EdgeB, NewEdge, Edge, Etab2),
+    EdgeRec = Template#edge{ve=NewV,rtsu=NewEdge,ltpr=NewEdge},
+    Etab = array:set(Edge, EdgeRec, Etab3),
+    Htab = case gb_sets:is_member(Edge, Htab0) of
+	       false -> Htab0;
+	       true -> gb_sets:insert(NewEdge, Htab0)
+	   end,
+    We2 = We1#we{es=Etab,vc=Vct,vp=Vtab,he=Htab},
+
+    %% Now interpolate and set vertex attributes.
     Weight = if
-		 Pos0 == default -> 0.5;
-		 VstartPos == VendPos -> 0.5;
-		 Pos0 == VstartPos -> 0.0;
-		 Pos0 == VendPos -> 1.0;
+		 Pos0 =:= default -> 0.5;
+		 VstartPos =:= VendPos -> 0.5;
+		 Pos0 =:= VstartPos -> 0.0;
+		 Pos0 =:= VendPos -> 1.0;
 		 true ->
 		     ADist = e3d_vec:dist(Pos0, VstartPos),
 		     BDist = e3d_vec:dist(Pos0, VendPos),
 		     ADist/(ADist+BDist)
 	     end,
-    NewColA = wings_color:mix(Weight, AColOther, ACol),
-    NewColB = wings_color:mix(Weight, BCol, BColOther),
+    AttrMidLeft = wings_va:edge_attrs(Edge, left, Weight, We1),
+    AttrMidRight = wings_va:edge_attrs(Edge, right, Weight, We1),
+    AttrEndLeft = wings_va:edge_attrs(Edge, right, We1),
 
-    NewEdgeRec = Template#edge{vs=NewV,a=NewColA,ltsu=Edge,rtpr=Edge},
-    Etab1 = gb_trees:insert(NewEdge, NewEdgeRec, Etab0),
-    Etab2 = patch_edge(EdgeA, NewEdge, Edge, Etab1),
-    Etab3 = patch_edge(EdgeB, NewEdge, Edge, Etab2),
-    EdgeRec = Template#edge{ve=NewV,b=NewColB,rtsu=NewEdge,ltpr=NewEdge},
-    Etab = gb_trees:update(Edge, EdgeRec, Etab3),
+    We3 = wings_va:set_edge_attrs(Edge, right, AttrMidRight, We2),
+    We = wings_va:set_both_edge_attrs(NewEdge, AttrMidLeft, AttrEndLeft, We3),
 
-    Htab = case gb_sets:is_member(Edge, Htab0) of
-	       false -> Htab0;
-	       true -> gb_sets:insert(NewEdge, Htab0)
-	   end,
-    {We#we{es=Etab,vc=Vct,vp=Vtab,he=Htab},NewV}.
-
-get_vtx_color(Edge, Face, Etab) ->
-    case gb_trees:get(Edge, Etab) of
-	#edge{lf=Face,a=Col} -> Col;
-	#edge{rf=Face,b=Col} -> Col
-    end.
+    {We,NewV}.
 
 %% screaming_cut(Edge, Position, We0) -> {We,NewVertex,NewEdge}
 %%  Cut an edge in two parts screamlingly fast. Does not handle
-%%  vertex colors or UV coordinates are distorted.
+%%  vertex colors or UV coordinates.
 
 screaming_cut(Edge, NewVPos, We0) ->
     {NewEdge=NewV,We} = wings_we:new_ids(1, We0),
     #we{es=Etab0,vc=Vct0,vp=Vtab0,he=Htab0} = We,
-    Template = gb_trees:get(Edge, Etab0),
+    Template = array:get(Edge, Etab0),
     #edge{ve=Vend,ltpr=EdgeA,rtsu=EdgeB} = Template,
-    Vct1 = gb_trees:update(Vend, NewEdge, Vct0),
-    Vct = gb_trees:insert(NewV, NewEdge, Vct1),
-    Vtab = gb_trees:insert(NewV, NewVPos, Vtab0),
+    Vct1 = array:set(Vend, NewEdge, Vct0),
+    Vct = array:set(NewV, NewEdge, Vct1),
+    Vtab = array:set(NewV, NewVPos, Vtab0),
 
     NewEdgeRec = Template#edge{vs=NewV,ltsu=Edge,rtpr=Edge},
-    Etab1 = gb_trees:insert(NewEdge, NewEdgeRec, Etab0),
+    Etab1 = array:set(NewEdge, NewEdgeRec, Etab0),
     Etab2 = patch_edge(EdgeA, NewEdge, Edge, Etab1),
     Etab3 = patch_edge(EdgeB, NewEdge, Edge, Etab2),
     EdgeRec = Template#edge{ve=NewV,rtsu=NewEdge,ltpr=NewEdge},
-    Etab = gb_trees:update(Edge, EdgeRec, Etab3),
+    Etab = array:set(Edge, EdgeRec, Etab3),
 
     Htab = case gb_sets:is_member(Edge, Htab0) of
 	       false -> Htab0;
@@ -167,10 +164,10 @@ dissolve_edge(Edge, We) ->
 
 dissolve_edges(Edges0, We0) when is_list(Edges0) ->
     #we{es=Etab} = We1 = foldl(fun internal_dissolve_edge/2, We0, Edges0),
-    case [E || E <- Edges0, gb_trees:is_defined(E, Etab)] of
+    case [E || E <- Edges0, array:get(E, Etab) =/= undefined] of
 	Edges0 ->
 	    %% No edge was deleted in the last pass. We are done.
-	    We = wings_we:rebuild(We0#we{vc=undefined}),
+	    We = wings_we:rebuild(We0),
 	    wings_we:validate_mirror(We);
 	Edges ->
 	    dissolve_edges(Edges, We1)
@@ -179,16 +176,17 @@ dissolve_edges(Edges, We) ->
     dissolve_edges(gb_sets:to_list(Edges), We).
 
 internal_dissolve_edge(Edge, #we{es=Etab}=We0) ->
-    case gb_trees:lookup(Edge, Etab) of
-	none -> We0;
-	{value,#edge{ltpr=Same,ltsu=Same,rtpr=Same,rtsu=Same}} ->
-	    Empty = gb_trees:empty(),
-	    We0#we{vc=Empty,vp=Empty,es=Empty,fs=Empty,he=gb_sets:empty()};
-	{value,#edge{rtpr=Back,ltsu=Back}=Rec} ->
+    case array:get(Edge, Etab) of
+	undefined -> We0;
+	#edge{ltpr=Same,ltsu=Same,rtpr=Same,rtsu=Same} ->
+	    EmptyGbTree = gb_trees:empty(),
+	    Empty = array:new(),
+	    We0#we{vc=Empty,vp=Empty,es=Empty,fs=EmptyGbTree,he=gb_sets:empty()};
+	#edge{rtpr=Back,ltsu=Back}=Rec ->
 	    merge_edges(backward, Edge, Rec, We0);
-	{value,#edge{rtsu=Forward,ltpr=Forward}=Rec} ->
+	#edge{rtsu=Forward,ltpr=Forward}=Rec ->
 	    merge_edges(forward, Edge, Rec, We0);
-	{value,Rec} ->
+	Rec ->
 	    try dissolve_edge_1(Edge, Rec, We0) of
 		We -> We
 	    catch
@@ -198,17 +196,40 @@ internal_dissolve_edge(Edge, #we{es=Etab}=We0) ->
 
 %% dissolve_edge_1(Edge, EdgeRecord, We) -> We
 %%  Remove an edge and a face. If one of the faces is degenerated
-%%  (only consists of two edges), remove that one. Otherwise, it
-%%  doesn't matter which face we remove.
+%%  (only consists of two edges), remove that one. If no face is
+%%  degenerated, prefer to keep an invisible face (if an edge
+%%  bordering a hole is dissolved, we except except the hole to
+%%  expand). Otherwise, it does not matter which face we keep.
+%%
 dissolve_edge_1(Edge, #edge{lf=Remove,rf=Keep,ltpr=Same,ltsu=Same}=Rec, We) ->
     dissolve_edge_2(Edge, Remove, Keep, Rec, We);
-dissolve_edge_1(Edge, #edge{lf=Keep,rf=Remove}=Rec, We) ->
-    dissolve_edge_2(Edge, Remove, Keep, Rec, We).
+dissolve_edge_1(Edge, #edge{lf=Keep,rf=Remove,rtpr=Same,rtsu=Same}=Rec, We) ->
+    dissolve_edge_2(Edge, Remove, Keep, Rec, We);
+dissolve_edge_1(Edge, #edge{lf=Lf,rf=Rf}=Rec, We) ->
+    if
+	Lf < 0 ->
+	    %% Keep left face.
+	    if
+		Rf < 0 ->
+		    %% The right face is also hidden. (Probably unusual
+		    %% in practice.) It might also be a hole.
+		    Holes = ordsets:del_element(Rf, We#we.holes),
+		    dissolve_edge_2(Edge, Rf, Lf, Rec, We#we{holes=Holes});
+		true ->
+		    dissolve_edge_2(Edge, Rf, Lf, Rec, We)
+	    end;
+	Rf < 0 ->
+	    %% Keep the right face. Remove the (visible) left face.
+	    dissolve_edge_2(Edge, Lf, Rf, Rec, We);
+	true ->
+	    %% It does not matter which one we keep.
+	    dissolve_edge_2(Edge, Rf, Lf, Rec, We)
+    end.
 
 dissolve_edge_2(Edge, FaceRemove, FaceKeep,
-		#edge{ltpr=LP,ltsu=LS,rtpr=RP,rtsu=RS},
-		#we{fs=Ftab0,es=Etab0,he=Htab0}=We0) ->
-    %% First change face for all edges surrounding the face we will remove.
+		#edge{vs=Va,ve=Vb,ltpr=LP,ltsu=LS,rtpr=RP,rtsu=RS},
+		#we{fs=Ftab0,es=Etab0,vc=Vct0,he=Htab0}=We0) ->
+    %% First change the face for all edges surrounding the face we will remove.
     Etab1 = wings_face:fold(
 	      fun (_, E, _, IntEtab) when E =:= Edge -> IntEtab;
 		  (_, E, R, IntEtab) ->
@@ -218,9 +239,9 @@ dissolve_edge_2(Edge, FaceRemove, FaceKeep,
 			  #edge{rf=FaceRemove,lf=FaceKeep} ->
 			      throw(hole);
 			  #edge{lf=FaceRemove} ->
-			      gb_trees:update(E, R#edge{lf=FaceKeep}, IntEtab);
+			      array:set(E, R#edge{lf=FaceKeep}, IntEtab);
 			  #edge{rf=FaceRemove} ->
-			      gb_trees:update(E, R#edge{rf=FaceKeep}, IntEtab)
+			      array:set(E, R#edge{rf=FaceKeep}, IntEtab)
 		      end
 	      end, Etab0, FaceRemove, We0),
 
@@ -231,18 +252,44 @@ dissolve_edge_2(Edge, FaceRemove, FaceKeep,
     Etab5 = patch_edge(RS, LP, Edge, Etab4),
 
     %% Remove the edge.
-    Etab = gb_trees:delete(Edge, Etab5),
+    Etab = array:reset(Edge, Etab5),
     Htab = hardness(Edge, soft, Htab0),
 
-    %% Remove the face. Patch the face entry for the remaining face.
+    %% Update the incident vertex table for both vertices
+    %% to make sure they point to the correct existing edges.
+    %% 
+    %% We used to simply set the 'vc' field to 'undefined' to
+    %% force a complete rebuild of the vertex table, but that
+    %% could cause Extrude (for regions) to become slow for certain
+    %% selection shapes, as the Extrude command internally does a
+    %% collapse of one edge in a triangle face, which in turns causes
+    %% a dissolve of one of the remaining edges.
+    Vct = case Vct0 of
+	      undefined ->
+		  Vct0;
+	      _ ->
+		  %% For the vertices Va and Vb, pick one of the still existing
+		  %% edges emanating from the vertex.
+		  %%
+		  %% The edges LS ('ltsu') and RP ('rtpr') emanate from Va ('vs').
+		  %% The edges LP ('ltpr') and RS ('rtsu') emanate from Vb ('ve').
+		  Vct1 = array:set(Va, LS, Vct0),
+		  array:set(Vb, RS, Vct1)
+	  end,
+
+    %% Remove the face. Update the incident face to make sure
+    %% the face points to an existing edge.
     Ftab1 = gb_trees:delete(FaceRemove, Ftab0),
     We1 = wings_facemat:delete_face(FaceRemove, We0),
-    Ftab = gb_trees:update(FaceKeep, LP, Ftab1),
+    AnEdge = LP,
+    Ftab = gb_trees:update(FaceKeep, AnEdge, Ftab1),
 
-    %% Return result.
-    We = We1#we{es=Etab,fs=Ftab,vc=undefined,he=Htab},
-    AnEdge = gb_trees:get(FaceKeep, Ftab),
-    case gb_trees:get(AnEdge, Etab) of
+    %% Store all updated tables.
+    We = We1#we{es=Etab,fs=Ftab,vc=Vct,he=Htab},
+
+    %% If the kept face (FaceKeep) has become a two-edge face,
+    %% we must get rid of that face by dissolving one of its edges.
+    case array:get(AnEdge, Etab) of
 	#edge{lf=FaceKeep,ltpr=Same,ltsu=Same} ->
 	    internal_dissolve_edge(AnEdge, We);
 	#edge{rf=FaceKeep,rtpr=Same,rtsu=Same} ->
@@ -267,11 +314,11 @@ dissolve_isolated_vs([], We) -> We.
 %% edge table for vertices updated, we'll need to lookup
 %% all incident edges now before we have started to dissolve.
 dissolve_isolated_vs_1([V|Vs], #we{vc=Vct}=We, Acc) ->
-    case gb_trees:lookup(V, Vct) of
-	none ->
+    case array:get(V, Vct) of
+	undefined ->
 	    %% A previous pass has already removed this vertex.
 	    dissolve_isolated_vs_1(Vs, We, Acc);
-	{value,Edge} ->
+	Edge ->
 	    dissolve_isolated_vs_1(Vs, We, [{V,Edge}|Acc])
     end;
 dissolve_isolated_vs_1([], We, Vc) ->
@@ -308,23 +355,23 @@ dissolve_isolated_vs_2([], We0, Vs) ->
 %%  (after rebuilding the incident table) since there might be more
 %%  work to do. 
 dissolve_vertex(V, Edge, #we{es=Etab}=We0) ->
-    case gb_trees:lookup(Edge, Etab) of
-	{value,#edge{vs=V,ltsu=AnEdge,rtpr=AnEdge}=Rec} ->
+    case array:get(Edge, Etab) of
+	#edge{vs=V,ltsu=AnEdge,rtpr=AnEdge}=Rec ->
 	    merge_edges(backward, Edge, Rec, We0);
-	{value,#edge{ve=V,rtsu=AnEdge,ltpr=AnEdge}=Rec} ->
+	#edge{ve=V,rtsu=AnEdge,ltpr=AnEdge}=Rec ->
 	    merge_edges(forward, Edge, Rec, We0);
 
 	%% Handle the case that the incident edge is correct for
 	%% the given vertex, but the vertex is NOT isolated.
-	{value,#edge{vs=V}} -> done;
-	{value,#edge{ve=V}} -> done;
+	#edge{vs=V} -> done;
+	#edge{ve=V} -> done;
 
 	%% The incident edge is either non-existing or no longer
 	%% references the given edge. In this case, we'll need
 	%% to try dissolving the vertex again in the next
 	%% pass after the incident table has been rebuilt.
-	none -> We0;
-	{value,_} -> We0
+	undefined -> We0;
+	_ -> We0
     end.
 
 %%
@@ -334,8 +381,8 @@ dissolve_vertex(V, Edge, #we{es=Etab}=We0) ->
 %%
 
 merge_edges(Dir, Edge, Rec, #we{es=Etab}=We) ->
-    {Va,Vb,_,_,_,_,To,To} = half_edge(Dir, Rec),
-    case gb_trees:get(To, Etab) of
+    {Va,Vb,_,_,To,To} = half_edge(Dir, Rec),
+    case array:get(To, Etab) of
 	#edge{vs=Va,ve=Vb} ->
 	    del_2edge_face(Dir, Edge, Rec, To, We);
 	#edge{vs=Vb,ve=Va} ->
@@ -344,23 +391,36 @@ merge_edges(Dir, Edge, Rec, #we{es=Etab}=We) ->
 	    merge_1(Dir, Edge, Rec, To, We)
     end.
 
-merge_1(Dir, Edge, Rec, To, #we{es=Etab0,fs=Ftab0,he=Htab0}=We) ->
+merge_1(Dir, Edge, Rec, To, #we{es=Etab0,fs=Ftab0,he=Htab0}=We0) ->
     OtherDir = reverse_dir(Dir),
-    {Vkeep,Vdelete,Lf,Rf,A,B,L,R} = half_edge(OtherDir, Rec),
+    {Vkeep,Vdelete,Lf,Rf,L,R} = half_edge(OtherDir, Rec),
     Etab1 = patch_edge(L, To, Edge, Etab0),
     Etab2 = patch_edge(R, To, Edge, Etab1),
-    Etab3 = patch_half_edge(To, Vkeep, Lf, A, L, Rf, B, R, Vdelete, Etab2),
+    Etab3 = patch_half_edge(To, Vkeep, Lf, L, Rf, R, Vdelete, Etab2),
     Htab = hardness(Edge, soft, Htab0),
-    Etab = gb_trees:delete(Edge, Etab3),
+    Etab = array:reset(Edge, Etab3),
     #edge{lf=Lf,rf=Rf} = Rec,
     Ftab1 = update_face(Lf, To, Edge, Ftab0),
     Ftab = update_face(Rf, To, Edge, Ftab1),
-    merge_2(To, We#we{es=Etab,fs=Ftab,he=Htab,vc=undefined}).
+    We1 = We0#we{es=Etab,fs=Ftab,he=Htab,vc=undefined},
+    We = case {wings_va:any_attributes(We1),Dir} of
+	     {false,_} ->
+		 We1;
+	     {_,backward} ->
+		 Attr = wings_va:edge_attrs(Edge, right, We0),
+		 We2 = wings_va:set_edge_attrs(To, Rf, Attr, We1),
+		 wings_va:del_edge_attrs(Edge, We2);
+	     {_,forward} ->
+		 Attr = wings_va:edge_attrs(Edge, left, We0),
+		 We2 = wings_va:set_edge_attrs(To, Lf, Attr, We1),
+		 wings_va:del_edge_attrs(Edge, We2)
+	 end,
+    merge_2(To, We).
 
 merge_2(Edge, #we{es=Etab}=We) ->
     %% If the merged edge is part of a two-edge face, we must
     %% remove that edge too.
-    case gb_trees:get(Edge, Etab) of
+    case array:get(Edge, Etab) of
 	#edge{ltpr=Same,ltsu=Same} ->
 	    internal_dissolve_edge(Edge, We);
 	#edge{rtpr=Same,rtsu=Same} ->
@@ -375,23 +435,23 @@ update_face(Face, Edge, OldEdge, Ftab) ->
     end.
 
 del_2edge_face(Dir, EdgeA, RecA, EdgeB,
-	       #we{es=Etab0,fs=Ftab0,he=Htab0}=We) ->
-    {_,_,Lf,Rf,_,_,_,_} = half_edge(reverse_dir(Dir), RecA),
-    RecB = gb_trees:get(EdgeB, Etab0),
+	       #we{es=Etab0,fs=Ftab0,he=Htab0,holes=Holes0}=We) ->
+    {_,_,Lf,Rf,_,_} = half_edge(reverse_dir(Dir), RecA),
+    RecB = array:get(EdgeB, Etab0),
     Del = gb_sets:from_list([EdgeA,EdgeB]),
     EdgeANear = stabile_neighbor(RecA, Del),
     EdgeBNear = stabile_neighbor(RecB, Del),
     Etab1 = patch_edge(EdgeANear, EdgeBNear, EdgeA, Etab0),
     Etab2 = patch_edge(EdgeBNear, EdgeANear, EdgeB, Etab1),
-    Etab3 = gb_trees:delete(EdgeA, Etab2),
-    Etab = gb_trees:delete(EdgeB, Etab3),
+    Etab3 = array:reset(EdgeA, Etab2),
+    Etab = array:reset(EdgeB, Etab3),
 
     %% Patch hardness table.
     Htab1 = hardness(EdgeA, soft, Htab0),
     Htab = hardness(EdgeB, soft, Htab1),
 
     %% Patch the face table.
-    #edge{lf=Klf,rf=Krf} = gb_trees:get(EdgeANear, Etab),
+    #edge{lf=Klf,rf=Krf} = array:get(EdgeANear, Etab),
     KeepFaces = ordsets:from_list([Klf,Krf]),
     EdgeAFaces = ordsets:from_list([Lf,Rf]),
     [DelFace] = ordsets:subtract(EdgeAFaces, KeepFaces),
@@ -400,8 +460,12 @@ del_2edge_face(Dir, EdgeA, RecA, EdgeB,
     Ftab2 = update_face(KeepFace, EdgeANear, EdgeA, Ftab1),
     Ftab = update_face(KeepFace, EdgeBNear, EdgeB, Ftab2),
 
+    %% It is probably unusual that 2 edge face is a hole,
+    %% but better safe than sorry.
+    Holes = ordsets:del_element(DelFace, Holes0),
+
     %% Return result.
-    We#we{vc=undefined,es=Etab,fs=Ftab,he=Htab}.
+    We#we{vc=undefined,es=Etab,fs=Ftab,he=Htab,holes=Holes}.
 
 stabile_neighbor(#edge{ltpr=Ea,ltsu=Eb,rtpr=Ec,rtsu=Ed}, Del) ->
     [Edge] = foldl(fun(E, A) ->
@@ -438,14 +502,15 @@ select_region(#st{selmode=edge}=St) ->
     wings_sel:set(face, Sel, St);
 select_region(St) -> St.
 
-select_region(Edges, #we{id=Id}=We, Acc) ->
-    Part = wings_edge_loop:partition_edges(Edges, We),
+select_region(Edges0, #we{id=Id}=We, Acc) ->
+    Part = wings_edge_loop:partition_edges(Edges0, We),
+    Edges = select_region_borders(Edges0, We),
     FaceSel0 = select_region_1(Part, Edges, We, []),
     FaceSel = gb_sets:from_ordset(wings_we:visible(FaceSel0, We)),
     [{Id,FaceSel}|Acc].
 
 select_region_1([[AnEdge|_]|Ps], Edges, #we{es=Etab}=We, Acc) ->
-    #edge{lf=Lf,rf=Rf} = gb_trees:get(AnEdge, Etab),
+    #edge{lf=Lf,rf=Rf} = array:get(AnEdge, Etab),
     Left = collect_faces(Lf, Edges, We),
     Right = collect_faces(Rf, Edges, We),
 
@@ -499,6 +564,19 @@ select_region_2(P) ->
 	    Fs
     end.
 
+select_region_borders(Edges0, #we{mirror=Mirror,holes=Holes}=We) ->
+    Bs = case Mirror of
+	     none -> Holes;
+	     _ -> [Mirror|Holes]
+	 end,
+    case Bs of
+	[] ->
+	    Edges0;
+	[_|_] ->
+	    BorderEdges = wings_face:to_edges(Bs, We),
+	    gb_sets:union(gb_sets:from_list(BorderEdges), Edges0)
+    end.
+
 make_digraph(G, [Es|T]) ->
     make_digraph_1(G, Es),
     make_digraph(G, T);
@@ -564,10 +642,10 @@ select_edge_ring_decr(St) -> St.
 
 build_selection(Edges, #we{id=Id}=We, ObjAcc) ->
     Init = init_edge_ring([],unknown,Edges,We,0,[]),
-    Stops0 = lists:foldl(fun(#r{id=MyId,ls=O},S0) ->
-				 foldl(fun(E,S) -> [{E,MyId} | S] end,
-				       S0, gb_sets:to_list(O))
-			 end,[],Init),
+    Stops0 = foldl(fun(#r{id=MyId,ls=O},S0) ->
+			   gb_sets:fold(fun(E,S) -> [{E,MyId} | S] end,
+					S0, O)
+		   end,[],Init),
     Stop = gb_trees:from_orddict(lists:sort(Stops0)),
     Sel0 = grow_rings(Init,[],Stop,We,gb_sets:empty()),
     Sel = wings_we:visible_edges(Sel0, We),
@@ -666,7 +744,7 @@ replace_edge(Edge,#r{l=L,r=R,ls=O} = EI,We) ->
     end.
 
 opposing_edge(Edge, #we{es=Es}=We, Side) ->
-    #edge{lf=Left,rf=Right} = gb_trees:get(Edge, Es),
+    #edge{lf=Left,rf=Right} = array:get(Edge, Es),
     Face = case Side of
                left -> Left;
                right -> Right
@@ -678,16 +756,17 @@ opposing_edge(Edge, #we{es=Es}=We, Side) ->
     end.
 
 next_edge(Edge, Face, #we{es=Etab})->
-    case gb_trees:get(Edge, Etab) of
+    case array:get(Edge, Etab) of
         #edge{lf=Face,ltsu=NextEdge} -> NextEdge;
         #edge{rf=Face,rtsu=NextEdge} -> NextEdge
     end.
 
 incr_ring_selection(Edges, #we{id=Id}=We, ObjAcc) ->
-    [{Id,foldl(fun(Edge, EdgeAcc) ->
-		       Es = incr_from_edge(Edge, We, EdgeAcc),
-		       wings_we:visible_edges(Es, We)
-	       end, gb_sets:empty(), gb_sets:to_list(Edges))}|ObjAcc].
+    [{Id,gb_sets:fold(
+	   fun(Edge, EdgeAcc) ->
+		   Es = incr_from_edge(Edge, We, EdgeAcc),
+		   wings_we:visible_edges(Es, We)
+	   end, gb_sets:empty(), Edges)}|ObjAcc].
 
 incr_from_edge(Edge, We, Acc) ->
     Selected = gb_sets:add(Edge, Acc),
@@ -702,9 +781,10 @@ incr_from_edge(Edge, We, Acc) ->
     end.
 
 decr_ring_selection(Edges, #we{id=Id} = We, ObjAcc) ->
-    [{Id,foldl(fun(Edge, EdgeAcc) ->
-		       decr_from_edge(Edge, We, Edges, EdgeAcc)
-	       end, Edges, gb_sets:to_list(Edges))}|ObjAcc].
+    [{Id,gb_sets:fold(
+	   fun(Edge, EdgeAcc) ->
+		   decr_from_edge(Edge, We, Edges, EdgeAcc)
+	   end, Edges, Edges)}|ObjAcc].
 
 decr_from_edge(Edge, We, Orig, Acc) ->
     Left = opposing_edge(Edge, We, left),
@@ -729,26 +809,26 @@ decr_from_edge(Edge, We, Orig, Acc) ->
 reverse_dir(forward) -> backward;
 reverse_dir(backward) -> forward.
 
-half_edge(backward, #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf,a=A,b=B,ltsu=L,rtpr=R}) ->
-    {Va,Vb,Lf,Rf,A,B,L,R};
-half_edge(forward, #edge{ve=Va,vs=Vb,lf=Lf,rf=Rf,a=A,b=B,ltpr=L,rtsu=R}) ->
-    {Va,Vb,Lf,Rf,A,B,L,R}.
+half_edge(backward, #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf,ltsu=L,rtpr=R}) ->
+    {Va,Vb,Lf,Rf,L,R};
+half_edge(forward, #edge{ve=Va,vs=Vb,lf=Lf,rf=Rf,ltpr=L,rtsu=R}) ->
+    {Va,Vb,Lf,Rf,L,R}.
 
-patch_half_edge(Edge, V, FaceA, A, Ea, FaceB, B, Eb, OrigV, Etab) ->
-    New = case gb_trees:get(Edge, Etab) of
+patch_half_edge(Edge, V, FaceA, Ea, FaceB, Eb, OrigV, Etab) ->
+    New = case array:get(Edge, Etab) of
 	      #edge{vs=OrigV,lf=FaceA,rf=FaceB}=Rec ->
-		  Rec#edge{a=A,vs=V,ltsu=Ea,rtpr=Eb};
+		  Rec#edge{vs=V,ltsu=Ea,rtpr=Eb};
 	      #edge{vs=OrigV,lf=FaceB,rf=FaceA}=Rec ->
-		  Rec#edge{a=B,vs=V,ltsu=Eb,rtpr=Ea};
+		  Rec#edge{vs=V,ltsu=Eb,rtpr=Ea};
 	      #edge{ve=OrigV,lf=FaceA,rf=FaceB}=Rec ->
-		  Rec#edge{b=B,ve=V,ltpr=Ea,rtsu=Eb};
+		  Rec#edge{ve=V,ltpr=Ea,rtsu=Eb};
 	      #edge{ve=OrigV,lf=FaceB,rf=FaceA}=Rec ->
-		  Rec#edge{b=A,ve=V,ltpr=Eb,rtsu=Ea}
+		  Rec#edge{ve=V,ltpr=Eb,rtsu=Ea}
 	  end,
-    gb_trees:update(Edge, New, Etab).
+    array:set(Edge, New, Etab).
 
 patch_edge(Edge, ToEdge, OrigEdge, Etab) ->
-    New = case gb_trees:get(Edge, Etab) of
+    New = case array:get(Edge, Etab) of
 	      #edge{ltsu=OrigEdge}=R ->
 		  R#edge{ltsu=ToEdge};
 	      #edge{ltpr=OrigEdge}=R ->
@@ -758,10 +838,10 @@ patch_edge(Edge, ToEdge, OrigEdge, Etab) ->
 	      #edge{rtpr=OrigEdge}=R ->
 		  R#edge{rtpr=ToEdge}
 	  end,
-    gb_trees:update(Edge, New, Etab).
+    array:set(Edge, New, Etab).
 
 patch_edge(Edge, ToEdge, Face, OrigEdge, Etab) ->
-    New = case gb_trees:get(Edge, Etab) of
+    New = case array:get(Edge, Etab) of
 	      #edge{lf=Face,ltsu=OrigEdge}=R ->
 		  R#edge{ltsu=ToEdge};
 	      #edge{lf=Face,ltpr=OrigEdge}=R ->
@@ -771,4 +851,4 @@ patch_edge(Edge, ToEdge, Face, OrigEdge, Etab) ->
 	      #edge{rf=Face,rtpr=OrigEdge}=R ->
 		  R#edge{rtpr=ToEdge}
 	  end,
-    gb_trees:update(Edge, New, Etab).
+    array:set(Edge, New, Etab).
