@@ -45,8 +45,9 @@ menu(X, Y, St) ->
 	    {?__(10,"Dissolve"), dslv(),
 		 {?__(11,"Eliminate selected edges"), [],
 		  ?__(50,"Eliminate selected edges and remove remaining isolated verts")},[]},
-		{?__(12,"Collapse"),collapse,
-	     ?__(13,"Delete edges, replacing them with vertices")},
+		{?__(12,"Collapse"),collapse_fun(),
+	     {?__(13,"Delete edges, replacing them with vertices"), [],
+		  ?__(21,"Delete edges, replacing them with vertices and remove any newly created isolated vertices")},[]},
 	    separator,
 	    {?__(14,"Hardness"),
 	     {hardness,[{?__(15,"Soft"),soft},
@@ -65,7 +66,12 @@ dslv() ->
 	(3, _Ns) -> {edge,clean_dissolve};
 	(_, _) -> ignore
     end.
-
+collapse_fun() ->
+    fun
+	(1, _Ns) -> {edge,collapse};
+	(3, _Ns) -> {edge,clean_collapse};
+	(_, _) -> ignore
+    end.
 cut_line(#st{sel=[{_,Es}]}) ->
     case gb_sets:size(Es) of
 	1 -> cut_fun();
@@ -131,7 +137,9 @@ command(clean_dissolve, St) ->
 command(dissolve, St) ->
     {save_state,dissolve(St)};
 command(collapse, St) ->
-    {save_state,wings_collapse:collapse(St)};
+    {save_state, wings_collapse:uniform_collapse(St)};
+command(clean_collapse, St) ->
+    {save_state, wings_collapse:clean_uniform_collapse(St)};
 command({hardness,Type}, St) ->
     {save_state,hardness(Type, St)};
 command(loop_cut, St) ->
@@ -195,26 +203,9 @@ remove_nonconnectable([], _, _, Acc) -> Acc.
 
 set_color(Color, St) ->
     wings_sel:map(fun(Es, We) ->
-			  set_color_1(gb_sets:to_list(Es), Color,
-				      We#we{mode=vertex})
+			  wings_va:set_edge_color(Es, Color, We)
 		  end, St).
 
-set_color_1([E|Es], Color, #we{es=Etab0}=We) ->
-    Rec0 = #edge{vs=Va,ve=Vb,rtpr=Rp,ltpr=Lp} = gb_trees:get(E, Etab0),
-    Rec = Rec0#edge{a=Color,b=Color},
-    Etab1 = gb_trees:update(E, Rec, Etab0),
-    Etab2 = set_color_2(Rp, Va, Color, Etab1),
-    Etab = set_color_2(Lp, Vb, Color, Etab2),
-    set_color_1(Es, Color, We#we{es=Etab});
-set_color_1([], _, We) -> We.
-
-set_color_2(E, V, Color, Etab) ->
-    Rec = case gb_trees:get(E, Etab) of
-	      #edge{vs=V}=Rec0 -> Rec0#edge{a=Color};
-	      #edge{ve=V}=Rec0 -> Rec0#edge{b=Color}
-	  end,
-    gb_trees:update(E, Rec, Etab).
-
 %%%
 %%% The Cut command.
 %%%
@@ -230,10 +221,10 @@ cut(N, #st{selmode=edge}=St0) when N > 1 ->
 cut(_, St) -> St.
 
 cut_edges(Edges, N, We0) ->
-    foldl(fun(Edge, W0) ->
-		  {We,_} = wings_edge:cut(Edge, N, W0),
-		  We
-	  end, We0, gb_sets:to_list(Edges)).
+    gb_sets:fold(fun(Edge, W0) ->
+			 {We,_} = wings_edge:cut(Edge, N, W0),
+			 We
+		 end, We0, Edges).
 
 %%%
 %%% Cut at an arbitrary position.
@@ -257,9 +248,9 @@ cut_pick_error() ->
     wings_u:error(?__(1,"Only one edge can be cut at an arbitrary position.")).
 
 cut_pick_make_tvs(Edge, #we{id=Id,es=Etab,vp=Vtab,next_id=NewV}=We) ->
-    #edge{vs=Va,ve=Vb} = gb_trees:get(Edge, Etab),
-    Start = gb_trees:get(Va, Vtab),
-    End = gb_trees:get(Vb, Vtab),
+    #edge{vs=Va,ve=Vb} = array:get(Edge, Etab),
+    Start = array:get(Va, Vtab),
+    End = array:get(Vb, Vtab),
     Dir = e3d_vec:sub(End, Start),
     Char = {7,7,3,3,7,0,
 	    <<2#01111100,
@@ -276,7 +267,7 @@ cut_pick_make_tvs(Edge, #we{id=Id,es=Etab,vp=Vtab,next_id=NewV}=We) ->
 cut_pick_marker([I], D, Edge, We0, Start, Dir, Char) ->
     {X,Y,Z} = Pos = e3d_vec:add_prod(Start, Dir, I),
     {MM,PM,ViewPort} = wings_u:get_matrices(0, original),
-    {Sx,Sy,_} = glu:project(X, Y, Z, MM, PM, ViewPort),
+    {Sx,Sy,_} = wings_gl:project(X, Y, Z, MM, PM, ViewPort),
     Dl = gl:genLists(1),
     gl:newList(Dl, ?GL_COMPILE),
     gl:pushAttrib(?GL_ALL_ATTRIB_BITS),
@@ -309,13 +300,20 @@ cut_pick_marker({finish,[I]}, D0, Edge, We, Start, Dir, Char) ->
 %%%
 
 clean_dissolve(St0) ->
-    St = wings_sel:map(fun(Es, We) ->
- 				     IsolatedVs1 = wings_vertex:isolated(We),
- 				     We1 = wings_edge:dissolve_edges(Es, We),
- 				     IsolatedVs2 = wings_vertex:isolated(We1),
- 				     C = IsolatedVs2 -- IsolatedVs1,
- 				     wings_edge:dissolve_isolated_vs(C, We1)
-		       end, St0),
+    St = wings_sel:map(fun(Es, #we{es=Etab}=We) ->
+            E1 = gb_sets:size(Es),
+            E2 = wings_util:array_entries(Etab),
+            case E1 =:= E2 of
+              false ->
+                    IsolatedVs1 = wings_vertex:isolated(We),
+                    We1 = wings_edge:dissolve_edges(Es, We),
+                    IsolatedVs2 = wings_vertex:isolated(We1),
+                    C = IsolatedVs2 -- IsolatedVs1,
+                    wings_edge:dissolve_isolated_vs(C, We1);
+              true ->
+                   #we{}
+            end
+    end, St0),
     wings_sel:clear(St).
 
 %%%
@@ -323,9 +321,16 @@ clean_dissolve(St0) ->
 %%%
 
 dissolve(St0) ->
-    St = wings_sel:map(fun(Es, We) ->
-			       wings_edge:dissolve_edges(Es, We)
-		       end, St0),
+    St = wings_sel:map(fun(Es, #we{es=Etab}=We) ->
+            E1 = gb_sets:size(Es),
+            E2 = wings_util:array_entries(Etab),
+            case E1 =:= E2 of
+              false ->
+                   wings_edge:dissolve_edges(Es, We);
+              true ->
+                   #we{}
+            end
+    end, St0),
     wings_sel:clear(St).
 
 %%%
@@ -538,16 +543,16 @@ slide_part_loop(Es,We) ->
 
 slide_gather_info([Edge|Es],We=#we{es=Etab,vp=Vtab},Acc) ->
     #edge{vs=V1,ve=V2,ltpr=LP,ltsu=LS,lf=LF,rtpr=RP,rtsu=RS,rf=RF} =
-	gb_trees:get(Edge, Etab),
-    A1 = other(V1,gb_trees:get(RP, Etab)),
-    B1 = other(V1,gb_trees:get(LS, Etab)),
-    A2 = other(V2,gb_trees:get(RS, Etab)),
-    B2 = other(V2,gb_trees:get(LP, Etab)),
+	array:get(Edge, Etab),
+    A1 = other(V1, array:get(RP, Etab)),
+    B1 = other(V1, array:get(LS, Etab)),
+    A2 = other(V2, array:get(RS, Etab)),
+    B2 = other(V2, array:get(LP, Etab)),
     N1 = wings_face:normal(LF,We), N2 = wings_face:normal(RF,We),
     N = norm(average(N2,N1)),
-    V1pos = gb_trees:get(V1, Vtab),V2pos = gb_trees:get(V2, Vtab),
-    A1pos = gb_trees:get(A1, Vtab),A2pos = gb_trees:get(A2, Vtab),
-    B1pos = gb_trees:get(B1, Vtab),B2pos = gb_trees:get(B2, Vtab),
+    V1pos = array:get(V1, Vtab), V2pos = array:get(V2, Vtab),
+    A1pos = array:get(A1, Vtab), A2pos = array:get(A2, Vtab),
+    B1pos = array:get(B1, Vtab), B2pos = array:get(B2, Vtab),
     E1v1  = sub(A1pos,V1pos), E2v1 = sub(B1pos,V1pos),
     E1v2  = sub(A2pos,V2pos), E2v2 = sub(B2pos,V2pos),
     NE1v1 = norm(E1v1), NE2v1 = norm(E2v1),
@@ -558,7 +563,7 @@ slide_gather_info([Edge|Es],We=#we{es=Etab,vp=Vtab},Acc) ->
 	   V2, {V2pos,{NE1v2,len(E1v2)},{NE2v2,len(E2v2)}},
 	   E1, E2, N},
     slide_gather_info(Es, We, [{Edge,New}|Acc]);
-slide_gather_info([],_,Acc) ->
+slide_gather_info([], _, Acc) ->
     gb_trees:from_orddict(sort(Acc)).
 
 slide_dir([Edge|R],Es,Up0,Dw0,N0,Acc) ->
