@@ -20,15 +20,13 @@
 	 face_normal_cw/2,face_normal_ccw/2,
 	 good_normal/2,is_thin/2,
 	 center/2,area/2,
-	 vinfo_cw/2,vinfo_cw/3,
-	 vinfo_ccw/2,vinfo_ccw/3,
+	 mirror_matrix/2,
 	 vertices_cw/2,vertices_cw/3,
 	 vertices_ccw/2,vertices_ccw/3,
 	 vertex_positions/2,vertex_positions/3,
-	 vertex_info/2,vertex_info/3,
 	 extend_border/2,
 	 inner_edges/2,outer_edges/2,inner_outer_edges/2,
-	 fold/4,fold/5,fold_vinfo/4,fold_faces/4,
+	 fold/4,fold/5,fold_faces/4,
 	 iterator/2,skip_to_edge/2,skip_to_cw/2,skip_to_ccw/2,
 	 next_cw/1,next_ccw/1,
 	 iter2etab/1,
@@ -46,7 +44,7 @@ from_edges(Es, We) ->
     from_edges(gb_sets:to_list(Es), We).
     
 from_edges_1([E|Es], Etab, Acc) ->
-    #edge{lf=Lf,rf=Rf} = gb_trees:get(E, Etab),
+    #edge{lf=Lf,rf=Rf} = array:get(E, Etab),
     from_edges_1(Es, Etab, [Lf,Rf|Acc]);
 from_edges_1([], _, Acc) -> gb_sets:from_list(Acc).
 
@@ -101,7 +99,7 @@ face_normal_cw(Vs, Vtab) ->
     face_normal_cw(Vs, Vtab, []).
 
 face_normal_cw([V|Vs], Vtab, Acc) ->
-    face_normal_cw(Vs, Vtab, [gb_trees:get(V, Vtab)|Acc]);
+    face_normal_cw(Vs, Vtab, [array:get(V, Vtab)|Acc]);
 face_normal_cw([], _Vtab, Acc) ->
     e3d_vec:normal(Acc).
 
@@ -114,7 +112,7 @@ face_normal_ccw(Vs, Vtab) ->
     face_normal_ccw(Vs, Vtab, []).
 
 face_normal_ccw([V|Vs], Vtab, Acc) ->
-    face_normal_ccw(Vs, Vtab, [gb_trees:get(V, Vtab)|Acc]);
+    face_normal_ccw(Vs, Vtab, [array:get(V, Vtab)|Acc]);
 face_normal_ccw([], _Vtab, Acc) ->
     e3d_vec:normal(reverse(Acc)).
 
@@ -122,7 +120,7 @@ face_normal_ccw([], _Vtab, Acc) ->
 good_normal(Face, #we{vp=Vtab}=We) ->
     [Va,Vb|_] = Vpos =
 	fold(fun(V, _, _, A) ->
-		     [gb_trees:get(V, Vtab)|A]
+		     [array:get(V, Vtab)|A]
 	     end, [], Face, We),
     D = e3d_vec:sub(Va, Vb),
     good_normal(D, Vpos, Vpos).
@@ -203,45 +201,25 @@ center(Face, We) ->
 area(Face, We) ->
     #we{vp=Vtab} = We,
     Vs = wings_face:vertices_ccw(Face, We),
-    Vcoords = [gb_trees:get(V, Vtab) || V <- Vs],
+    Vcoords = [array:get(V, Vtab) || V <- Vs],
     FaceVs = lists:seq(0, length(Vs)-1),
     E3dFaces = [#e3d_face{vs=FaceVs}],
     [Area] = e3d_mesh:face_areas(E3dFaces, Vcoords),
     Area.
 
-%% Vertex info for drawing.
-
-vinfo_cw(Face, #we{fs=Ftab}=We) ->
-    Edge = gb_trees:get(Face, Ftab),
-    vinfo_cw(Face, Edge, We).
-
-vinfo_cw(Face, Edge, #we{es=Etab}) ->
-    vinfo_cw_1(Edge, Etab, Face, Edge, []).
-
-vinfo_cw_1(LastEdge, _, _, LastEdge, Acc) when Acc =/= [] -> Acc;
-vinfo_cw_1(Edge, Etab, Face, LastEdge, Acc) ->
-    case gb_trees:get(Edge, Etab) of
-	#edge{vs=V,a=Col,lf=Face,ltpr=NextEdge} ->
-	    vinfo_cw_1(NextEdge, Etab, Face, LastEdge, [[V|Col]|Acc]);
-	#edge{ve=V,b=Col,rtpr=NextEdge} ->
-	    vinfo_cw_1(NextEdge, Etab, Face, LastEdge, [[V|Col]|Acc])
-    end.
-
-vinfo_ccw(Face, #we{fs=Ftab}=We) ->
-    Edge = gb_trees:get(Face, Ftab),
-    vinfo_ccw(Face, Edge, We).
-
-vinfo_ccw(Face, Edge, #we{es=Etab}) ->
-    vinfo_ccw_1(Edge, Etab, Face, Edge, []).
-
-vinfo_ccw_1(LastEdge, _, _, LastEdge, Acc) when Acc =/= [] -> Acc;
-vinfo_ccw_1(Edge, Etab, Face, LastEdge, Acc) ->
-    case gb_trees:get(Edge, Etab) of
-	#edge{vs=V,a=Col,lf=Face,ltsu=NextEdge} ->
-	    vinfo_ccw_1(NextEdge, Etab, Face, LastEdge, [[V|Col]|Acc]);
-	#edge{ve=V,b=Col,rtsu=NextEdge} ->
-	    vinfo_ccw_1(NextEdge, Etab, Face, LastEdge, [[V|Col]|Acc])
-    end.
+%% mirror_matrix(Face, We) -> Matrix
+%%  Calculate the matrix needed to reflect a point or
+%%  vector in the plane of the face Face.
+mirror_matrix(Face, We) ->
+    VsPos = vertex_positions(Face, We),
+    N = e3d_vec:normal(VsPos),
+    Center = e3d_vec:average(VsPos),
+    RotBack = e3d_mat:rotate_to_z(N),
+    Rot = e3d_mat:transpose(RotBack),
+    Mat0 = e3d_mat:mul(e3d_mat:translate(Center), Rot),
+    Mat1 = e3d_mat:mul(Mat0, e3d_mat:scale(1.0, 1.0, -1.0)),
+    Mat = e3d_mat:mul(Mat1, RotBack),
+    e3d_mat:mul(Mat, e3d_mat:translate(e3d_vec:neg(Center))).
 
 vertices_cw(Face, #we{es=Etab,fs=Ftab}) ->
     Edge = gb_trees:get(Face, Ftab),
@@ -252,7 +230,7 @@ vertices_cw(Face, Edge, #we{es=Etab}) ->
 
 vertices_cw_1(LastEdge, _, _, LastEdge, Acc) when Acc =/= [] -> Acc;
 vertices_cw_1(Edge, Etab, Face, LastEdge, Acc) ->
-    case gb_trees:get(Edge, Etab) of
+    case array:get(Edge, Etab) of
 	#edge{vs=V,lf=Face,ltpr=NextEdge} ->
 	    vertices_cw_1(NextEdge, Etab, Face, LastEdge, [V|Acc]);
 	#edge{ve=V,rf=Face,rtpr=NextEdge} ->
@@ -268,7 +246,7 @@ vertices_ccw(Face, Edge, #we{es=Etab}) ->
 
 vertices_ccw_1(LastEdge, _, _, LastEdge, Acc) when Acc =/= [] -> Acc;
 vertices_ccw_1(Edge, Etab, Face, LastEdge, Acc) ->
-    case gb_trees:get(Edge, Etab) of
+    case array:get(Edge, Etab) of
 	#edge{vs=V,lf=Face,ltsu=NextEdge} ->
 	    vertices_ccw_1(NextEdge, Etab, Face, LastEdge, [V|Acc]);
 	#edge{ve=V,rf=Face,rtsu=NextEdge} ->
@@ -284,43 +262,28 @@ vertex_positions(Face, Edge, #we{es=Etab,vp=Vtab}) ->
 
 vertex_positions_1(LastEdge, _, _, _, LastEdge, Acc) when Acc =/= [] -> Acc;
 vertex_positions_1(Edge, Etab, Vtab, Face, LastEdge, Acc) ->
-    case gb_trees:get(Edge, Etab) of
+    case array:get(Edge, Etab) of
 	#edge{vs=V,lf=Face,ltsu=NextEdge} ->
-	    Pos = gb_trees:get(V, Vtab),
+	    Pos = array:get(V, Vtab),
 	    vertex_positions_1(NextEdge, Etab, Vtab, Face, LastEdge, [Pos|Acc]);
 	#edge{ve=V,rf=Face,rtsu=NextEdge} ->
-	    Pos = gb_trees:get(V, Vtab),
+	    Pos = array:get(V, Vtab),
 	    vertex_positions_1(NextEdge, Etab, Vtab, Face, LastEdge, [Pos|Acc])
     end.
 
-vertex_info(Face, #we{fs=Ftab}=We) ->
-    Edge = gb_trees:get(Face, Ftab),
-    vertex_info(Face, Edge, We).
-
-vertex_info(Face, Edge, #we{es=Etab}) ->
-    vertex_info_1(Edge, Etab, Face, Edge, []).
-
-vertex_info_1(LastEdge, _, _, LastEdge, Acc) when Acc =/= [] -> Acc;
-vertex_info_1(Edge, Etab, Face, LastEdge, Acc) ->
-    case gb_trees:get(Edge, Etab) of
-	#edge{a=Info,lf=Face,ltsu=NextEdge} ->
-	    vertex_info_1(NextEdge, Etab, Face, LastEdge, [Info|Acc]);
-	#edge{b=Info,rf=Face,rtsu=NextEdge} ->
-	    vertex_info_1(NextEdge, Etab, Face, LastEdge, [Info|Acc])
-    end.
-
 %% extend_border(FacesGbSet, We) -> FacesGbSet'
 %%  Extend the the given set of faces to include all faces not in the
 %%  set that share at least one edge with a face in the set.
 extend_border(Fs0, We) ->
-    foldl(fun(Face, S0) ->
-		  fold(fun(_, _, #edge{lf=Lf,rf=Rf}, S1) ->
-			       if
-				   Lf =/= Face -> gb_sets:add(Lf, S1);
-				   true -> gb_sets:add(Rf, S1)
-			       end
-		       end, S0, Face, We)
-	  end, Fs0, gb_sets:to_list(Fs0)).
+    gb_sets:fold(
+      fun(Face, S0) ->
+	      fold(fun(_, _, #edge{lf=Lf,rf=Rf}, S1) ->
+			   if
+			       Lf =/= Face -> gb_sets:add(Lf, S1);
+			       true -> gb_sets:add(Rf, S1)
+			   end
+		   end, S0, Face, We)
+      end, Fs0, Fs0).
 
 %% inner_edges(Faces, We) -> [Edge]
 %%  Given a set of faces, return all inner edges.
@@ -359,7 +322,6 @@ inner_outer_edges_1([E|T], In, Out) ->
 inner_outer_edges_1([], In, Out) ->
     {reverse(In),reverse(Out)}.
 
-
 %% Fold over all edges surrounding a face.
 
 fold(F, Acc, Face, #we{es=Etab,fs=Ftab}) ->
@@ -371,31 +333,15 @@ fold(F, Acc, Face, Edge, #we{es=Etab}) ->
 
 fold(LastEdge, _, _, Acc, _, LastEdge, done) -> Acc;
 fold(Edge, Etab, F, Acc0, Face, LastEdge, _) ->
-    case gb_trees:get(Edge, Etab) of
-	#edge{ve=V,lf=Face,ltsu=NextEdge}=E ->
+    case array:get(Edge, Etab) of
+	#edge{vs=V,lf=Face,ltsu=NextEdge}=E ->
 	    Acc = F(V, Edge, E, Acc0),
 	    fold(NextEdge, Etab, F, Acc, Face, LastEdge, done);
-	#edge{vs=V,rf=Face,rtsu=NextEdge}=E ->
+	#edge{ve=V,rf=Face,rtsu=NextEdge}=E ->
 	    Acc = F(V, Edge, E, Acc0),
 	    fold(NextEdge, Etab, F, Acc, Face, LastEdge, done)
     end.
 
-%% Fold over all edges surrounding a face.
-
-fold_vinfo(F, Acc, Face, #we{es=Etab,fs=Ftab}) ->
-    Edge = gb_trees:get(Face, Ftab),
-    fold_vinfo(F, Acc, Face, Edge, Edge, Etab, not_done).
-
-fold_vinfo(_F, Acc, _Face, LastEdge, LastEdge, _Etab, done) -> Acc;
-fold_vinfo(F, Acc0, Face, Edge, LastEdge, Etab, _) ->
-    Acc = case gb_trees:get(Edge, Etab) of
-	      #edge{vs=V,a=VInfo,lf=Face,ltsu=NextEdge} ->
-		  F(V, VInfo, Acc0);
-	      #edge{ve=V,b=VInfo,rf=Face,rtsu=NextEdge} ->
-		  F(V, VInfo, Acc0)
-	  end,
-    fold_vinfo(F, Acc, Face, NextEdge, LastEdge, Etab, done).
-
 %% Fold over a set of faces.
 
 fold_faces(F, Acc0, [Face|Faces], #we{es=Etab,fs=Ftab}=We) ->
@@ -408,7 +354,7 @@ fold_faces(F, Acc, Faces, We) ->
 
 fold_faces_1(LastEdge, _, _, Acc, _, LastEdge, done) -> Acc;
 fold_faces_1(Edge, Etab, F, Acc0, Face, LastEdge, _) ->
-    case gb_trees:get(Edge, Etab) of
+    case array:get(Edge, Etab) of
 	#edge{ve=V,lf=Face,ltsu=NextEdge}=E ->
 	    Acc = F(Face, V, Edge, E, Acc0),
 	    fold_faces_1(NextEdge, Etab, F, Acc, Face, LastEdge, done);
@@ -432,7 +378,7 @@ to_edges_raw([], _, _, Acc) -> Acc.
 
 to_edges_raw_1(LastEdge, _, Acc, _, LastEdge, done) -> Acc;
 to_edges_raw_1(Edge, Etab, Acc, Face, LastEdge, _) ->
-    case gb_trees:get(Edge, Etab) of
+    case array:get(Edge, Etab) of
 	#edge{lf=Face,ltsu=NextEdge} ->
 	    to_edges_raw_1(NextEdge, Etab, [Edge|Acc], Face, LastEdge, done);
 	#edge{rf=Face,rtsu=NextEdge} ->
@@ -470,7 +416,7 @@ iter2etab({face_iterator,_,_,Etab}) -> Etab.
 %% Return next edge clockwise.
 
 next_cw({face_iterator,Edge,Face,Etab}) ->
-    case gb_trees:get(Edge, Etab) of
+    case array:get(Edge, Etab) of
 	#edge{ve=V,lf=Face,ltsu=NextEdge}=Rec ->
 	    {V,Edge,Rec,{face_iterator,NextEdge,Face,Etab}};
 	#edge{vs=V,rf=Face,rtsu=NextEdge}=Rec ->
@@ -480,7 +426,7 @@ next_cw({face_iterator,Edge,Face,Etab}) ->
 %% Return next edge counter-clockwise.
 
 next_ccw({face_iterator,Edge,Face,Etab}) ->
-    case gb_trees:get(Edge, Etab) of
+    case array:get(Edge, Etab) of
 	#edge{ve=V,lf=Face,ltpr=NextEdge}=Rec ->
 	    {V,Edge,Rec,{face_iterator,NextEdge,Face,Etab}};
 	#edge{vs=V,rf=Face,rtpr=NextEdge}=Rec ->
@@ -496,7 +442,7 @@ delete_bad_faces(Fs, We) ->
 bad_edges([F|Fs], Ftab, Etab, Acc) ->
     case gb_trees:lookup(F, Ftab) of
 	{value,Edge} ->
-	    case gb_trees:get(Edge, Etab) of
+	    case array:get(Edge, Etab) of
 		#edge{ltpr=Same,ltsu=Same,rtpr=Same,rtsu=Same} ->
 		    erlang:error({internal_error,one_edged_face,F});
 		#edge{ltpr=Same,ltsu=Same} ->
