@@ -12,8 +12,15 @@
 %%
 
 -module(wings_face_cmd).
+
+%% User interface.
 -export([menu/3,command/2]).
--export([mirror_faces/2,set_color/2,force_bridge/5]).
+
+%% Useful functions.
+-export([mirror_faces/2,set_color/2,force_bridge/5,
+	 extrude_faces/1,extrude_region/1,extract_region/1]).
+	 % not currently used outside this module (but could be)
+	 % extract_faces/1, shell_extrude/1
 
 -include("wings.hrl").
 -import(lists, [foldl/3,reverse/1,sort/1,member/2]).
@@ -25,8 +32,8 @@ menu(X, Y, St) ->
 	    wings_menu_util:scale(St),
 	    separator,
 	    {?__(3,"Extrude"),{extrude,Dir}},
-	    {?__(4,"Extrude Region"),{extrude_region,Dir}},
-	    {?__(5,"Extract Region"),{extract_region,Dir}},
+	    {?__(35,"Extract"),{extract,Dir}},
+	    {?__(36,"Shell Extrude"),{shell_extrude,Dir}},
 	    separator,
 	    wings_menu_util:flatten(),
 	    separator,
@@ -57,6 +64,10 @@ menu(X, Y, St) ->
 	    {?__(29,"Tesselate"),{tesselate,wings_tesselation:submenu()}},
 	    separator,
 	    {?__(32,"Hide"),hide,?__(33,"Hide the selected faces")},
+	    {?__(37,"Hole"),hole_fun(),
+	     {?__(38,"Create hole"),
+	      [],
+	      ?__(39,"Remove holes adjacent to selected faces")},[]},
 	    separator] ++ wings_material:material_menu(St) ++
 	[{?__(30,"Vertex Color"),vertex_color,
 	  ?__(31,"Apply vertex colors to selected faces")}],
@@ -97,12 +108,18 @@ mirror_fun() ->
        (_, _) -> ignore
     end.
 
-command({extrude,Type}, St) ->
-    ?SLOW(extrude(Type, St));
-command({extrude_region,Type}, St) ->
-    ?SLOW(extrude_region(Type, St));
-command({extract_region,Type}, St) ->
-    extract_region(Type, St);
+hole_fun() ->
+    fun(1, _Ns) -> {face,create_hole};
+       (3, _Ns) -> {face,remove_hole};
+       (_, _Ns) -> ignore
+    end.
+
+command({extrude, Dir}, St) ->
+    ?SLOW(extrude(Dir, St));
+command({extract, Dir}, St) ->
+    ?SLOW(extract(Dir, St));
+command({shell_extrude, Dir}, St) ->
+    ?SLOW(shell_extrude(Dir, St));
 command(bump, St) ->
     ?SLOW(wings_extrude_edge:bump(St));
 command({flatten,Plane}, St) ->
@@ -148,28 +165,33 @@ command(vertex_color, St) ->
 			       set_color(Color, St)
 		       end);
 command(hide, St) ->
-    {save_state,hide_faces(St)}.
-    
+    {save_state,hide_faces(St)};
+command(create_hole, St) ->
+    {save_state,create_hole(St)};
+command(remove_hole, St) ->
+    {save_state,remove_hole(St)}.
 
 %%%
-%%% Extrude, Extrude Region, and Inset commands.
+%%% Extrude individual faces or regions.
 %%%
 
-extrude(Type, St) ->
-    wings_move:setup(Type, extrude_faces(St)).
+extrude({faces, Axis}, St) ->
+    wings_move:setup(Axis, extrude_faces(St));
+extrude({region, Axis}, St0) ->
+    St = extrude_region(St0),
+    wings_move:setup(Axis, St).
 
+%%% Extrude Faces
 extrude_faces(St) ->
     wings_sel:map(fun(Faces, We) ->
 			  wings_extrude_face:faces(Faces, We)
 		  end, St).
 
 %%% Extrude the selected regions.
+extrude_region(St) ->
+    wings_sel:map(fun extrude_region/2, St).
 
-extrude_region(Type, St0) ->
-    St = wings_sel:map(fun extrude_region_0/2, St0),
-    wings_move:setup(Type, St).
-
-extrude_region_0(Faces0, We0) ->
+extrude_region(Faces0, We0) ->
     %% We KNOW that a gb_set with fewer elements sorts before
     %% a gb_set with more elements.
     Rs = sort(wings_sel:face_regions(Faces0, We0)),
@@ -201,26 +223,81 @@ extrude_region_vmirror(OldWe, #we{mirror=Face0}=We0) ->
 	false ->
 	    Dissolve = gb_sets:insert(Face0, Dissolve0),
 	    We1 = wings_dissolve:faces(Dissolve, We0),
-	    [Face] = NewFace = wings_we:new_items_as_ordset(face, We0, We1),
-	    We = wings_facemat:assign('_hole_', NewFace, We1),
+	    [Face1] = FaceList = wings_we:new_items_as_ordset(face, We0, We1),
+	    We2 = wings_we:hide_faces(FaceList, We1),
+	    Face = -Face1 - 1,
+	    We = wings_facemat:assign(default, [Face], We2),
 	    wings_we:mirror_flatten(OldWe, We#we{mirror=Face})
     end.
 
+%%% Shell Extrude
+shell_extrude(Axis, St0) ->
+    St = shell_extrude(St0),
+    wings_move:setup(Axis, St).
+
+shell_extrude(St0) ->
+    #st{sel=Sel0}=St1 = extract_1(St0, St0),
+    #st{sel=Sel}=St2 = recreate_face_topology(St1#st{sel=sort(Sel0)}),
+    wings_sel:set(Sel, St2).
+
 %%%
-%%% The Extract Region command.
+%%% The Extract command.
 %%%
 
-extract_region(Type, St0) ->
-    St1 = wings_sel:fold(
-	    fun(Faces, We0, #st{sel=Sel0,onext=Oid}=S0) ->
-		    We = wings_dissolve:complement(Faces, We0),
-		    S = wings_shape:insert(We, extract, S0),
-		    Sel = [{Oid,Faces}|Sel0],
-		    S#st{sel=Sel}
-	    end, St0#st{sel=[]}, St0),
-    Sel = St1#st.sel,
-    St = wings_sel:set(Sel, St1),
-    wings_move:setup(Type, St).
+extract({region, Axis}, St0) ->
+    St = extract_region(St0),
+    wings_move:setup(Axis, St);
+
+extract({faces, Axis}, St0) ->
+    St = extract_faces(St0),
+    wings_move:setup(Axis, St).
+
+%% extract_faces/1 and extract_region/1 takes St0 and returns NewSt.
+extract_faces(St0) ->
+    St1 = wings_sel:map(fun(Faces, We) ->
+				wings_extrude_face:faces(Faces, We)
+			end, St0),
+    #st{sel=Sel}=St2 = extract_1(St1, St0),
+    wings_sel:set(Sel, St2).
+
+extract_region(St0) ->
+    #st{sel=Sel}=St = extract_1(St0, St0),
+    wings_sel:set(Sel, St).
+
+extract_1(St, St0) ->
+    wings_sel:fold(
+	    fun(Faces, We0, S0) ->
+		    Regions = wings_sel:face_regions(Faces, We0),
+		    extract_2(Regions, We0, S0)
+	    end, St0#st{sel=[]}, St).
+
+extract_2([Faces|Regions], We0, #st{sel=Sel0,onext=Oid}=St0) ->
+    We = wings_dissolve:complement(Faces, We0),
+    St = wings_shape:insert(We, extract, St0),
+    Sel = [{Oid,Faces}|Sel0],
+    extract_2(Regions, We0, St#st{sel=Sel});
+extract_2([],_,St) ->
+    St.
+
+recreate_face_topology(St) ->
+    wings_sel:map(fun(Fs, #we{fs=AllFs0}=We0) ->
+        AllFs = gb_sets:from_ordset(gb_trees:keys(AllFs0)),
+        Inverse = gb_sets:difference(AllFs,Fs),
+        intrude_extract(Inverse, We0)
+        end, St).
+
+intrude_extract(Faces0, #we{es=Etab,fs=Ftab,next_id=Wid}=We0) ->
+    Faces = gb_sets:to_list(Faces0),
+    RootSet0 = lists:foldl(
+         fun(F, A) ->
+             Edge = gb_trees:get(F, Ftab),
+             #edge{vs=V} = array:get(Edge, Etab),
+             [{face,F},{vertex,V}|A]
+         end, [], Faces),
+    {We1,RootSet} = wings_we:renumber(We0, Wid, RootSet0),
+    We2 = wings_we:invert_normals(We1),
+    We3 = wings_we:merge(We0, We2),
+    intrude_bridge(RootSet0, RootSet, We3).
 
 %%%
 %%% The Dissolve command.
@@ -230,10 +307,18 @@ dissolve(St0) ->
     {St,Sel} = wings_sel:mapfold(fun dissolve_sel/3, [], St0),
     wings_sel:set(Sel, St).
 
-dissolve_sel(Faces, #we{id=Id}=We0, Acc) ->
-    We = wings_dissolve:faces(Faces, We0),
-    Sel = wings_we:new_items_as_gbset(face, We0, We),
-    {We,[{Id,Sel}|Acc]}.
+dissolve_sel(Faces, #we{id=Id,fs=Ftab,holes=Holes,mirror=M}=We0, Acc) ->
+    Mirror = if M =:= none -> 0; true -> 1 end,
+    F1 = gb_sets:size(Faces) + length(Holes) + Mirror,
+    F2 = gb_trees:size(Ftab),
+    case F1 =:= F2 of
+      false ->
+        We = wings_dissolve:faces(Faces, We0),
+        Sel = wings_we:new_items_as_gbset(face, We0, We),
+        {We,[{Id,Sel}|Acc]};
+      true ->
+        {#we{},Acc}
+    end.
 
 %%%
 %%% Clean Dissolve
@@ -243,44 +328,60 @@ clean_dissolve(St0) ->
     {St,Sel} = wings_sel:mapfold(fun clean_dissolve_sel/3, [], St0),
     wings_sel:set(Sel, St).
 
-clean_dissolve_sel(Faces, #we{id=Id}=We0, Acc) ->
-    IsolatedVs1 = wings_vertex:isolated(We0),
-    We1 = wings_dissolve:faces(Faces, We0),
-    IsolatedVs2 = wings_vertex:isolated(We1),
-    C = IsolatedVs2 -- IsolatedVs1,
-    We = wings_edge:dissolve_isolated_vs(C, We1),
-    Sel = wings_we:new_items_as_gbset(face, We0, We),
-    {We,[{Id,Sel}|Acc]}.
+clean_dissolve_sel(Faces, #we{id=Id,fs=Ftab,holes=Holes,mirror=M}=We0, Acc) ->
+    Mirror = if M =:= none -> 0; true -> 1 end,
+    F1 = gb_sets:size(Faces) + length(Holes) + Mirror,
+    F2 = gb_trees:size(Ftab),
+    case F1 =:= F2 of
+      false ->
+         IsolatedVs1 = wings_vertex:isolated(We0),
+         We1 = wings_dissolve:faces(Faces, We0),
+         IsolatedVs2 = wings_vertex:isolated(We1),
+         C = IsolatedVs2 -- IsolatedVs1,
+         We = wings_edge:dissolve_isolated_vs(C, We1),
+         Sel = wings_we:new_items_as_gbset(face, We0, We),
+         {We,[{Id,Sel}|Acc]};
+      true ->
+        {#we{},Acc}
+    end.
 
 %%%
 %%% The Intrude command.
 %%%
 
 intrude(St0) ->
-    St1 = dissolve(St0),
-    {St,Sel} = wings_sel:mapfold(fun intrude/3, [], St1),
+    {St,Sel} = wings_sel:mapfold(fun intrude/3, [], St0),
     wings_move:setup(intrude, wings_sel:set(Sel, St)).
 
-intrude(Faces0, #we{id=Id,es=Etab,fs=Ftab,next_id=Wid}=We0, SelAcc) ->
-    Faces = gb_sets:to_list(Faces0),
+intrude(Faces0, We0, SelAcc) ->
+    We1 = wings_dissolve:faces(Faces0, We0),
+    Faces = wings_we:new_items_as_ordset(face, We0, We1),
+    #we{id=Id,es=Etab,fs=Ftab,next_id=Wid} = We1,
     RootSet0 = foldl(
 		 fun(F, A) ->
 			 Edge = gb_trees:get(F, Ftab),
-			 #edge{vs=V} = gb_trees:get(Edge, Etab),
+			 #edge{vs=V} = array:get(Edge, Etab),
 			 [{face,F},{vertex,V}|A]
 		 end, [], Faces),
-    {We1,RootSet} = wings_we:renumber(We0, Wid, RootSet0),
-    We2 = wings_we:invert_normals(We1),
-    We3 = wings_we:merge(We0, We2),
-    Sel0 = wings_we:new_items_as_gbset(face, We0, We3),
-    BridgeFaces = [F || {face,F} <- RootSet0 ++ RootSet],
-    Sel = gb_sets:difference(Sel0, gb_sets:from_list(BridgeFaces)),
-    We4 = intrude_bridge(RootSet0, RootSet, We3),
-    We = restore_mirror(We4, We0),
-    {We#we{mode=We0#we.mode},[{Id,Sel}|SelAcc]}.
+    {We2,RootSet} = wings_we:renumber(We1, Wid, RootSet0),
+    We3 = wings_we:invert_normals(We2),
+    We4 = wings_we:merge(We1, We3),
+    Sel0 = wings_we:new_items_as_gbset(face, We1, We4),
+    Exclude = [F || {face,F} <- RootSet0 ++ RootSet] ++ We4#we.holes,
+    Sel = gb_sets:difference(Sel0, gb_sets:from_list(Exclude)),
+    case gb_sets:is_empty(Sel) of
+	false ->
+	    We5 = intrude_bridge(RootSet0, RootSet, We4),
+	    We = restore_mirror(We5, We0),
+	    {We,[{Id,Sel}|SelAcc]};
+	true ->
+	    wings_u:error(?__(1,"Intrude does not work with all faces selected."))
+    end.
 
 restore_mirror(We, #we{mirror=none}) -> We;
-restore_mirror(We, #we{mirror=Face}) -> We#we{mirror=Face}.
+restore_mirror(We, #we{mirror=Mir}) ->
+    Face = -(Mir + 1),
+    wings_we:create_mirror(Face, We).
 
 intrude_bridge([{face,FaceA},{vertex,Va}|FsA],
 	       [{face,FaceB},{vertex,Vb}|FsB], We0) ->
@@ -309,12 +410,11 @@ mirror(St0) ->
     St = wings_sel:map(fun mirror_faces/2, St0),
     wings_sel:clear(St).
 
-mirror_faces(Faces, #we{mode=Mode}=We0) when is_list(Faces) ->
+mirror_faces(Faces, We0) when is_list(Faces) ->
     OrigWe = wings_we:invert_normals(We0),
-    We = foldl(fun(Face, WeAcc) ->
-		       mirror_face(Face, OrigWe, WeAcc)
-	       end, We0, Faces),
-    We#we{mode=Mode};
+    foldl(fun(Face, WeAcc) ->
+		  mirror_face(Face, OrigWe, WeAcc)
+	  end, We0, Faces);
 mirror_faces(Faces, We) ->
     mirror_faces(gb_sets:to_list(Faces), We).
 
@@ -334,20 +434,9 @@ mirror_face(Face, #we{fs=Ftab}=OrigWe, #we{next_id=Id}=We0) ->
     N = wings_face:vertices(Face, We),
     mirror_weld(N, IterA, Face, IterB, FaceNew, We, We).
 
-mirror_vs(Face, #we{vp=Vtab0}=We) ->
-    Normal = wings_face:normal(Face, We),
-    Center = wings_face:center(Face, We),
-    Vtab1 = foldl(fun(Vtx, A) ->
-			  mirror_move_vs(Vtx, Normal, Center, A)
-		  end, [], gb_trees:to_list(Vtab0)),
-    Vtab = gb_trees:from_orddict(reverse(Vtab1)),
-    We#we{vp=Vtab}.
-
-mirror_move_vs({V,Pos0}, PlaneNormal, Center, A) ->
-    ToCenter = e3d_vec:sub(Center, Pos0),
-    Dot = e3d_vec:dot(ToCenter, PlaneNormal),
-    Pos = wings_util:share(e3d_vec:add_prod(Pos0, PlaneNormal, 2.0*Dot)),
-    [{V,Pos}|A].
+mirror_vs(Face, We) ->
+    MirrorMatrix = wings_face:mirror_matrix(Face, We),
+    wings_we:transform_vs(MirrorMatrix, We).
 
 mirror_weld(0, _IterA0, FaceA, _IterB0, FaceB, _WeOrig, #we{fs=Ftab0}=We0) ->
     Ftab1 = gb_trees:delete(FaceA, Ftab0),
@@ -359,23 +448,15 @@ mirror_weld(N, IterA0, FaceA, IterB0, FaceB, WeOrig, We0) ->
     %% surrounding FaceB.
     {_,EdgeA,RecA0,IterA} = wings_face:next_cw(IterA0),
     {_,EdgeB,RecB0,IterB} = wings_face:next_ccw(IterB0),
-    Col = case RecB0 of
-	      #edge{lf=FaceB,b=Col0} -> Col0;
-	      #edge{rf=FaceB,a=Col0} -> Col0
-	  end,
-    RecA1 = case RecA0 of
-		#edge{lf=FaceA}=R -> R#edge{a=Col};
-		#edge{rf=FaceA}=R -> R#edge{b=Col}
-	    end,
     RecB = turn_edge(RecB0),
-    {RecA,Pred,Succ} =
-	case RecA1 of
+    {RecA,NewFace,Pred,Succ} =
+	case RecA0 of
 	    #edge{lf=FaceA} ->
-		update_edge(RecA1, RecB,
+		update_edge(RecA0, RecB,
 			    #edge.lf, #edge.ltpr, #edge.ltsu,
 			    #edge.rtpr, #edge.rtsu);
 	    #edge{rf=FaceA} ->
-		update_edge(RecA1, RecB,
+		update_edge(RecA0, RecB,
 			    #edge.rf, #edge.rtpr, #edge.rtsu,
 			    #edge.ltpr, #edge.ltsu)
 	    end,
@@ -387,7 +468,7 @@ mirror_weld(N, IterA0, FaceA, IterB0, FaceB, WeOrig, We0) ->
 		   #edge{rf=FaceB,rtpr=D0,rtsu=D1} -> [D0,D1]
 	      end,
     {Etab1,Htab} = delete_edges(DelEdges, Etab0, Htab0),
-    Etab2 = gb_trees:update(EdgeA, RecA, Etab1),
+    Etab2 = array:set(EdgeA, RecA, Etab1),
     Etab3 = cond_patch_edge(Pred, EdgeA, EdgeB, Etab2),
     Etab4 = cond_patch_edge(Succ, EdgeA, EdgeB, Etab3),
 
@@ -398,12 +479,17 @@ mirror_weld(N, IterA0, FaceA, IterB0, FaceB, WeOrig, We0) ->
     Etab = replace_vertex(VendB, VendA, WeOrig, Etab5),
 
     %% Update face table
-    Ftab1 = wings_face:patch_face(wings_face:other(FaceA, RecA1),
+    Ftab1 = wings_face:patch_face(wings_face:other(FaceA, RecA0),
 				  EdgeA, Ftab0),
     Ftab = wings_face:patch_face(wings_face:other(FaceB, RecB), EdgeA, Ftab1),
 
     %% Next edge.
-    We = We0#we{es=Etab,fs=Ftab,vc=undefined,he=Htab},
+    We1 = We0#we{es=Etab,fs=Ftab,vc=undefined,he=Htab},
+
+    %% Update vertex attributes.
+    Attrs = wings_va:edge_attrs(EdgeB, {other,FaceB}, WeOrig),
+    We = wings_va:set_edge_attrs(EdgeA, NewFace, Attrs, We1),
+
     mirror_weld(N-1, IterA, FaceA, IterB, FaceB, WeOrig, We).
 
 update_edge(New0, Old, FaceP, PrP, SuP, OPrP, OSuP) ->
@@ -415,20 +501,21 @@ update_edge(New0, Old, FaceP, PrP, SuP, OPrP, OSuP) ->
 	       {Succ,Succ} -> New1;
 	       {Succ,_} -> setelement(SuP, New1, Succ)
 	   end,
-    New = setelement(FaceP, New2, element(FaceP, Old)),
-    {New,Pred,Succ}.
+    NewFace = element(FaceP, Old),
+    New = setelement(FaceP, New2, NewFace),
+    {New,NewFace,Pred,Succ}.
 
 cond_patch_edge(Edge, New, Orig, Etab) ->
-    case gb_trees:is_defined(Edge, Etab) of
+    case array:get(Edge, Etab) =/= undefined of
 	true -> wings_edge:patch_edge(Edge, New, Orig, Etab);
 	false -> Etab
     end.
 
 delete_edges(Edges, Etab0, Htab0) ->
     foldl(fun(Edge, {Et0,Ht0}=Acc) ->
-		  case gb_trees:is_defined(Edge, Et0) of
+		  case array:get(Edge, Et0) =/= undefined of
 		      true ->
-			  Et = gb_trees:delete(Edge, Et0),
+			  Et = array:reset(Edge, Et0),
 			  Ht = wings_edge:hardness(Edge, soft, Ht0),
 			  {Et,Ht};
 		      false -> Acc
@@ -442,11 +529,11 @@ turn_edge(Rec) ->
 replace_vertex(Old, New, We, Etab0) ->
     wings_vertex:fold(
       fun(Edge, _, _, Et0) ->
-	      case gb_trees:lookup(Edge, Et0) of
-		  {value,#edge{vs=Old}=Rec} ->
-		      gb_trees:update(Edge, Rec#edge{vs=New}, Et0);
-		  {value,#edge{ve=Old}=Rec} ->
-		      gb_trees:update(Edge, Rec#edge{ve=New}, Et0);
+	      case array:get(Edge, Et0) of
+		  #edge{vs=Old}=Rec ->
+		      array:set(Edge, Rec#edge{vs=New}, Et0);
+		  #edge{ve=Old}=Rec ->
+		      array:set(Edge, Rec#edge{ve=New}, Et0);
 		  _Other -> Et0			%Deleted or already modified.
 	      end
       end, Etab0, Old, We).
@@ -494,9 +581,9 @@ do_flatten(Faces, PlaneNormal, We) ->
     wings_vertex:flatten(Vs, PlaneNormal, Center, We).
 
 do_flatten_normal(Faces, Center, We) ->
-    N0 = foldl(fun(Face, A) ->
-		       [wings_face:normal(Face, We)|A]
-	       end, [], gb_sets:to_list(Faces)),
+    N0 = gb_sets:fold(fun(Face, A) ->
+			      [wings_face:normal(Face, We)|A]
+		      end, [], Faces),
     N = e3d_vec:norm(e3d_vec:add(N0)),
     Vs = wings_face:to_vertices(Faces, We),
     wings_vertex:flatten(Vs, N, Center, We).
@@ -536,9 +623,11 @@ all_edges(Faces, We) ->
 		{[],[]}, Faces, We),
     {ordsets:from_list(Vs),ordsets:from_list(Es)}.
 
-smooth_connect(Vs, Faces0, #we{mirror=Mirror}=We0) ->
+smooth_connect(Vs, Faces0, #we{mirror=Mirror,holes=Holes0}=We0) ->
     Faces1 = ordsets:add_element(Mirror, Faces0),
-    Faces = sofs:from_external(Faces1, [face]),
+    Faces2 = sofs:from_external(Faces1, [face]),
+    Holes = sofs:from_external(Holes0, [face]),
+    Faces = sofs:union(Faces2, Holes),
     FaceVs0 = wings_vertex:per_face(Vs, We0),
     FaceVs1 = sofs:from_external(FaceVs0, [{face,[vertex]}]),
     FaceVs2 = sofs:drestriction(FaceVs1, Faces),
@@ -547,13 +636,8 @@ smooth_connect(Vs, Faces0, #we{mirror=Mirror}=We0) ->
     wings_we:hide_faces(Hide, We).
 
 smooth_connect_0([{Face,Vs}|Fvs], Hide0, We0) ->
-    case wings_facemat:face(Face, We0) of
-	'_hole_' ->
-	    smooth_connect_0(Fvs, Hide0, We0);
-	_ ->
-	    {We,Hide} = smooth_connect_1(Face, Vs, Hide0, We0),
-	    smooth_connect_0(Fvs, Hide, We)
-    end;
+    {We,Hide} = smooth_connect_1(Face, Vs, Hide0, We0),
+    smooth_connect_0(Fvs, Hide, We);
 smooth_connect_0([], Hide, We) -> {We,Hide}.
 
 smooth_connect_1(Face, [V], Hide, We) ->
@@ -598,13 +682,10 @@ smooth_connect_3(Va, Vb, Face, Hide, We0) ->
 bridge(#st{shapes=Shapes0,sel=[{IdA,FacesA},{IdB,FacesB}]}=St0) ->
     case {gb_sets:to_list(FacesA),gb_sets:to_list(FacesB)} of
 	{[FA],[FB0]} ->
-	    #we{next_id=Id}=WeA0 = gb_trees:get(IdA, Shapes0),
+	    #we{next_id=Id}=WeA = gb_trees:get(IdA, Shapes0),
 	    #we{}=WeB0 = gb_trees:get(IdB, Shapes0),
-	    {WeB1,[{face,FB}]} = wings_we:renumber(WeB0, Id, [{face,FB0}]),
-	    Mode = unify_modes(WeA0, WeB1),
-	    WeA = bridge_null_uvs(Mode, WeA0),
-	    WeB = bridge_null_uvs(Mode, WeB1),
-	    We = (wings_we:merge(WeA, WeB))#we{mode=Mode},
+	    {WeB,[{face,FB}]} = wings_we:renumber(WeB0, Id, [{face,FB0}]),
+	    We = wings_we:merge(WeA, WeB),
 	    Shapes1 = gb_trees:delete(IdB, Shapes0),
 	    Shapes = gb_trees:update(IdA, We, Shapes1),
 	    Sel = [{IdA,gb_sets:from_list([FA,FB])}],
@@ -627,21 +708,6 @@ bridge(#st{shapes=Shapes0,sel=[{Id,Faces}]}=St) ->
 bridge(_St) ->
     bridge_error().
 
-unify_modes(#we{mode=Mode}, #we{mode=Mode}) -> Mode;
-unify_modes(_, _) ->
-    wings_u:error(?__(1,
-			 "An object with vertex colors cannot be bridged with an object with materials.")).
-
-bridge_null_uvs(Mode, #we{mode=Mode}=We) -> We;
-bridge_null_uvs(uv, #we{es=Etab0}=We) ->
-    Etab = bridge_null_uvs_1(gb_trees:to_list(Etab0), {0.0,0.0}, []),
-    We#we{es=Etab}.
-
-bridge_null_uvs_1([{E,Rec}|Es], UV, Acc) ->
-    bridge_null_uvs_1(Es, UV, [{E,Rec#edge{a=UV,b=UV}}|Acc]);
-bridge_null_uvs_1([], _, Acc) ->
-    gb_trees:from_orddict(reverse(Acc)).
-
 bridge(FaceA, FaceB, #we{vp=Vtab}=We) ->
     VsA = wings_face:vertices_ccw(FaceA, We),
     VsB = wings_face:vertices_ccw(FaceB, We),
@@ -678,7 +744,7 @@ try_bridge(0, _Len, _Va, _FaceA, _IterA, _Vb, _FaceB, _IterB, _, _, {_,We}) ->
     We;
 try_bridge(N, Len, Va0, FaceA, IterA0, Vb, FaceB, IterB, Ids, We0,
 	   {EdgeSum0,_}=Best0) ->
-    We = do_bridge(Len, Va0, FaceA, IterA0, Vb, FaceB, IterB, Ids, We0),
+    We = do_bridge(Len, Va0, FaceA, IterA0, Vb, FaceB, IterB, Ids, We0, We0),
     Best = case sum_edge_lens(Len, Ids, We, 0) of
 		  Min when Min < EdgeSum0 -> {Min,We};
 	       _ -> Best0
@@ -691,9 +757,9 @@ try_bridge(N, Len, Va0, FaceA, IterA0, Vb, FaceB, IterB, Ids, We0,
 sum_edge_lens(0, _Ids, _We, Sum) -> Sum;
 sum_edge_lens(N, Ids0, #we{es=Etab,vp=Vtab}=We, Sum) ->
     Edge = wings_we:id(0, Ids0),
-    #edge{vs=Va,ve=Vb} = gb_trees:get(Edge, Etab),
-    VaPos = gb_trees:get(Va, Vtab),
-    VbPos = gb_trees:get(Vb, Vtab),
+    #edge{vs=Va,ve=Vb} = array:get(Edge, Etab),
+    VaPos = array:get(Va, Vtab),
+    VbPos = array:get(Vb, Vtab),
     Dist = e3d_vec:dist(VaPos, VbPos),
     Ids = wings_we:bump_id(Ids0),
     sum_edge_lens(N-1, Ids, We, Sum + Dist).
@@ -703,67 +769,73 @@ force_bridge(FaceA, Va, FaceB, Vb, We0) ->
     {Ids,We} = wings_we:new_wrap_range(Len, 2, We0),
     IterA = wings_face:skip_to_cw(Va, wings_face:iterator(FaceA, We)),
     IterB = wings_face:skip_to_ccw(Vb, wings_face:iterator(FaceB, We)),
-    do_bridge(Len, Va, FaceA, IterA, Vb, FaceB, IterB, Ids, We).
+    do_bridge(Len, Va, FaceA, IterA, Vb, FaceB, IterB, Ids, We, We).
 
-do_bridge(0, _Va, FaceA, _IterA, _Vb, FaceB, _IterB, _, #we{fs=Ftab0}=We) ->
+do_bridge(0, _Va, FaceA, _IterA, _Vb, FaceB, _IterB, _, _,
+	  #we{fs=Ftab0,holes=Holes0}=We) ->
     Ftab1 = gb_trees:delete(FaceA, Ftab0),
     Ftab = gb_trees:delete(FaceB, Ftab1),
-    wings_facemat:delete_faces([FaceA,FaceB], We#we{fs=Ftab});
-do_bridge(N, Va0, FaceA, IterA0, Vb0, FaceB, IterB0, Ids0, We0) ->
+    Holes = ordsets:subtract(Holes0, ordsets:from_list([FaceA,FaceB])),
+    wings_facemat:delete_faces([FaceA,FaceB], We#we{fs=Ftab,holes=Holes});
+do_bridge(N, Va0, FaceA, IterA0, Vb0, FaceB, IterB0, Ids0, OrigWe, We0) ->
     #we{es=Etab0,fs=Ftab0} = We0,
     NewEdge = wings_we:id(2, Ids0),
     RightFace = wings_we:id(3, Ids0),
     RightEdge = wings_we:id(4, Ids0),
     
     {_,EdgeA,RecA0,IterA} = wings_face:next_cw(IterA0),
+    ColA = wings_va:edge_attrs(EdgeA, {other,FaceA}, We0),
     RecA = case RecA0 of
-	       #edge{b=ColA,lf=FaceA,rf=OfA,rtpr=ColEdgeA} ->
-		   ColA0 = bridge_color(ColEdgeA, OfA, IterA),
-		   RecA0#edge{a=ColA0,lf=RightFace,
-			      ltpr=NewEdge,ltsu=RightEdge};
-	       #edge{a=ColA,rf=FaceA,lf=OfA,ltpr=ColEdgeA} ->
-		   ColA0 = bridge_color(ColEdgeA, OfA, IterA),
-		   RecA0#edge{b=ColA0,rf=RightFace,
-			      rtpr=NewEdge,rtsu=RightEdge}
+	       #edge{lf=FaceA,rf=OfA,rtpr=ColEdgeA} ->
+		   ColA0 = wings_va:edge_attrs(ColEdgeA, OfA, OrigWe),
+		   RecA0#edge{lf=RightFace,ltpr=NewEdge,ltsu=RightEdge};
+	       #edge{rf=FaceA,lf=OfA,ltpr=ColEdgeA} ->
+		   ColA0 = wings_va:edge_attrs(ColEdgeA, OfA, OrigWe),
+		   RecA0#edge{rf=RightFace,rtpr=NewEdge,rtsu=RightEdge}
 	   end,
-    Etab1 = gb_trees:update(EdgeA, RecA, Etab0),
+    Etab1 = array:set(EdgeA, RecA, Etab0),
 
     {_,EdgeB,RecB0,IterB} = wings_face:next_ccw(IterB0),
+    ColB = wings_va:edge_attrs(EdgeB, {other,FaceB}, We0),
     RecB = case RecB0 of
-	       #edge{b=ColB,lf=FaceB,rf=OfB,rtpr=ColEdgeB} ->
-		   ColB0 = bridge_color(ColEdgeB, OfB, IterA),
-		   RecB0#edge{a=ColB0,lf=RightFace,
-			      ltpr=RightEdge,ltsu=NewEdge};
-	       #edge{a=ColB,rf=FaceB,lf=OfB,ltpr=ColEdgeB} ->
-		   ColB0 = bridge_color(ColEdgeB, OfB, IterA),
-		   RecB0#edge{b=ColB0,rf=RightFace,
-			      rtpr=RightEdge,rtsu=NewEdge}
+	       #edge{lf=FaceB,rf=OfB,rtpr=ColEdgeB} ->
+		   ColB0 = wings_va:edge_attrs(ColEdgeB, OfB, OrigWe),
+		   RecB0#edge{lf=RightFace,ltpr=RightEdge,ltsu=NewEdge};
+	       #edge{rf=FaceB,lf=OfB,ltpr=ColEdgeB} ->
+		   ColB0 = wings_va:edge_attrs(ColEdgeB, OfB, OrigWe),
+		   RecB0#edge{rf=RightFace,rtpr=RightEdge,rtsu=NewEdge}
 	   end,
-    Etab2 = gb_trees:update(EdgeB, RecB, Etab1),
+    Etab2 = array:set(EdgeB, RecB, Etab1),
 
     RightRec0 = get_edge(RightEdge, Etab0),
-    RightRec = RightRec0#edge{a=ColB,lf=RightFace,ltpr=EdgeA,ltsu=EdgeB},
-    Etab3 = gb_trees:enter(RightEdge, RightRec, Etab2),
+    RightRec = RightRec0#edge{lf=RightFace,ltpr=EdgeA,ltsu=EdgeB},
+    Etab3 = array:set(RightEdge, RightRec, Etab2),
     
     NewRec0 = get_edge(NewEdge, Etab0),
-    NewRec = NewRec0#edge{ve=Va0,vs=Vb0,b=ColA,
-			  rf=RightFace,rtpr=EdgeB,rtsu=EdgeA},
-    Etab = gb_trees:enter(NewEdge, NewRec, Etab3),
+    NewRec = NewRec0#edge{ve=Va0,vs=Vb0,rf=RightFace,rtpr=EdgeB,rtsu=EdgeA},
+    Etab = array:set(NewEdge, NewRec, Etab3),
 
     Mat = wings_facemat:face(FaceA, We0),
     We1 = wings_facemat:assign(Mat, [RightFace], We0),
     Ftab = gb_trees:insert(RightFace, NewEdge, Ftab0),
     
-    We = We1#we{es=Etab,fs=Ftab},
+    We2 = We1#we{es=Etab,fs=Ftab},
     Ids = wings_we:bump_id(Ids0),
     Va = wings_vertex:other(Va0, RecA0),
     Vb = wings_vertex:other(Vb0, RecB0),
-    do_bridge(N-1, Va, FaceA, IterA, Vb, FaceB, IterB, Ids, We).
+
+    %% Update vertex attributes.
+    We3 = wings_va:set_edge_attrs(EdgeA, RightFace, ColA0, We2),
+    We4 = wings_va:set_edge_attrs(EdgeB, RightFace, ColB0, We3),
+    We5 = wings_va:set_edge_attrs(RightEdge, left, ColB, We4),
+    We = wings_va:set_edge_attrs(NewEdge, right, ColA, We5),
+
+    do_bridge(N-1, Va, FaceA, IterA, Vb, FaceB, IterB, Ids, OrigWe, We).
 
 get_edge(Edge, Etab) ->
-    case gb_trees:lookup(Edge, Etab) of
-	{value,Erec} -> Erec;
-	none -> #edge{}
+    case array:get(Edge, Etab) of
+	undefined-> #edge{};
+	Erec -> Erec
     end.
 
 bridge_error() ->
@@ -772,13 +844,6 @@ bridge_error() ->
 bridge_error(Error) ->
     wings_u:error(Error).
 
-bridge_color(Edge, Face, Iter) ->
-    Etab = wings_face:iter2etab(Iter),
-    case gb_trees:get(Edge, Etab) of
-	#edge{lf=Face,a=Col} -> Col;
-	#edge{rf=Face,b=Col} -> Col
-    end.
-
 %%%
 %%% The Lift command.
 %%%
@@ -868,7 +933,7 @@ lift_from_edge(Dir, Faces, Edges, We0, Tv) ->
     end.
 
 lift_from_edge_1(Dir, [{Face,Edge}|T], #we{es=Etab}=OrigWe, We0, Tv0) ->
-    Side = case gb_trees:get(Edge, Etab) of
+    Side = case array:get(Edge, Etab) of
 	       #edge{lf=Face} -> left;
 	       #edge{rf=Face} -> right
 	   end,
@@ -878,7 +943,7 @@ lift_from_edge_1(_Dir, [], _OrigWe, We, Tv) -> {We,Tv}.
 
 lift_from_edge_2(Dir, Face, Edge, Side, #we{id=Id,es=Etab}=We0, Tv) ->
     FaceVs0 = ordsets:from_list(wings_face:vertices_ccw(Face, We0)),
-    #edge{vs=Va0,ve=Vb0} = gb_trees:get(Edge, Etab),
+    #edge{vs=Va0,ve=Vb0} = array:get(Edge, Etab),
     {Va,Ea} = lift_edge_vs(Va0, FaceVs0, We0),
     {Vb,Eb} = lift_edge_vs(Vb0, FaceVs0, We0),
     We1 = wings_collapse:collapse_edge(Ea, We0),
@@ -1134,6 +1199,49 @@ hide_faces_fun(Fs, We0) ->
 	false -> We
     end.
 
+%%%
+%%% The Hole command.
+%%%
+
+create_hole(St0) ->
+    St = wings_sel:map(fun create_hole_1/2, St0),
+    wings_sel:clear(St).
+
+create_hole_1(Fs0, #we{holes=Holes0}=We0) ->
+    %% Adjacent holes should be coalesced. The easiest way to do that
+    %% is to include existing holes in the selection when dissolving.
+    Fs = ordsets:union(gb_sets:to_list(Fs0), Holes0),
+    We1 = wings_dissolve:faces(Fs, We0#we{holes=[]}),
+    case wings_util:array_is_empty(We1#we.es) of
+	false ->
+	    Holes = wings_we:new_items_as_ordset(face, We0, We1),
+	    We2 = wings_facemat:assign(default, Holes, We1),
+	    We = wings_va:remove(all, Holes, We2),
+	    wings_we:create_holes(Holes, We);
+	true ->
+	    wings_u:error(?__(1,"A hole cannot comprise all faces in an object."))
+    end.
+
+remove_hole(St0) ->
+    St = wings_sel:map(fun remove_hole_1/2, St0),
+    wings_sel:clear(St).
+
+remove_hole_1(Fs, #we{holes=Holes0}=We) ->
+    %% Find all hidden faces adjacent to the selection.
+    Find = fun(Face, _, _, #edge{lf=Face,rf=Hole}, A) when Hole < 0 ->
+		   [Hole|A];
+	      (Face, _, _, #edge{rf=Face,lf=Hole}, A) when Hole < 0 ->
+		   [Hole|A];
+	      (_, _, _, _, A) -> A
+	   end,
+    RemoveHoles0 = wings_face:fold_faces(Find, [], Fs, We),
+    RemoveHoles1 = ordsets:from_list(RemoveHoles0),
+    RemoveHoles = ordsets:intersection(RemoveHoles1, Holes0),
+
+    %% Update the list of holes and unhide the hole faces.
+    Holes = ordsets:subtract(Holes0, RemoveHoles),
+    wings_we:show_faces(RemoveHoles, We#we{holes=Holes}).
+
 %%
 %% Common help function.
 %%
@@ -1144,7 +1252,7 @@ on_target(face, Face, We) ->
     Center = wings_vertex:center(Vs, We),
     {N,Center};
 on_target(edge, Edge, #we{es=Etab}=We) ->
-    #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf} = gb_trees:get(Edge, Etab),
+    #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf} = array:get(Edge, Etab),
     N = e3d_vec:norm(e3d_vec:add([wings_face:normal(Lf, We),
 				  wings_face:normal(Rf, We)])),
     Center = wings_vertex:center([Va,Vb], We),
@@ -1160,18 +1268,5 @@ on_target(vertex, V, We) ->
 
 set_color(Color, St) ->
     wings_sel:map(fun(Fs, We) ->
-			  set_color_1(gb_sets:to_list(Fs), Color,
-				      We#we{mode=vertex})
+			  wings_va:set_face_color(Fs, Color, We)
 		  end, St).
-
-set_color_1([F|Fs], Color, #we{es=Etab0}=We) ->
-    Etab = wings_face:fold(
-	     fun(_V, Edge, Rec0, Es) ->
-		     Rec = case Rec0 of
-			       #edge{lf=F} -> Rec0#edge{a=Color};
-			       #edge{rf=F} -> Rec0#edge{b=Color}
-			   end,
-		     gb_trees:update(Edge, Rec, Es)
-	     end, Etab0, F, We),
-    set_color_1(Fs, Color, We#we{es=Etab});
-set_color_1([], _, We) -> We.
