@@ -3,7 +3,7 @@
 %%
 %%     This module contain the functions for reading and writing .wings files.
 %%
-%%  Copyright (c) 2001-2008 Bjorn Gustavsson
+%%  Copyright (c) 2001-2009 Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -16,7 +16,7 @@
 
 -include("wings.hrl").
 -include("e3d_image.hrl").
--import(lists, [sort/1,reverse/1,foldl/3]).
+-import(lists, [sort/1,reverse/1,foldl/3,any/2,keymember/3,keyfind/3]).
 
 -define(WINGS_HEADER, "#!WINGS-1.0\r\n\032\04").
 
@@ -31,7 +31,7 @@ import_1(Name, St0) ->
     case file:read_file(Name) of
 	{ok,<<?WINGS_HEADER,Sz:32,Data/binary>>} when byte_size(Data) =:= Sz ->
 	    wings_pb:update(0.08, ?__(1,"converting binary")),
-	    case catch binary_to_term(Data) of
+	    try binary_to_term(Data) of
 		{wings,0,_Shapes} ->
                     {error, ?__(2,"Pre-0.80 Wings format no longer supported.")};
 		{wings,1,_,_,_} ->
@@ -45,6 +45,9 @@ import_1(Name, St0) ->
 		Other ->
 		    io:format("~P\n", [Other,20]),
                     {error,?__(5,"corrupt Wings file")}
+	    catch
+		error:badarg ->
+                    {error,?__(5,"corrupt Wings file")}
 	    end;
 	{ok,_Bin} ->
 	    {error,?__(6,"not a Wings file (or old Wings format)")};
@@ -52,11 +55,17 @@ import_1(Name, St0) ->
 	    {error,file:format_error(Reason)}
     end.
 
+-record(va, {color_lt=none,
+	     color_rt=none,
+	     uv_lt=none,
+	     uv_rt=none}).
+
 import_vsn2(Shapes, Materials0, Props, Dir, St0) ->
     wings_pb:update(0.10, ?__(1,"images and materials")),
     Images = import_images(Dir,Props),
     Materials1 = translate_materials(Materials0),
-    Materials  = translate_map_images(Materials1, Images),
+    Materials2 = translate_map_images(Materials1, Images),
+    Materials = translate_object_modes(Materials2, Shapes),
     {St1,NameMap0} = wings_material:add_materials(Materials, St0),
     NameMap1 = gb_trees:from_orddict(sort(NameMap0)),
     NameMap = optimize_name_map(Materials, NameMap1, []),
@@ -80,65 +89,79 @@ import_objects(Shapes, NameMap, #st{selmode=Mode,shapes=Shs0,onext=Oid0}=St) ->
 
 import_objects([Sh0|Shs], Mode, NameMap, Oid, ShAcc) ->
     {object,Name,{winged,Es,Fs,Vs,He},Props} = Sh0,
-    ObjMode = import_object_mode(Props),
-    Etab = import_edges(Es, #edge{}, 0, []),
-    FaceMat = import_face_mat(Fs, NameMap, 0, []),
+    Etab = import_edges(Es, 0, []),
+    %% The 'default' material saved in this .wings file might not
+    %% match the current default material, so it could have been
+    %% renamed (to 'default2', for instance). We must make sure
+    %% that we use the correctly named default material on faces
+    %% without explicit material.
+    DefaultMat = case gb_trees:lookup(default, NameMap) of
+		     none -> default;
+		     {value,DefaultMat0} -> DefaultMat0
+		 end,
+    FaceMat = import_face_mat(Fs, NameMap, DefaultMat, 0, []),
     Vtab = import_vs(Vs, 0, []),
     Htab = gb_sets:from_list(He),
     Perm = import_perm(Props),
     Mirror = proplists:get_value(mirror_face, Props, none),
+    Holes = proplists:get_value(holes, Props, []),
     Pst0 = proplists:get_value(plugin_states, Props, []),
     Pst = try gb_trees:from_orddict(Pst0)
 	  catch error:_ -> gb_trees:empty()
 	  end,
-    We = #we{es=Etab,vp=Vtab,he=Htab,perm=Perm,pst=Pst,
-	     id=Oid,name=Name,mode=ObjMode,mirror=Mirror,mat=FaceMat},
+    We = #we{he=Htab,perm=Perm,holes=Holes,pst=Pst,
+	     id=Oid,name=Name,mirror=Mirror,mat=FaceMat},
     HiddenFaces = proplists:get_value(num_hidden_faces, Props, 0),
-    import_objects(Shs, Mode, NameMap, Oid+1, [{HiddenFaces,We}|ShAcc]);
+    import_objects(Shs, Mode, NameMap, Oid+1, [{HiddenFaces,We,{Vtab,Etab}}|ShAcc]);
 import_objects([], _Mode, _NameMap, Oid, Objs0) ->
     %%io:format("flat_size: ~p\n", [erts_debug:flat_size(Objs0)]),
     Objs = share_list(Objs0),
     %%io:format("size: ~p\n", [erts_debug:size(Objs)]),
     {Objs,Oid}.
     
-import_edges([E|Es], Template, Edge, Acc) ->
-    Rec = import_edge(E, Template),
-    import_edges(Es, Template, Edge+1, [{Edge,Rec}|Acc]);
-import_edges([], _Template, _Edge, Acc) -> reverse(Acc).
+import_edges([[{edge,Va,Vb,Lf,Rf,Ltpr,Ltsu,Rtpr,Rtsu}]|Es], Edge, Acc) ->
+    Rec = #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf,
+		ltpr=Ltpr,ltsu=Ltsu,rtpr=Rtpr,rtsu=Rtsu},
+    EdgeData = {Rec,none},
+    import_edges(Es, Edge+1, [{Edge,EdgeData}|Acc]);
+import_edges([E|Es], Edge, Acc) ->
+    EdgeData = import_edge(E, none, #va{}),
+    import_edges(Es, Edge+1, [{Edge,EdgeData}|Acc]);
+import_edges([], _Edge, Acc) -> reverse(Acc).
 
-import_edge([{edge,Va,Vb,Lf,Rf,Ltpr,Ltsu,Rtpr,Rtsu}|T], Rec0) ->
-    Rec = Rec0#edge{vs=Va,ve=Vb,lf=Lf,rf=Rf,
-		    ltpr=Ltpr,ltsu=Ltsu,rtpr=Rtpr,rtsu=Rtsu},
-    import_edge(T, Rec);
-import_edge([{uv_lt,<<U/float,V/float>>}|T], Rec) ->
-    import_edge(T, Rec#edge{a={U,V}});
-import_edge([{uv_rt,<<U/float,V/float>>}|T], Rec) ->
-    import_edge(T, Rec#edge{b={U,V}});
-import_edge([{color_lt,<<R:32/float,G:32/float,B:32/float>>}|T], Rec) ->
-    import_edge(T, Rec#edge{a={R,G,B}});
-import_edge([{color_rt,<<R:32/float,G:32/float,B:32/float>>}|T], Rec) ->
-    import_edge(T, Rec#edge{b={R,G,B}});
-import_edge([{color,Bin}|T], Rec) ->
+import_edge([{edge,Va,Vb,Lf,Rf,Ltpr,Ltsu,Rtpr,Rtsu}|T], _, Attrs) ->
+    Rec = #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf,
+		ltpr=Ltpr,ltsu=Ltsu,rtpr=Rtpr,rtsu=Rtsu},
+    import_edge(T, Rec, Attrs);
+import_edge([{uv_lt,<<U/float,V/float>>}|T], Rec, Attrs) ->
+    import_edge(T, Rec, Attrs#va{uv_lt={U,V}});
+import_edge([{uv_rt,<<U/float,V/float>>}|T], Rec, Attrs) ->
+    import_edge(T, Rec, Attrs#va{uv_rt={U,V}});
+import_edge([{color_lt,<<R:32/float,G:32/float,B:32/float>>}|T], Rec, Attrs) ->
+    import_edge(T, Rec, Attrs#va{color_lt={R,G,B}});
+import_edge([{color_rt,<<R:32/float,G:32/float,B:32/float>>}|T], Rec, Attrs) ->
+    import_edge(T, Rec, Attrs#va{color_rt={R,G,B}});
+import_edge([{color,Bin}|T], Rec, Attrs) ->
     %% Old-style vertex colors (pre 0.98.15).
     <<R1/float,G1/float,B1/float,R2/float,G2/float,B2/float>> = Bin,
-    import_edge(T, Rec#edge{a={R1,G1,B1},b={R2,G2,B2}});
-import_edge([{uv,Bin}|T], Rec) ->
+    import_edge(T, Rec, Attrs#va{color_lt={R1,G1,B1},color_rt={R2,G2,B2}});
+import_edge([{uv,Bin}|T], Rec, Attrs) ->
     %% Old-style UV coordinates (pre 0.98.15).
     <<U1/float,V1/float,U2/float,V2/float>> = Bin,
-    import_edge(T, Rec#edge{a={U1,V1},b={U2,V2}});
-import_edge([_|T], Rec) ->
-    import_edge(T, Rec);
-import_edge([], Rec) -> Rec.
+    import_edge(T, Rec, Attrs#va{uv_lt={U1,V1},uv_rt={U2,V2}});
+import_edge([_|T], Rec, Attrs) ->
+    import_edge(T, Rec, Attrs);
+import_edge([], Rec, Attrs) -> {Rec,Attrs}.
 
-import_face_mat([F|Fs], NameMap, Face, Acc) ->
-    Mat = import_face_mat_1(F, NameMap, default),
-    import_face_mat(Fs, NameMap, Face+1, [{Face,Mat}|Acc]);
-import_face_mat([], _, _, Acc) -> reverse(Acc).
+import_face_mat([F|Fs], NameMap, Default, Face, Acc) ->
+    Mat = import_face_mat_1(F, NameMap, Default),
+    import_face_mat(Fs, NameMap, Default, Face+1, [{Face,Mat}|Acc]);
+import_face_mat([], _, _, _, Acc) -> reverse(Acc).
 
-import_face_mat_1([{material,Name}|T], NameMap, _) ->
+import_face_mat_1([{material,Name}|T], NameMap, Default) ->
     %% Silently ignore materials not found in the name map.
     Mat = case gb_trees:lookup(Name, NameMap) of
-	      none ->  default;
+	      none -> Default;
 	      {value,Other} -> Other
 	  end,
     import_face_mat_1(T, NameMap, Mat);
@@ -167,15 +190,6 @@ import_perm(Props) ->
 	_Unknown -> 0
     end.
 
-import_object_mode(Ps) ->
-    case proplists:get_value(mode, Ps, material) of
-	undefined ->
-	    io:format(?__(1,"Changed undefined mode to material\n")),
-	    material;
-	uv -> material;
-	Other -> Other
-    end.
-
 import_props([{selection,{Mode,Sel0}}|Ps], St) ->
     Sel = import_sel(Sel0, St),
     import_props(Ps, St#st{selmode=Mode,sel=Sel});
@@ -341,30 +355,52 @@ translate_map_images_2([], _, _) -> [].
 %%%
 
 share_list(Wes) ->
-    Tabs0 = [{Vtab,Etab} || {_,#we{vp=Vtab,es=Etab}} <- Wes],
+    Tabs0 = [Tabs || {_,_,{_,_}=Tabs} <- Wes],
     Floats = share_floats(Tabs0, tuple_to_list(wings_color:white())),
     Tabs = share_list_1(Tabs0, Floats, gb_trees:empty(), []),
     share_list_2(Tabs, Wes, []).
 
 share_list_1([{Vtab0,Etab0}|Ts], Floats, Tuples0, Acc) ->
     Vtab = share_vs(Vtab0, Floats, []),
-    {Etab,Tuples} = share_es(Etab0, Floats, [], Tuples0),
-    share_list_1(Ts, Floats, Tuples, [{Vtab,Etab}|Acc]);
+    {Etab,Attr,Tuples} = share_es(Etab0, Floats, [], [], Tuples0),
+    share_list_1(Ts, Floats, Tuples, [{Vtab,Etab,Attr}|Acc]);
 share_list_1([], _, _, Ts) -> reverse(Ts).
 
-share_list_2([{Vtab0,Etab0}|Ts],
-	     [{NumHidden,#we{id=Id,mat=FaceMat}=We0}|Wes], Acc) ->
-    Vtab = gb_trees:from_orddict(Vtab0),
-    Etab = gb_trees:from_orddict(Etab0),
+share_list_2([{Vtab0,Etab0,Attr}|Ts],
+	     [{NumHidden,#we{id=Id,mat=FaceMat}=We0,_}|Wes], Acc) ->
+    Vtab = array:from_orddict(Vtab0),
+    Etab = array:from_orddict(Etab0),
     We1 = wings_we:rebuild(We0#we{vp=Vtab,es=Etab,mat=default}),
     We2 = wings_facemat:assign(FaceMat, We1),
+
+    %% Hide invisible faces and set holes.
     We3 = if
-	     NumHidden =:= 0 -> We2;
-	     true ->
-		 Hidden = lists:seq(0, NumHidden-1),
-		 wings_we:hide_faces(Hidden, We2)
+	      NumHidden =:= 0 -> We2;
+	      true ->
+		  Hidden = lists:seq(0, NumHidden-1),
+		  Holes = ordsets:from_list([-F-1 || F <- We2#we.holes]),
+		  wings_we:hide_faces(Hidden, We2#we{holes=Holes})
+	  end,
+    We4 = translate_old_holes(We3),
+    We5 = validate_holes(We4),
+
+    %% Very old Wings files can have invalid mirror faces for some reason.
+    We6 = wings_we:validate_mirror(We5),
+
+    %% Set attributes (if any) for all edges.
+    We7 = foldl(fun({E,Lt,Rt}, W) ->
+			wings_va:set_both_edge_attrs(E, Lt, Rt, W)
+		end, We6, Attr),
+
+    %% At last, hide the virtual mirror face.
+    We = case We7 of
+	     #we{mirror=none} ->
+		 We7;
+	     #we{mirror=MirrorFace} ->
+		 %% Hide the virtual mirror face.
+		 We8 = wings_we:hide_faces([MirrorFace], We7),
+		 We8#we{mirror=-MirrorFace-1}
 	 end,
-    We = ensure_valid_mirror_face(We3),
     share_list_2(Ts, Wes, [{Id,We}|Acc]);
 share_list_2([], [], Wes) -> sort(Wes).
 
@@ -381,9 +417,13 @@ share_floats_1([{_,{A,B,C}}|T], Shared) ->
     share_floats_1(T, [A,B,C|Shared]);
 share_floats_1([], Shared) -> Shared.
 
-share_floats_2([{_,#edge{a=A,b=B}}|T], Shared0) ->
-    Shared1 = share_floats_3(A, Shared0),
-    Shared = share_floats_3(B, Shared1),
+share_floats_2([{_,{_,none}}|T], Shared) ->
+    share_floats_2(T, Shared);
+share_floats_2([{_,{_,#va{}=Va}}|T], Shared0) ->
+    Shared1 = share_floats_3(Va#va.color_lt, Shared0),
+    Shared2 = share_floats_3(Va#va.color_rt, Shared1),
+    Shared3 = share_floats_3(Va#va.uv_lt, Shared2),
+    Shared = share_floats_3(Va#va.uv_rt, Shared3),
     share_floats_2(T, Shared);
 share_floats_2([], Shared) -> Shared.
 
@@ -404,14 +444,21 @@ share_vs([{V,{X0,Y0,Z0}}|Vs], Floats, Acc) ->
     share_vs(Vs, Floats, [{V,{X,Y,Z}}|Acc]);
 share_vs([], _, Acc) -> reverse(Acc).
 
-share_es([{E,#edge{a=Same0,b=Same0}=Rec}|Vs], Floats, Acc, Shared0) ->
-    {Same,Shared} = share_tuple(Same0, Floats, Shared0),
-    share_es(Vs, Floats, [{E,Rec#edge{a=Same,b=Same}}|Acc], Shared);
-share_es([{E,#edge{a=A0,b=B0}=Rec}|Vs], Floats, Acc, Shared0) ->
-    {A,Shared1} = share_tuple(A0, Floats, Shared0),
-    {B,Shared} = share_tuple(B0, Floats, Shared1),
-    share_es(Vs, Floats, [{E,Rec#edge{a=A,b=B}}|Acc], Shared);
-share_es([], _, Acc, Shared) -> {reverse(Acc),Shared}.
+share_es([{E,{Rec,none}}|Vs], Floats, Acc, AttrAcc, Shared) ->
+    share_es(Vs, Floats, [{E,Rec}|Acc], AttrAcc, Shared);
+share_es([{E,{Rec,Va0}}|Vs], Floats, Acc, AttrAcc0, Shared0) ->
+    #va{color_lt=ColLt0,color_rt=ColRt0,
+	uv_lt=UvLt0,uv_rt=UvRt0} = Va0,
+    {ColLt,Shared1} = share_tuple(ColLt0, Floats, Shared0),
+    {ColRt,Shared2} = share_tuple(ColRt0, Floats, Shared1),
+    {UvLt,Shared3} = share_tuple(UvLt0, Floats, Shared2),
+    {UvRt,Shared} = share_tuple(UvRt0, Floats, Shared3),
+    LtAttr = wings_va:new_attr(ColLt, UvLt),
+    RtAttr = wings_va:new_attr(ColRt, UvRt),
+    AttrAcc = [{E,LtAttr,RtAttr}|AttrAcc0],
+    share_es(Vs, Floats, [{E,Rec}|Acc], AttrAcc, Shared);
+share_es([], _, Acc, AttrAcc, Shared) ->
+    {reverse(Acc),AttrAcc,Shared}.
 
 share_tuple({A0,B0}=Tuple0, Floats, Shared) ->
     case gb_trees:lookup(Tuple0, Shared) of
@@ -434,14 +481,6 @@ share_tuple({A0,B0,C0}=Tuple0, Floats, Shared) ->
     end;
 share_tuple(none, _, Shared) -> {none,Shared}.
 
-%% Very old Wings files can have invalid mirror faces for some reason.
-ensure_valid_mirror_face(#we{mirror=none}=We) -> We;
-ensure_valid_mirror_face(#we{mirror=Face,fs=Ftab}=We) ->
-    case gb_trees:is_defined(Face, Ftab) of
-	false -> We#we{mirror=none};
-	true -> We
-    end.
-
 %%%
 %%% Import of old materials format (up to and including wings-0.94.02).
 %%%
@@ -476,6 +515,82 @@ translate_material([], _, OpenGL, Maps) ->
     [{opengl,OpenGL},{maps,Maps}].
 
 trans({Key,{R,G,B}}, Opac) -> {Key,{R,G,B,Opac}}.
+
+%%
+%% Object modes were removed after the 1.1.7 release and
+%% replaced with information about vertex colors in the
+%% materials. At the same time the 'default' material was
+%% changed to show vertex colors for the faces it was applied
+%% to.
+%%
+%% Left alone, there would be two annoyances when loading
+%% old models:
+%%
+%% 1. Vertex colors would not be shown.
+%%
+%% 2. Since the 'default' materials do not match, the 'default'
+%%    material in the file will be renamed to 'default2' (or
+%%    something similar) and there would be a new 'default'
+%%    material.
+%%
+%% We will avoid both those annoyances by changing the 'default'
+%% material in the file so that it is more likely to match
+%% current 'default' material. We will only do this change if
+%% the file contains an implicit object mode for some object,
+%% i.e. was saved by 1.1.7 or earlier.
+%%
+translate_object_modes(Mats, Objects) ->
+    OldFile = any(fun(Obj) ->
+			  {object,_Name,_Winged,Props} = Obj,
+			  keymember(mode, 1, Props)
+		  end, Objects),
+    case OldFile of
+	false -> Mats;
+	true -> [translate_object_mode(M) || M <- Mats]
+    end.
+
+translate_object_mode({default=Name,Props0}) ->
+    OpenGL0 = proplists:get_value(opengl, Props0, []),
+    OpenGL = [{vertex_colors,set}|OpenGL0],
+    Props = [{opengl,OpenGL}|lists:keydelete(opengl, 1, Props0)],
+    {Name,Props};
+translate_object_mode(Mat) -> Mat.
+
+%%
+%% There used to be a '_hole_' material up to the 1.1.10 release.
+%% The '_hole_' material was pre-defined and was specially handled
+%% when exporting (faces having the material would not be exported)
+%% and importing (missing faces would be created and assigned the
+%% material).
+%%
+%% Translate faces with the '_hole_' material to the new type
+%% of holes introduced after the 1.1.10 release.
+%%
+translate_old_holes(#we{holes=[]}=We) ->
+    case wings_facemat:is_material_used('_hole_', We) of
+	false -> We;
+	true -> translate_old_holes_1(We)
+    end;
+translate_old_holes(We) -> We.
+
+translate_old_holes_1(#we{fs=Ftab}=We0) ->
+    MatFaces = wings_facemat:mat_faces(gb_trees:to_list(Ftab), We0),
+    {_,Holes0} = keyfind('_hole_', 1, MatFaces),
+    Holes = [F || {F,_} <- Holes0],
+    We1 = wings_dissolve:faces(Holes, We0),
+    NewHoleFaces = wings_we:new_items_as_ordset(face, We0, We1),
+    We = wings_facemat:assign(default, NewHoleFaces, We1),
+    HiddenNewHoleFaces = ordsets:from_list([-F-1 || F <- NewHoleFaces]),
+    wings_we:hide_faces(NewHoleFaces, We#we{holes=HiddenNewHoleFaces}).
+
+%% validate_holes(We0) -> We
+%%  Remove any invalid entries in We#we.holes. Ideally, there should
+%%  be no invalid entries, but because of bugs there could be.
+%%
+validate_holes(#we{fs=Ftab,holes=Holes0}=We) ->
+    %% Only keep faces that exist and are invisible.
+    Holes = [F || F <- Holes0, F < 0, gb_trees:is_defined(F, Ftab)],
+    We#we{holes=Holes}.
     
 %%%
 %%% Save a Wings file (in version 2).
@@ -484,7 +599,7 @@ trans({Key,{R,G,B}}, Opac) -> {Key,{R,G,B,Opac}}.
 export(Name, St0) ->
     wings_pb:start( ?__(1,"saving")),
     wings_pb:update(0.01, ?__(2,"lights")),
-    Lights = wings_light:export(St0),
+    Lights = wings_light:export_bc(St0),
     Materials = case wings_pref:get_value(save_unused_materials) of
         true -> 
             #st{mat=Mat} = St0,
@@ -496,8 +611,10 @@ export(Name, St0) ->
     remove_lights(St0),
     Sel0 = collect_sel(St),
     wings_pb:update(0.65, ?__(3,"renumbering")),
-    {Shs1,Sel} = renumber(gb_trees:to_list(Shs0), Sel0, 0, [], []),
-    Shs = foldl(fun shape/2, [], Shs1),
+    Shs1 = [{Id,show_mirror_face(We)} ||
+	       {Id,We} <- gb_trees:to_list(Shs0)],
+    {Shs2,Sel} = renumber(Shs1, Sel0, 0, [], []),
+    Shs = foldl(fun shape/2, [], Shs2),
     wings_pb:update(0.98, ?__(4,"objects")),
     Props0 = export_props(Sel),
     Props1 = case Lights of
@@ -546,6 +663,12 @@ collect_sel_groups([{{Mode,Name},Sel}|Gs], St, Acc0) ->
     collect_sel_groups(Gs, St, Acc);
 collect_sel_groups([], _, Acc) -> Acc.
 
+show_mirror_face(#we{mirror=none}=We) -> We;
+show_mirror_face(#we{mirror=Face}=We) ->
+    %% The mirror face should not be hidden in a .wings file.
+    %% (For compatibility with previous versions.)
+    wings_we:show_faces([Face], We#we{mirror=-Face-1}).
+
 renumber([{Id,We0}|Shs], [{Id,Root0}|Sel], NewId, WeAcc, RootAcc) ->
     Hidden = wings_we:num_hidden(We0),
     {We,Root} = wings_we:renumber(We0, 0, Root0),
@@ -592,21 +715,22 @@ write_file(Name, Bin) ->
 	{error,Reason} -> {error,file:format_error(Reason)}
     end.
 
-shape({Hidden,#we{mode=ObjMode,name=Name,vp=Vs0,es=Es0,he=Htab,pst=Pst}=We}, Acc) ->
-    Vs1 = foldl(fun export_vertex/2, [], gb_trees:values(Vs0)),
+shape({Hidden,#we{name=Name,vp=Vs0,es=Es0,he=Htab,pst=Pst}=We}, Acc) ->
+    Vs1 = foldl(fun export_vertex/2, [], array:sparse_to_list(Vs0)),
     Vs = reverse(Vs1),
     UvFaces = gb_sets:from_ordset(wings_we:uv_mapped_faces(We)),
-    Es1 = foldl(fun(E, A) ->
-			export_edge(E, UvFaces, A)
-		end, [], gb_trees:values(Es0)),
+    Es1 = array:sparse_foldl(fun(E, Rec, A) ->
+				     export_edge(E, Rec, UvFaces, We, A)
+			     end, [], Es0),
     Es = reverse(Es1),
     Fs1 = foldl(fun export_face/2, [], wings_facemat:all(We)),
     Fs = reverse(Fs1),
     He = gb_sets:to_list(Htab),
-    Props0 = [{mode,ObjMode}|export_perm(We)],
+    Props0 = export_perm(We),
     Props1 = hidden_faces(Hidden, Props0),
     Props2 = mirror(We, Props1),
-    Props  = export_pst(Pst,Props2),
+    Props3 = export_holes(We, Props2),
+    Props  = export_pst(Pst, Props3),
     [{object,Name,{winged,Es,Fs,Vs,He},Props}|Acc].
 
 mirror(#we{mirror=none}, Props) -> Props;
@@ -615,6 +739,9 @@ mirror(#we{mirror=Face}, Props) -> [{mirror_face,Face}|Props].
 hidden_faces(0, Props) -> Props;
 hidden_faces(N, Props) -> [{num_hidden_faces,N}|Props].
 
+export_holes(#we{holes=[]}, Props) -> Props;
+export_holes(#we{holes=Holes}, Props) -> [{holes,Holes}|Props].
+
 export_perm(#we{perm=[]}) ->
     [{state,hidden_locked}];	     %Only for backward compatibility.
 export_perm(#we{perm=0}) -> [];
@@ -624,29 +751,40 @@ export_perm(#we{perm=3}) -> [{state,hidden_locked}];
 export_perm(#we{perm={Mode,Elems}}) ->
     [{state,{hidden,Mode,gb_sets:to_list(Elems)}}].
 
-export_edge(Rec, UvFaces, Acc) ->
+export_edge(E, Rec, UvFaces, We, Acc) ->
     #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf,
 	  ltpr=Ltpr,ltsu=Ltsu,rtpr=Rtpr,rtsu=Rtsu} = Rec,
     Data0 = [{edge,Va,Vb,Lf,Rf,Ltpr,Ltsu,Rtpr,Rtsu}],
-    Data = edge_data(Rec, UvFaces, Data0),
+    Data = edge_data(E, Rec, We, UvFaces, Data0),
     [Data|Acc].
     
-edge_data(#edge{lf=Lf,rf=Rf,a=A,b=B}, UvFaces, Acc0) ->
-    Acc = edge_data_1(left, Lf, A, UvFaces, Acc0),
-    edge_data_1(right, Rf, B, UvFaces, Acc).
+edge_data(E, #edge{lf=Lf,rf=Rf}, We, UvFaces, Acc0) ->
+    A = wings_va:edge_attrs(E, left, We),
+    B = wings_va:edge_attrs(E, right, We),
 
-edge_data_1(Side, Face, {U,V}, UvFaces, Acc) ->
+    %% If there are both vertex colors and UV coordinates,
+    %% we want them in the following order:
+    %%   [{color_*,_},{uv_*,_}]
+    %% On import in an old version of Wings, the UV coordinates
+    %% will be used.
+    Acc1 = edge_data_uv(left, Lf, wings_va:attr(uv, A), UvFaces, Acc0),
+    Acc2 = edge_data_uv(right, Rf, wings_va:attr(uv, B), UvFaces, Acc1),
+    Acc = edge_data_color(left, wings_va:attr(color, A), Acc2),
+    edge_data_color(right, wings_va:attr(color, B), Acc).
+
+edge_data_uv(Side, Face, {U,V}, UvFaces, Acc) ->
     case gb_sets:is_member(Face, UvFaces) of
 	false -> Acc;
-	true when Side == left  -> [{uv_lt,<<U/float,V/float>>}|Acc];
-	true when Side == right -> [{uv_rt,<<U/float,V/float>>}|Acc]
+	true when Side =:= left  -> [{uv_lt,<<U/float,V/float>>}|Acc];
+	true when Side =:= right -> [{uv_rt,<<U/float,V/float>>}|Acc]
     end;
-edge_data_1(_Side, _Face, {1.0,1.0,1.0}, _UvFaces, Acc) -> Acc;
-edge_data_1(left, _Face, {R,G,B}, _UvFaces, Acc) ->
+edge_data_uv(_, _, _, _, Acc) -> Acc.
+
+edge_data_color(left, {R,G,B}, Acc) ->
     [{color_lt,<<R:32/float,G:32/float,B:32/float>>}|Acc];
-edge_data_1(right, _Face, {R,G,B}, _UvFaces, Acc) ->
+edge_data_color(right, {R,G,B}, Acc) ->
     [{color_rt,<<R:32/float,G:32/float,B:32/float>>}|Acc];
-edge_data_1(_, _, _, _, Acc) -> Acc.
+edge_data_color(_, _, Acc) -> Acc.
 
 export_face({_,default}, Acc) -> [[]|Acc];
 export_face({_,Mat}, Acc) -> [[{material,Mat}]|Acc].
