@@ -257,7 +257,7 @@ new(#st{saved=true}=St0) ->
     St = clean_images(wings_undo:init(St1)),
     wings_u:caption(St),
     {new,St#st{saved=true}};
-new(St0) ->			     %File is not saved or autosaved.
+new(#st{}=St0) ->		      %File is not saved or autosaved.
     wings_u:caption(St0#st{saved=false}), 
     wings_u:yes_no_cancel(str_save_changes(),
 			  fun() -> {file,{save,{file,new}}} end,
@@ -578,7 +578,8 @@ recent_files(Tail) ->
 	    recent_files_1(Files, 1, Help, [separator|Tail])
     end.
 
-recent_files_1([{Base,_}|T], I, Help, Tail) ->
+recent_files_1([{Base0,_}|T], I, Help, Tail) ->
+    Base = wings_u:pretty_filename(Base0),
     [{Base,I,Help}|recent_files_1(T, I+1, Help, Tail)];
 recent_files_1([], _, _, Tail) -> Tail.
 
