@@ -13,7 +13,8 @@
 
 -module(wings_help).
 -export([menu/1,command/2]).
--export([cmd/1,help_window/2,no_more_basic_menus/0]).
+-export([cmd/1,help_window/2,no_more_basic_menus/0,
+	 not_possible_to_save_prefs/0]).
 
 -define(NEED_OPENGL, 1).
 -define(NEED_ESDL, 1).
@@ -112,11 +113,22 @@ no_more_basic_menus() ->
 				     ignore
 			     end).
 
+not_possible_to_save_prefs() ->
+    H = [{label,"The Wings3D.exe program was not able to locate the folder "
+	  "for applications settings."},
+	 panel,
+	 {label,"This is not supposed to happen. "
+	  "You should report this bug."},
+	 panel,
+	 {hframe,[{button,ok,[ok]}]}],
+    Qs = {vframe,H},
+    wings_ask:dialog("", Qs, fun(_) -> ignore end).
+
 getting_started(Head) ->
     B = "(",
     E = ")",
     H = [?__(1,
-	      "When learning Wings, keep an eye at the information line at the bottom of the screen."
+	      "When learning Wings, keep an eye on the information line at the bottom of the screen."
 	      " Generally it shows what the mouse buttons will do at any given moment."),
 
 	 ?__(2,
@@ -248,18 +260,18 @@ hotkeys(Head) ->
 def_hotkeys(Head) ->
     Help = [?__(1,"Any command that appears in a menu, can be assigned a keyboard short-cut (hotkey)."),
 	    ?__(4,"To assign a hotkey to a command:"),
-	    ?__(5,"1. Open the menu containing the command and high-light the command."),
+	    ?__(5,"1. Open the menu containing the command and highlight the command."),
 	    ?__(6,"2. Press the [Insert] or [/] key."),
-	    ?__(7,"3. If the command does different things depending "
-		"on which mouse button is pressed (when Advanced menus are enabled), "
-		"there will be a message in the information line asking you to "
-		"press a mouse button to choose which variant of the command to "
-		"assign to the hotkey. (Note: The mouse cursor must remain inside "
+	    ?__(10,"3. If the command does different things depending on "
+		"which mouse button is pressed there will be a message in "
+		"the information line asking you to press a mouse button "
+		"to choose which variant of the command to assign to "
+		"the hotkey. (Note: The mouse cursor must remain inside "
 		"the menu when you click.)"),
 	    ?__(8,"4. The information line asks you to press the key that the command "
 		"should be assigned to."),
 
- 	    ?__(9,"To delete a hotkey, similarly high-light the command in a menu,"
+	    ?__(9,"To delete a hotkey, similarly highlight the command in a menu,"
 		" and press the [Del] or [\\] key. A dialog box listing all keys "
 		"bound to the command will appear. "
 		"Check all hotkeys you want to delete.")],
@@ -278,15 +290,13 @@ lights(Head) ->
 opengl_info(Head) ->
     gl:getError(),			%Clear any previous error.
     [{_,VerTuple}] = ets:lookup(wings_gl_ext, version),
-    {SdlVersion0,_} = sdl_video:wm_getInfo(),
-    SdlVersion1 = [integer_to_list(V) || V <- tuple_to_list(SdlVersion0)],
-    SdlVersion = string:join(SdlVersion1, "."),
+    BackendVersion = "2D-API: " ++ wings_io:version_info(),
     Help = [
 	    ?__(1,"Vendor: ") ++ gl:getString(?GL_VENDOR) ++ "\n" ++
 	    ?__(2,"Renderer: ") ++ gl:getString(?GL_RENDERER) ++ "\n" ++
 	    ?__(3,"Version: ") ++ gl:getString(?GL_VERSION) ++ "\n" ++
 	    ?__(4,"Version tuple: ") ++ lists:flatten(io_lib:format("~p\n", [VerTuple])) ++
-	    "SDL: " ++ SdlVersion,
+	    BackendVersion,
 
 	    get_info([{?__(5,"Red bits"),?GL_RED_BITS},
 		      {?__(6,"Green bits"),?GL_GREEN_BITS},
@@ -319,6 +329,7 @@ opengl_info(Head) ->
 		      {?__(32,"Color buffers store indices"),?GL_INDEX_MODE},
 		      {?__(33,"Double buffering"),?GL_DOUBLEBUFFER},
 		      {?__(34,"Stereo buffers"),?GL_STEREO},
+		      {?__(45,"Sample buffers"),?GL_SAMPLE_BUFFERS},
 		      {?__(35,"Range of aliased point sizes"),?GL_ALIASED_POINT_SIZE_RANGE},
 		      {?__(36,"Range of antialised point sizes"),?GL_SMOOTH_POINT_SIZE_RANGE},
 		      {?__(37,"Range of aliased line widths"),?GL_ALIASED_LINE_WIDTH_RANGE},
@@ -356,6 +367,7 @@ has_one_elem(?GL_MAX_ELEMENTS_INDICES) -> true;
 has_one_elem(?GL_NUM_COMPRESSED_TEXTURE_FORMATS) -> true;
 has_one_elem(?GL_MAX_VERTEX_UNITS_ARB) -> true;
 has_one_elem(?GL_MAX_3D_TEXTURE_SIZE) -> true;
+has_one_elem(?GL_SAMPLE_BUFFERS) -> true;
 has_one_elem(_) -> false.
 
 extensions() ->
