@@ -12,10 +12,13 @@
 %%
 
 -module(wings_light).
--export([light_types/0,menu/3,command/2,is_any_light_selected/1,info/1,
+-export([light_types/0,menu/3,command/2,is_any_light_selected/1,
+	 any_enabled_lights/0,info/1,
 	 create/2,update_dynamic/2,update_matrix/2,update/1,render/1,
-	 modeling_lights/2,global_lights/0,camera_lights/0,
-	 export/1,export_camera_lights/0,import/2,import/1,shape_materials/2]).
+	 global_lights/0,camera_lights/0,camera_lights/1,
+	 export/1,export_bc/1,export_camera_lights/0,
+	 import/2,import/1,shape_materials/2,
+	 light_pos/1]).
 
 -define(NEED_OPENGL, 1).
 -include("wings.hrl").
@@ -29,8 +32,11 @@
 -define(DEF_POS, {?DEF_X,?DEF_Y,?DEF_Z}).
 
 %% Light record in We.
+%%
+%% The type field must be the first field, since it is used by
+%% the light testing macros in wings.hrl.
 -record(light,
-	{type,
+	{type,					%Type. (DO NOT MOVE.)
 	 diffuse={1.0,1.0,1.0,1.0},
 	 ambient={0.0,0.0,0.0,1.0},
 	 specular={1.0,1.0,1.0,1.0},
@@ -138,6 +144,11 @@ is_any_light_selected([{Id,_}|Sel], Shs) ->
     end;
 is_any_light_selected([], _) -> false.
 
+any_enabled_lights() ->
+    wings_dl:fold(fun(#dlo{src_we=We}, Bool) ->
+			  Bool orelse ?IS_ANY_LIGHT(We)
+		  end, false).
+
 info(#we{name=Name,light=#light{type=Type}=L}=We) ->
     Info0 = wings_util:format(?__(1,"Light ~s"), [Name]),
     case Type of
@@ -305,7 +316,7 @@ edit(_) -> wings_u:error(?__(1,"Select only one light.")).
 
 edit(Id, #st{shapes=Shs}=St) ->
     We = #we{light=#light{type=Type}} = gb_trees:get(Id, Shs),
-    {Name,Prop} = get_light(We),
+    {Name,Prop} = get_light(We, false),
     case Type of
 	ambient ->
 	    {dialog,Qs,Fun} = edit_ambient_dialog(Name, Prop, We, Shs, St),
@@ -441,7 +452,7 @@ create(Type, #st{onext=Oid}=St) ->
 %%% Updating, drawing and rendering lights.
 %%%
 update_dynamic(#dlo{src_we=We0}=D, Vtab0) ->
-    Vtab = gb_trees:from_orddict(sort(Vtab0)),
+    Vtab = array:from_orddict(sort(Vtab0)),
     We = We0#we{vp=Vtab},
     List = update_1(We, D),
     D#dlo{work=List,src_we=We}.
@@ -554,9 +565,17 @@ render(#dlo{work=Light}) ->
 %%% Exporting lights.
 %%%
 
-export(#st{shapes=Shs}) ->
+%% For exporters.
+export(St) ->
+    export(St, false).
+
+%% For saving in .wings files.
+export_bc(St) ->
+    export(St, true).
+
+export(#st{shapes=Shs}, BackwardsCompatible) ->
     L = foldl(fun(We, A) when ?IS_ANY_LIGHT(We) ->
-		      [get_light(We)|A];
+		      [get_light(We, BackwardsCompatible)|A];
 		 (_, A) -> A
 	      end, [], gb_trees:values(Shs)),
     reverse(L).
@@ -575,22 +594,22 @@ export_camera_lights() ->
     GL = fun({Name,Li = #light{aim=Diff}}) ->
 		 LPos = e3d_vec:add(CameraPos,Diff),
 		 We = #we{name = Name,
-			  vp = gb_trees:from_orddict([{1, LPos}]),
+			  vp = array:from_orddict([{1, LPos}]),
 			  light = Li#light{aim=Aim}},
-		 get_light(We)
+		 get_light(We, false)
 	 end,
     [GL(Light) || Light <- [Amb|Ls]].
 
-get_light(#we{name=Name,perm=P}=We) ->
-    Ps0 = get_light_1(We),
+get_light(#we{name=Name,perm=P}=We, BC) ->
+    Ps0 = get_light_1(We, BC),
     Ps = export_perm(P, Ps0),
     {Name,Ps}.
 
-get_light_1(#we{light=#light{type=ambient,ambient=Amb,prop=Prop}}=We) ->
+get_light_1(#we{light=#light{type=ambient,ambient=Amb,prop=Prop}}=We, _) ->
     P = light_pos(We),
     OpenGL = [{type,ambient},{ambient,Amb},{position,P}],
     [{opengl,OpenGL}|Prop];
-get_light_1(#we{light=L}=We) ->
+get_light_1(#we{light=L}=We, BC) ->
     #light{type=Type,diffuse=Diff,ambient=Amb,specular=Spec,
 	   aim=Aim,spot_angle=Angle,spot_exp=SpotExp,
 	   lin_att=LinAtt,quad_att=QuadAtt,prop=Prop} = L,
@@ -610,7 +629,7 @@ get_light_1(#we{light=L}=We) ->
 		 true -> OpenGL0
 	     end,
     OpenGL = case Type of
-		 area -> [{mesh,wings_export:make_mesh(We, [])}|OpenGL1];
+		 area -> [{mesh,export_mesh(We, BC)}|OpenGL1];
 		 _ -> OpenGL1
 	     end,
     [{opengl,OpenGL}|Prop].
@@ -622,6 +641,34 @@ export_perm({_,_}, Ps) ->
 export_perm(P, Ps) when is_integer(P) ->
     [{visible,P < 2},{locked,(P band 1) =/= 0}|Ps].
 
+%% This is the classic definition of e3d_face{}, as it was defined
+%% before 1.1.9.
+-record(classic_e3d_face,
+	{vs=[],				        %List of vertex indices.
+	 vc=[],					%Vertex color indices.
+	 tx=[],				        %List of texture indices.
+	 ns=[],				        %List of normal indices.
+	 mat=[],				%Materials for face.
+	 vis=-1}).				%Visible edges (as in 3DS).
+
+export_mesh(We, BC) ->
+    #e3d_mesh{fs=Fs0} = Mesh = wings_export:make_mesh(We, []),
+    Fs = case BC of
+	     false ->
+		 Fs0;
+	     true ->
+		 [export_fix_face(F) || F <- Fs0]
+	 end,
+    Mesh#e3d_mesh{fs=Fs}.
+
+export_fix_face(#e3d_face{vs=Vs,mat=Mat}) ->
+    %% Fix the face record so that it looks like the classic
+    %% definition of #e3d_face{} (before 1.1.9).
+    FaceRec = #classic_e3d_face{vs=Vs,mat=Mat},
+
+    %% Patch the record type.
+    setelement(1, FaceRec, e3d_face).
+
 %%%
 %%% Importing lights.
 %%%
@@ -661,72 +708,52 @@ import_we(#light{type=area}=Light, OpenGL, {X,Y,Z}) ->
 	    none ->
 		#e3d_mesh{type=polygon,
 			  fs=[#e3d_face{vs=[0,1,2,3],
-					mat=[default]},
-			      #e3d_face{vs=[3,2,1,0],
-					mat=['_hole_']}],
+					mat=[default]}],
 			  vs=[{X+1.0,Y,Z+1.0},{X-1.0,Y,Z+1.0},
 			      {X-1.0,Y,Z-1.0},{X+1.0,Y,Z-1.0}]};
-	    {mesh,M} -> e3d_mesh:transform(e3d_mesh:clean_faces(M))
+	    {mesh,M} -> import_fix_mesh(M)
 	end,
     We = wings_import:import_mesh(material, Mesh),
-    We#we{light=Light,has_shape=true};
-import_we(#light{}=Light, _OpenGL, Pos) ->
+    We#we{light=Light};
+import_we(#light{}=Light, _OpenGL, {X,Y,Z}) ->
+    %% We used to put all vertices at the same position, but with
+    %% then rewritten pick handling we need a vertex array for picking.
+    %% The cube will be slightly larger than the sphere that is shown
+    %% for the light. The position of the light will be the centroid
+    %% of the cube.
     Fs = [[0,3,2,1],[2,3,7,6],[0,4,7,3],[1,2,6,5],[4,5,6,7],[0,1,5,4]],
-    Vs = lists:duplicate(8, Pos),
+    S = 0.07,
+    Vs = [{X-S,Y-S,Z+S},{X-S,Y+S,Z+S},{X+S,Y+S,Z+S},{X+S,Y-S,Z+S},
+	  {X-S,Y-S,Z-S},{X-S,Y+S,Z-S},{X+S,Y+S,Z-S},{X+S,Y-S,Z-S}],
     We = wings_we:build(Fs, Vs),
-    We#we{light=Light,has_shape=false}.
+    We#we{light=Light}.
 
+import_fix_mesh(#e3d_mesh{fs=Fs0}=Mesh0) ->
+    Fs = [import_fix_face(F) || F <- Fs0],
+    Mesh1 = Mesh0#e3d_mesh{fs=Fs},
+    Mesh = e3d_mesh:clean_faces(Mesh1),
+    e3d_mesh:transform(Mesh).
 
+import_fix_face(FaceRec) when is_tuple(FaceRec) ->
+    %% Different versions of Wings can have #e3d_face{}
+    %% records of different size (for example, in Wings 1.1.9
+    %% a new 'sg' field was added to #e3d_face{}). We know
+    %% that the fields we are interested in are in the same
+    %% place, so we can retrieve them using element/2.
+    %%
+    e3d_face = element(1, FaceRec),		%Crash on unknown record type.
+    Vs = element(#e3d_face.vs, FaceRec),
+    Mat = element(#e3d_face.mat, FaceRec),
+    #e3d_face{vs=Vs,mat=Mat}.
 
 %%%
 %%% Setting up lights.
 %%%
 
 camera_lights() ->
-    setup_lights(camera).
+    camera_lights(wings_pref:get_value(number_of_lights)).
 
-global_lights() ->
-    setup_lights(global).
-
-setup_lights(CoordType) ->
-    case wings_pref:get_value(scene_lights) of
-	false -> 
-	    NoOfLights = wings_pref:get_value(number_of_lights),
-	    modeling_lights(CoordType, NoOfLights);
-	true -> scene_lights(CoordType)
-    end.
-
-camera_ambient() ->
-    #light{type = ambient, 
-	   aim = {0.0,0.0,0.0},
-	   ambient = {0.1,0.1,0.1,1.0}}.
-camera_infinite_1_0() ->
-    #light{type = infinite, 
-	   diffuse  = {0.7,0.7,0.7,1},
-	   specular = {0.2,0.2,0.2,1},
-	   ambient  = {0,0,0,1.0},
-	   aim      = {0.110,0.0,0.994}
-	  }.
-camera_infinite_2_0() ->
-    #light{type = infinite, 
-	   diffuse  = {1,1,1,1},
-	   specular = {0.3,0.3,0.3,1},
-	   ambient  = {0,0,0,1.0},
-	   aim      = {0.71,0.71,0.0}
-	  }.
-camera_infinite_2_1() ->
-    #light{type = infinite, 
-	   diffuse  = {0.5,0.5,0.5,0.5},
-	   specular = {0.3,0.3,0.3,1},
-	   ambient  = {0,0,0,1.0},
-	   aim      = {-0.71,-0.71,0.0}
-	  }.
-
-
-infinite({X,Y,Z}) -> {X,Y,Z,0.0}.
-
-modeling_lights(global, _Type) -> ok;
-modeling_lights(camera, Type) ->
+camera_lights(Type) ->
     gl:enable(?GL_LIGHT0),
     gl:disable(?GL_LIGHT2),
     gl:disable(?GL_LIGHT3),
@@ -762,12 +789,39 @@ modeling_lights(camera, Type) ->
 	    gl:lightfv(?GL_LIGHT0, ?GL_POSITION, {1,1,2,0})
     end.
 
-scene_lights(camera) -> ok;
-scene_lights(global) ->
+global_lights() ->
     gl:lightModelfv(?GL_LIGHT_MODEL_AMBIENT, {0.0,0.0,0.0,1.0}),
     Lnum = wings_dl:fold(fun scene_lights_fun/2, ?GL_LIGHT0),
     disable_from(Lnum).
 
+camera_ambient() ->
+    #light{type = ambient, 
+	   aim = {0.0,0.0,0.0},
+	   ambient = {0.1,0.1,0.1,1.0}}.
+camera_infinite_1_0() ->
+    #light{type = infinite, 
+	   diffuse  = {0.7,0.7,0.7,1},
+	   specular = {0.2,0.2,0.2,1},
+	   ambient  = {0,0,0,1.0},
+	   aim      = {0.110,0.0,0.994}
+	  }.
+camera_infinite_2_0() ->
+    #light{type = infinite, 
+	   diffuse  = {1,1,1,1},
+	   specular = {0.3,0.3,0.3,1},
+	   ambient  = {0,0,0,1.0},
+	   aim      = {0.71,0.71,0.0}
+	  }.
+camera_infinite_2_1() ->
+    #light{type = infinite, 
+	   diffuse  = {0.5,0.5,0.5,0.5},
+	   specular = {0.3,0.3,0.3,1},
+	   ambient  = {0,0,0,1.0},
+	   aim      = {-0.71,-0.71,0.0}
+	  }.
+
+infinite({X,Y,Z}) -> {X,Y,Z,0.0}.
+
 disable_from(Lnum) when Lnum > ?GL_LIGHT7 -> ok;
 disable_from(Lnum) ->
     gl:disable(Lnum),
@@ -783,14 +837,15 @@ scene_lights_fun(#dlo{transparent=#we{light=L}=We}, Lnum) ->
 scene_lights_fun(#dlo{drag=Drag,src_we=We0}=D, Lnum) ->
     %% Area lights handled here in all selection modes +
     %% other lights in vertex/edge/face modes.
-    We = if ?HAS_SHAPE(We0) -> 
+    We = case We0 of
+	     #we{light=#light{type=area}} ->
 		 %% For an area light it looks better in vertex/edge/face
 		 %% modes to emulate with the static non-splitted shape
 		 %% during drag. It would be more correct if the area light
 		 %% updating would use the resulting #we{}, but it does not
 		 %% exist until the drag is done.
 		 wings_draw:original_we(D);
-	    true -> 
+	     _ -> 
 		 %% Non-area lights drag the whole shape so they can use
 		 %% the dynamic part of the splitted shape 
 		 %% (which is the whole shape).
@@ -877,18 +932,17 @@ setup_attenuation(Lnum, #light{lin_att=Lin,quad_att=Quad}) ->
 light_pos(#we{light=#light{type=area}}=We) ->
     {Pos,_,_} = arealight_posdirexp(We),
     Pos;
-light_pos(#we{vp=Vtab}) ->
-    gb_trees:get(1, Vtab).
+light_pos(We) ->
+    wings_vertex:center(We).
 
 arealight_posdirexp(#we{light=#light{type=area}}=We) ->
-    FaceMats = wings_facemat:all(We),
+    Faces = wings_we:visible(We),
     ANCs = 
 	[begin 
 	     {wings_face:area(Face, We),
 	      wings_face:normal(Face, We),
 	      wings_face:center(Face, We)}
-	 end || {Face,Mat} <- FaceMats, 
-		(Mat =/= '_hole_') andalso wings_face:good_normal(Face, We)],
+	 end || Face <- Faces, wings_face:good_normal(Face, We)],
     Area = foldl(fun ({A,_,_}, Acc) -> A+Acc end, 0.0, ANCs),
     AreaNormal = 
 	foldl(fun ({A,N,_}, Acc) -> 
@@ -915,22 +969,16 @@ arealight_posdirexp(#we{light=#light{type=area}}=We) ->
     end.
 
 move_light(Pos, #we{vp=Vtab0}=We) ->
-    Vtab1 = [{V,Pos} || V <- gb_trees:keys(Vtab0)],
-    Vtab = gb_trees:from_orddict(Vtab1),
+    Vtab = array:sparse_map(fun(V, _) -> {V,Pos} end, Vtab0),
     We#we{vp=Vtab}.
 
-shape_materials(#light{diffuse={_,_,_,Af}=Front,ambient={_,_,_,Ab}=Back}, St) ->
+shape_materials(#light{diffuse={_,_,_,Af}=Front}, St) ->
     BlackF = {0.0,0.0,0.0,Af},
-    BlackB = {0.0,0.0,0.0,Ab},
     Default = 
 	sort([{opengl,sort([{diffuse,BlackF},{ambient,BlackF},{specular,BlackF},
 			    {emission,Front},{shininess,0.0}])},
 	      {maps,[]}]),
-    Hole = 
-	sort([{opengl,sort([{diffuse,BlackB},{ambient,BlackB},{specular,BlackB},
-			    {emission,Back},{shininess,0.0}])},
-	      {maps,[]}]),
-    wings_material:update_materials([{default,Default},{'_hole_',Hole}], St).
+    wings_material:update_materials([{default,Default}], St).
 
 mul_point({1.0,0.0,0.0, 0.0,1.0,0.0, 0.0,0.0,1.0, Tx,Ty,Tz}, {X,Y,Z}) ->
     {X+Tx,Y+Ty,Z+Tz};
