@@ -16,8 +16,9 @@
 	 add_materials/2,add_materials/3,
 	 update_materials/2,
 	 update_image/4,used_images/1,
-	 used_materials/1,
-	 apply_material/2,is_transparent/2]).
+	 used_materials/1,has_texture/2,
+	 apply_material/3,is_transparent/2,
+	 needed_attributes/2]).
 
 -define(NEED_OPENGL, 1).
 -define(NEED_ESDL, 1).
@@ -25,7 +26,7 @@
 -include("e3d_image.hrl").
 
 -import(lists, [sort/1,foldl/3,reverse/1,
-		keyreplace/4,keydelete/3,keysearch/3,flatten/1]).
+		keyreplace/4,keydelete/3,keyfind/3,flatten/1]).
 
 material_menu(St) ->
     [{?__(4,"Material"),{material,material_fun(St)}}].
@@ -83,7 +84,7 @@ command({duplicate,MatList}, St) ->
 command({delete,MatList}, St) ->
     delete_material(MatList, St);
 command({rename,MatList0}, St) ->
-    case MatList0 -- ["default","_hole_"] of
+    case MatList0 -- ["default"] of
 	[] -> St;
 	MatList -> rename(MatList, St)
     end;
@@ -117,8 +118,6 @@ duplicate_material([], St) -> St.
 
 delete_material(["default"|Ms], St) ->
     delete_material(Ms, St);
-delete_material(["_hole_"|Ms], St) ->
-    delete_material(Ms, St);
 delete_material([M0|Ms], #st{mat=Mat0}=St0) ->
     M = list_to_atom(M0),
     Mat = gb_trees:delete(M, Mat0),
@@ -183,49 +182,52 @@ make_fake_selection_1([#we{id=Id,mat=MatTab}|Shs], OldMat) ->
     end;
 make_fake_selection_1([], _) -> [].
 
-select_material(Mat, St) ->
-    %% XXX Works but is slow.
-    wings_sel:make(fun(_, #we{mat=AtomMat}) when is_atom(AtomMat) ->
-			   Mat =:= AtomMat;
-		      (Face, #we{mat=MatTab}) when Face >= 0 ->
-			   case keysearch(Face, 1, MatTab) of
-			       false -> false;
-			       {value,{_,Mat}} -> true;
-			       {value,_} -> false
-			   end;
-			   (_, _) -> false
-		   end, face, St).
+select_material(Mat, #st{shapes=Shs}=St) ->
+    Sel = foldl(fun(We, A) ->
+			select_material_1(We, Mat, A)
+		end, [], gb_trees:values(Shs)),
+    wings_sel:set(Sel, St).
+
+select_material_1(#we{id=Id,fs=Ftab}=We, Mat, Acc) ->
+    MatFaces = wings_facemat:mat_faces(gb_trees:to_list(Ftab), We),
+    case keyfind(Mat, 1, MatFaces) of
+	false ->
+	    Acc;
+	{Mat,FaceInfoList} ->
+	    Sel = [F || {F,_} <- FaceInfoList],
+	    [{Id,gb_sets:from_ordset(Sel)}|Acc]
+    end.
 
 set_material(Mat, #st{selmode=face}=St) ->
     wings_sel:map(fun(Faces, We) ->
-			  wings_facemat:assign(Mat, Faces, We#we{mode=material})
+			  wings_facemat:assign(Mat, Faces, We)
 		  end, St);
 set_material(Mat, #st{selmode=body}=St) ->
     wings_sel:map(fun(_, #we{fs=Ftab}=We) ->
-			  wings_facemat:assign(Mat, gb_trees:keys(Ftab),
-					       We#we{mode=material})
+			  wings_facemat:assign(Mat, gb_trees:keys(Ftab), We)
 		  end, St);
 set_material(_, St) -> St.
 
 default() ->
     Dm = wings_pref:get_value(material_default),
-    Hm = wings_pref:get_value(material_hole),
-    M = [{default,make_default(Dm, 1.0)},
-        {'_hole_',make_default(Hm, 0.50)}],
+    M = [{default,make_default(Dm, 1.0, [{vertex_colors,set}])}],
     gb_trees:from_orddict(sort(M)).
 
-make_default({R,G,B}, Opacity) ->
+make_default(Color, Opacity) ->
+    make_default(Color, Opacity, []).
+
+make_default({R,G,B}, Opacity, More) ->
     Color = {R,G,B,Opacity},
     White = {1.0,1.0,1.0,1.0},
     Mat = [{opengl,[{diffuse,Color},{ambient,Color},{specular,White},
-		    {emission,{0.0,0.0,0.0,0.0}},{shininess,1.0}]},
+		    {emission,{0.0,0.0,0.0,0.0}},{shininess,1.0}|More]},
 	   {maps,[]}],
     sort([{K,sort(L)} || {K,L} <- Mat]).
 
 update_image(MatName, MapType, Image, #st{mat=Mtab}) ->
     Mat = gb_trees:get(MatName, Mtab),
     Maps = prop_get(maps, Mat, []),
-    {value,{MapType,ImageId}} = keysearch(MapType, 1, Maps),
+    {MapType,ImageId} = keyfind(MapType, 1, Maps),
     wings_image:update(ImageId, Image).
 
 add_materials(Ms, St) ->
@@ -260,11 +262,18 @@ add_defaults(Props0) ->
 
 add_defaults_1(P) ->
     Def = {1.0,1.0,1.0,1.0},
+    VertexColor = valid_vertex_color(prop_get(vertex_colors, P, ignore)),
     [{diffuse,norm(prop_get(diffuse, P, Def))},
      {ambient,norm(prop_get(ambient, P, Def))},
      {specular,norm(prop_get(specular, P, Def))},
      {emission,norm(prop_get(emission, P, {0.0,0.0,0.0,0.0}))},
-     {shininess,prop_get(shininess, P, 1.0)}].
+     {shininess,prop_get(shininess, P, 1.0)},
+     {vertex_colors,VertexColor}].
+
+%% For future compatibility, ignore anything that we don't recognize.
+valid_vertex_color(multiply=A) -> A;
+valid_vertex_color(set=A) -> A;
+valid_vertex_color(_) -> ignore.
 
 update_materials([{Name,Mat0}|Ms], St) ->
     Mat1 = add_defaults(Mat0),
@@ -325,11 +334,6 @@ load_map_1(File0, Dir) ->
 	    none
     end.
     
-add('_hole_'=Name, Mat, #st{mat=MatTab}=St) ->
-    case gb_trees:is_defined(Name, MatTab) of
-	true -> St;
-	false -> St#st{mat=gb_trees:insert(Name, Mat, MatTab)}
-    end;
 add(Name, Mat0, #st{mat=MatTab}=St) ->
     Mat = sort([{K,sort(L)} || {K,L} <- Mat0]),
     case gb_trees:lookup(Name, MatTab) of
@@ -350,17 +354,59 @@ new_name(Name0, Tab) ->
     Name = wings_util:unique_name(Name0, Names),
     list_to_atom(Name).
 
-apply_material(Name, Mtab) when is_atom(Name) ->
+has_texture(Name, Mtab) ->
+    Mat = gb_trees:get(Name, Mtab),
+    has_texture(Mat).
+
+has_texture(Mat) ->
+    Maps = prop_get(maps, Mat, []),
+    none =/= prop_get(diffuse, Maps, none).
+
+apply_material(Name, Mtab, ActiveVertexColors) when is_atom(Name) ->
     Mat = gb_trees:get(Name, Mtab),
     OpenGL = prop_get(opengl, Mat),
-    gl:materialfv(?GL_FRONT_AND_BACK, ?GL_DIFFUSE, prop_get(diffuse, OpenGL)), 
-    gl:materialfv(?GL_FRONT_AND_BACK, ?GL_AMBIENT, prop_get(ambient, OpenGL)),
     gl:materialfv(?GL_FRONT_AND_BACK, ?GL_SPECULAR, prop_get(specular, OpenGL)),
     Shine = prop_get(shininess, OpenGL)*128,
     gl:materialf(?GL_FRONT_AND_BACK, ?GL_SHININESS, Shine),
     gl:materialfv(?GL_FRONT_AND_BACK, ?GL_EMISSION, prop_get(emission, OpenGL)),
-    Maps = prop_get(maps, Mat, []),
-    apply_texture(prop_get(diffuse, Maps, none)).
+    Maps0 = prop_get(maps, Mat, []),
+    VertexColors = case ActiveVertexColors of
+		       false -> ignore;
+		       true -> prop_get(vertex_colors, OpenGL, ignore)
+		   end,
+    Def = fun() -> ok end,
+    {Maps,DeApply} =
+	case VertexColors of
+	    ignore ->
+		%% Ignore vertex colors. If the hemispherical lighting
+		%% shader is enabled, it is not enough to only disable
+		%% COLOR_MATERIAL, but we must also disable the color
+		%% array.
+		gl:disable(?GL_COLOR_MATERIAL),
+		case ActiveVertexColors of
+		    true ->
+			{Maps0,fun() ->
+				       gl:enableClientState(?GL_COLOR_ARRAY)
+			       end};
+		    false ->
+			{Maps0,Def}
+		   end;
+	    set ->
+		%% Vertex colors overrides diffuse and ambient color
+		%% and suppresses any texture.
+		gl:colorMaterial(?GL_FRONT_AND_BACK, ?GL_AMBIENT_AND_DIFFUSE),
+		gl:enable(?GL_COLOR_MATERIAL),
+		{[],Def};
+	    multiply ->
+		%% Vertex colors are multiplied with the texture.
+		gl:colorMaterial(?GL_FRONT_AND_BACK, ?GL_AMBIENT_AND_DIFFUSE),
+		gl:enable(?GL_COLOR_MATERIAL),
+		{Maps0,Def}
+	end,
+    gl:materialfv(?GL_FRONT_AND_BACK, ?GL_DIFFUSE, prop_get(diffuse, OpenGL)),
+    gl:materialfv(?GL_FRONT_AND_BACK, ?GL_AMBIENT, prop_get(ambient, OpenGL)),
+    apply_texture(prop_get(diffuse, Maps, none)),
+    DeApply.
 
 apply_texture(none) -> no_texture();
 apply_texture(Image) ->
@@ -387,7 +433,8 @@ apply_texture_1(Image, TxId) ->
     case wings_gl:is_ext({1,2}) of
 	true ->
 	    %% Calculate specular color correctly on textured models.
-	    gl:lightModeli(?GL_LIGHT_MODEL_COLOR_CONTROL,?GL_SEPARATE_SPECULAR_COLOR);
+	    gl:lightModeli(?GL_LIGHT_MODEL_COLOR_CONTROL,
+			   ?GL_SEPARATE_SPECULAR_COLOR);
 	false -> ok
     end,
     case wings_image:info(Image) of
@@ -452,21 +499,44 @@ is_mat_transparent(Mat) ->
 		  end, false, OpenGL),
     Trans.
 
-%% case Trans of
-%% 	true -> true;
-%% 	false -> 
-%% 	    Maps = prop_get(maps, Mat),    
-%% 	    case prop_get(diffuse, Maps, none) of
-%% 		none -> false;
-%% 		DiffMap -> 		    
-%% 		    case wings_image:info(DiffMap) of
-%% 			#e3d_image{bytes_pp=4} -> true;
-%% 			#e3d_image{type=a8} -> true;
-%% 			_ -> false
-%% 		    end
-%% 	    end
-%%     end.
-%%% The material editor.
+%% needed_attributes(We, St) -> [Attr]
+%%     Attr = uv|color
+%%  Return a ordered list of the type of attributes that are needed
+%%  according to the materials.
+needed_attributes(We, #st{mat=Mat}) ->
+    Used = wings_facemat:used_materials(We),
+    needed_attributes_1(Used, Mat, false, false).
+
+needed_attributes_1(_, _, true, true) -> [color,uv];
+needed_attributes_1([M|Ms], MatTab, Col0, UV0) ->
+    Mat = gb_trees:get(M, MatTab),
+    UV = UV0 orelse needs_uvs(Mat),
+    Col = Col0 orelse needs_vertex_colors(Mat),
+    needed_attributes_1(Ms, MatTab, Col, UV);
+needed_attributes_1([], _, Col, UV) ->
+    L = case UV of
+	    true -> [uv];
+	    false -> []
+	end,
+    case Col of
+	true -> [color|L];
+	false -> L
+    end.
+
+needs_vertex_colors(Mat) ->
+    OpenGL = prop_get(opengl, Mat),
+    prop_get(vertex_colors, OpenGL, ignore) =/= ignore.
+
+needs_uvs(Mat) ->
+    OpenGL = prop_get(opengl, Mat),
+    case prop_get(vertex_colors, OpenGL, ignore) of
+	set ->
+	    %% Vertex colors overrides the texture (if any).
+	    false;
+	_ ->
+	    %% We need UV coordinates if there is a diffuse texture.
+	    has_texture(Mat)
+    end.
 
 -define(PREVIEW_SIZE, 100).
 
@@ -479,6 +549,7 @@ edit(Name, Assign, #st{mat=Mtab}=St) ->
 
 edit_dialog(Name, Assign, St=#st{mat=Mtab0}, Mat0) ->
     OpenGL0 = prop_get(opengl, Mat0),
+    VertexColors0 = prop_get(vertex_colors, OpenGL0, ignore),
     {Diff0,Opacity0} = ask_prop_get(diffuse, OpenGL0),
     {Amb0,_} = ask_prop_get(ambient, OpenGL0),
     {Spec0,_} = ask_prop_get(specular, OpenGL0),
@@ -488,6 +559,12 @@ edit_dialog(Name, Assign, St=#st{mat=Mtab0}, Mat0) ->
     Preview = fun(A,S,D,F,G) ->
 		      mat_preview(A,S,D,F,G,prop_get(maps,Mat0))
 	      end,
+    Hook = {hook,fun(is_disabled, {_Var,_I,Sto}) ->
+			 gb_trees:get(vertex_colors, Sto) =/= ignore;
+		    (_, _) -> void
+		 end},
+    AnyTexture = has_texture(Mat0),
+    VtxColMenu = vertex_color_menu(AnyTexture, VertexColors0),
     Qs1 = [{vframe,
 	    [
 	     {hframe, 
@@ -496,14 +573,17 @@ edit_dialog(Name, Assign, St=#st{mat=Mtab0}, Mat0) ->
 		[{label,?__(1,"Diffuse")},
 		 {label,?__(2,"Ambient")},
 		 {label,?__(3,"Specular")},
-		 {label,?__(4,"Emission")}]
+		 {label,?__(4,"Emission")},
+		 {label,"Vertex Colors"}
+		]
 	       },
 	       {vframe,
-		[{slider,{color,Diff0,[{key,diffuse}]}},
+		[{slider,{color,Diff0,
+			  [{key,diffuse},Hook]}},
 		 {slider,{color,Amb0,[{key,ambient}]}},
 		 {slider,{color,Spec0,[{key,specular}]}},
-		 {slider,{color,Emiss0,[{key,emission}]}}
-		]}]},
+		 {slider,{color,Emiss0,[{key,emission}]}},
+		 VtxColMenu]}]},
 	     {hframe, [{vframe, [{label,?__(5,"Shininess")},
 				 {label,?__(6,"Opacity")}]},
 		       {vframe, [{slider,{text,Shine0,
@@ -518,13 +598,18 @@ edit_dialog(Name, Assign, St=#st{mat=Mtab0}, Mat0) ->
     Qs = {hframe,[{vframe,Qs2},
 		  {vframe,[{button,?__(7,"OK"),done,[ok,{key,material_editor_ok}]},
 			   {button,wings_s:cancel(),cancel,[cancel]}]}]},
-    Ask = fun([{diffuse,Diff},{ambient,Amb},{specular,Spec},
-	       {emission,Emiss},{shininess,Shine},{opacity,Opacity}|More0]) ->
+    Ask = fun([{diffuse,Diff},
+	       {ambient,Amb},
+	       {specular,Spec},
+	       {emission,Emiss},
+	       {vertex_colors,VertexColors},
+	       {shininess,Shine},{opacity,Opacity}|More0]) ->
 		  OpenGL = [ask_prop_put(diffuse, Diff, Opacity),
 			    ask_prop_put(ambient, Amb, Opacity),
 			    ask_prop_put(specular, Spec, Opacity),
 			    ask_prop_put(emission, Emiss, Opacity),
-			    {shininess,Shine}],
+			    {shininess,Shine},
+			    {vertex_colors,VertexColors}],
 		  Mat1 = keyreplace(opengl, 1, Mat0, {opengl,OpenGL}),
 		  {Mat2,More} = update_maps(Mat1, More0),
 		  case plugin_results(Name, Mat2, More) of
@@ -536,6 +621,23 @@ edit_dialog(Name, Assign, St=#st{mat=Mtab0}, Mat0) ->
 	  end,
     {dialog,Qs,Ask}.
 
+vertex_color_menu(MultiplyPossible, Def0) ->
+    Def = case MultiplyPossible of
+	      true -> Def0;
+	      false when Def0 =:= multiply -> set;
+	      false -> Def0
+	  end,
+    {menu,[{"Ignore",ignore,[{info,"Ignore vertex colors"}]},
+	   {"Set",set,[{info,"Show vertex colors"}]}|
+	   case MultiplyPossible of
+	       true ->
+		   [{"Multiply",multiply,
+		     [{info,"Multiply texture colors with vertex colors"}]}];
+	       false -> []
+	   end],Def,
+     [{info,"Choose how to use vertex colors"},
+      {key,vertex_colors}]}.
+
 maybe_assign(false, _, St) -> St;
 maybe_assign(true, Name, St) -> set_material(Name, St).
 
@@ -597,7 +699,7 @@ mat_preview(X, Y, _W, _H, Common, Maps) ->
     MM = gl:getDoublev(?GL_MODELVIEW_MATRIX),
     PM = gl:getDoublev(?GL_PROJECTION_MATRIX),
     ViewPort = wings_wm:viewport(),
-    {Ox,Oy,_} = glu:project(X, Y+?PREVIEW_SIZE, 0, MM, PM, ViewPort),
+    {Ox,Oy,_} = wings_gl:project(X, Y+?PREVIEW_SIZE, 0, MM, PM, ViewPort),
     gl:pushAttrib(?GL_ALL_ATTRIB_BITS),
     gl:viewport(trunc(Ox), trunc(Oy), ?PREVIEW_SIZE, ?PREVIEW_SIZE),
     gl:matrixMode(?GL_PROJECTION),
@@ -608,7 +710,7 @@ mat_preview(X, Y, _W, _H, Common, Maps) ->
     gl:pushMatrix(),
     gl:loadIdentity(),
     gl:translatef(0.0, 0.0, -2.0),
-    wings_light:modeling_lights(camera,mat_preview),
+    wings_light:camera_lights(mat_preview),
     gl:shadeModel(?GL_SMOOTH),
     Alpha = gb_trees:get(opacity, Common),
     Amb = preview_mat(ambient, Common, Alpha),
