@@ -3,12 +3,12 @@
 %%
 %%     Maintains the outliner window.
 %%
-%%  Copyright (c) 2003-2004 Bjorn Gustavsson
+%%  Copyright (c) 2003-2008 Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 %%
-%%     $Id: wings_outliner.erl 178 2008-01-01 10:11:00Z bjorng $
+%%     $Id$
 %%
 
 -module(wings_outliner).
@@ -18,8 +18,7 @@
 -define(NEED_OPENGL, 1).
 -include("wings.hrl").
 -include("e3d_image.hrl").
--import(lists, [map/2,foldl/3,reverse/1,reverse/2,
-		keymember/3,keysearch/3,keydelete/3,sort/1]).
+-import(lists, [foldl/3,reverse/1,keydelete/3]).
 
 %%%
 %%% Outliner window.
@@ -81,7 +80,7 @@ event(got_focus, _) ->
     Msg = wings_msg:button_format(?__(1,"Select"), [],
 				  ?__(2,"Show outliner menu (if selection) or creation menu (if no selection)")),
     wings_wm:message(Msg),
-    wings_wm:dirty();
+    keep;
 event({current_state,St}, Ost0) ->
     Ost = update_state(St, Ost0),
     update_scroller(Ost),
