@@ -15,13 +15,14 @@
 -export([event/2,event/3,hilite_event/3, hilite_event/4]).
 -export([do_pick/3]).
 -export([marquee_pick/3,paint_pick/3]).
+
 -define(NEED_OPENGL, 1).
 -define(NEED_ESDL, 1).
 -include("wings.hrl").
 -include("e3d.hrl").
 
--import(lists, [foreach/2,reverse/2,sort/1,map/2,min/1,
-		keysearch/3,member/2]).
+-import(lists, [foreach/2,reverse/2,sort/1,usort/1,map/2,min/1,
+		keyfind/3,member/2,keysort/2]).
 
 %% For ordinary picking.
 -record(pick,
@@ -195,9 +196,9 @@ insert_hilite_dl_1(#dlo{hilite=none}=D, _, _) -> D;
 insert_hilite_dl_1(D, _, _) -> D#dlo{hilite=none}.
 
 hilite_color({Id,Item}, #st{sel=Sel}) ->
-    Key = case keysearch(Id, 1, Sel) of
+    Key = case keyfind(Id, 1, Sel) of
 	      false -> unselected_hlite;
-	      {value,{Id,Items}} ->
+	      {Id,Items} ->
 		  case gb_sets:is_member(Item, Items) of
 		      false -> unselected_hlite;
 		      true -> selected_hlite
@@ -208,36 +209,36 @@ hilite_color({Id,Item}, #st{sel=Sel}) ->
 hilit_draw_sel(vertex, V, #dlo{src_we=#we{vp=Vtab}}) ->
     gl:pointSize(wings_pref:get_value(selected_vertex_size)),
     gl:'begin'(?GL_POINTS),
-    gl:vertex3fv(gb_trees:get(V, Vtab)),
+    gl:vertex3fv(array:get(V, Vtab)),
     gl:'end'();
 hilit_draw_sel(edge, Edge, #dlo{src_we=#we{es=Etab,vp=Vtab}}) ->
-    #edge{vs=Va,ve=Vb} = gb_trees:get(Edge, Etab),
+    #edge{vs=Va,ve=Vb} = array:get(Edge, Etab),
     gl:lineWidth(wings_pref:get_value(selected_edge_width)),
     gl:'begin'(?GL_LINES),
-    wpc_ogla:two(gb_trees:get(Va, Vtab),
-		 gb_trees:get(Vb, Vtab)),
+    gl:vertex3fv(array:get(Va, Vtab)),
+    gl:vertex3fv(array:get(Vb, Vtab)),
     gl:'end'();
-hilit_draw_sel(face, Face, D) ->
+hilit_draw_sel(face, Face, #dlo{vab=#vab{face_map=Map, face_vs=Vs}}) ->
     case wings_pref:get_value(selection_style) of
 	stippled -> gl:enable(?GL_POLYGON_STIPPLE);
 	solid -> ok
     end,
     gl:polygonMode(?GL_FRONT_AND_BACK, ?GL_FILL),
-    gl:'begin'(?GL_TRIANGLES),
-    wings_draw_util:unlit_face(Face, D),
-    gl:'end'(),
+    wings_draw_setup:enableVertexPointer(Vs),
+    {Start,NoElements} = array:get(Face,Map),
+    gl:drawArrays(?GL_TRIANGLES, Start, NoElements),
+    wings_draw_setup:disableVertexPointer(Vs),
     gl:disable(?GL_POLYGON_STIPPLE);
-hilit_draw_sel(body, _, #dlo{src_we=We}=D) ->
+hilit_draw_sel(body, _, #dlo{vab=#vab{face_vs=Vs}}=D) ->
     case wings_pref:get_value(selection_style) of
 	stippled -> gl:enable(?GL_POLYGON_STIPPLE);
 	solid -> ok
     end,
     gl:polygonMode(?GL_FRONT_AND_BACK, ?GL_FILL),
-    gl:'begin'(?GL_TRIANGLES),
-    foreach(fun(Face) ->
-		    wings_draw_util:unlit_face(Face, D)
-	    end, wings_we:visible(We)),
-    gl:'end'(),
+    wings_draw_setup:enableVertexPointer(Vs),
+    Count = wings_draw_setup:face_vertex_count(D),
+    gl:drawArrays(?GL_TRIANGLES, 0, Count),
+    wings_draw_setup:disableVertexPointer(Vs),
     gl:disable(?GL_POLYGON_STIPPLE).
 
 enhanced_hl_info(Base,#hl{redraw=#st{sel=[],shapes=Shs},prev={SelMode,_,{Obj,Elem}}})->
@@ -252,7 +253,7 @@ enhanced_hl_info(Base,#hl{redraw=#st{sel=[],shapes=Shs},prev={SelMode,_,{Obj,Ele
                                 wings_util:nice_float(Y),
                                 wings_util:nice_float(Z)])];
           edge -> 
-            #edge{vs=Va,ve=Vb} = gb_trees:get(Elem, We#we.es),
+            #edge{vs=Va,ve=Vb} = array:get(Elem, We#we.es),
             {Xa,Ya,Za} = wings_vertex:pos(Va, We),
             {Xb,Yb,Zb} = wings_vertex:pos(Vb, We),
             Length = e3d_vec:dist({Xa,Ya,Za}, {Xb,Yb,Zb}),
@@ -417,7 +418,7 @@ marquee_convert_1(Faces0, face, Rect, #we{vp=Vtab}=We) ->
 	     end, [], Faces0, We),
     Vfs = wings_util:rel2fam(Vfs0),
     Kill0 = [Fs || {V,Fs} <- Vfs,
-		   not is_inside_rect(gb_trees:get(V, Vtab), Rect)],
+		   not is_inside_rect(array:get(V, Vtab), Rect)],
     Kill1 = sofs:set(Kill0, [[face]]),
     Kill = sofs:union(Kill1),
     Faces1 = sofs:from_external(Faces0, [face]),
@@ -425,16 +426,16 @@ marquee_convert_1(Faces0, face, Rect, #we{vp=Vtab}=We) ->
     sofs:to_external(Faces);
 marquee_convert_1(Faces, vertex, Rect, #we{vp=Vtab}=We) ->
     Vs = wings_face:to_vertices(Faces, We),
-    [V || V <- Vs, is_inside_rect(gb_trees:get(V, Vtab), Rect)];
+    [V || V <- Vs, is_inside_rect(array:get(V, Vtab), Rect)];
 marquee_convert_1(Faces, edge, Rect, #we{vp=Vtab}=We) ->
     Es0 = wings_face:fold_faces(fun(_, _, E, Rec, A) ->
 					[{E,Rec}|A]
 				end, [], Faces, We),
     Es = ordsets:from_list(Es0),
     [E || {E,#edge{vs=Va,ve=Vb}} <- Es,
-	  is_all_inside_rect([gb_trees:get(Va, Vtab),gb_trees:get(Vb, Vtab)], Rect)];
+	  is_all_inside_rect([array:get(Va, Vtab),array:get(Vb, Vtab)], Rect)];
 marquee_convert_1(_Faces, body, Rect, #we{vp=Vtab}) ->
-    case is_all_inside_rect(gb_trees:values(Vtab), Rect) of
+    case is_all_inside_rect(array:sparse_to_list(Vtab), Rect) of
 	true -> [0];
 	false -> []
     end.
@@ -444,7 +445,7 @@ is_all_inside_rect([P|Ps], Rect) ->
 is_all_inside_rect([], _Rect) -> true.
 
 is_inside_rect({Px,Py,Pz}, {MM,PM,ViewPort,X1,Y1,X2,Y2}) ->
-    {Sx,Sy,_} = glu:project(Px, Py, Pz, MM, PM, ViewPort),
+    {Sx,Sy,_} = wings_gl:project(Px, Py, Pz, MM, PM, ViewPort),
     X1 < Sx andalso Sx < X2 andalso
 	Y1 < Sy andalso Sy < Y2.
 
@@ -468,9 +469,14 @@ marquee_update_sel(Op, Hits0, #st{selmode=body}=St) ->
     Zero = sofs:from_term([0], [data]),
     Hits = sofs:constant_function(Hits2, Zero),
     marquee_update_sel_1(Op, Hits, St);
-marquee_update_sel(Op, Hits0, St) ->
-    Hits1 = sofs:relation(Hits0, [{id,data}]),
-    Hits = sofs:relation_to_family(Hits1),
+marquee_update_sel(Op, Hits0, #st{selmode=Mode}=St) ->
+    Hits1 = wings_util:rel2fam(Hits0),
+    Hits2 = [{Id,begin
+		     Items1 = gb_sets:from_ordset(Items0),
+		     Items = expand_light_items(Mode, Id, Items1, St),
+		     gb_sets:to_list(Items)
+		 end} || {Id,Items0} <- Hits1],
+    Hits = sofs:from_external(Hits2, [{id,[data]}]),
     marquee_update_sel_1(Op, Hits, St).
 
 marquee_update_sel_1(add, Hits, #st{sel=Sel0}=St) ->
@@ -517,176 +523,109 @@ pick_event(#mousebutton{button=B,state=?SDL_RELEASED}, #pick{st=St})
     pop;
 pick_event(_, _) -> keep.
 
+%% do_pick(X, Y, St0) -> {add|delete,{Id,Item,original|mirror},St}.
+%%  Pick the item (body, face, edge, or vertex, depending on the selection
+%%  mode in St0) and either add it to the selection or delete it from
+%%  the selection.
+%%
+%%  The first element in the returned tuple will indicate whether
+%%  something was added or deleted from the selection, and the second
+%%  element will indicate which item in which object was picked.
+%%
+%%  NOTE: Lights will only be selected in their entirety.
+%%  So in face mode, for instance, a light will have either all of
+%%  its faces selected or none.
+%%
+-type vm_mirror_side() :: 'original'|'mirror'.
+-spec do_pick(non_neg_integer(), non_neg_integer(), #st{}) ->
+    'none' | {'add'|'delete',{non_neg_integer(),non_neg_integer(),
+			      vm_mirror_side()},#st{}}.
 do_pick(X, Y, St) ->
     case raw_pick(X, Y, St) of
 	none ->
 	    none;
-	Hit ->	    % Hit is {edge,original,{1,1}}
-	    %Mod = sdl_keyboard:getModState(),
-	    %LKey = Mod band ?KMOD_LALT =/= 0,
-	    %RKey = Mod band ?KMOD_RALT =/= 0,
-	    %% LKey = Mod band ?KMOD_LMETA =/= 0,
-	    %% RKey = Mod band ?KMOD_RMETA =/= 0,
-	    %if
-		%LKey ->
-		%    {_,_,St2} = update_selection(Hit, St),
-		%    St3 = wings_edge_loop:select_loop(St2),
-		%    {add,original,St3};
-		%RKey ->
-		%    {_,_,St2} = update_selection(Hit, St),
-		%    St3 = wings_edge:select_edge_ring(St2),
-		%    {add,original,St3};
-		%true ->
-		    update_selection(Hit, St)
-	    %end
+	Hit ->
+	    update_selection(Hit, St)
     end.
 
-raw_pick(X0, Y0, #st{selmode=Mode}=St) ->
-    HitBuf = get(wings_hitbuf),
-    gl:selectBuffer(?HIT_BUF_SIZE, HitBuf),
-    gl:renderMode(?GL_SELECT),
-    gl:initNames(),
-    gl:matrixMode(?GL_PROJECTION),
-    gl:loadIdentity(),
+raw_pick(X0, Y0, St) ->
     {W,H} = wings_wm:win_size(),
     X = float(X0),
     Y = H-float(Y0),
     S = 5,
-    glu:pickMatrix(X, Y, S, S, {0,0,W,H}),
-    wings_view:projection(),
-    wings_view:modelview(),
+    set_pick_matrix(X, Y, S, S, W, H),
     case wings_wm:lookup_prop(select_backface) of
 	{value,true} ->
-	    draw();
+	    %% Only in AutoUV windows.
+	    wpc_pick:cull(false);
 	_ ->
-	    gl:enable(?GL_CULL_FACE),
-	    draw(),
-	    gl:disable(?GL_CULL_FACE)
+	    wpc_pick:cull(true)
     end,
-    Hits = get_hits(HitBuf),
-    case best_face_hit(Hits, Mode) of
-	none -> none;
-	{Id,Face} -> convert_hit(Id, Face, X, Y, St)
+    case dlo_pick(St, true) of
+	[] -> none;
+	[{Id,Face}] -> convert_hit(Id, Face, X, Y, St)
     end.
 
-update_selection({Mode,MM,{Id,Item}}, #st{sel=Sel0}=St) ->
-    {Type,Sel} = update_selection(Id, Item, Sel0, []),
-    {Type,MM,St#st{selmode=Mode,sel=Sel,sh=false}}.
+set_pick_matrix(X, Y, Xs, Ys, W, H) ->
+    gl:matrixMode(?GL_PROJECTION),
+    gl:loadIdentity(),
+    wpc_pick:pick_matrix(X, Y, Xs, Ys, {0,0,W,H}),
+    wings_view:projection(),
+    wings_view:modelview(),
+    set_pick_matrix().
 
-update_selection(Id, Item, [{I,_}=H|T], Acc) when Id > I ->
-    update_selection(Id, Item, T, [H|Acc]);
-update_selection(Id, Item, [{I,_}|_]=T, Acc) when Id < I ->
-    {add,reverse(Acc, [{Id,gb_sets:singleton(Item)}|T])};
-update_selection(Id, Item, [{_,Items0}|T0], Acc) -> %Id == I
-    case gb_sets:is_member(Item, Items0) of
+set_pick_matrix() ->
+    ModelMatrix = gl:getDoublev(?GL_MODELVIEW_MATRIX),
+    ProjMatrix = gl:getDoublev(?GL_PROJECTION_MATRIX),
+    wpc_pick:matrix(ModelMatrix, ProjMatrix).
+
+%% update_selection({Mode,MM,{Id,Item}}, St0) ->
+%%                 {add|delete,{Id,Item,MM},St0}.
+%%    Mode = body|face|edge|vertex
+%%    MM = original|mirror
+%%  If the item (body/face/edge/vertex) was not selected, add it
+%%  to the selection, otherwise delete it from the selection.
+%%  The first element in the returned tuple will indicate whether
+%%  something was added or deleted from the selection.
+%%
+%%  NOTE: Lights must only be selected in their entirety.
+%%  So in face mode, for instance, a light will have either all of
+%%  its faces selected or none.
+%%
+update_selection({Mode,MM,{Id,Item}}, #st{sel=Sel0}=St) ->
+    Items = expand_light_items(Mode, Id, gb_sets:singleton(Item), St),
+    {Type,Sel} = update_selection(Id, Items, Sel0, []),
+    {Type,{Id,Item,MM},St#st{selmode=Mode,sel=Sel,sh=false}}.
+
+update_selection(Id, Items, [{I,_}=H|T], Acc) when Id > I ->
+    update_selection(Id, Items, T, [H|Acc]);
+update_selection(Id, Items, [{I,_}|_]=T, Acc) when Id < I ->
+    {add,reverse(Acc, [{Id,Items}|T])};
+update_selection(Id, Items1, [{_,Items0}|T0], Acc) -> %Id == I
+    case gb_sets:is_disjoint(Items1, Items0) of
 	true ->
-	    Items = gb_sets:delete(Item, Items0),
+	    %% Add to selection.
+	    Items = gb_sets:union(Items0, Items1),
+	    {add,reverse(Acc, [{Id,Items}|T0])};
+	false ->
+	    %% Delete from selection.
+	    Items = gb_sets:difference(Items0, Items1),
 	    T = case gb_sets:is_empty(Items) of
 		    true -> T0;
 		    false -> [{Id,Items}|T0]
 		end,
-	    {delete,reverse(Acc, T)};
-	false ->
-	    Items = gb_sets:insert(Item, Items0),
-	    {add,reverse(Acc, [{Id,Items}|T0])}
+	    {delete,reverse(Acc, T)}
     end;
-update_selection(Id, Item, [], Acc) ->
-    {add,reverse(Acc, [{Id,gb_sets:singleton(Item)}])}.
+update_selection(Id, Items, [], Acc) ->
+    {add,reverse(Acc, [{Id,Items}])}.
 
-%%%
-%%% Pick up raw hits.
-%%%
-
-get_hits(HitBuf) ->
-    gl:flush(),
-    case gl:renderMode(?GL_RENDER) of
-	0 -> [];
-	NumHits ->
- 	    HitData = sdl_util:read(HitBuf, 5*NumHits),
- 	    get_hits_1(NumHits, HitData, [])
+expand_light_items(Mode, Id, Items, #st{shapes=Shs}) ->
+    We = gb_trees:get(Id, Shs),
+    case ?IS_LIGHT(We) of
+	false -> Items;
+	true -> wings_sel:get_all_items(Mode, We)
     end.
 
-get_hits_1(0, _, Acc) -> Acc;
-get_hits_1(N, [2,_,_,A,B|T], Acc) ->
-    get_hits_1(N-1, T, [{A,B}|Acc]).
-
-%%%
-%%% Filter face hits to obtain just one hit.
-%%%
-
-best_face_hit([], _) -> none;
-best_face_hit([Hit], _) -> Hit;
-best_face_hit([{Id,_}|T]=Hits, body) ->
-    %% If all hits are in the same body we can return any hit.
-    best_face_hit_body(T, Id, Hits);
-best_face_hit(Hits, _) -> best_face_hit_1(Hits).
-
-best_face_hit_body([{Id,_}|T], Id, Hits) ->
-    best_face_hit_body(T, Id, Hits);
-best_face_hit_body([_|_], _, Hits) ->
-    %% Different bodies.  Must find the nearest face the in the usual way.
-    best_face_hit_1(Hits);
-best_face_hit_body([], _, [Hit|_]) -> Hit.
-
-best_face_hit_1(Hits0) ->
-    Hits = sort([{abs(Id),Id,Face} || {Id,Face} <- Hits0]),
-    {_,_,W,H} =  wings_wm:viewport(),
-    Model = gl:getDoublev(?GL_MODELVIEW_MATRIX),
-    Proj = gl:getDoublev(?GL_PROJECTION_MATRIX),
-    ViewPort = {0,0,W,H},
-    Orig = glu:unProject(0, 0, 0, Model, Proj, ViewPort),
-    Dir0 = glu:unProject(0, 0, 0.5, Model, Proj, ViewPort),
-    Dir = e3d_vec:norm_sub(Dir0, Orig),
-    ViewRay = {Orig,Dir},
-    {_,Best,_} = wings_dl:fold(fun(D, A) ->
-				       best_face_hit_1(D, ViewRay, A) end,
-			       {Hits,none,infinite}),
-    Best.
-
-best_face_hit_1(#dlo{src_we=#we{id=Id}=We,ns=Ns}, EyePoint,
-		{[{Id,_,_}|_],_,_}=A) ->
-    best_face_hit_2(We, Ns, EyePoint, A);
-best_face_hit_1(_, _, A) -> A.
-
-best_face_hit_2(#we{id=Id,vp=Vtab}=We, Ns, Ray, {[{Id,Id,Face}|Hits],Hit0,T0})
-  when ?IS_LIGHT(We) ->
-    {Orig,Dir} = Ray,
-    P = gb_trees:get(1, Vtab),
-    T = e3d_vec:dot(Dir, P) - e3d_vec:dot(Dir, Orig),
-    A = if
-	    T < T0 ->
-		{Hits,{Id,Face},T};
-	    true ->
-		{Hits,Hit0,T0}
-	end,
-    best_face_hit_2(We, Ns, Ray, A);
-best_face_hit_2(#we{id=AbsId}=We, Ns, Ray, {[{AbsId,Id,Face}|Hits],Hit0,T0}) ->
-    case gb_trees:get(Face, Ns) of
-	[N|[P0|_]] -> ok;
-	{N,_,[P0|_]} -> ok
-    end,
-    P = if 
-	    Id < 0 ->
-		e3d_mat:mul_point(wings_dl:mirror_matrix(AbsId), P0);
-	    true ->
-		P0
-	end,
-    {Orig,Dir} = Ray,
-    A = case e3d_vec:dot(N, Dir) of
-	    Den when abs(Den) < 1.0e-20 ->
-		{Hits,Hit0,T0};
-	    Den ->
-		T = (e3d_vec:dot(N, P) - e3d_vec:dot(N, Orig)) / Den,
-		if
-		    T < T0 ->
-			{Hits,{Id,Face},T};
-		    true ->
-			{Hits,Hit0,T0}
-		end
-	end,
-    best_face_hit_2(We, Ns, Ray, A);
-best_face_hit_2(_, _, _, A) -> A.
-
 %%
 %% Given a face selection hit, return the correct vertex/edge/face/body.
 %%
@@ -774,8 +713,8 @@ find_edge(Face, We, Cx, Cy, Trans) ->
 
 project_vertex(V, We, {ModelMatrix,ProjMatrix,ViewPort}) ->
     {Px,Py,Pz} = wings_vertex:pos(V, We),
-    {Xs,Ys,_} = glu:project(Px, Py, Pz, ModelMatrix,
-			    ProjMatrix, ViewPort),
+    {Xs,Ys,_}  = wings_gl:project(Px, Py, Pz, ModelMatrix,
+				  ProjMatrix, ViewPort),
     {Xs,Ys}.
 
 check_restriction({Mode,MM,_}=Hilite, Id, V, Edge, Face) ->
@@ -817,160 +756,156 @@ restrict_hilite(face, Modes, Id, V, Edge, _Face, MM) ->
 pick_all(_DrawFaces, _X, _Y, W, H, St) when W < 1.0; H < 1.0 ->
     {none,St};
 pick_all(DrawFaces, X, Y0, W, H, St) ->
-    HitBuf = get(wings_hitbuf),
-    gl:selectBuffer(?HIT_BUF_SIZE, HitBuf),
-    gl:renderMode(?GL_SELECT),
-    gl:initNames(),
-    gl:matrixMode(?GL_PROJECTION),
-    gl:loadIdentity(),
     {Ww,Wh} = wings_wm:win_size(),
     Y = Wh-Y0,
-    glu:pickMatrix(X, Y, W, H, [0,0,Ww,Wh]),
-    wings_view:projection(),
-    wings_view:modelview(),
+    set_pick_matrix(X, Y, W, H, Ww, Wh),
     case DrawFaces of
 	true ->
-	    gl:enable(?GL_CULL_FACE),
-	    draw(),
-	    gl:disable(?GL_CULL_FACE);
-	false -> marquee_draw(St)
-    end,
-    {get_hits(HitBuf),St}.
+	    wpc_pick:cull(true),
+	    {dlo_pick(St, false),St};
+	false ->
+	    wpc_pick:cull(false),
+	    {marquee_pick(St),St}
+    end.
 
-marquee_draw(#st{selmode=edge}) ->
-    Draw = fun(#we{es=Etab,vp=Vtab}=We) ->
-		   Vis = gb_sets:from_ordset(wings_we:visible(We)),
-		   marquee_draw_edges(gb_trees:to_list(Etab), Vtab, Vis)
-	   end,
-    marquee_draw_1(Draw);
-marquee_draw(#st{selmode=vertex}) ->
-    Draw = fun(#we{vp=Vtab}=We) ->
-		   case wings_we:any_hidden(We) of
-		       false ->
-			   marquee_draw_all_vs(gb_trees:to_list(Vtab));
-		       true ->
-			   marquee_draw_some_vs(wings_we:visible_vs(We),
-						Vtab)
-		   end
-	   end,
-    marquee_draw_1(Draw);
-marquee_draw(_) -> draw().
+marquee_pick(#st{selmode=edge}) ->
+    PickFun = fun(#we{es=Etab,vp=Vtab,id=Id}=We, Acc) ->
+		      Vis = gb_sets:from_ordset(wings_we:visible(We)),
+		      EsPos = visible_edges(array:sparse_to_orddict(Etab),
+					    Vtab, Vis, []),
+		      case wpc_pick:edges(EsPos) of
+			  [] ->
+			      Acc;
+			  Picked -> [{Id,E} || E <- Picked] ++ Acc
+		      end
+	      end,
+    setup_pick_context(PickFun);
+marquee_pick(#st{selmode=vertex}) ->
+    PickFun = fun(#we{vp=Vtab,id=Id}=We, Acc) ->
+		      VsPos = case wings_we:is_open(We) of
+				  false ->
+				      array:sparse_to_orddict(Vtab);
+				  true ->
+				      Vs = wings_we:visible_vs(We),
+				      [{V,array:get(V, Vtab)} || V <- Vs]
+			      end,
+		      case wpc_pick:vertices(VsPos) of
+			  [] ->
+			      Acc;
+			  Picked ->
+			      [{Id,V} || V <- Picked] ++ Acc
+		      end
+	      end,
+    setup_pick_context(PickFun);
+marquee_pick(St) -> dlo_pick(St, false).
 
-marquee_draw_all_vs([{V,Pos}|VsPos]) ->
-    gl:loadName(V),
-    gl:'begin'(?GL_POINTS),
-    gl:vertex3fv(Pos),
-    gl:'end'(),
-    marquee_draw_all_vs(VsPos);
-marquee_draw_all_vs([]) -> ok.
-
-marquee_draw_some_vs([V|Vs], Vtab) ->
-    gl:loadName(V),
-    gl:'begin'(?GL_POINTS),
-    gl:vertex3fv(gb_trees:get(V, Vtab)),
-    gl:'end'(),
-    marquee_draw_some_vs(Vs, Vtab);
-marquee_draw_some_vs([], _) -> ok.
-
-marquee_draw_edges([{Edge,#edge{vs=Va,ve=Vb,lf=Lf,rf=Rf}}|Es], Vtab, Vis) ->
+visible_edges([{Edge,#edge{vs=Va,ve=Vb,lf=Lf,rf=Rf}}|Es], Vtab, Vis, Acc) ->
     case gb_sets:is_member(Lf, Vis) orelse gb_sets:is_member(Rf, Vis) of
-	false -> ok;
+	false ->
+	    visible_edges(Es, Vtab, Vis, Acc);
 	true ->
-	    gl:loadName(Edge),
-	    gl:'begin'(?GL_LINES),
-	    wpc_ogla:two(gb_trees:get(Va, Vtab),
-			 gb_trees:get(Vb, Vtab)),
-	    gl:'end'()
-    end,
-    marquee_draw_edges(Es, Vtab, Vis);
-marquee_draw_edges([], _, _) -> ok.
+	    E = {Edge,{array:get(Va, Vtab),array:get(Vb, Vtab)}},
+	    visible_edges(Es, Vtab, Vis,  [E|Acc])
+    end;
+visible_edges([], _, _, Acc) -> Acc.
 
-marquee_draw_1(Draw) ->
-    wings_dl:fold(fun(D, _) -> marquee_draw_fun(D, Draw) end, []).
+setup_pick_context(PickFun) ->
+    Res = wings_dl:fold(fun(D, Acc) ->
+				setup_pick_context_fun(D, PickFun, Acc)
+			end, []),
+    sort(Res).
 
-marquee_draw_fun(#dlo{src_we=#we{perm=Perm}}, _) when not ?IS_SELECTABLE(Perm) -> ok;
-marquee_draw_fun(#dlo{mirror=Mirror,src_we=#we{id=Id}=We}, Draw) ->
-    List = gl:genLists(1),
-    gl:newList(List, ?GL_COMPILE),
-    gl:pushName(0),
-    Draw(We),
-    gl:popName(),
-    gl:endList(),
-    gl:pushName(Id),
-    case Mirror of
-	none ->
-	    wings_dl:call(List);
-	Matrix ->
-	    wings_dl:call(List),
-	    gl:pushMatrix(),
-	    gl:multMatrixf(Matrix),
-	    wings_dl:call(List),
-	    gl:popMatrix()
-    end,
-    gl:popName(),
-    gl:deleteLists(List, 1).
+setup_pick_context_fun(#dlo{src_we=#we{perm=Perm}}, _PickFun, Acc)
+  when ?IS_NOT_SELECTABLE(Perm) ->
+    Acc;
+setup_pick_context_fun(#dlo{mirror=none,src_we=We}, PickFun, Acc) ->
+    PickFun(We, Acc);
+setup_pick_context_fun(#dlo{mirror=Matrix,src_we=We}, PickFun, Acc0) ->
+    Acc1 = PickFun(We, Acc0),
+    gl:pushMatrix(),
+    gl:multMatrixf(Matrix),
+    set_pick_matrix(),
+    wpc_pick:front_face(cw),
+    Acc = PickFun(We, Acc1),
+    wpc_pick:front_face(ccw),
+    gl:popMatrix(),
+    set_pick_matrix(),
+    Acc.
 
 %%
 %% Draw for the purpose of picking the items that the user clicked on.
 %%
 
-draw() ->
-    wings_dl:map(fun draw_fun/2, []).
+dlo_pick(St, OneHit) ->
+    Hits0 = wings_dl:map(fun(D, Acc) ->
+				 do_dlo_pick(D, St, OneHit, Acc)
+			 end, []),
+    case Hits0 of
+	{_,Hits} ->
+	    %% Only one hit.
+	    Hits;
+	_ ->
+	    %% All hits.
+	    usort(Hits0)
+    end.
 
-draw_fun(#dlo{work=Work,src_we=#we{id=Id,perm=Perm}=We}=D, _)
-  when ?IS_LIGHT(We), ?IS_SELECTABLE(Perm) ->
-    gl:pushName(Id),
-    gl:pushName(1),
-    wings_dl:call(Work),
-    gl:popName(),
-    gl:popName(),
-    D;
-draw_fun(#dlo{pick=none}=D, _) ->
-    List = gl:genLists(1),
-    gl:newList(List, ?GL_COMPILE),
-    draw_1(D),
-    gl:endList(),
-    draw_dlist(D#dlo{pick=List});
-draw_fun(D, _) -> draw_dlist(D).
-
-draw_dlist(#dlo{mirror=none,pick=Pick,src_we=#we{id=Id}}=D) ->
-    gl:pushName(Id),
-    wings_dl:call(Pick),
-    gl:popName(),
-    D;
-draw_dlist(#dlo{mirror=Matrix,pick=Pick,src_we=#we{id=Id}}=D) ->
-    gl:pushName(Id),
-    wings_dl:call(Pick),
-    gl:loadName(-Id),
-    gl:frontFace(?GL_CW),
+do_dlo_pick(#dlo{src_we=#we{perm=Perm}}=D, _St, _OneHit, Acc)
+  when ?IS_NOT_SELECTABLE(Perm) ->
+    {D,Acc};
+do_dlo_pick(D=#dlo{vab=none}, St, OneHit, Acc) ->
+    do_dlo_pick(wings_draw_setup:work(D, St), St, OneHit, Acc);
+do_dlo_pick(D=#dlo{vab=#vab{face_vs=none}}, St, OneHit, Acc) ->
+    do_dlo_pick(wings_draw_setup:work(D, St), St, OneHit, Acc);
+do_dlo_pick(#dlo{mirror=none,src_we=#we{id=Id}=We}=D, _, OneHit, Acc)
+  when ?IS_AREA_LIGHT(We) ->
+    wpc_pick:cull(false),
+    Res = do_dlo_pick_0(Id, D, OneHit, Acc),
+    wpc_pick:cull(true),
+    Res;
+do_dlo_pick(#dlo{mirror=none,src_we=#we{id=Id}}=D, _, OneHit, Acc) ->
+    do_dlo_pick_0(Id, D, OneHit, Acc);
+do_dlo_pick(#dlo{mirror=Matrix,src_we=#we{id=Id}}=D0, _, OneHit, Acc0) ->
+    {D1,Acc1} = do_dlo_pick_0(Id, D0, OneHit, Acc0),
     gl:pushMatrix(),
     gl:multMatrixf(Matrix),
-    wings_dl:call(Pick),
+    set_pick_matrix(),
+    wpc_pick:front_face(cw),
+    {D,Acc} = do_dlo_pick_0(-Id, D1, OneHit, Acc1),
+    wpc_pick:front_face(ccw),
     gl:popMatrix(),
-    gl:popName(),
-    gl:frontFace(?GL_CCW),
-    D.
+    set_pick_matrix(),
+    {D,Acc}.
 
-draw_1(#dlo{ns=Ns0,src_we=#we{perm=Perm}=We})
-  when ?IS_SELECTABLE(Perm) ->
-    Ns = wings_we:visible(gb_trees:to_list(Ns0), We),
-    gl:pushName(0),
-    foreach(fun({Face,Info}) ->
-		    gl:loadName(Face),
-		    face(Info)
-	    end, Ns),
-    gl:popName();
-draw_1(_) -> ok.
+do_dlo_pick_0(Id, #dlo{vab=#vab{face_vs=Vs,face_map=Map0}}=D0, OneHit, Acc0) ->
+    case wpc_pick:faces(Vs, OneHit) of
+	[] ->
+	    %% No hit.
+	    {D0,Acc0};
+	{Hit0,Depth} ->
+	    %% OneHit is true - only the nearest hit is returned.
+	    case Acc0 of
+		{PrevDepth,_} when PrevDepth < Depth ->
+		    {D0,Acc0};
+		_ ->
+		    {D,Map} = dlo_tri_map(D0, Map0),
+		    Acc = do_dlo_pick_1([Hit0], Map, Id, []),
+		    {D,{Depth,Acc}}
+	    end;
+	RawHits ->
+	    %% OneHit is false - all hits are returned.
+	    {D,Map} = dlo_tri_map(D0, Map0),
+	    {D,do_dlo_pick_1(RawHits, Map, Id, Acc0)}
+    end.
 
-face([_|[A,B,C]]) ->
-    gl:'begin'(?GL_TRIANGLES),
-    wpc_ogla:tri(A, B, C),
-    gl:'end'();
-face([_|[A,B,C,D]]) ->
-    gl:'begin'(?GL_QUADS),
-    wpc_ogla:quad(A, B, C, D),
-    gl:'end'();
-face({_,Fs,VsPos}) ->
-    gl:'begin'(?GL_TRIANGLES),
-    wings__du:plain_face(Fs, VsPos),
-    gl:'end'().
+do_dlo_pick_1([H|Hits], [{Face,{Start,Num}}|_]=T, Id, Acc)
+  when Start =< H, H < Start+Num ->
+    do_dlo_pick_1(Hits, T, Id, [{Id,Face}|Acc]);
+do_dlo_pick_1([_|_]=Hits, [_|T], Id, Acc) ->
+    do_dlo_pick_1(Hits, T, Id, Acc);
+do_dlo_pick_1([], [_|_], _, Acc) -> Acc;
+do_dlo_pick_1([], [], _, Acc) -> Acc.
+
+dlo_tri_map(#dlo{tri_map=none}=D, Map0) ->
+    Map1 = keysort(2, array:sparse_to_orddict(Map0)),
+    {D#dlo{tri_map=Map1},Map1};
+dlo_tri_map(#dlo{tri_map=TriMap}=D, _) -> {D,TriMap}.
