@@ -3,7 +3,7 @@
 %%
 %%     Preference management.
 %%
-%%  Copyright (c) 2001-2008 Bjorn Gustavsson
+%%  Copyright (c) 2001-2009 Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -22,7 +22,7 @@
 
 -define(NEED_ESDL, 1).	  %% Some keybindings
 -include("wings.hrl").
--import(lists, [foreach/2,keysearch/3,reverse/1,sort/1,last/1,foldl/3]).
+-import(lists, [foreach/2,reverse/1,sort/1,last/1,foldl/3]).
 
 -define(MAC_PREFS, "Library/Preferences/Wings 3D Preferences.txt").
 -define(WIN32_OLD_PREFS, "Preferences").
@@ -82,7 +82,17 @@ finish() ->
     foreach(fun({Key,Val}) ->
 		    set_value(Key, Val)
 	    end, ets:tab2list(wings_delayed_update)),
-    PrefFile = new_pref_file(),
+    case new_pref_file() of
+	none ->
+	    %% No location for saving the preferences.
+	    %% The user has already been warned about this.
+	    %% Do nothing.
+	    ok;
+	PrefFile ->
+	    finish_save_prefs(PrefFile)
+    end.
+
+finish_save_prefs(PrefFile) ->
     List0 = ets:tab2list(wings_state),
     List = prune_defaults(List0),
     Format = "~p. \n",
@@ -128,7 +138,7 @@ prune_defaults(List) ->
 win32_save_maximized() ->
     case os:type() of
 	{win32,_} ->
-	    set_value(win32_start_maximized, sdl_video:wm_isMaximized());
+	    set_value(win32_start_maximized, wings_io:is_maximized());
 	_ ->
 	    ok
     end.
@@ -152,15 +162,36 @@ unix_pref() ->
     try_location(os:getenv("HOME"), ".wings").
 
 win32_pref() ->
+    case win32_appdata() of
+	none ->
+	    W = "Warning: Wings3D.exe provided no location "
+		"for the AppData folder.\n"
+		"Trying to locate preferences using the registry...\n",
+	    io:put_chars(W),
+	    self() ! {external,not_possible_to_save_prefs},
+	    win32_pref_registry(["AppData","Personal"]);
+	AppData ->
+	    case try_location(AppData, ?WIN32_PREFS) of
+		none ->
+		    %% We did not find a preference file
+		    %% in AppData. Continue searching using
+		    %% the registry.
+		    win32_pref_registry(["Personal"]);
+		File -> File
+	    end
+    end.
+
+%% Try to find old preferences using the registry.
+win32_pref_registry(RegistryLocations) ->
     {ok,R} = win32reg:open([read]),
-    Res = win32_pref_1(R, ["AppData","Personal"]),
+    Res = win32_pref_1(R, RegistryLocations),
     ok = win32reg:close(R),
     Res.
 
 %% Search for a preference file in "special folders", such as "AppData"
 %% and "My Documents".
 win32_pref_1(R, [FolderType|T]) ->
-    case win32_special_folder(R, FolderType) of
+    case wings_u:win32_special_folder(R, FolderType) of
 	none -> win32_pref_1(R, T);
 	Path ->
 	    case try_location(Path, ?WIN32_PREFS) of
@@ -168,93 +199,10 @@ win32_pref_1(R, [FolderType|T]) ->
 		File -> File
 	    end
     end;
-win32_pref_1(R, []) ->
-    case try_location(wings_util:lib_dir(wings), ?WIN32_PREFS) of
-	none -> win32_pref_2(R);
-	File -> File
-    end.
-
-%% No preferences found so far. Search in old installations of
-%% Wings for preference files.
-win32_pref_2(R) ->
-    case win32_9816(R) of
-	none -> win32_pref_pre9816(R);
-	File -> File
-    end.
-
-win32_9816(R) ->
-    %% Search for a preference file in a Wings installation in 0.98.16.
-    %% (Too bad... in a special place in this release only.)
-    case win32reg:change_key(R, "\\hklm\\SOFTWARE\\Wings 3D") of
-	ok ->
-	    case win32reg:sub_keys(R) of
-		{ok,SubKeys0} ->
-		    SubKeys = reverse(sort(SubKeys0)),
-		    {ok,Curr} = win32reg:current_key(R),
-		    win32_9816_1(R, SubKeys, Curr);
-		{error,_Error} ->
-		    %% Can't read sub keys - strange.
-		    none
-	    end;
-	{error,_Error} ->
-	    %% No 'Wings 3D' key (this is STRANGE)
-	    none
-    end.
-
-win32_9816_1(R, [K|Keys], Curr) ->
-    ok = win32reg:change_key(R, K),
-    Dir = reg_get_default(R),
-    ok = win32reg:change_key(R, Curr),
-    WingsDirs = filelib:wildcard(Dir++"/lib/wings-*"),
-    case try_locations(WingsDirs, ?WIN32_OLD_PREFS) of
-	none -> win32_9816_1(R, Keys, Curr);
-	File -> File
-    end;
-win32_9816_1(_, [], _) -> none.
-
-win32_pref_pre9816(R) ->
-    %% Search for a preference file in a Wings installation older than 0.98.16
-    %% using the uninstall string.
-    case win32reg:change_key(R, "\\hklm\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\Wings 3D") of
-	ok ->
-	    case win32reg:value(R, "UninstallString") of
-		{ok,Str0} ->
-		    Str = strip_quotes(Str0),
-		    try_location(filename:dirname(Str), ?WIN32_OLD_PREFS);
-		{error,_} -> none
-	    end;
-	{error,_} -> none
-    end.
-
-win32_special_folder(R, FolderType) ->
-    Key = "\\hkcu\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders",
-    case win32reg:change_key(R, Key) of
-	ok ->
-	    case win32reg:value(R, FolderType) of
-		{error,_} -> none;
-		{ok,Value} -> Value
-	    end;
-	_ -> error
-    end.
-
-reg_get_default(R) ->
-    %% There seems to be a bug in win32reg:value/2 preventing
-    %% us from retrieving the default value. Workaround follows.
-    case win32reg:values(R) of
-	{ok,Values} ->
-	    case keysearch(default, 1, Values) of
-		{value,{default,Val}} -> Val;
-		false -> ""
-	    end;
-	{error,_} -> ""
-    end.
-
-strip_quotes([$"|T0]) ->
-    case reverse(T0) of
-	[$"|T] -> reverse(T);
-	_ -> T0
-    end;
-strip_quotes(S) -> S.
+win32_pref_1(_, []) ->
+    %% This should normally never happen. See the "desperate
+    %% fallback" in win32_new_pref_1/2 below.
+    try_location(wings_util:lib_dir(wings), ?WIN32_PREFS).
 
 %%%
 %%% Return a suitable path for a new preference file.
@@ -271,33 +219,29 @@ new_pref_file() ->
     end.
 
 win32_new_pref() ->
-    {ok,R} = win32reg:open([read]),
-    Res = win32_new_pref_1(R, ["AppData","Personal"]),
-    ok = win32reg:close(R),
-    Res.
-
-win32_new_pref_1(R, [FolderType|T]) ->
-    case win32_special_folder(R, FolderType) of
-	none -> win32_new_pref_1(R, T);
-	Path ->
-	    File = filename:join(Path, ?WIN32_PREFS),
+    case win32_appdata() of
+	none ->
+	    none;
+	AppData ->
+	    File = filename:join(AppData, ?WIN32_PREFS),
 	    filelib:ensure_dir(File),
 	    File
-    end;
-win32_new_pref_1(_, []) ->
-    %% Desperate fallback... (no standard folders found - should not happen).
-    filename:join(wings_util:lib_dir(wings), ?WIN32_PREFS).
+    end.
 
 %%%
 %%% Utilities.
 %%%
 
-try_locations([D|Ds], File) ->
-    case try_location(D, File) of
-	none -> try_locations(Ds, File);
-	Name -> Name
-    end;
-try_locations([], _) -> none.
+win32_appdata() ->
+    case init:get_plain_arguments() of
+	[AppData] ->
+	    AppData;
+	[] ->
+	    %% No AppData location provided. Probably because
+	    %% a developer started Wings using a script that
+	    %% does not provide the location.
+	    none
+    end.
 
 try_location(Dir, File) ->
     Name = filename:join(Dir, File),
@@ -430,7 +374,6 @@ defaults() ->
      {normal_vector_color,{0.0,0.0,0.65}},
      {smart_highlighting,false},
      {material_default,{1.0,1.0,1.0}},
-     {material_hole,{0.0,0.0,0.9}},
 
      {selection_style,solid},
      {hide_sel_while_dragging,false},
@@ -443,6 +386,7 @@ defaults() ->
      {dummy_axis_letter,false},
      {polygon_offset_f,1.0},
      {polygon_offset_r,1.0},
+     {multisample,undefined},
 
      %% Advanced features.
      {default_commands,false},
@@ -453,7 +397,7 @@ defaults() ->
      {clip_plane_size,1.5},
      {highlight_aim_at_selected,false},
      {highlight_aim_at_unselected,true},
-	 {hilite_select,false},
+     {hilite_select,false},
      {start_in_tweak, false},
 
      %% Constraints preferences.
@@ -522,7 +466,9 @@ defaults() ->
      {title_passive_color,{0.325,0.4,0.325,1.0}},
      {title_text_color,{1.0,1.0,1.0}},
      {menu_bar_bg,{0.52,0.52,0.52}},
+     {menubar_text,{0.0,0.0,0.0}},
      {info_line_bg,{0.52,0.52,0.52}},
+     {info_line_text,{0.0,0.0,0.0}},
      {no_progress_bar,false},
      {interface_icons,bluecube},
      {objects_in_outliner,true},
@@ -541,7 +487,10 @@ defaults() ->
      {num_undo_levels,32},
 
      %% Languages.
-     {lang,en}
+     {lang,en},
+
+     %% Develop menu.
+     {show_develop_menu,false}
     ].
 
 clean(List) ->
@@ -563,6 +512,21 @@ clean([{{bindkey,_}=Bk,{view,{virtual_mirror,Cmd}},user}|T], Acc) ->
     clean(T, [{Bk,{tools,{virtual_mirror,Cmd}},user}|Acc]);
 clean([{{bindkey,_}=Bk,{tools,tweak},user}|T], Acc) ->
     clean(T, [{Bk,{tools,{tweak,false}},user}|Acc]);
+
+clean([{{bindkey,_,_}=Bk,{face,{Cmd,Dir}},user}|T], Acc)
+        when Cmd==extrude; Cmd==extrude_region; Cmd==extract_region; Cmd==extract_faces ->
+    {Type,Mode} = case Cmd of
+        extrude_region -> {extrude,region};
+        extrude -> {extrude,faces};
+        extract_region -> {extract,region};
+        extract_faces -> {extract,faces}
+    end,
+    C = case Dir of
+      {E,_} when E==region; E==faces -> Dir;
+      Dir -> {Mode,Dir}
+    end,
+    clean(T, [{Bk,{face,{Type,C}},user}|Acc]);
+
 clean([{{bindkey,_},Cmd,user}=Bk|T], Acc) ->
     case bad_command(Cmd) of
 	false -> clean(T, [Bk|Acc]);
@@ -605,7 +569,8 @@ not_bad(wire_edge_color, _) -> false;
 not_bad(show_wire_backfaces, _) -> false;	%Now a window property.
 not_bad(smoothed_preview_cage, _) -> false;
 not_bad(smoothed_preview_edges, _) -> false;
-not_bad(contour,_) -> false;
+not_bad(contour, _) -> false;
+not_bad(material_hole, _) -> false;
 
 %% Crashes have occurred.
 not_bad(last_axis, Val) -> is_wings_vector(Val);
