@@ -4,7 +4,7 @@
 %%
 %%     Preference management.
 %%
-%%  Copyright (c) 2001-2008 Bjorn Gustavsson
+%%  Copyright (c) 2001-2009 Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -18,7 +18,7 @@
 -include("wings.hrl").
 -import(wings_pref, [get_value/1,set_value/2]).
 
--import(lists, [reverse/1]).
+-import(lists, [reverse/1,foreach/2]).
 -define(NONZERO,1.0e-6).
 
 menu(_St) ->
@@ -121,14 +121,11 @@ gen_prefs() ->
 	    {color,?__(46,"Color"),normal_vector_color,
 	     [{info,?__(47,"Color of normals")}]}]}],
 	 [{title,?__(48,"Normals Display")}]},
-  {vframe,
-	[{label_column,
-	 [{color,?__(51,"Default"),material_default,
-	  [{info,?__(52,"Sets the colour of the default material")}]},
-	  {color,?__(53,"Hole"),material_hole,
-	  [{info,?__(54,"Sets the colour of the hole material")}]}]}],
+	{vframe,
+	 [{label_column,
+	   [{color,?__(51,"Default"),material_default,
+	     [{info,?__(52,"Sets the colour of the default material")}]}]}],
 	 [{title,?__(55,"Materials")}]}]},
-	
 	  {hframe,
 	   [{vframe,
 	 [{?__(36,"Show Axis Letters"),show_axis_letters},
@@ -152,12 +149,19 @@ gen_prefs() ->
 	   [{info,?__(34,"Always show the grid when the view is aligned along one of the major axes")}]},
 	    {?__(aa_ortho1,"Force Axis-Aligned Ortho"),force_ortho_along_axis,
 	   [{info, ?__(aa_ortho2,"Always go into orthogonal mode when the view is aligned along one of the major axes")}]}],
-     [{title,?__(35,"Grid")}]},
+     [{title,?__(35,"Grid")}]}|multisampling()]}.
      
-  {hframe,
-	 [{"Enable Multi-Sampling",multisample, 
-	  [{info, "Anti-Alias the whole interface, including polygons."}]}],
-     [{title,"Anti-Aliasing"}]}]}.
+multisampling() ->
+    case wings_pref:get_value(multisample) of
+	undefined -> [];
+	_ ->
+	    [{hframe,
+	      [{"Enable Multi-Sampling",multisample,
+		[{info,"Anti-Alias the whole interface, "
+		  "including polygons."}]}],
+	      [{title,"Anti-Aliasing"}]}]
+    end.
+
 
 advanced_prefs() ->
 %%     DisableHook = fun (is_disabled, {_Var,_I,Store}) ->
@@ -424,7 +428,8 @@ ui_prefs() ->
 	      {label,?__(11,"Title (Passive) Background")},
 	      {label,?__(10,"Title Text")},
 	      {label,?__(2,"Desktop/Geometry Background")},
-	      {label,?__(28,"Menu Bar Background")},
+	      {label,?__(28,"Menubar Background")},
+	      {label,?__(31,"Menubar Text")},
 	      {label,?__(6,"Menu Background")},
 	      {label,?__(3,"Menu Text")},
 	      {label,?__(4,"Menu Highlight")},
@@ -432,13 +437,15 @@ ui_prefs() ->
 	      {label,?__(9,"Dialog Background")},
 	      {label,?__(7,"Dialog Text")},
 	      {label,?__(8,"Dialog (Disabled) Text")},
-	      {label,?__(29,"Info Line Background")}]},
+	      {label,?__(29,"Info Line Background")},
+	      {label,?__(30,"Info Line Text")}]},
 	    {vframe,
 	     [{color,title_active_color},
 	      {color,title_passive_color},
 	      {color,title_text_color},
 	      {color,background_color},
 	      {color,menu_bar_bg},
+	      {color,menubar_text},
 	      {color,menu_color},
 	      {color,menu_text},
 	      {color,menu_hilite},
@@ -446,7 +453,8 @@ ui_prefs() ->
 	      {color,dialog_color},
 	      {color,dialog_text},
 	      {color,dialog_disabled},
-	      {color,info_line_bg}]}],
+	      {color,info_line_bg},
+	      {color,info_line_text}]}],
 	   [{title,?__(13,"Colors")}]}
 	 ]},
 	{vframe,
@@ -577,7 +585,10 @@ misc_prefs() ->
 		 {text,polygon_offset_f,[{range,{1.0,100.0}}]},
 		 {text,polygon_offset_r,[{range,{1.0,100.0}}]}],
 	 [{title,?__(22,"Edge display problems?")}]}],
-       [{title,?__(21,"Workarounds")}]}
+       [{title,?__(21,"Workarounds")}]},
+      {vframe,
+       [{hframe,[{?__(24,"Show Develop menu"),show_develop_menu,
+		  [{info,?__(25,"Show a menu with tools for the Wings developers")}]}]}]}
      ]}.
 
 workaround(L) ->
@@ -645,9 +656,11 @@ smart_set_value_1(Key, Val, St) ->
 		material_default ->
 		    delayed_set_value(Key, OldVal, Val),
 		    wings_u:message(?__(3,"The change to the default material color will take\neffect the next time Wings 3D is started."));
-		material_hole ->
-		    delayed_set_value(Key, OldVal, Val),
-		    wings_u:message(?__(5,"The change to the hole material color will take\neffect the next time Wings 3D is started."));
+		show_develop_menu ->
+		    wings:init_menubar(),
+		    foreach(fun(W) ->
+				    wings_wm:send(W, language_changed) end,
+			    wings_wm:windows());
 		_Other -> ok
 	    end
     end.
@@ -668,8 +681,9 @@ clear_vertex_dlist(D, _) -> D#dlo{vs=none}.
 clear_proxy_edges(St) ->
     wings_dl:map(fun(D, _) -> clear_proxy_edges(D, St) end, []).
 
-clear_proxy_edges(D, St) ->
-    wings_proxy:update(D#dlo{proxy_edges=none}, St).
+clear_proxy_edges(#dlo{proxy_data=Data}=D, St) ->
+    PD = wings_proxy:invalidate(Data, edges),
+    wings_proxy:update(D#dlo{proxy_data=PD}, St).
 
 make_query({'VALUE',Val}) ->
     Val;
