@@ -13,8 +13,9 @@
 %%
 
 -module(wings_render).
--export([init/0, render/1,polygonOffset/1,
-	 enable_lighting/0, disable_lighting/0]).
+-export([init/0,
+	 render/1,polygonOffset/1,
+	 enable_lighting/1,disable_lighting/0]).
 
 -import(erlang, [max/2]).
 
@@ -22,16 +23,9 @@
 -include("wings.hrl").
 
 init() ->
-    case wings_gl:support_shaders() of
-	true ->
-	    try
-		wings_shaders:init()
-	    catch _:_Err ->
-		ok
-	    end;
-	false ->
-	    ok
-    end.
+    init_shaders(),
+    init_multisample(),
+    init_polygon_stipple().
 
 %% render(St)
 %%  Render the entire contents of a Geometry or AutoUV window,
@@ -48,19 +42,17 @@ render(#st{selmode=Mode}=St) ->
     gl:enable(?GL_DEPTH_TEST),
     gl:enable(?GL_CULL_FACE),
     case wings_pref:get_value(multisample) of
-	true ->
-	    gl:enable(?GL_MULTISAMPLE);
-	false ->
-	    gl:disable(?GL_MULTISAMPLE);
-	undefined ->
-	    wings_pref:set_default(multisample, true)
+	true -> gl:enable(?GL_MULTISAMPLE);
+	false -> gl:disable(?GL_MULTISAMPLE);
+	undefined -> ok
     end,
-    wings_view:load_matrices(true),
+    SceneLights = wings_view:load_matrices(true),
     ground_and_axes(),
     mini_axis_icon(),
     show_saved_bb(St),
+    show_bb_center(St),
     user_clipping_planes(on),
-    render_objects(Mode),
+    render_objects(Mode, SceneLights),
     user_clipping_planes(off),
     axis_letters(),
     gl:lineWidth(1),
@@ -68,7 +60,8 @@ render(#st{selmode=Mode}=St) ->
     gl:polygonMode(?GL_FRONT_AND_BACK, ?GL_LINE),
     {W,H} = wings_wm:win_size(),
     gl:rectf(W-0.5, 0.5, 0.5, H-0.5),
-    gl:popAttrib().
+    gl:popAttrib(),
+    wings_develop:gl_error_check("Rendering scene").
 
 polygonOffset(M) ->
     case get(polygon_offset) of
@@ -85,62 +78,107 @@ polygonOffset(M) ->
 %%% Internal functions follow.
 %%%
 
-render_objects(Mode) ->
+init_shaders() ->
+    case wings_gl:support_shaders() of
+	true ->
+	    try
+		wings_shaders:init()
+	    catch _:_Err ->
+		ok
+	    end;
+	false ->
+	    ok
+    end.
+
+init_multisample() ->
+    case wings_gl:is_ext('GL_ARB_multisample') of
+	false ->
+	    %% Not supported.
+	    %%
+	    %% It would have been better to use something
+	    %% clearer like 'not_supported' here, but that
+	    %% would cause older versions of Wings from
+	    %% 0.99.54 through 1.1.2 to crash.
+	    wings_pref:set_value(multisample, undefined);
+	true ->
+	    wings_pref:set_default(multisample, true)
+    end.
+
+init_polygon_stipple() ->
+    P = <<16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77,
+	 16#DD,16#DD,16#DD,16#DD,16#77,16#77,16#77,16#77>>,
+    gl:polygonStipple(P).
+
+render_objects(Mode, SceneLights) ->
     Dls = wings_dl:display_lists(),
     case wings_wm:get_prop(workmode) of
 	false ->
-	    render_smooth_objects(Dls, Mode, false),
-	    render_smooth_objects(Dls, Mode, true);
+	    render_smooth_objects(Dls, Mode, false, SceneLights),
+	    render_smooth_objects(Dls, Mode, true, SceneLights);
 	true ->
-	    render_work_objects(Dls, Mode)
+	    render_work_objects(Dls, Mode, SceneLights)
     end.
 
-render_smooth_objects([D|Dls], Mode, RenderTrans) ->
-    render_object(D, Mode, false, RenderTrans),
-    render_smooth_objects(Dls, Mode, RenderTrans);
-render_smooth_objects([], _, _) -> ok.
+render_smooth_objects([D|Dls], Mode, RenderTrans, SceneLights) ->
+    render_object(D, Mode, false, RenderTrans, SceneLights),
+    render_smooth_objects(Dls, Mode, RenderTrans, SceneLights);
+render_smooth_objects([], _, _, _) -> ok.
 
-render_work_objects([D|Dls], Mode) ->
-    render_object(D, Mode, true, false),
-    render_work_objects(Dls, Mode);
-render_work_objects([], _) -> ok.
+render_work_objects([D|Dls], Mode, SceneLights) ->
+    render_object(D, Mode, true, false, SceneLights),
+    render_work_objects(Dls, Mode, SceneLights);
+render_work_objects([], _, _) -> ok.
 
-render_object(#dlo{drag={matrix,_,_,Matrix}}=D, Mode, Work, RT) ->
+render_object(#dlo{drag={matrix,_,_,Matrix}}=D, Mode, Work, RT, SceneLights) ->
     gl:pushMatrix(),
     gl:multMatrixf(Matrix),
-    render_object_1(D, Mode, Work, RT),
+    render_object_1(D, Mode, Work, RT, SceneLights),
     gl:popMatrix();
-render_object(D, Mode, Work, RT) ->
-    render_object_1(D, Mode, Work, RT).
+render_object(D, Mode, Work, RT, SceneLights) ->
+    render_object_1(D, Mode, Work, RT, SceneLights).
 
-render_object_1(#dlo{mirror=none}=D, Mode, Work, RenderTrans) ->
-    render_object_2(D, Mode, Work, RenderTrans);
-render_object_1(#dlo{mirror=Matrix}=D, Mode, Work, RenderTrans) ->
-    render_object_2(D, Mode, Work, RenderTrans),
+render_object_1(#dlo{mirror=none}=D, Mode, Work, RenderTrans, SceneLights) ->
+    render_object_2(D, Mode, Work, RenderTrans, SceneLights);
+render_object_1(#dlo{mirror=Matrix}=D, Mode, Work, RenderTrans, SceneLights) ->
+    render_object_2(D, Mode, Work, RenderTrans, SceneLights),
     gl:frontFace(?GL_CW),
     gl:pushMatrix(),
     gl:multMatrixf(Matrix),
-    render_object_2(D, Mode, Work, RenderTrans),
+    render_object_2(D, Mode, Work, RenderTrans, SceneLights),
     gl:popMatrix(),
     gl:frontFace(?GL_CCW).
 
-render_object_2(#dlo{src_we=We}=D, _, _, false) when ?IS_LIGHT(We) ->
+render_object_2(#dlo{src_we=We}=D, _, _, false, _) when ?IS_LIGHT(We) ->
     wings_light:render(D);
-render_object_2(#dlo{src_we=We}, _, _, true) when ?IS_LIGHT(We) ->
+render_object_2(#dlo{src_we=We}, _, _, true, _) when ?IS_LIGHT(We) ->
     ok;
-render_object_2(D, Mode, true, _) ->
-    render_plain(D, Mode);
-render_object_2(#dlo{transparent=true}=D, _, false, false) ->
+render_object_2(D, Mode, true, _, SceneLights) ->
+    render_plain(D, Mode, SceneLights);
+render_object_2(#dlo{transparent=true}=D, _, false, false, SceneLights) ->
     gl:disable(?GL_CULL_FACE),
-    render_smooth(D, false),
+    render_smooth(D, false, SceneLights),
     gl:enable(?GL_CULL_FACE);
-render_object_2(#dlo{transparent=true}=D, _, false, true) ->
-    render_smooth(D, true);
-render_object_2(#dlo{transparent=false}=D, _, false, RenderTrans) ->
-    render_smooth(D, RenderTrans).
+render_object_2(#dlo{transparent=true}=D, _, false, true, SceneLights) ->
+    render_smooth(D, true, SceneLights);
+render_object_2(#dlo{transparent=false}=D, _, false, RenderTrans, SceneLights) ->
+    render_smooth(D, RenderTrans, SceneLights).
 
 render_plain(#dlo{work=Faces,edges=Edges,open=Open,
-		  src_we=We,proxy_data=none}=D, SelMode) ->
+		  src_we=We,proxy=false}=D, SelMode, SceneLights) ->
     %% Draw faces for winged-edge-objects.
     Wire = wire(We),
     case Wire of
@@ -149,11 +187,12 @@ render_plain(#dlo{work=Faces,edges=Edges,open=Open,
 	    gl:enable(?GL_POLYGON_OFFSET_FILL),
 	    polygonOffset(2),
 	    gl:shadeModel(?GL_SMOOTH),
-	    enable_lighting(),
+	    enable_lighting(SceneLights),
 	    case Open of
 		false ->
 		    wings_dl:call(Faces);
 		true ->
+		    gl:lightModeli(?GL_LIGHT_MODEL_TWO_SIDE, ?GL_TRUE),
 		    gl:disable(?GL_CULL_FACE),
 		    wings_dl:call(Faces),
 		    gl:enable(?GL_CULL_FACE)
@@ -197,9 +236,9 @@ render_plain(#dlo{work=Faces,edges=Edges,open=Open,
 	    end
     end,
     render_plain_rest(D, Wire, SelMode);
-render_plain(#dlo{src_we=We}=D, SelMode) ->
+render_plain(#dlo{src_we=We}=D, SelMode, SceneLights) ->
     Wire = wire(We),
-    wings_proxy:draw(D, Wire),
+    wings_proxy:draw(D, Wire, SceneLights),
     render_plain_rest(D, Wire, SelMode).
 
 render_plain_rest(#dlo{}=D, Wire, SelMode) ->
@@ -220,18 +259,26 @@ render_plain_rest(#dlo{}=D, Wire, SelMode) ->
     draw_vertices(D, SelMode),
     draw_hard_edges(D, SelMode),
     draw_normals(D),
-	draw_plugins(plain,D,SelMode). %% arbitrary placement in the grand scheme of things
+    draw_plugins(plain,D,SelMode). %% arbitrary placement in the grand scheme of things
 
-render_smooth(#dlo{work=Work,edges=Edges,smooth=Smooth,transparent=Trans,
-		   src_we=We,proxy_data=Pd,open=Open}=D,
-	      RenderTrans) ->
+render_smooth(#dlo{work=Work,edges=Edges,smooth=Smooth0,transparent=Trans0,
+		   src_we=We,proxy=Proxy,proxy_data=PD,open=Open}=D,
+	      RenderTrans, SceneLights) ->
     gl:shadeModel(?GL_SMOOTH),
     gl:polygonMode(?GL_FRONT_AND_BACK, ?GL_FILL),
-    enable_lighting(),
+    enable_lighting(SceneLights),
     gl:enable(?GL_POLYGON_OFFSET_FILL),
     gl:polygonOffset(2, 2),
 
-    case Trans of
+    case Proxy of
+	true ->
+	    {Smooth, Trans} = wings_proxy:smooth_dl(PD);
+	false ->
+	    Smooth = Smooth0,
+	    Trans  = Trans0
+    end,
+
+    case Trans orelse Open of
 	false -> gl:lightModeli(?GL_LIGHT_MODEL_TWO_SIDE, ?GL_FALSE);
 	true -> gl:lightModeli(?GL_LIGHT_MODEL_TWO_SIDE, ?GL_TRUE)
     end,
@@ -250,19 +297,13 @@ render_smooth(#dlo{work=Work,edges=Edges,smooth=Smooth,transparent=Trans,
 	false -> ok;
 	true -> gl:disable(?GL_CULL_FACE)
     end,
-
     case {Smooth,RenderTrans} of
-	{none,false} ->
-	    if
-		Pd =:= none ->
-		    wings_dl:call(Work);
-		true ->
-		    wings_proxy:draw(D, wire(We))
-	    end;
+	{none,false} ->   wings_dl:call(Work);
 	{[Op,_],false} -> wings_dl:call(Op);
-	{[_,Tr],true} -> wings_dl:call(Tr);
+	{[_,Tr],true} ->  wings_dl:call(Tr);
 	{_,_} -> ok
     end,
+
     gl:enable(?GL_CULL_FACE),
 
     gl:disable(?GL_POLYGON_OFFSET_FILL),
@@ -270,7 +311,7 @@ render_smooth(#dlo{work=Work,edges=Edges,smooth=Smooth,transparent=Trans,
     disable_lighting(),
     gl:shadeModel(?GL_FLAT),
     case wire(We) of
-	true when Pd =:= none ->
+	true when Proxy =:= false ->
 	    gl:color3fv(wings_pref:get_value(edge_color)),
 	    gl:lineWidth(1),
 	    gl:polygonMode(?GL_FRONT_AND_BACK, ?GL_LINE),
@@ -284,7 +325,7 @@ render_smooth(#dlo{work=Work,edges=Edges,smooth=Smooth,transparent=Trans,
     draw_hilite(D),
     draw_sel(D),
     draw_orig_sel(D),
-	draw_plugins(smooth,D,none).
+    draw_plugins(smooth,D,none).
 
 wire(#we{id=Id}) ->
     W = wings_wm:get_prop(wireframed_objects),
@@ -572,7 +613,7 @@ show_saved_bb(#st{bb=[{X1,Y1,Z1},{X2,Y2,Z2}]}) ->
 	true ->
 	    gl:enable(?GL_LINE_STIPPLE),
 	    gl:lineStipple(4, 2#1110111011101110),
-	    gl:color3f(0, 0, 1),
+	    gl:color3fv(wings_pref:get_value(active_vector_color)),
 	    gl:'begin'(?GL_LINE_STRIP),
 	    gl:vertex3f(X1, Y1, Z1),
 	    gl:vertex3f(X2, Y1, Z1),
@@ -597,21 +638,47 @@ show_saved_bb(#st{bb=[{X1,Y1,Z1},{X2,Y2,Z2}]}) ->
     end;
 show_saved_bb(_) -> ok.
 
-enable_lighting() ->
+show_bb_center(#st{bb=[_,_]=BB}) ->
+    case wings_pref:get_value(show_bb_center) of
+      false -> ok;
+      true ->
+        {Cx,Cy,Cz} = Center = e3d_vec:average(BB),
+        Colour = gl:color3fv(wings_pref:get_value(active_vector_color)),
+        Colour,
+        gl:pointSize(8.0),
+        gl:'begin'(?GL_POINTS),
+        gl:vertex3fv(Center),
+        gl:'end'(),
+        gl:'begin'(?GL_LINES),
+        Colour,
+        gl:vertex3fv({Cx,Cy+0.2,Cz}),
+        gl:vertex3fv({Cx,Cy-0.2,Cz}),
+        gl:vertex3fv({Cx+0.2,Cy,Cz}),
+        gl:vertex3fv({Cx-0.2,Cy,Cz}),
+        gl:vertex3fv({Cx,Cy,Cz+0.2}),
+        gl:vertex3fv({Cx,Cy,Cz-0.2}),
+        gl:'end'()
+
+    end;
+show_bb_center(_) -> ok.
+
+enable_lighting(SceneLights) ->
     Progs = get(light_shaders),
-    NumLights = wings_pref:get_value(number_of_lights),
-    NumShaders = wings_pref:get_value(number_of_shaders),
-    UseProg = (Progs /= undefined) andalso
-	      (not wings_pref:get_value(scene_lights)) andalso
-	      (NumLights == 2),
+    UseProg = Progs =/= undefined andalso
+	      not SceneLights andalso
+	      wings_pref:get_value(number_of_lights) =:= 2,
     case UseProg of
 	false ->
 	    gl:enable(?GL_LIGHTING);
 	true ->
+	    NumShaders = wings_pref:get_value(number_of_shaders),
 	    {Prog,_Name} = element(NumShaders, Progs),
-	    gl:color4ub(255,255,255,255), %% Reset color needed by crappy drivers.
-	    %% We put it here and not in apply_material because we can't use some
-	    %% optimizations (i.e. reuse display lists) when drawing selected objects
+
+	    %% Reset color. Needed by some drivers.
+	    %% We put it here and not in apply_material, because we
+	    %% can't use some optimizations (e.g. reuse display lists)
+	    %% when drawing selected objects.
+	    gl:color4ub(255, 255, 255, 255), 
 	    gl:useProgram(Prog)
     end.
 
