@@ -19,7 +19,8 @@
 -export([select_all/1]).
 
 -include("wings.hrl").
--import(lists, [map/2,foldl/3,reverse/1,keymember/3,keysearch/3]).
+-import(lists, [map/2,foldl/3,reverse/1,keymember/3,keyfind/3,usort/1]).
+-import(erlang, [max/2]).
 
 menu(St) ->
     [{?__(1,"Deselect"),deselect,?__(2,"Clear the selection")},
@@ -52,7 +53,10 @@ menu(St) ->
 	{?__(22,"Grow Edge Ring"),edge_ring_incr,
 	 ?__(23,"Grow edge selection by one edge in ring direction")},
 	{?__(24,"Shrink Edge Ring"),edge_ring_decr,
-	 ?__(25,"Shrink edge selection by one edge in ring directions")}]}},
+	 ?__(25,"Shrink edge selection by one edge in ring directions")},
+	separator,
+	{?__(96,"To Complete Loops"),complete_loops,
+	 ?__(97,"Switches to Edge mode and completes loop selection")}]}},
      separator,
      {?__(26,"Adjacent"),
       {adjacent,[{?__(27,"Vertices"),vertex},
@@ -249,6 +253,8 @@ command({edge_loop,prev_edge_loop}, St) ->
     {save_state,wings_edge_loop:select_prev(St)};
 command({edge_loop,edge_loop_to_region}, St) ->
     {save_state,wings_edge:select_region(St)};
+command({edge_loop,complete_loops}, St) ->
+    {save_state,complete_loops(St)};
 command(deselect, St) ->
     {save_state,wings_sel:reset(St)};
 command(more, St) ->
@@ -347,29 +353,28 @@ by_command({vertices_with,N}, St) ->
     {save_state,wings_sel:make(Sel, vertex, St)};
 by_command({non_quad,all}, St) ->
     Sel = fun(Face, We) ->
-		  wings_face:vertices(Face, We) =/= 4 andalso Face >= 0
+		  wings_face:vertices(Face, We) =/= 4
 	    end,
     {save_state,wings_sel:make(Sel, face, St)};
 by_command({non_quad,odd}, St) ->
     Sel = fun(Face, We) ->
-		  Half = wings_face:vertices(Face, We)/2,
-		  trunc(Half) /= Half andalso Face >= 0
-	    end,
+		  wings_face:vertices(Face, We) rem 2 =/= 0
+	  end,
     {save_state,wings_sel:make(Sel, face, St)};
 by_command({non_quad,even}, St) ->
     Sel = fun(Face, We) ->
-		  Half = wings_face:vertices(Face, We)/2,
-		  (Half /= 2) and (trunc(Half) == Half) andalso Face >= 0
-	    end,
+		  N = wings_face:vertices(Face, We),
+		  N =/= 4 andalso N rem 2 =:= 0
+	  end,
     {save_state,wings_sel:make(Sel, face, St)};
 by_command({faces_with,5}, St) ->
     Sel = fun(Face, We) ->
-		  wings_face:vertices(Face, We) >= 5 andalso Face >= 0
+		  wings_face:vertices(Face, We) >= 5
 	    end,
     {save_state,wings_sel:make(Sel, face, St)};
 by_command({faces_with,N}, St) ->
     Sel = fun(Face, We) ->
-		  N =:= wings_face:vertices(Face, We) andalso Face >= 0
+		  N =:= wings_face:vertices(Face, We)
 	  end,
     {save_state,wings_sel:make(Sel, face, St)};
 by_command(material_edges, St) ->
@@ -457,9 +462,9 @@ inverse(#st{selmode=Mode}=St) ->
 
 hide_selected(#st{selmode=Mode,shapes=Shs0,sel=Sel}=St) ->
     Shs1 = map(fun(#we{id=Id}=We) ->
-		       case keysearch(Id, 1, Sel) of
+		       case keyfind(Id, 1, Sel) of
 			   false -> {Id,We};
-			   {value,{_,Set}} -> {Id,We#we{perm={Mode,Set}}}
+			   {_,Set} -> {Id,We#we{perm={Mode,Set}}}
 		       end
 	       end, gb_trees:values(Shs0)),
     Shs = gb_trees:from_orddict(Shs1),
@@ -619,7 +624,7 @@ similar(#st{selmode=vertex}=St) ->
 				  [make_vertex_template(SelI, We) ||
 				      SelI <- gb_sets:to_list(Sel0)] ++ A
 			  end, [], St),
-    Templates = ordsets:from_list(Seed),
+    Templates = consolidate_templates(Seed),
     wings_sel:make(
       fun(V, W) ->
 	      match_templates(make_vertex_template(V, W), Templates)
@@ -629,7 +634,7 @@ similar(#st{selmode=edge}=St) ->
 				  [make_edge_template(SelI, We) ||
 				      SelI <- gb_sets:to_list(Sel0)] ++ A
 			  end, [], St),
-    Templates = ordsets:from_list(Seed),
+    Templates = consolidate_templates(Seed),
     wings_sel:make(
       fun(Edge, W) ->
 	      match_templates(make_edge_template(Edge, W), Templates)
@@ -639,22 +644,39 @@ similar(#st{selmode=face}=St) ->
 				  [make_face_template(SelI, We) ||
 				      SelI <- gb_sets:to_list(Sel0)] ++ A
 			  end, [], St),
-    Templates = ordsets:from_list(Seed),
+    Templates = consolidate_templates(Seed),
     wings_sel:make(
       fun(Face, WeI) ->
-	      Face >= 0 andalso match_templates(make_face_template(Face, WeI), Templates)
+	      match_templates(make_face_template(Face, WeI), Templates)
       end, face, St);
 similar(#st{selmode=body}=St) ->
     Template0 = wings_sel:fold(fun(_, #we{vp=Vtab,es=Etab,fs=Ftab}, Acc) ->
-				       [{gb_trees:size(Vtab),
-					 gb_trees:size(Etab),
+				       [{wings_util:array_entries(Vtab),
+					 wings_util:array_entries(Etab),
 					 gb_trees:size(Ftab)}|Acc]
 			       end, [], St),
     Template = ordsets:from_list(Template0),
     wings_sel:make(fun(_, We) -> match_body(Template, We) end, body, St).
 
+consolidate_templates(L) ->
+    case usort(L) of
+	[] -> [];
+	[H|T] -> consolidate_templates_1(H, T)
+    end.
+
+consolidate_templates_1(Templ0, [Templ|T]) ->
+    case match_template(Templ0, Templ) of
+	true ->
+	    consolidate_templates_1(Templ0, T);
+	false->
+	    [Templ0|consolidate_templates_1(Templ, T)]
+    end;
+consolidate_templates_1(Templ, []) -> [Templ].
+
 match_body(Template, #we{vp=Vtab,es=Etab,fs=Ftab}) ->
-    Sizes = {gb_trees:size(Vtab),gb_trees:size(Etab),gb_trees:size(Ftab)},
+    Sizes = {wings_util:array_entries(Vtab),
+	     wings_util:array_entries(Etab),
+	     gb_trees:size(Ftab)},
     match_body_1(Template, Sizes).
 
 match_body_1([Sizes|_], Sizes) -> true;
@@ -669,25 +691,15 @@ match_templates(F, [Template|Ts]) ->
 match_templates(_, []) -> false.
 
 match_template({Len,Ad,As}, {Len,Bd,Bs}) ->
-    case rel_compare(Ad, Bd, 1.0E-5) of
-	true -> rel_compare(As, Bs, 1.0E-5);
-	false -> false
-    end;
+    compare(Ad, Bd) andalso compare(As, Bs);
 match_template(_, _) -> false.
 
-make_face_template(Face, #we{vp=Vtab}=We) ->
-    Vs = wings_face:fold(
-	   fun(V, _, _, Acc0) ->
-		   [V|Acc0]
-	   end, [], Face, We),
-    {DotSum,SqSum} = face_dots_and_sqlens(Vs, Vtab),
-    {length(Vs),DotSum,SqSum}.
+make_face_template(Face, We) ->
+    VsPos = wings_face:vertex_positions(Face, We),
+    {DotSum,SqSum} = face_dots_and_sqlens(VsPos),
+    {length(VsPos),DotSum,SqSum}.
 
-face_dots_and_sqlens(Vs, Vtab) ->
-    Vpos = [gb_trees:get(P, Vtab) || P <- Vs],
-    face_dots_and_sqlens_1(Vpos).
-
-face_dots_and_sqlens_1([Va,Vb|_]=Vpos) ->
+face_dots_and_sqlens([Va,Vb|_]=Vpos) ->
     D = e3d_vec:sub(Va, Vb),
     face_dots_and_sqlens_2(D, Vpos, Vpos, 0, 0).
 
@@ -702,9 +714,9 @@ face_dots_and_sqlens_2(_D1, _Other, _More, Dot, Sq) -> {Dot,Sq}.
 
 make_edge_template(Edge, #we{vp=Vtab,es=Etab}=We) ->
     #edge{vs=Va,ve=Vb,ltpr=LP,ltsu=LS,rtpr=RP,rtsu=RS} =
-	gb_trees:get(Edge, Etab),
-    VaPos = gb_trees:get(Va, Vtab),
-    VbPos = gb_trees:get(Vb, Vtab),
+	array:get(Edge, Etab),
+    VaPos = array:get(Va, Vtab),
+    VbPos = array:get(Vb, Vtab),
     Vec = e3d_vec:sub(VaPos, VbPos),
     DotSum = edge_dot(LP, Vb, VbPos, Vec, We) +
 	edge_dot(RS, Vb, VbPos, Vec, We) +
@@ -713,13 +725,13 @@ make_edge_template(Edge, #we{vp=Vtab,es=Etab}=We) ->
     {0,DotSum,e3d_vec:dot(Vec, Vec)}.
 
 edge_dot(Edge, V, Pos, Vec, #we{es=Etab}=We) ->
-    Rec = gb_trees:get(Edge, Etab),
+    Rec = array:get(Edge, Etab),
     OtherPos = wings_vertex:other_pos(V, Rec, We),
     ThisVec = e3d_vec:sub(Pos, OtherPos),
     abs(e3d_vec:dot(ThisVec, Vec)).
 
 make_vertex_template(V, #we{vp=Vtab}=We) ->
-    Center = gb_trees:get(V, Vtab),
+    Center = array:get(V, Vtab),
     Vecs = wings_vertex:fold(
 	     fun(_, _, Rec, Acc0) ->
 		     Pos = wings_vertex:other_pos(V, Rec, Vtab),
@@ -737,12 +749,15 @@ vertex_dots_and_sqlens(Vecs, [VecB|_], Dot, Sq) ->
     vertex_dots_and_sqlens(Vecs++[VecB], [], Dot, Sq);
 vertex_dots_and_sqlens(_Other, _More, Dot, Sq) -> {Dot,Sq}.
 
-rel_compare(A, B, Tresh) when abs(A) < Tresh ->
-    abs(B) < Tresh;
-rel_compare(A, B, Tresh) when abs(A) > abs(B) ->
-    abs(A-B)/abs(A) < Tresh;
-rel_compare(A, B, Tresh) ->
-    abs(A-B)/abs(B) < Tresh.
+-define(EPSILON, 1.0E-5).
+compare(A, B) ->
+    %% Comparison with a relative tolerance for large
+    %% values and an absolute tolerance for small values
+    %% as described by Christer Ericson in
+    %% http://realtimecollisiondetection.net/blog/?p=89
+    %% and in his book "Real-Time Collision Detection".
+    %%
+    abs(A-B) =< ?EPSILON*max(1.0, max(abs(A), abs(B))).
 
 %%
 %% Select Random.
@@ -769,9 +784,9 @@ short_edges([Tolerance], St0) ->
     {save_state,St#st{selmode=edge}}.
 
 short_edge(Tolerance, Edge, #we{es=Etab,vp=Vtab}) ->
-    #edge{vs=Va,ve=Vb} = gb_trees:get(Edge, Etab),
-    VaPos = gb_trees:get(Va, Vtab),
-    VbPos = gb_trees:get(Vb, Vtab),
+    #edge{vs=Va,ve=Vb} = array:get(Edge, Etab),
+    VaPos = array:get(Va, Vtab),
+    VbPos = array:get(Vb, Vtab),
     abs(e3d_vec:dist(VaPos, VbPos)) < Tolerance.
 
 %%
@@ -782,33 +797,30 @@ material_edges(St) ->
     wings_sel:make(fun material_edges_fun/2, edge, St).
 
 material_edges_fun(E, #we{es=Etab}=We) ->
-    #edge{lf=Lf,rf=Rf} = gb_trees:get(E, Etab),
+    #edge{lf=Lf,rf=Rf} = array:get(E, Etab),
     wings_facemat:face(Lf, We) =/= wings_facemat:face(Rf, We).
 
 %%
-%% Select all faces that have UV coordinates.
+%% Select all faces that have (proper) UV coordinates.
 %%
 
-uv_mapped_faces(St) ->
-    wings_sel:make(fun is_uv_mapped_face/2, face, St).
-
-is_uv_mapped_face(Face, We) when Face >= 0->
-    is_uv_mapped_face_1(wings_face:vertex_info(Face, We));
-is_uv_mapped_face(_,_) -> false.
-
-is_uv_mapped_face_1([{_,_}|T]) ->
-    is_uv_mapped_face_1(T);
-is_uv_mapped_face_1([_|_]) -> false;
-is_uv_mapped_face_1([]) -> true.
+uv_mapped_faces(#st{shapes=Shs}=St) ->
+    Sel = foldl(fun(#we{id=Id}=We, A) ->
+			case wings_we:uv_mapped_faces(We) of
+			    [] -> A;
+			    Fs -> [{Id,gb_sets:from_ordset(Fs)}|A]
+			end
+		end, [], gb_trees:values(Shs)),
+    wings_sel:set(face, Sel, St).
 
 %%
 %% Select by numerical item id.
 %%
 
-by_id(#st{selmode=body}=St) ->
+by_id(#st{selmode=body}) ->
     ask([{"Object Id",1}],
 	fun([Id]) ->
-		valid_sel("", [{Id,gb_sets:singleton(0)}], St)
+		{"",[{Id,gb_sets:singleton(0)}]}
 	end);
 by_id(#st{selmode=vertex}=St) ->
     item_by_id("Vertex Id", St);
@@ -881,12 +893,11 @@ select_lights_1([We|Shs], Mode) when not ?IS_LIGHT(We) ->
 select_lights_1([#we{id=Id}|Shs], body) ->
     [{Id,gb_sets:singleton(0)}|select_lights_1(Shs, body)];
 select_lights_1([#we{id=Id,vp=Vtab,es=Etab,fs=Ftab}|Shs], Mode) ->
-    Tab = case Mode of
-	      vertex -> Vtab;
-	      edge -> Etab;
-	      face -> Ftab
+    Sel = case Mode of
+	      vertex -> wings_util:array_keys(Vtab);
+	      edge -> wings_util:array_keys(Etab);
+	      face -> gb_trees:keys(Ftab)
 	  end,
-    Sel = gb_trees:keys(Tab),
     [{Id,gb_sets:from_ordset(Sel)}|select_lights_1(Shs, Mode)];
 select_lights_1([], _) -> [].
 
@@ -920,7 +931,7 @@ nonplanar_faces(Ask, _St) when is_atom(Ask) ->
 		     fun(Res) -> {select,{by,{nonplanar_faces,Res}}} end);
 nonplanar_faces([Tolerance], St) ->
     Sel = fun(Face, We) ->
-	      not wings_face:is_planar(Tolerance,Face,We) andalso Face >= 0
+		  not wings_face:is_planar(Tolerance,Face,We)
 	  end,
     {save_state,wings_sel:make(Sel, face, St)}.
 
@@ -967,17 +978,19 @@ oriented_faces([Tolerance,false,Save], St) ->
     {save_state,wings_sel:make(Sel, face, St)};
 
 oriented_faces([Tolerance,true,Save], St0) ->
-    wings_pref:set_value(similar_normals_angle,{Save,Tolerance}),
-    wings_pref:set_value(similar_normals_connected,true),
+    wings_pref:set_value(similar_normals_angle, {Save,Tolerance}),
+    wings_pref:set_value(similar_normals_connected, true),
     CosTolerance = math:cos(Tolerance * (math:pi() / 180.0)),
-    Sel = wings_sel:fold(fun(Faces, #we{id=Id}=We, A) ->
-	        Normals0 = foldl(fun(F,Acc) ->
-			        [wings_face:normal(F, We)|Acc]
-		    end,[],gb_sets:to_list(Faces)),
-            Normals = lists:usort(Normals0),
-            [{Id,norm_search(Faces,Normals,CosTolerance,We,Faces)}|A]
-    end, [], St0),
-    wings_sel:set(face,Sel,St0).
+    Sel = wings_sel:fold(
+	    fun(Faces, #we{id=Id}=We, A) ->
+		    Normals0 = gb_sets:fold(
+				 fun(F,Acc) ->
+					 [wings_face:normal(F, We)|Acc]
+				 end, [], Faces),
+		    Normals = lists:usort(Normals0),
+		    [{Id,norm_search(Faces, Normals, CosTolerance, We, Faces)}|A]
+	    end, [], St0),
+    wings_sel:set(face, Sel, St0).
 
 norm_search(Faces,Normals,CosTolerance,We,LastSel) ->
     Fs0 = wings_face:extend_border(LastSel, We),
@@ -1020,29 +1033,31 @@ similar_material(_, #st{selmode=Mode}) when Mode =/= face ->
 
 similar_material(Ask, _St) when is_atom(Ask) ->
     Connected = wings_pref:get_value(similar_materials_connected,false),
-    Qs = [{?__(2,"Connected Faces Only"),Connected}],
+    Qs = [{?__(2,"Connected Faces Only"),Connected},
+	  {vradio,[{?__(5,"Materials"),material},
+		   {?__(6,"Vertex Color"),vertex_color}],
+	   material}],
     wings_ask:dialog(Ask, ?__(4,"Select Faces with the same Material"),
     [{vframe,Qs}],
     fun(Res) -> {select,{similar_material,Res}} end);
 
-similar_material([Connected], #st{selmode=face, sel=[]}) ->
+similar_material([Connected|_], #st{selmode=face,sel=[]}) ->
     wings_pref:set_value(similar_materials_connected,Connected),
     wings_u:error(?__(3,"At least one face must be selected"));
 
-similar_material([false], St) ->
+similar_material([false,Mode], St) ->
     Materials = wings_sel:fold(fun
-       (Faces, #we{mode=vertex}=We, A) ->
-        foldl(fun(F,Acc) ->
-               [wings_color:average([C || [_|C]
-                 <- wings_face:vinfo_ccw(F, We)])|Acc]
-        end,A,gb_sets:to_list(Faces));
+       (Faces, We, A) when Mode =:= vertex_color ->
+        gb_sets:fold(fun(F, Acc) ->
+                   [average_colors(F, We)|Acc]
+        end,A,Faces);
        (Faces, We, A) ->
                [wings_facemat:face(SelI, We) ||
                    SelI <- gb_sets:to_list(Faces)] ++ A
              end, [], St),
     Sel = fun	
-        (F, #we{mode=vertex}=We) ->
-          Col = wings_color:average([C||[_|C] <- wings_face:vinfo_ccw(F, We)]),
+        (F, We) when Mode =:= vertex_color ->
+          Col = average_colors(F, We),
           any_matching_material(Col,Materials);
         (Face, We) ->
           Mat = wings_facemat:face(Face,We),
@@ -1051,50 +1066,51 @@ similar_material([false], St) ->
     wings_pref:set_value(similar_materials_connected,false),
     {save_state,wings_sel:make(Sel, face, St)};
 
-similar_material([true], St0) ->
+similar_material([true,Mode], St0) ->
     Selection = wings_sel:fold(fun
-        (Faces, #we{id=Id,mode=vertex}=We, A) ->
-            AllCols = foldl(fun(F,Acc) ->
-                Cols = [C || [_|C] <- wings_face:vinfo_ccw(F, We)],
-                [wings_color:average(Cols)|Acc]
-            end,A,gb_sets:to_list(Faces)),
+        (Faces, #we{id=Id}=We, A) when Mode =:= vertex_color ->
+            AllCols = gb_sets:fold(fun(F,Acc) ->
+                [average_colors(F, We)|Acc]
+            end,A,Faces),
             Colours = lists:usort(AllCols),
-            [{Id,mat_search(Faces,Colours,We,Faces)}|A];
+            [{Id,mat_search(Faces,Colours,We,Mode,Faces)}|A];
         (Faces, #we{id=Id}=We, A) ->
             AllMats = [wings_facemat:face(Face, We) || Face <- gb_sets:to_list(Faces)],
             Materials = lists:usort(AllMats),
-            [{Id,mat_search(Faces,Materials,We,Faces)}|A]
+            [{Id,mat_search(Faces,Materials,We,Mode,Faces)}|A]
     end, [], St0),
     wings_sel:set(face,Selection,St0).
 
-mat_search(Faces,Colours,We,LastSel) ->
+average_colors(Face, We) ->
+    wings_color:average([C || C <- wings_va:face_attr(color, Face, We)]).
+
+mat_search(Faces,Colours,We,Mode,LastSel) ->
     Fs0 = wings_face:extend_border(LastSel, We),
     Fs1 = gb_sets:subtract(Fs0,Faces),
-    AddSel = check_face_colours(Fs1,Colours,We,gb_sets:empty()),
+    AddSel = check_face_colours(Fs1,Colours,We,Mode,gb_sets:empty()),
     case gb_sets:is_empty(AddSel) of
         true -> Faces;
         false ->
             Faces1 = gb_sets:union(AddSel,Faces),
-            mat_search(Faces1,Colours,We,AddSel)
+            mat_search(Faces1,Colours,We,Mode,AddSel)
     end.
 
-check_face_colours(Fs0,Colours,We,Selection)->
+check_face_colours(Fs0,Colours,We,Mode,Selection)->
     case gb_sets:is_empty(Fs0) of
       true -> Selection;
       false ->
         {F,Fs1} = gb_sets:take_smallest(Fs0),
-        Colour = face_info(F,We),
+        Colour = face_info(F,We,Mode),
         Sel = case any_matching_material(Colour,Colours) of
           true -> gb_sets:add(F,Selection);
           false -> Selection
         end,
-        check_face_colours(Fs1,Colours,We,Sel)
+        check_face_colours(Fs1,Colours,We,Mode,Sel)
     end.
 
-face_info(F,#we{mode=vertex}=We) ->
-    Cols = [C || [_|C] <- wings_face:vinfo_ccw(F, We)],
-    wings_color:average(Cols);
-face_info(F,We) ->
+face_info(F,We,vertex_color) ->
+    average_colors(F, We);
+face_info(F,We,_Mode) ->
     wings_facemat:face(F, We).
 
 any_matching_material(_,[]) ->
@@ -1125,7 +1141,7 @@ sharp_edges([Tolerance], St0) ->
     {save_state,St}.
 
 sharp_edge(CosTolerance, Edge, #we{es=Etab}=We) ->
-    #edge{lf=Lf,rf=Rf} = gb_trees:get(Edge, Etab),
+    #edge{lf=Lf,rf=Rf} = array:get(Edge, Etab),
     Lfn = wings_face:normal(Lf, Edge, We),
     Rfn = wings_face:normal(Rf, Edge, We),
     e3d_vec:dot(Lfn,Rfn) < CosTolerance.
@@ -1136,7 +1152,7 @@ sharp_edge(CosTolerance, Edge, #we{es=Etab}=We) ->
 
 shortest_path(Method, St) ->
     #st{shapes=Shapes,selmode=Mode,sel=Sel} = St,
-    case (Mode==vertex) and (length(Sel)==1) of
+    case (Mode==vertex) andalso (length(Sel)==1) of
 	true -> ok;
 	false -> wings_u:error(?__(1,"Exactly two vertices must be\n selected on the same object."))
     end,
@@ -1149,10 +1165,10 @@ shortest_path(Method, St) ->
     [Pa,Pb] = [wings_vertex:pos(V, We) || V <- gb_sets:to_list(SelectedVs)],
     #we{es=Etab,vp=Vtab} = We,
     Graph = digraph:new(),
-    Add_Edge = fun(EdgeIdx) ->
-		    build_digraph(Graph, gb_trees:get(EdgeIdx,Etab), Vtab)
+    Add_Edge = fun(_, EdgeRec, _) ->
+		       build_digraph(Graph, EdgeRec, Vtab)
 	       end,
-    lists:foreach(Add_Edge, gb_trees:keys(Etab)),
+    array:sparse_foldl(Add_Edge, [], Etab),
     PathVs = find_path_verts(Method, Graph, Pa, Pb),
     digraph:delete(Graph),
     SelFun = fun(Vert, We2) -> is_vert_in_path(PathVs, Vert, We2) end,
@@ -1294,7 +1310,7 @@ astar_relax_all(U, ONs, Gcosts, Open, PrevNodes, Pb) ->
 
 is_vert_in_path(PathVs, Vert, We) ->
     #we{vp=Vtab} = We,
-    Vpos = gb_trees:get(Vert, Vtab),
+    Vpos = array:get(Vert, Vtab),
     lists:member(Vpos, PathVs).
 
 get_path(none, _) -> [];
@@ -1356,9 +1372,18 @@ similar_area([Tolerance], St) ->
     St2 = wings_sel:make(SelFun, face, St),
     {save_state,St2}.
 
-is_area_similar(Area1, Tolerance, Face, We) when Face >= 0 ->
+is_area_similar(Area1, Tolerance, Face, We) ->
     Area2 = wings_face:area(Face, We),
     AreaDiff = abs(Area1-Area2),
-    AreaDiff =< Tolerance;
-is_area_similar(_,_,_,_) ->
-    false.
+    AreaDiff =< Tolerance.
+
+%%%
+%%% Change selection to complete loops from any mode
+%%%
+complete_loops(#st{sel=[]}=St) ->
+    St;
+complete_loops(#st{selmode=edge}=St) ->
+    wings_edge_loop:select_loop(St);
+complete_loops(St0) ->
+    St = wings_sel_conv:mode(edge,St0),
+    wings_edge_loop:select_loop(St).
