@@ -3,7 +3,7 @@
 %%
 %%     Various utility functions that not obviously fit somewhere else.
 %%
-%%  Copyright (c) 2001-2008 Bjorn Gustavsson
+%%  Copyright (c) 2001-2009 Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -16,11 +16,11 @@
 	 geom_windows/0,menu_restriction/2,
 	 yes_no/2,yes_no/3,yes_no_cancel/3,
 	 export_we/2,win_crash/1,crash_log/2,crash_log/3,
-	 caption/1]).
+	 pretty_filename/1,caption/1,win32_special_folder/2]).
 
 -define(NEED_OPENGL, 1).
 -include("wings.hrl").
--import(lists, [member/2,foreach/2]).
+-import(lists, [member/2,foreach/2,all/2]).
 
 -spec error([any()]) -> no_return().
 
@@ -131,25 +131,44 @@ report_stacktrace(_, []) ->
     %% (There will be no stacktrace if we were called from the
     %% process that runs wings:halt_loop/1.)
     ok.
-    
+
 caption(#st{file=undefined}=St) ->
     Caption = wings(),
-    sdl_video:wm_setCaption(Caption, Caption),
+    wings_io:set_title(Caption),
     St;
 caption(#st{saved=true,file=Name}=St) ->
     Caption = wings() ++ " - " ++ filename:basename(Name),
-    sdl_video:wm_setCaption(Caption, Caption),
+    wings_io:set_title(Caption),
     St;
 caption(#st{saved=auto,file=Name}=St) ->
     Caption = wings() ++ " - " ++ filename:basename(Name) ++
 	"* [" ++ ?__(1,"auto-saved") ++ "]",
-    sdl_video:wm_setCaption(Caption, Caption),
+    wings_io:set_title(Caption),
     St;
 caption(#st{file=Name}=St) ->
     Caption = wings() ++ " - " ++ filename:basename(Name) ++ "*",
-    sdl_video:wm_setCaption(Caption, Caption),
+    wings_io:set_title(Caption),
     St.
 
+%% pretty_filename(Name) -> PrettyName
+%%  If the Name *seems* to be encoded in UTF8, convert to a
+%%  list of Unicode characters. Leave it alone otherwise.
+%%
+pretty_filename(Name0) ->
+    case all(fun(C) -> C < 128 end, Name0) of
+	true ->
+	    %% Only 7-bit ASCII characters. Already fine.
+	    Name0;
+	false ->
+	    %% Try to convert it from UTF8 to a list of Unicode characters.
+	    Name1 = list_to_binary(Name0),
+	    case unicode:characters_to_list(Name1, utf8) of
+		Name when is_list(Name) -> unicode_combine(Name);
+		_ -> Name0
+	    end
+    end.
+
+
 wings() ->
     case ?wings_branch of
 	"" -> debug("Wings3D");
@@ -162,6 +181,17 @@ debug(Caption) -> Caption ++ " [debug]".
 debug(Caption) -> Caption.
 -endif.
 
+win32_special_folder(R, FolderType) ->
+    Key = "\\hkcu\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders",
+    case win32reg:change_key(R, Key) of
+	ok ->
+	    case win32reg:value(R, FolderType) of
+		{error,_} -> none;
+		{ok,Value} -> Value
+	    end;
+	_ -> error
+    end.
+
 %%%
 %%% Local functions.
 %%%
@@ -190,23 +220,18 @@ log_file_dir() ->
 	_Other -> "."
     end.
 
-log_file_dir({unix,_}) -> os:getenv("HOME");
+log_file_dir({unix,_}) ->
+    os:getenv("HOME");
 log_file_dir({win32,_}) ->
-    Root = code:root_dir(),
-    case filelib:is_file("Wings3D.exe") of
-        true -> Root;
-        false ->
-            %% Development system.
-            case code:which(?MODULE) of
-                Name0 when is_list(Name0) ->
-                    Name = filename:dirname(Name0),
-                    case filename:basename(Name) of
-                        "ebin" -> filename:dirname(Name);
-                        "patches" -> filename:dirname(Name);
-                        _Other -> Name
-                    end
-            end
-    end.
+    {ok,R} = win32reg:open([read]),
+    Res = case win32_special_folder(R, "Desktop") of
+	      none ->
+		  {ok,[Drive,$:|_]} = file:get_cwd(),
+		  [Drive,$:,$/];
+	      Path -> Path
+	  end,
+    ok = win32reg:close(R),
+    Res.
 	
 open_log_file(Name) ->
     {ok,F} = file:open(Name, [write]),
@@ -232,15 +257,16 @@ try_arg(F, #st{shapes=Shapes}, N) ->
 try_arg(F, #we{}=We, N) ->
     arg(F, N),
     dump_we(F, We);
-try_arg(F, {I,_}=GbTree, N) when is_integer(I) ->
-    case catch gb_trees:to_list(GbTree) of
-	{'EXIT',_} -> ok;
+try_arg(F, Tab, N) ->
+    try array:sparse_to_orddict(Tab) of
 	[{_,#edge{}}|_]=Es ->
 	    arg(F, N),
 	    dump_edges(F, Es);
-	_ -> ok
-    end;
-try_arg(_, _, _) -> ok.
+	_ ->
+	    ok
+    catch _:_ ->
+	    ok
+    end.
 
 arg(F, N) ->
     io:format(F, "Argument #~p:\n", [N]).
@@ -248,14 +274,14 @@ arg(F, N) ->
 dump_shape(F, #we{}=We) ->
     dump_we(F, We).
 
-dump_we(F, #we{name=Name,id=Id,mode=Mode,es=Etab,fs=Ftab,
+dump_we(F, #we{name=Name,id=Id,es=Etab,fs=Ftab,
 	       next_id=Next}) ->
     io:put_chars(F, "\n"),
     io:format(F, "OBJECT ~p: ~p\n", [Id,Name]),
     io:format(F, "=======================\n", []),
-    io:format(F, "   mode=~p next_id=~p\n", [Mode,Next]),
+    io:format(F, "   next_id=~p\n", [Next]),
     dump_faces(F, gb_trees:to_list(Ftab)),
-    dump_edges(F, gb_trees:to_list(Etab)).
+    dump_edges(F, array:sparse_to_orddict(Etab)).
     
 dump_edges(F, Es) ->
     io:put_chars(F, "\n"),
@@ -274,12 +300,53 @@ dump_faces(F, Fs) ->
 %% Dumping of data structures.
 %% 
 
-show_edge(F, Edge, #edge{vs=Vs,ve=Ve,a=A,b=B,lf=Lf,rf=Rf,ltpr=Lpred,ltsu=Lsucc,
+show_edge(F, Edge, #edge{vs=Vs,ve=Ve,lf=Lf,rf=Rf,ltpr=Lpred,ltsu=Lsucc,
 			 rtpr=Rpred,rtsu=Rsucc}) ->
     io:format(F, "~p: vs=~p ve=~p\n", [Edge,Vs,Ve]),
-    io:format(F, "  a=~p b=~p\n", [A,B]),
     io:format(F, "  left: face=~p pred=~p succ=~p\n", [Lf,Lpred,Lsucc]),
     io:format(F, "  right: face=~p pred=~p succ=~p\n", [Rf,Rpred,Rsucc]).
 
 show_face(F, Face, Edge) ->
     io:format(F, "~p: edge=~p\n", [Face,Edge]).
+
+
+%% unicode_combine(String0) -> String
+%%  Combine diacritical characters with the previous character.
+%%  (We need to do this because Mac OS X encodes filenames in
+%%  this way. Sigh!)
+%%
+%%  For instance, the characters 'LATIN CAPITAL LETTER A' (U+0041) and
+%%  'COMBINING DIAERESIS' (U+0308) combines to form the letter
+%%  'LATIN CAPITAL LETTER A WITH DIAERESIS' (U+00C4).
+%%
+%%  Currently, we only handle a few of the possible character combinations,
+%%  namely those that occur in Swedish/Finnish, German, and Russian.
+%%
+unicode_combine([$A,16#0308|T]) ->
+    [16#C4|unicode_combine(T)];			%"Ä"
+unicode_combine([$a,16#0308|T]) ->
+    [16#E4|unicode_combine(T)];			%"ä"
+unicode_combine([$A,16#030A|T]) ->
+    [16#C5|unicode_combine(T)];			%"Å"
+unicode_combine([$a,16#030A|T]) ->
+    [16#E5|unicode_combine(T)];			%"å"
+unicode_combine([$O,16#0308|T]) ->
+    [16#D6|unicode_combine(T)];			%"Ö"
+unicode_combine([$o,16#0308|T]) ->
+    [16#F6|unicode_combine(T)];			%"ö"
+unicode_combine([$U,16#0308|T]) ->
+    [16#DC|unicode_combine(T)];			%"Ü"
+unicode_combine([$u,16#0308|T]) ->
+    [16#FC|unicode_combine(T)];			%"ü"
+%% Russian.
+unicode_combine([16#0415,16#0308|T]) ->
+    [16#0401|unicode_combine(T)];		%"Ё"
+unicode_combine([16#0435,16#0308|T]) ->
+    [16#0451|unicode_combine(T)];		%"ё"
+unicode_combine([16#0418,16#0306|T]) ->
+    [16#0419|unicode_combine(T)];		%"Й"
+unicode_combine([16#0438,16#0306|T]) ->
+    [16#0439|unicode_combine(T)];		%"й"
+unicode_combine([H|T]) ->
+    [H|unicode_combine(T)];
+unicode_combine([]) -> [].
