@@ -21,6 +21,8 @@
 	 cap/1,upper/1,stringify/1,quote/1,
 	 add_vpos/2,update_vpos/2,
 	 gb_trees_smallest_key/1,gb_trees_largest_key/1,
+	 array_keys/1,array_smallest_key/1,array_greatest_key/1,
+	 array_is_empty/1,array_entries/1,
 	 nice_float/1,
 	 unique_name/2,
 	 lib_dir/1,
@@ -70,6 +72,7 @@ make_vector(z) -> {0.0,0.0,1.0};
 make_vector(free) -> free;
 make_vector(normal) -> normal;
 make_vector(intrude) -> normal;
+make_vector({region,normal}) -> normal;
 make_vector(Axis) when Axis == last_axis; Axis == default_axis ->
     {_,Vec} = wings_pref:get_value(Axis),
     Vec.
@@ -124,15 +127,15 @@ upper([]) -> [].
 add_vpos(Vs, #we{vp=Vtab}) -> add_vpos(Vs, Vtab);
 add_vpos(Vs, Vtab) ->
     foldl(fun(V, A) ->
-		  [{V,gb_trees:get(V, Vtab)}|A]
+		  [{V,array:get(V, Vtab)}|A]
 	  end, [], Vs).
 
 update_vpos(Vs, #we{vp=Vtab}) -> update_vpos(Vs, Vtab);
 update_vpos(Vs, Vtab) ->
     foldl(fun({V,_}, A) ->
-		  [{V,gb_trees:get(V, Vtab)}|A];
+		  [{V,array:get(V, Vtab)}|A];
 	     ({V,_,Dist,Inf}, A) ->
-		  [{V,gb_trees:get(V, Vtab),Dist,Inf}|A]
+		  [{V,array:get(V, Vtab),Dist,Inf}|A]
 	  end, [], reverse(Vs)).
 
 gb_trees_smallest_key(Tree) ->
@@ -143,6 +146,42 @@ gb_trees_largest_key(Tree) ->
     {Key,_Val} = gb_trees:largest(Tree),
     Key.
 
+array_keys(Array) ->
+    array:sparse_foldr(fun(I, _, A) -> [I|A] end, [], Array).
+
+array_smallest_key(Array) ->
+    try
+	array:sparse_foldl(fun(I, _, _) -> throw(I) end, [], Array),
+	erlang:error(empty_array)
+    catch
+	throw:I when is_integer(I) ->
+	    I
+    end.
+
+array_greatest_key(Array) ->
+    try
+	array:sparse_foldr(fun(I, _, _) -> throw(I) end, [], Array),
+	erlang:error(empty_array)
+    catch
+	throw:I when is_integer(I) ->
+	    I
+    end.
+
+array_is_empty(Array) ->
+    try
+	array:sparse_foldr(fun(_, _, _) -> throw(false) end, [], Array),
+	throw(true)
+    catch
+	throw:Empty ->
+	    Empty
+    end.
+
+%% array_entries(Array) -> NumberOfEntries
+%%  Return the number of non-default entries in the array.
+%%
+array_entries(Array) ->
+    array:sparse_foldl(fun(_, _, N) -> N + 1 end, 0, Array).
+			        
 nice_float(F) when is_float(F) ->
     simplify_float(lists:flatten(io_lib:format("~f", [F]))).
 
