@@ -3,7 +3,7 @@
 %%
 %%     This module implements "vectors" and the secondary selection mode.
 %%
-%%  Copyright (c) 2002-2008 Bjorn Gustavsson.
+%%  Copyright (c) 2002-2009 Bjorn Gustavsson.
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -219,17 +219,7 @@ handle_event_4({action,{view,highlight_aim}}, Ss, St0) ->
       false ->
 	    handle_event_4({action,{view,aim}}, Ss, St0);
       true ->
-        HL0 = wings_pref:get_value(highlight_aim_at_unselected),
-        HL1 = wings_pref:get_value(highlight_aim_at_selected),
-        {_,X,Y} = wings_wm:local_mouse_state(),
-        {{_,Cmd},St1} = case wings_pick:do_pick(X, Y, St0) of
-          {add,_,St} when HL0 =:= true ->
-              {{view,highlight_aim},{add,St0,St}};
-          {delete,_,St} when HL1 =:= true ->
-              {{view,highlight_aim},{delete,St0,St}};
-          _Other -> 
-              {{view,aim}, St0}
-        end,
+        {{_,Cmd},St1} = wings:highlight_aim_setup(St0),
         St2 = wings_view:command(Cmd,St1),
         get_event(Ss, St2)
     end;
@@ -521,7 +511,7 @@ guard_string() ->
 get_pos(Vs,We,Acc) ->
     Positions = lists:foldl(fun(Vert,A) ->
                 #we{vp=Vtab} = We,
-                Pos = gb_trees:get(Vert,Vtab),
+                Pos = array:get(Vert, Vtab),
                 [Pos|A]
                 end,[],Vs),
     [Positions|Acc].
@@ -543,7 +533,7 @@ get_vec(vertex, [Va, Vb], [We0, We1]) ->
     VaPos = wings_vertex:pos(Va, We0),
     VbPos = wings_vertex:pos(Vb, We1),
     Vec = e3d_vec:norm_sub(VaPos, VbPos),
-    Center = e3d_vec:average([VaPos,VbPos]),
+    Center = e3d_vec:average(VaPos, VbPos),
     Normal = e3d_vec:norm(e3d_vec:add(wings_vertex:normal(Va, We0),
               wings_vertex:normal(Vb, We1))),
     [{{Center,Vec},
@@ -561,7 +551,7 @@ get_vec(edge, [{Edges1},{Edges2}], [We1,We2]) ->
     {[Vs1],[Vs2]} ->
         LoopCenter1 = wings_vertex:center(Vs1, We1),
         LoopCenter2 = wings_vertex:center(Vs2, We2),
-        Center = e3d_vec:average([LoopCenter1, LoopCenter2]),
+        Center = e3d_vec:average(LoopCenter1, LoopCenter2),
         Vec = e3d_vec:norm_sub(LoopCenter1, LoopCenter2),
         [{{Center,Vec},?__(28,"Axis between edge loop centers saved as axis.")}];		
     _Other ->
@@ -569,15 +559,15 @@ get_vec(edge, [{Edges1},{Edges2}], [We1,We2]) ->
     end;
 
 get_vec(edge, [Edge1,Edge2], [#we{es=Etab0,vp=Vtab0},#we{es=Etab1,vp=Vtab1}]) ->
-    #edge{vs=Va1,ve=Vb1} = gb_trees:get(Edge1, Etab0),
-    #edge{vs=Va2,ve=Vb2} = gb_trees:get(Edge2, Etab1),
-    Va1Pos = gb_trees:get(Va1, Vtab0),
-    Vb1Pos = gb_trees:get(Vb1, Vtab0),
-    Va2Pos = gb_trees:get(Va2, Vtab1),
-    Vb2Pos = gb_trees:get(Vb2, Vtab1),
-    Center1 = e3d_vec:average([Va1Pos,Vb1Pos]),
-    Center2 = e3d_vec:average([Va2Pos,Vb2Pos]),
-    Center = e3d_vec:average([Center1,Center2]),
+    #edge{vs=Va1,ve=Vb1} = array:get(Edge1, Etab0),
+    #edge{vs=Va2,ve=Vb2} = array:get(Edge2, Etab1),
+    Va1Pos = array:get(Va1, Vtab0),
+    Vb1Pos = array:get(Vb1, Vtab0),
+    Va2Pos = array:get(Va2, Vtab1),
+    Vb2Pos = array:get(Vb2, Vtab1),
+    Center1 = e3d_vec:average(Va1Pos, Vb1Pos),
+    Center2 = e3d_vec:average(Va2Pos, Vb2Pos),
+    Center = e3d_vec:average(Center1, Center2),
     Vec = e3d_vec:norm_sub(Center1, Center2),
     Vec1 = e3d_vec:norm_sub(Va1Pos,Vb1Pos),
     Vec2 = e3d_vec:norm_sub(Va2Pos,Vb2Pos),
@@ -593,7 +583,7 @@ get_vec(edge, [Edge1,Edge2], [#we{es=Etab0,vp=Vtab0},#we{es=Etab1,vp=Vtab1}]) ->
 get_vec(face, [Face1, Face2], [We0, We1]) ->
     Center1 = wings_face:center(Face1, We0),
     Center2 = wings_face:center(Face2, We1),
-    Center = e3d_vec:average([Center1,Center2]),
+    Center = e3d_vec:average(Center1, Center2),
     Vec = e3d_vec:norm_sub(Center1, Center2),
     Face1n = wings_face:normal(Face1, We0),
     Face2n = wings_face:normal(Face2, We1),
@@ -607,9 +597,9 @@ get_vec(face, [Face1, Face2], [We0, We1]) ->
 
 %% Use single edge as axis
 get_vec(edge, [Edge], #we{es=Etab,vp=Vtab}=We) ->
-    #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf} = gb_trees:get(Edge, Etab),
-    VaPos = gb_trees:get(Va, Vtab),
-    VbPos = gb_trees:get(Vb, Vtab),
+    #edge{vs=Va,ve=Vb,lf=Lf,rf=Rf} = array:get(Edge, Etab),
+    VaPos = array:get(Va, Vtab),
+    VbPos = array:get(Vb, Vtab),
     Vec = e3d_vec:norm_sub(VbPos, VaPos),
     Center = wings_vertex:center([Va,Vb], We),
     Ln = wings_face:normal(Lf, We),
@@ -621,19 +611,19 @@ get_vec(edge, [Edge], #we{es=Etab,vp=Vtab}=We) ->
       {?__(3,"Edge normal saved as axis"),?__(4,"Save edge direction")}}];
 %% Use direction between two edges
 get_vec(edge, [Edge1,Edge2], #we{es=Etab,vp=Vtab}) ->
-    #edge{vs=Va1,ve=Vb1} = gb_trees:get(Edge1, Etab),
-    #edge{vs=Va2,ve=Vb2} = gb_trees:get(Edge2, Etab),
-    Va1Pos = gb_trees:get(Va1, Vtab),
-    Vb1Pos = gb_trees:get(Vb1, Vtab),
-    Va2Pos = gb_trees:get(Va2, Vtab),
-    Vb2Pos = gb_trees:get(Vb2, Vtab),
-    Center1 = e3d_vec:average([Va1Pos,Vb1Pos]),
-    Center2 = e3d_vec:average([Va2Pos,Vb2Pos]),
-    Center = e3d_vec:average([Center1,Center2]),
+    #edge{vs=Va1,ve=Vb1} = array:get(Edge1, Etab),
+    #edge{vs=Va2,ve=Vb2} = array:get(Edge2, Etab),
+    Va1Pos = array:get(Va1, Vtab),
+    Vb1Pos = array:get(Vb1, Vtab),
+    Va2Pos = array:get(Va2, Vtab),
+    Vb2Pos = array:get(Vb2, Vtab),
+    Center1 = e3d_vec:average(Va1Pos, Vb1Pos),
+    Center2 = e3d_vec:average(Va2Pos, Vb2Pos),
+    Center = e3d_vec:average(Center1, Center2),
     Vec = e3d_vec:norm_sub(Center1, Center2),
-    Vec1 = e3d_vec:norm_sub(Va1Pos,Vb1Pos),
-    Vec2 = e3d_vec:norm_sub(Va2Pos,Vb2Pos),
-    Cross = e3d_vec:cross(Vec1,Vec2),
+    Vec1 = e3d_vec:norm_sub(Va1Pos, Vb1Pos),
+    Vec2 = e3d_vec:norm_sub(Va2Pos, Vb2Pos),
+    Cross = e3d_vec:cross(Vec1, Vec2),
     [{{Center,Vec},
       {?__(5,"Direction between edges saved as axis."),?__(24,"Save cross vector")}},
      {{Center,Cross},
@@ -648,7 +638,7 @@ get_vec(edge, Edges, #we{vp=Vtab}=We) ->
 	[Vs1,Vs2] ->
 	    LoopCenter1 = wings_vertex:center(Vs1, We),
 	    LoopCenter2 = wings_vertex:center(Vs2, We),
-		Center = e3d_vec:average([LoopCenter1, LoopCenter2]),
+		Center = e3d_vec:average(LoopCenter1, LoopCenter2),
 	    Vec = e3d_vec:norm_sub(LoopCenter1, LoopCenter2),
 	    [{{Center,Vec},?__(28,"Axis between edge loop centers saved as axis.")}];		
 	_Other ->
@@ -702,7 +692,7 @@ get_vec(face, [Face], We) ->
 get_vec(face, [Face1,Face2], We) ->
     Center1 = wings_face:center(Face1, We),
     Center2 = wings_face:center(Face2, We),
-    Center = e3d_vec:average([Center1,Center2]),
+    Center = e3d_vec:average(Center1, Center2),
     Vec = e3d_vec:norm_sub(Center1, Center2),
     Face1n = wings_face:normal(Face1, We),
     Face2n = wings_face:normal(Face2, We),
@@ -766,10 +756,9 @@ kill_mirror(#st{shapes=Shs0}=St) ->
     Shs = kill_mirror_1(gb_trees:values(Shs0), []),
     St#st{shapes=Shs}.
 
-kill_mirror_1([#we{id=Id,mirror=none}=We|Wes], Acc) ->
+kill_mirror_1([#we{id=Id}=We0|Wes], Acc) ->
+    We = wings_we:break_mirror(We0),
     kill_mirror_1(Wes, [{Id,We}|Acc]);
-kill_mirror_1([#we{id=Id}=We|Wes], Acc) ->
-    kill_mirror_1(Wes, [{Id,We#we{mirror=none}}|Acc]);
 kill_mirror_1([], Acc) ->
     gb_trees:from_orddict(reverse(Acc)).
 
