@@ -42,11 +42,11 @@ from_edges(Es, We) ->
 %%
 
 fold(F, Acc, V, #we{vc=Vct}=We) ->
-    Edge = gb_trees:get(V, Vct),
+    Edge = array:get(V, Vct),
     fold(F, Acc, V, Edge, We).
 
 fold(F, Acc0, V, Edge, #we{es=Etab}) ->
-    Acc = case gb_trees:get(Edge, Etab) of
+    Acc = case array:get(Edge, Etab) of
 	      #edge{vs=V,lf=Face,rf=Other,rtpr=NextEdge}=E ->
 		  F(Edge, Face, E, Acc0);
 	      #edge{ve=V,lf=Face,rf=Other,rtsu=NextEdge}=E ->
@@ -56,7 +56,7 @@ fold(F, Acc0, V, Edge, #we{es=Etab}) ->
 
 fold(_F, Acc, _V, _Face, Last, Last, _Etab) -> Acc;
 fold(F, Acc0, V, Face, Edge, LastEdge, Etab) ->
-    Acc = case gb_trees:get(Edge, Etab) of
+    Acc = case array:get(Edge, Etab) of
 	      #edge{vs=V,lf=Face,rf=Other,rtpr=NextEdge}=E ->
 		  F(Edge, Face, E, Acc0);
 	      #edge{ve=V,lf=Face,rf=Other,rtsu=NextEdge}=E ->
@@ -74,16 +74,16 @@ fold(F, Acc0, V, Face, Edge, LastEdge, Etab) ->
 %%
 
 until(F, Acc, V, #we{vc=Vct}=We) ->
-    Edge = gb_trees:get(V, Vct),
+    Edge = array:get(V, Vct),
     until(F, Acc, V, Edge, We).
 
 until(F, Acc, V, Edge, #we{es=Etab}) ->
-    #edge{lf=Face} = gb_trees:get(Edge, Etab),
+    #edge{lf=Face} = array:get(Edge, Etab),
     until(F, Acc, V, Face, Edge, Edge, Etab, not_done).
 
 until(_F, Acc, _V, _Face, Last, Last, _Etab, done) -> Acc;
 until(F, Acc0, V, Face, Edge, LastEdge, Etab, _) ->
-    Acc = case gb_trees:get(Edge, Etab) of
+    Acc = case array:get(Edge, Etab) of
 	      #edge{vs=V,lf=Face,rf=Other,rtpr=NextEdge}=E ->
 		  F(Edge, Face, E, Acc0);
 	      #edge{ve=V,lf=Face,rf=Other,rtsu=NextEdge}=E ->
@@ -107,9 +107,9 @@ other(V, #edge{ve=V,vs=Other}) -> Other.
 %% pos(Vertex, VtabOrWe) -> {X,Y,Z}
 %%  Return the three co-ordinates for a vertex.
 pos(V, #we{vp=Vtab}) ->
-    gb_trees:get(V, Vtab);
+    array:get(V, Vtab);
 pos(V, Vtab) ->
-    gb_trees:get(V, Vtab).
+    array:get(V, Vtab).
 
 %% other_pos(Vertex, EdgeRecord, VtabOrWe) -> {X,Y,Z}
 %%  Pick up the position for the "other vertex" from an edge record.
@@ -119,18 +119,8 @@ other_pos(V, #edge{ve=V,vs=Other}, Tab) -> pos(Other, Tab).
 %% center(We) -> {CenterX,CenterY,CenterZ}
 %%  Find the geometric center of a body.
 center(#we{vp=Vtab}=We) ->
-    Center = e3d_vec:average(gb_trees:values(Vtab)),
-    center_1(Center, We).
-
-center_1(Center, #we{mirror=none}) -> Center;
-center_1(Center, #we{mirror=Face}=We) ->
-    %% Slide the center point down to the nearest point on the mirror plane.
-    MirrorNormal = wings_face:normal(Face, We),
-    FaceVs = wings_face:to_vertices(gb_sets:singleton(Face), We),
-    Origin = wings_vertex:center(FaceVs, We),
-    M0 = e3d_mat:translate(Origin),
-    M = e3d_mat:mul(M0, e3d_mat:project_to_plane(MirrorNormal)),
-    Flatten = e3d_mat:mul(M, e3d_mat:translate(e3d_vec:neg(Origin))),
+    Center = e3d_vec:average(array:sparse_to_list(Vtab)),
+    Flatten = wings_we:mirror_projection(We),
     e3d_mat:mul_point(Flatten, Center).
 
 %% center(VertexGbSet, We) -> {CenterX,CenterY,CenterZ}
@@ -151,7 +141,7 @@ bounding_box(We) ->
     bounding_box(We, none).
 
 bounding_box(#we{vp=Vtab}=We, BB) ->
-    do_bounding_box(gb_trees:values(Vtab), We, BB);
+    do_bounding_box(array:sparse_to_list(Vtab), We, BB);
 bounding_box(Vs, We) ->
     bounding_box(Vs, We, none).
     
@@ -162,7 +152,7 @@ bounding_box(Vs, We, BB) ->
 
 bounding_box_1(Vs0, #we{vp=Vtab}=We, BB) ->
     Vs1 = sofs:from_external(Vs0, [vertex]),
-    R = sofs:from_external(gb_trees:to_list(Vtab), [{vertex,data}]),
+    R = sofs:from_external(array:sparse_to_orddict(Vtab), [{vertex,data}]),
     I = sofs:image(R, Vs1),
     Vs = sofs:to_external(I),
     do_bounding_box(Vs, We, BB).
@@ -231,10 +221,10 @@ flatten_matrix(Origin, PlaneNormal) ->
     M = e3d_mat:mul(M0, e3d_mat:project_to_plane(PlaneNormal)),
     e3d_mat:mul(M, e3d_mat:translate(e3d_vec:neg(Origin))).
 
-flatten_move(V, Matrix, Tab0) ->
-    Pos0 = gb_trees:get(V, Tab0),
+flatten_move(V, Matrix, Vtab0) ->
+    Pos0 = array:get(V, Vtab0),
     Pos = e3d_mat:mul_point(Matrix, Pos0),
-    gb_trees:update(V, Pos, Tab0).
+    array:set(V, Pos, Vtab0).
 
 %% dissolve_isolated_vs([Vertex], We) -> We'
 %%  Remove all isolated vertices ("winged vertices", or vertices
@@ -385,7 +375,7 @@ nearest_pair_smart_1([], _, _, _, _) ->
 nearest_pair(Face, AllVs, #we{vp=Vtab}=We) ->
     Vs0 = ordsets:from_list(wings_face:vertices_ccw(Face, We)),
     Vs = ordsets:intersection(Vs0, AllVs),
-    VsPos = [{V,gb_trees:get(V, Vtab)} || V <- Vs],
+    VsPos = [{V,array:get(V, Vtab)} || V <- Vs],
     nearest_pair(VsPos, Face, We, []).
 
 nearest_pair([{V,Pos}|VsPos], Face, We, Acc0) ->
@@ -445,12 +435,19 @@ try_connect_1(Va, Vb, Face, We0) ->
 	false -> Res
     end.
 
+%% force_connect(Vstart, Vend, Face, We0) -> We
+%%  Create a new edge between the vertices Vstart and Vend
+%%  in face Face, also creating a new face.
+%%
+%%  The edges between Vstart and Vend (in the clockwise direction)
+%%  and the left side of the new edge will belong to the new face.
+%%
 force_connect(Vstart, Vend, Face, #we{es=Etab0,fs=Ftab0}=We0) ->
-    {NewFace,We} = wings_we:new_ids(1, We0),
+    {NewFace,We1} = wings_we:new_ids(1, We0),
     NewEdge = NewFace,
     NeRec0 = #edge{vs=Vstart,ve=Vend,lf=NewFace,rf=Face},
 
-    Iter0 = wings_face:iterator(Face, We),
+    Iter0 = wings_face:iterator(Face, We1),
     Iter1 = wings_face:skip_to_cw(Vstart, Iter0),
     {_,_,_,Iter2} = wings_face:next_ccw(Iter1),
     {Etab1,NeRec1,Iter3} = connect_1(Iter2, Vstart, NewEdge, NeRec0, Etab0),
@@ -460,21 +457,27 @@ force_connect(Vstart, Vend, Face, #we{es=Etab0,fs=Ftab0}=We0) ->
 
     Ftab1 = gb_trees:insert(NewFace, NewEdge, Ftab0),
     Ftab = gb_trees:update(Face, NewEdge, Ftab1),
-    Mat = wings_facemat:face(Face, We),
-    {wings_facemat:assign(Mat, [NewFace], We#we{es=Etab,fs=Ftab}),NewFace}.
+    Mat = wings_facemat:face(Face, We1),
+    We2 = We1#we{es=Etab,fs=Ftab},
+
+    LeftAttr = wings_va:vtx_attrs(Vstart, Face, We0),
+    RightAttr = wings_va:vtx_attrs(Vend, Face, We0),
+    We = wings_va:set_both_edge_attrs(NewEdge, LeftAttr, RightAttr, We2),
+
+    {wings_facemat:assign(Mat, [NewFace], We),NewFace}.
 
 %% connect_1(Iter0, Vstart, NewEdge, NeRec0, Etab0) -> {Etab,NeRec,Iter}
 %%  Connect the edge immediately before Vstart.
 connect_1(Iter0, Vstart, NewEdge, NeRec0, Etab) ->
     case wings_face:next_cw(Iter0) of
-	{_,Edge,#edge{b=ColB,ve=Vstart}=Rec0,Iter} ->
-	    NeRec = NeRec0#edge{a=ColB,rtpr=Edge},
+	{_,Edge,#edge{ve=Vstart}=Rec0,Iter} ->
+	    NeRec = NeRec0#edge{rtpr=Edge},
 	    Rec = Rec0#edge{rtsu=NewEdge};
-	{_,Edge,#edge{a=ColA,vs=Vstart}=Rec0,Iter} ->
-	    NeRec = NeRec0#edge{a=ColA,rtpr=Edge},
+	{_,Edge,#edge{vs=Vstart}=Rec0,Iter} ->
+	    NeRec = NeRec0#edge{rtpr=Edge},
 	    Rec = Rec0#edge{ltsu=NewEdge}
     end,
-    {gb_trees:update(Edge, Rec, Etab),NeRec,Iter}.
+    {array:set(Edge, Rec, Etab),NeRec,Iter}.
 
 %% connect_2(Iter0, Vstart, NewEdge, NeRec0, Etab) -> {Etab,NeRec}
 %%  Connect the edge immediately after Vstart.
@@ -487,7 +490,7 @@ connect_2(Iter, Vstart, NewEdge, NeRec0, Etab) ->
 	    NeRec = NeRec0#edge{ltsu=Edge},
 	    Rec = Rec0#edge{ltpr=NewEdge}
     end,
-    {gb_trees:update(Edge, Rec, Etab),NeRec}.
+    {array:set(Edge, Rec, Etab),NeRec}.
 
 %% connect_3(Iter, Face, Vend, NewFace, Etab0) -> {Etab,Iter}
 %%  Replace the face for all edges between Vstart and Vend.
@@ -497,11 +500,11 @@ connect_3(Iter0, Face, Vend, NewFace, Etab0) ->
 
     %% Ignore the record returned by the iterator, because it
     %% is stale for the edge that was updated by connect_2/5.
-    Rec = case gb_trees:get(Edge, Etab0) of
+    Rec = case array:get(Edge, Etab0) of
 	      #edge{lf=Face}=Rec0 -> Rec0#edge{lf=NewFace};
 	      #edge{rf=Face}=Rec0 -> Rec0#edge{rf=NewFace}
 	  end,
-    Etab = gb_trees:update(Edge, Rec, Etab0),
+    Etab = array:set(Edge, Rec, Etab0),
     case Rec of
 	#edge{vs=Vend} -> {Etab,Iter0};
 	#edge{ve=Vend} -> {Etab,Iter0};
@@ -512,15 +515,15 @@ connect_3(Iter0, Face, Vend, NewFace, Etab0) ->
 %%  Patches the final two edges.
 connect_4(Iter0, Vend, NewEdge, NeRec0, Etab0) ->
     {_,Edge,_,Iter} = wings_face:next_cw(Iter0),
-    Rec = case gb_trees:get(Edge, Etab0) of
-	      #edge{b=ColB,ve=Vend}=Rec0 ->
-		  NeRec1 = NeRec0#edge{b=ColB,ltpr=Edge},
+    Rec = case array:get(Edge, Etab0) of
+	      #edge{ve=Vend}=Rec0 ->
+		  NeRec1 = NeRec0#edge{ltpr=Edge},
 		  Rec0#edge{rtsu=NewEdge};
-	      #edge{a=ColA,vs=Vend}=Rec0 ->
-		  NeRec1 = NeRec0#edge{b=ColA,ltpr=Edge},
+	      #edge{vs=Vend}=Rec0 ->
+		  NeRec1 = NeRec0#edge{ltpr=Edge},
 		  Rec0#edge{ltsu=NewEdge}
 	  end,
-    Etab1 = gb_trees:update(Edge, Rec, Etab0),
+    Etab1 = array:set(Edge, Rec, Etab0),
 
     %% Now for the final edge.
     FinalRec = case wings_face:next_cw(Iter) of
@@ -531,8 +534,8 @@ connect_4(Iter0, Vend, NewEdge, NeRec0, Etab0) ->
 		       NeRec = NeRec1#edge{rtsu=Final},
 		       FinalRec0#edge{ltpr=NewEdge}
 	       end,
-    Etab = gb_trees:update(Final, FinalRec, Etab1),
-    gb_trees:insert(NewEdge, NeRec, Etab).
+    Etab = array:set(Final, FinalRec, Etab1),
+    array:set(NewEdge, NeRec, Etab).
 
 %% outer_vertices_ccw(Faces, We) -> [V] | error
 %%  Faces (non-empty list or gb_set) must comprise a single face region
@@ -600,7 +603,7 @@ order_edges(_, [_,_|_], _, _) ->
 %%  edges from the given list of vertices.
 reachable(Vs0, #we{es=Etab,vc=Vct}) when is_list(Vs0) ->
     Es0 = foldl(fun(V, A) ->
-			[gb_trees:get(V, Vct)|A]
+			[array:get(V, Vct)|A]
 		end, [], Vs0),
     Es1 = gb_sets:from_list(Es0),
     Es = reachable_edges(Es1, Etab, gb_trees:empty()),
@@ -614,7 +617,7 @@ reachable_edges(Ws0, Etab, Reachable0) ->
 	true -> Reachable0;
 	false ->
 	    {Edge,Ws1} = gb_sets:take_smallest(Ws0),
-	    Rec = gb_trees:get(Edge, Etab),
+	    Rec = array:get(Edge, Etab),
 	    Reachable = gb_trees:insert(Edge, Rec, Reachable0),
 	    #edge{ltpr=LP,ltsu=LS,rtpr=RP,rtsu=RS} = Rec,
 	    reachable_edges_1([LP,LS,RP,RS], Etab, Ws1, Reachable)
@@ -636,7 +639,7 @@ reachable_edges_1([], Etab, Ws, Reachable) ->
 isolated(#we{vp=Vtab}=We) ->
     Vs0 = foldl(fun(V, A) -> 
 			isolated_1(V, We, A)
-		end, [], gb_trees:keys(Vtab)),
+		end, [], wings_util:array_keys(Vtab)),
     Vs1 = sofs:relation(Vs0),
     Fs0 = sofs:domain(Vs1),
     Fs = sofs:to_external(Fs0),
