@@ -13,7 +13,7 @@
 
 -module(wings_vertex_cmd).
 -export([menu/3,command/2,tighten/3,tighten/4,
-	 connect/2,bevel_vertex/2, flatten/2]).
+	 connect/2,bevel_vertex/2,flatten/2]).
 
 -export([set_color/2]).
 
@@ -128,9 +128,10 @@ extrude(Type, St0) ->
     wings_move:plus_minus(Type, Tvs, St).
 
 extrude_vertices(Vs, We0, Acc) ->
-    We = foldl(fun(V, A) ->
-		       ex_new_vertices(V, We0, A)
-	       end, We0, gb_sets:to_list(Vs)),
+    We = gb_sets:fold(
+	   fun(V, A) ->
+		   ex_new_vertices(V, We0, A)
+	   end, We0, Vs),
     NewVs = wings_we:new_items_as_ordset(vertex, We0, We),
     {We,[{Vs,NewVs,gb_sets:empty(),We}|Acc]}.
 
@@ -141,7 +142,7 @@ ex_new_vertices(V, OrigWe, #we{vp=Vtab}=We0) ->
 	  fun(Edge, Face, Rec, {W0,Vs}) ->
 		  OtherV = wings_vertex:other(V, Rec),
 		  R = edge_ratio(OtherV, OrigWe),
-		  Pos0 = gb_trees:get(OtherV, Vtab),
+		  Pos0 = array:get(OtherV, Vtab),
 		  Dir = e3d_vec:sub(Pos0, Center),
 		  Pos = e3d_vec:add(Center, e3d_vec:mul(Dir, R)),
 		  {W,NewV} = wings_edge:fast_cut(Edge, Pos, W0),
@@ -150,16 +151,22 @@ ex_new_vertices(V, OrigWe, #we{vp=Vtab}=We0) ->
     ex_connect(VsFaces, VsFaces, We).
 
 edge_ratio(V, #we{vp=Vtab}) ->
-    case gb_trees:is_defined(V, Vtab) of
-	false -> 1/3;
-	true -> 0.25
+    case array:get(V, Vtab) of
+	undefined -> 1/3;
+	_ -> 0.25
     end.
 
+ex_connect([_,Face|T], More, #we{mirror=Face}=We) ->
+    ex_connect(T, More, We);
 ex_connect([Va,Face|[Vb|_]=T], More, We0) ->
     {We,_} = wings_vertex:force_connect(Va, Vb, Face, We0),
     ex_connect(T, More, We);
+ex_connect([_,Face], _, #we{mirror=Face}=We) ->
+    We;
 ex_connect([Va,Face], [Vb|_], We0) ->
     {We,_} = wings_vertex:force_connect(Va, Vb, Face, We0),
+    We;
+ex_connect([], _, We) ->
     We.
 
 %%%
@@ -184,20 +191,20 @@ bevel_vertex(V, We0) ->
     case length(Es) of
 	2 -> We0;
 	NumEdges ->
-	    {We,_,_} = bevel_vertex_1(V, Es, NumEdges, [], We0, []),
+	    {We,_,_,_} = bevel_vertex_1(V, Es, NumEdges, [], We0, [], []),
 	    We
     end.
 
 bevel_1(VsSet, #we{id=Id}=We0, {Tvs,Fa}) ->
     Vs = gb_sets:to_list(VsSet),
-    {We,Tv,Fs0} = bevel_vertices(Vs, VsSet, We0, We0, [], []),
+    {We,Tv,WeTrans,Fs0} = bevel_vertices(Vs, VsSet, We0, We0, [], [], []),
     FaceSel = case Fs0 of
 		  [] -> Fa;
 		  _ -> [{Id,gb_sets:from_list(Fs0)}|Fa]
 	      end,
-    {We,{[{Id,Tv}|Tvs],FaceSel}}.
+    {We,{[{Id,WeTr} || WeTr <- WeTrans]++[{Id,Tv}|Tvs],FaceSel}}.
 
-bevel_vertices([V|Vs], VsSet, WeOrig, We0, Acc0, Facc) ->
+bevel_vertices([V|Vs], VsSet, WeOrig, We0, Acc0, Facc, WeTrans0) ->
     Adj = adjacent(V, VsSet, WeOrig),
     Es = wings_vertex:fold(
 	   fun(Edge, Face, Rec, Acc) ->
@@ -205,14 +212,15 @@ bevel_vertices([V|Vs], VsSet, WeOrig, We0, Acc0, Facc) ->
 	   end, [], V, We0),
     case length(Es) of
 	2 ->					%Winged vertex - ignore.
-	    bevel_vertices(Vs, VsSet, WeOrig, We0, Acc0, Facc);
+	    bevel_vertices(Vs, VsSet, WeOrig, We0, Acc0, Facc, WeTrans0);
 	NumEdges ->
-	    {We,Acc,Face} = bevel_vertex_1(V, Es, NumEdges, Adj, We0, Acc0),
-	    bevel_vertices(Vs, VsSet, WeOrig, We, Acc, [Face|Facc])
+	    {We,Acc,WeTrans,Face} =
+		bevel_vertex_1(V, Es, NumEdges, Adj, We0, Acc0, WeTrans0),
+	    bevel_vertices(Vs, VsSet, WeOrig, We, Acc, [Face|Facc], WeTrans)
     end;
-bevel_vertices([], _, _, We, Acc, Facc) -> {We,Acc,Facc}.
+bevel_vertices([], _, _, We, Acc, Facc, WeTrans) -> {We,Acc,WeTrans,Facc}.
 
-bevel_vertex_1(V, Es, NumEdges, Adj, We0, Vec0) ->
+bevel_vertex_1(V, Es, NumEdges, Adj, We0, Vec0, WeTrans0) ->
     {InnerFace,We1} = wings_we:new_id(We0),
     {Ids,We2} = wings_we:new_wrap_range(NumEdges, 2, We1),
     #we{es=Etab0,vc=Vct0,vp=Vtab0,fs=Ftab0} = We1,
@@ -220,14 +228,33 @@ bevel_vertex_1(V, Es, NumEdges, Adj, We0, Vec0) ->
     {_,Etab,Vec} = foldl(
 		     fun(E, {Ids0,Etab1,Vs0}) ->
 			     {Etab,Vec} = bevel(V, E, InnerFace, Ids0,
-						Adj, Vtab0, Etab0, Etab1),
+						Adj, Vtab0, Etab1),
 			     {wings_we:bump_id(Ids0),Etab,[Vec|Vs0]}
 		     end, {Ids,Etab0,Vec0}, Es),
     Mat = bevel_material(Es, We2),
     NewEdge = wings_we:id(1, Ids),
     Ftab = gb_trees:insert(InnerFace, NewEdge, Ftab0),
-    We = wings_facemat:assign(Mat, [InnerFace], We2),
-    {We#we{es=Etab,fs=Ftab,vc=Vct,vp=Vtab},Vec,InnerFace}.
+    We3 = wings_facemat:assign(Mat, [InnerFace], We2),
+    We4 = We3#we{es=Etab,fs=Ftab,vc=Vct,vp=Vtab},
+
+    %% Handling updating of vertex attributes.
+    {We,WeTrans} =
+	case wings_va:any_attributes(We4) of
+	    false ->
+		%% No vertex attributes.
+		{We4,WeTrans0};
+	    true ->
+		%% Define a fun to update the vertex attributes dynamically.
+		VaTrans = bevel_va_fun(V, Es, InnerFace, We0),
+		WeTrans1 = [{we,VaTrans}|WeTrans0],
+
+		%% Calculate initial values for vertex attributes.
+		%% This is necessary to get correct results in case two
+		%% neighboring vertices are being beveled.
+		We5 = VaTrans(We4, [0.0]),
+		{We5,WeTrans1}
+	end,
+    {We,Vec,WeTrans,InnerFace}.
 
 bevel_material(Es, We) ->
     bevel_material(Es, We, []).
@@ -242,7 +269,7 @@ bevel_material([], _, A0) ->
     [{_,Mat}|_] = sort([{-length(Fs),M} || {M,Fs} <- A]),
     Mat.
 
-bevel(V, {Edge,Face,Rec0}, InnerFace, Ids, Adj, Vtab, OrigEtab, Etab0) ->
+bevel(V, {Edge,Face,Rec0}, InnerFace, Ids, Adj, Vtab, Etab0) ->
     Vprev = wings_we:id(0, Ids),
     Eprev = wings_we:id(1, Ids),
     Va = wings_we:id(2, Ids),
@@ -251,58 +278,56 @@ bevel(V, {Edge,Face,Rec0}, InnerFace, Ids, Adj, Vtab, OrigEtab, Etab0) ->
     Enext = wings_we:id(5, Ids),
     {Rec,Curr} =
 	case Rec0 of
-	    #edge{vs=V,ve=Vother,rf=Face,rtpr=ColEdge} ->
-		Col = bevel_color(ColEdge, Face, OrigEtab),
+	    #edge{vs=V,ve=Vother,rf=Face} ->
 		{Rec0#edge{vs=Va,rtpr=Ecurr,ltsu=Eprev},
-		 Rec0#edge{vs=Vb,ve=Va,a=Col,lf=InnerFace,
+		 Rec0#edge{vs=Vb,ve=Va,lf=InnerFace,
 			   rtsu=Edge,ltpr=Eprev,ltsu=Enext}};
-	    #edge{vs=V,ve=Vother,rf=Of,rtpr=ColEdge} ->
-		Col = bevel_color(ColEdge, Of, OrigEtab),
+	    #edge{vs=V,ve=Vother} ->
 		{Rec0#edge{vs=Va,rtpr=Ecurr,ltsu=Enext},
-		 Rec0#edge{vs=Vprev,ve=Va,a=Col,lf=InnerFace,
+		 Rec0#edge{vs=Vprev,ve=Va,lf=InnerFace,
 			   rtsu=Edge,ltpr=Enext,ltsu=Eprev}};
-	    #edge{ve=V,vs=Vother,lf=Face,ltpr=ColEdge} ->
-		Col = bevel_color(ColEdge, Face, OrigEtab),
+	    #edge{ve=V,vs=Vother,lf=Face} ->
 		{Rec0#edge{ve=Va,ltpr=Ecurr,rtsu=Eprev},
-		 Rec0#edge{vs=Va,ve=Vb,b=Col,rf=InnerFace,
+		 Rec0#edge{vs=Va,ve=Vb,rf=InnerFace,
 			   ltsu=Edge,rtpr=Eprev,rtsu=Enext}};
-	    #edge{ve=V,vs=Vother,lf=Of,ltpr=ColEdge} ->
-		Col = bevel_color(ColEdge, Of, OrigEtab),
+	    #edge{ve=V,vs=Vother} ->
 		{Rec0#edge{ve=Va,ltpr=Ecurr,rtsu=Enext},
-		 Rec0#edge{vs=Va,ve=Vprev,b=Col,rf=InnerFace,
+		 Rec0#edge{vs=Va,ve=Vprev,rf=InnerFace,
 			   ltsu=Edge,rtpr=Enext,rtsu=Eprev}}
 	end,
-    Etab = gb_trees:update(Edge, Rec, Etab0),
-    Vpos = gb_trees:get(V, Vtab),
+    Etab = array:set(Edge, Rec, Etab0),
+    Vpos = array:get(V, Vtab),
     Vec = bevel_vec(Adj, Vother, Vpos, Vtab),
-    {gb_trees:insert(Ecurr, Curr, Etab),{Vec,Va}}.
+    {array:set(Ecurr, Curr, Etab),{Vec,Va}}.
 
 bevel_vec(Adj, Vother, Vpos, Vtab) ->
-    Opos = gb_trees:get(Vother, Vtab),
+    Opos = array:get(Vother, Vtab),
     case member(Vother, Adj) of
 	true ->
-	    e3d_vec:sub(e3d_vec:average([Opos,Vpos]), Vpos);
+	    e3d_vec:sub(e3d_vec:average(Opos, Vpos), Vpos);
 	false ->
 	    e3d_vec:sub(Opos, Vpos)
     end.
 
 bevel_vertices_1(V, Ids, N, Vct0, Vtab0) ->
-    Pos = gb_trees:get(V, Vtab0),
-    Vct = gb_trees:delete(V, Vct0),
-    Vtab = gb_trees:delete(V, Vtab0),
+    Pos = array:get(V, Vtab0),
+    Vct = array:reset(V, Vct0),
+    Vtab = array:reset(V, Vtab0),
     bevel_new_vertices(Ids, N, Pos, Vct, Vtab).
 
 bevel_new_vertices(Ids, N, Pos, Vct0, Vtab0) when N > 0 ->
     V = Id = wings_we:id(0, Ids),
     Edge = Id + 1,
-    Vct = gb_trees:insert(V, Edge, Vct0),
-    Vtab = gb_trees:insert(V, Pos, Vtab0),
+    Vct = array:set(V, Edge, Vct0),
+    Vtab = array:set(V, Pos, Vtab0),
     bevel_new_vertices(wings_we:bump_id(Ids), N-1, Pos, Vct, Vtab);
 bevel_new_vertices(_, _, _, Vct, Vtab) -> {Vct,Vtab}.
 
 bevel_normalize(Tvs) ->
     bevel_normalize(Tvs, 1.0E207, []).
 
+bevel_normalize([{_,{we,_}}=WeTrans|Tvs], Min, Acc) ->
+    bevel_normalize(Tvs, Min, [WeTrans|Acc]);
 bevel_normalize([{Id,VecVs0}|Tvs], Min0, Acc) ->
     {VecVs,Min} = bevel_normalize_1(VecVs0, Min0),
     bevel_normalize(Tvs, Min, [{Id,VecVs}|Acc]);
@@ -317,12 +342,6 @@ bevel_normalize_1(VecVs, Min0) ->
 		     {{e3d_vec:norm(Vec),[V]},Min}
 	     end, Min0, VecVs).
 
-bevel_color(ColEdge, Face, Etab0) ->
-    case gb_trees:get(ColEdge, Etab0) of
-	#edge{lf=Face,a=Col} -> Col;
-	#edge{rf=Face,b=Col} -> Col
-    end.
-
 adjacent(V, Vs, We) ->
     wings_vertex:fold(
       fun(_, _, Rec, A) ->
@@ -332,7 +351,78 @@ adjacent(V, Vs, We) ->
 		  false -> A
 	      end
       end, [], V, We).
-    
+
+%%%
+%%% Update vertex attributes for Bevel.
+%%%
+
+bevel_va_fun(V, Es0, InnerFace, OrigWe) ->
+    Es = bevel_va_preprocess_edges(Es0, V, OrigWe),
+    fun(We, [Move]) ->
+	    bevel_va(Es, Move, V, InnerFace, OrigWe, We)
+    end.
+
+%% Preprocess the edge information to optimize performance
+%% of the drag operation.
+bevel_va_preprocess_edges([{Edge,_,Rec}|T], V, OrigWe) ->
+    %% If performance becomes a real issue, we could precalculate
+    %% a lot more (at the expense of using more memory). Here we
+    %% basically just precalculates Scale (because it seems
+    %% ridiculously expensive to calculate it for every mouse
+    %% motion event).
+    case Rec of
+	#edge{vs=V,ve=OtherV} ->
+	    Scale = bevel_scale_factor(V, OtherV, OrigWe),
+	    [{vs,Edge,Scale}|bevel_va_preprocess_edges(T, V, OrigWe)];
+	#edge{ve=V,vs=OtherV} ->
+	    Scale = bevel_scale_factor(V, OtherV, OrigWe),
+	    [{ve,Edge,Scale}|bevel_va_preprocess_edges(T, V, OrigWe)]
+    end;
+bevel_va_preprocess_edges([], _, _) -> [].
+
+%% During the drag, update the vertex attributes.
+bevel_va([{vs,Edge,Scale}|Es], Move, V, InnerFace, OrigWe, #we{es=Etab}=We0) ->
+    W = Move * Scale,
+    LeftAttr = wings_va:edge_attrs(Edge, left, W, OrigWe),
+    RightAttr = wings_va:edge_attrs(Edge, right, W, OrigWe),
+    InnerAttr = wings_va:average_attrs(LeftAttr, RightAttr),
+    #edge{rtpr=Rtpr,rf=Rf} = array:get(Edge, Etab),
+    We1 = wings_va:set_edge_attrs(Rtpr, Rf, RightAttr, We0),
+    We2 = bevel_va_inner(Rtpr, InnerFace, InnerAttr, We1),
+    We = wings_va:set_edge_attrs(Edge, left, LeftAttr, We2),
+    bevel_va(Es, Move, V, InnerFace, OrigWe, We);
+bevel_va([{ve,Edge,Scale}|Es], Move, V, InnerFace, OrigWe, #we{es=Etab}=We0) ->
+    W = 1.0 - Move*Scale,
+    LeftAttr = wings_va:edge_attrs(Edge, left, W, OrigWe),
+    RightAttr = wings_va:edge_attrs(Edge, right, W, OrigWe),
+    InnerAttr = wings_va:average_attrs(LeftAttr, RightAttr),
+    #edge{ltpr=Ltpr,lf=Lf} = array:get(Edge, Etab),
+    We1 = wings_va:set_edge_attrs(Ltpr, Lf, LeftAttr, We0),
+    We2 = bevel_va_inner(Ltpr, InnerFace, InnerAttr, We1),
+    We = wings_va:set_edge_attrs(Edge, right, RightAttr, We2),
+    bevel_va(Es, Move, V, InnerFace, OrigWe, We);
+bevel_va([], _, _, _, _, We) -> We.
+
+bevel_va_inner(E, Inner, Attr, #we{es=Etab}=We) ->
+    case array:get(E, Etab) of
+	#edge{lf=Inner,ltpr=Ltpr} ->
+	    wings_va:set_edge_attrs(Ltpr, Inner, Attr, We);
+	#edge{rf=Inner,rtpr=Rtpr} ->
+	    wings_va:set_edge_attrs(Rtpr, Inner, Attr, We)
+    end.
+
+%% bevel_scale_factor(VertexA, VertexB) -> Scale
+%%  Calculate a scale factor to multiply the scale factor
+%%  to multiply Move by to reparameterize it to the
+%%  interval 0..1.
+bevel_scale_factor(Va, Vb, #we{vp=Vtab}) ->
+    PosA = array:get(Va, Vtab),
+    PosB = array:get(Vb, Vtab),
+    case e3d_vec:dist(PosB, PosA) of
+	Dist when Dist < 1.0E-10 -> 1.0;
+	Dist -> 1.0 / Dist
+    end.
+
 %%%
 %%% The Connect command.
 %%%
@@ -385,7 +475,7 @@ tighten_vec(V, #we{vp=Vtab,mirror=MirrorFace}=We) ->
 	      (_, _, _, A) -> A
 	   end, [], V, We),
     Center = e3d_vec:average(Cs),
-    e3d_vec:sub(Center, gb_trees:get(V, Vtab)).
+    e3d_vec:sub(Center, array:get(V, Vtab)).
 
 %%%
 %%% The magnetic version of Tighten.
@@ -435,18 +525,5 @@ dissolve(St0) ->
 
 set_color(Color, St) ->
     wings_sel:map(fun(Vs, We) ->
-			  set_color_1(gb_sets:to_list(Vs), Color,
-				      We#we{mode=vertex})
+			  wings_va:set_vertex_color(Vs, Color, We)
 		  end, St).
-
-set_color_1([V|Vs], Color, #we{es=Etab0}=We) ->
-    Etab = wings_vertex:fold(
-	     fun(Edge, _Face, Rec0, Es) ->
-		     Rec = case Rec0 of
-			       #edge{vs=V} -> Rec0#edge{a=Color};
-			       #edge{ve=V} -> Rec0#edge{b=Color}
-			   end,
-		     gb_trees:update(Edge, Rec, Es)
-	     end, Etab0, V, We),
-    set_color_1(Vs, Color, We#we{es=Etab});
-set_color_1([], _, We) -> We.
