@@ -17,7 +17,7 @@
 	 init/0,initial_properties/0,delete_all/1,
 	 current/0,set_current/1,
 	 load_matrices/1,projection/0,
-	 modelview/0,modelview/1,align_view_to_normal/1,
+	 modelview/0,align_view_to_normal/1,
 	 eye_point/0,export_views/1,import_views/2,camera_info/2,
 	 freeze_mirror/1]).
 
@@ -39,19 +39,21 @@ menu(#st{views={CurrentView,Views}}=St) ->
        crossmark(show_info_text)},
         separator,
       {?__(17,"Show Saved BB"),show_bb,?__(18,"Display any saved bounding box"),crossmark(show_bb)},
+      {?__(69,"Show BB Center"),show_bb_center,
+       ?__(70,"Display the Center of any saved bounding box"),crossmark(show_bb_center)},
       {?__(61,"Clipping Planes"),clip_plane,?__(62,"Toggle extra clipping planes (use default axis)"),
        crossmark(clip_plane)},
       {?__(23,"Show Normals"),show_normals,?__(24,"Show normals for selected elements"),
        crossmark(show_normals)},
         separator,
       {?__(42,"Show Colors"),show_colors,
-       ?__(43,"Show vertex colors on objects in \"vertex\" mode"),
+       ?__(43,"Show vertex colors"),
        crossmark(show_colors)},
       {?__(44,"Show Materials"),show_materials,
-       ?__(45,"Show materials on objects in \"material\" mode"),
+       ?__(45,"Show materials"),
        crossmark(show_materials)},
       {?__(46,"Show Textures"),show_textures,
-       ?__(47,"Show the texture on objects in \"material\" mode"),
+       ?__(47,"Show textures"),
        crossmark(show_textures)}]}},
      separator,
      {?__(7,"Wireframe"),wireframe,?__(8,"Display selected objects as a wireframe (same for all objects if nothing is selected)")},
@@ -298,22 +300,27 @@ command(orthogonal_view, St) ->
     St;
 command({show,show_textures}, St) ->
     toggle_option(show_textures),
-    wings_dl:map(fun(#dlo{src_we=#we{mode=material}}=D, _) ->
-			 D#dlo{work=none,smooth=none,proxy_faces=none};
+    wings_dl:map(fun(#dlo{proxy_data=PD}=D, _) ->
+			 %% Must invalidate vertex buffers.
+			 D#dlo{work=none,smooth=none,vab=none,
+			       proxy_data=wings_proxy:invalidate(PD, vab)};
 		    (D, _) -> D
 		 end, []),
     St;
 command({show,show_materials}, St) ->
     toggle_option(show_materials),
-    wings_dl:map(fun(#dlo{src_we=#we{mode=material}}=D, _) ->
-			 D#dlo{work=none,smooth=none,proxy_faces=none};
-		    (D, _) -> D
+    wings_dl:map(fun(#dlo{proxy_data=PD}=D, _) ->
+			 %% We only need to invalidate display lists.
+			 D#dlo{work=none,smooth=none,
+			       proxy_data=wings_proxy:invalidate(PD, dl)}
 		 end, []),
     St;
 command({show,show_colors}, St) ->
     toggle_option(show_colors),
-    wings_dl:map(fun(#dlo{src_we=#we{mode=vertex}}=D, _) ->
-			 D#dlo{work=none,smooth=none,proxy_faces=none};
+    wings_dl:map(fun(#dlo{proxy_data=PD}=D, _) ->
+			 %% Must invalidate vertex buffers.
+			 D#dlo{work=none,smooth=none,vab=none,
+			       proxy_data=wings_proxy:invalidate(PD, vab)};
 		    (D, _) -> D
 		 end, []),
     St;
@@ -335,12 +342,9 @@ command(show_edges, St) ->
 	    wings_dl:map(fun(D, _) -> D#dlo{hard=none} end, []),
 	    St
     end;
-command(highlight_aim, {add,St0,St}) ->
-    highlight_aim(St0,St),
-    St0;
-command(highlight_aim, {delete,St0,St}) ->
-    highlight_aim(St,St0),
-    St0;
+command({highlight_aim,{Type,{Selmode,Sel,MM}}}, St) ->
+    highlight_aim(Type, Selmode, Sel, MM, St),
+    St;
 command(highlight_aim, St) ->
     aim(St),
     St;
@@ -431,24 +435,26 @@ sel_to_set(#st{sel=Sel0}) ->
     Sel = foldl(fun({Id,_}, A) -> [Id|A] end, [], Sel0),
     gb_sets:from_list(Sel).
 
-virtual_mirror_fun(Faces, We) ->
+virtual_mirror_fun(Faces, We0) ->
     case gb_sets:to_list(Faces) of
 	[Face] ->
-	    We#we{mirror=Face};
+	    We = wings_we:create_mirror(Face, We0),
+	    wings_we:mirror_flatten(We, We);
 	_ ->
 	    wings_u:error(?__(1,"Only a single face must be selected per object."))
     end.
 
 break_mirror(#st{shapes=Shs0}=St) ->
-    Shs = foldl(fun(#we{id=Id}=We, Shs) ->
-			gb_trees:update(Id, We#we{mirror=none}, Shs)
+    Shs = foldl(fun(#we{id=Id}=We0, Shs) ->
+			We = wings_we:break_mirror(We0),
+			gb_trees:update(Id, We, Shs)
 		end, Shs0, sel_mirror_objects(St)),
     St#st{shapes=Shs}.
 
 freeze_mirror(#st{shapes=Shs0}=St) ->
-    Shs = foldl(fun(#we{id=Id,mirror=Face}=We0, Shs) ->
-			We = wings_face_cmd:mirror_faces([Face], We0),
-			gb_trees:update(Id, We#we{mirror=none}, Shs)
+    Shs = foldl(fun(#we{id=Id}=We0, Shs) ->
+			We = wings_we:freeze_mirror(We0),
+			gb_trees:update(Id, We, Shs)
 		end, Shs0, sel_mirror_objects(St)),
     St#st{shapes=Shs}.
 
@@ -848,6 +854,7 @@ init() ->
     wings_pref:set_default(number_of_shaders, 1),
     wings_pref:set_default(show_normals, false),
     wings_pref:set_default(show_bb, true),
+    wings_pref:set_default(show_bb_center, true),
     wings_pref:set_default(show_colors, true),
     wings_pref:set_default(show_materials, true),
     wings_pref:set_default(show_textures, true),
@@ -919,19 +926,30 @@ modelview(IncludeLights) ->
 	  elevation=El,pan_x=PanX,pan_y=PanY} = current(),
     gl:matrixMode(?GL_MODELVIEW),
     gl:loadIdentity(),
-    if
-	IncludeLights -> wings_light:camera_lights();
-	true -> ok
+
+    case IncludeLights of
+	true ->
+	    UseSceneLights = wings_pref:get_value(scene_lights) andalso
+		wings_light:any_enabled_lights(),
+	    case UseSceneLights of
+		false -> wings_light:camera_lights();
+		true -> ok
+	    end;
+	false ->
+	    UseSceneLights = false
     end,
+
     gl:translatef(PanX, PanY, -Dist),
     gl:rotatef(El, 1, 0, 0),
     gl:rotatef(Az, 0, 1, 0),
     {OX,OY,OZ} = Origin,
     gl:translatef(OX, OY, OZ),
-    if
-	IncludeLights -> wings_light:global_lights();
-	true -> ok
-    end.
+
+    case UseSceneLights of
+	false -> ok;
+	true -> wings_light:global_lights()
+    end,
+    UseSceneLights.
 
 %% Calculate the location of the viewer in 3D space.
 %% (The (0,0,0) point multiplied by the inverse model transformation matrix.)
@@ -945,53 +963,34 @@ view_matrix(#view{origin=Origin,distance=Dist,azimuth=Az,elevation=El,
     M2 = e3d_mat:mul(M1, e3d_mat:rotate(-El, {1.0,0.0,0.0})),
     e3d_mat:mul(M2, e3d_mat:translate(-PanX, -PanY, Dist)).
 
-%%%% Highlight Aim is called through the View|Aim command (see wings:do_hotkey/2)
-highlight_aim(St0, #st{selmode=body}=St) ->
-    highlight_aim(wings_sel_conv:mode(vertex,St0),wings_sel_conv:mode(vertex,St));
+%%%% Highlight Aim is called via event handler (see wings:do_hotkey/2)
+highlight_aim(add, Selmode, Sel0, MM, #st{selmode=body,sel=Sel1}=St) ->
+    Sel = lists:subtract(Sel0,Sel1),
+    aim(St#st{selmode=Selmode,sel=Sel},MM);
+highlight_aim(delete, Selmode, Sel0, MM, #st{selmode=body,sel=Sel1}=St) ->
+    Sel = lists:subtract(Sel1,Sel0),
+    aim(St#st{selmode=Selmode,sel=Sel},MM);
+highlight_aim(add, Selmode, Sel0, MM, #st{sel=Sel1}=St) ->
+    Sel = highlighted_element(Sel0, Sel1),
+    aim(St#st{selmode=Selmode,sel=Sel},MM);
+highlight_aim(delete, Selmode, Sel0, MM, #st{sel=Sel1}=St) ->
+    Sel = highlighted_element(Sel1, Sel0),
+    aim(St#st{selmode=Selmode,sel=Sel},MM).
 
-highlight_aim(#st{sel=Sel0}, #st{selmode=Mode, sel=Sel, shapes=Shs}) ->
-    Elems0 = highlight_aim_1(Sel0),
-    Elems = highlight_aim_1(Sel),
-    {TargetId,Target} = highlight_target(Elems0, Elems),
-    We = gb_trees:get(TargetId, Shs),
-    TargetVs = target_vs(Mode, Target, We),
-    Origin0 = wings_vertex:center(TargetVs,We),
-    Origin = e3d_vec:neg(Origin0),
-    #view{distance=Dist0} = View = current(),
-    Dist = case e3d_vec:dist(eye_point(), Origin0) of
-           D when D < Dist0 -> D;
-           _Other -> Dist0
-       end,
-    set_current(View#view{origin=Origin,distance=Dist,pan_x=0.0,pan_y=0.0}).
+highlighted_element([{Id,Sel0}], [{Id,Sel1}]) ->
+    [{Id,gb_sets:subtract(Sel0,Sel1)}];
+highlighted_element(Sel0, Sel1) ->
+    [{Id,_}] = lists:subtract(Sel0,Sel1),
+    S0 = get_sel_from_hl_obj(Id, Sel0),
+    S1 = get_sel_from_hl_obj(Id, Sel1),
+    [{Id,gb_sets:subtract(S0,S1)}].
 
-highlight_aim_1(Sel) ->
-    lists:foldl(fun({Id,Elements}, Acc) ->
-        [{Id, gb_sets:to_list(Elements)}|Acc]
-        end, [], Sel).
-
-highlight_target(Elems0, Elems) ->
-    case length(Elems0) == length(Elems) of
-      true -> 
-        highlight_aim_2(Elems0, Elems);
-      false -> 
-        [TargetElems] = Elems -- Elems0,
-        TargetElems
+get_sel_from_hl_obj(Id, Sel) ->
+    case orddict:find(Id, Sel) of
+      {_,SelSet} -> SelSet;
+      error -> gb_sets:empty()
     end.
 
-highlight_aim_2([Items0|Elems0],[Items|Elems]) when Items0 == Items ->
-    highlight_aim_2(Elems0,Elems);
-
-highlight_aim_2([{_,Items0}|_Elems0],[{Id,Items}|_Elems]) ->
-    Target = Items -- Items0,
-    {Id,Target}.
-
-target_vs(vertex, Target, _) ->
-    Target;
-target_vs(edge, Target, We) ->
-    wings_edge:to_vertices(Target,We);
-target_vs(face, Target, We) ->
-    wings_face:to_vertices(Target,We).
-
 %%%% Aim Camera
 aim(#st{sel=[]}) ->
     View = current(),
@@ -1006,9 +1005,39 @@ aim(St) ->
 	   end,
     set_current(View#view{origin=Origin,distance=Dist,pan_x=0.0,pan_y=0.0}).
 
+aim(#st{sel=[]}, _) ->
+    View = current(),
+    set_current(View#view{origin=e3d_vec:zero()});
+aim(St, {_, _, original}) ->
+    aim(St);
+aim(#st{shapes=Shapes}=St, {Id, _Elem, mirror}) ->
+    We = gb_trees:get(Id, Shapes),
+    #we{mirror=Face} = We,
+    MirNormal = wings_face:normal(Face,We),
+    Mirror = wings_face:center(Face,We),
+
+    Original = wings_sel:center(St),
+    Distance = dist_along_vector(Mirror,Original,MirNormal),
+    Origin0 = e3d_vec:add_prod(Original, MirNormal, Distance * 2),
+    Origin = e3d_vec:neg(Origin0),
+
+    #view{distance=Dist0} = View = current(),
+    Dist = case e3d_vec:dist(eye_point(), Origin0) of
+	       D when D < Dist0 -> D;
+	       _Other -> Dist0
+	   end,
+    set_current(View#view{origin=Origin,distance=Dist,pan_x=0.0,pan_y=0.0}).
+
+dist_along_vector(PosA,PosB,Vector) ->
+    %% Return Distance between PosA and PosB along Vector
+    {Xa,Ya,Za} = PosA,
+    {Xb,Yb,Zb} = PosB,
+    {Vx,Vy,Vz} = e3d_vec:norm(Vector),
+    Vx*(Xa-Xb)+Vy*(Ya-Yb)+Vz*(Za-Zb).
+
 frame(#st{sel=[],shapes=Shs}) ->
     BB = foldl(fun(#we{perm=P,vp=Vtab}=We, BB) when ?IS_VISIBLE(P) ->
-		       case gb_sets:is_empty(Vtab) of
+		       case wings_util:array_is_empty(Vtab) of
 			   false -> wings_vertex:bounding_box(We, BB);
 			   true -> BB
 		       end;
@@ -1026,8 +1055,8 @@ frame(St0) ->
     frame_1(wings_sel:bounding_box(St)).
 
 frame_1(none) -> ok;
-frame_1([A,B]=BB) ->
-    C = e3d_vec:average(BB),
+frame_1([A,B]) ->
+    C = e3d_vec:average(A, B),
     R = e3d_vec:len(e3d_vec:sub(A, B)) / 2,
     #view{fov=Fov} = View = current(),
     Dist = R/math:tan(Fov*math:pi()/2/180),
@@ -1038,10 +1067,9 @@ kill_mirror(#st{shapes=Shs0}=St) ->
     Shs = kill_mirror_1(gb_trees:values(Shs0), []),
     St#st{shapes=Shs}.
 
-kill_mirror_1([#we{id=Id,mirror=none}=We|Wes], Acc) ->
+kill_mirror_1([#we{id=Id}=We0|Wes], Acc) ->
+    We = wings_we:break_mirror(We0),
     kill_mirror_1(Wes, [{Id,We}|Acc]);
-kill_mirror_1([#we{id=Id}=We|Wes], Acc) ->
-    kill_mirror_1(Wes, [{Id,We#we{mirror=none}}|Acc]);
 kill_mirror_1([], Acc) ->
     gb_trees:from_orddict(lists:reverse(Acc)).
 
@@ -1250,7 +1278,7 @@ align_to_selection(#st{selmode=edge}=St) ->
     N = average_normals(
 	  fun(Edges, #we{es=Etab}=We, Acc) ->
 		  foldl(fun(Edge, A) ->
-				#edge{lf=Lf,rf=Rf} = gb_trees:get(Edge, Etab),
+				#edge{lf=Lf,rf=Rf} = array:get(Edge, Etab),
 				[wings_face:normal(Lf, We),
 				 wings_face:normal(Rf, We)|A]
 			end, Acc, Edges)
