@@ -467,7 +467,7 @@ global2local(X, Y) ->
     {X-Xorig,Y-(TopH-Yorig-H)}.
 
 local_mouse_state() ->
-    {B,X0,Y0} = sdl_mouse:getMouseState(),
+    {B,X0,Y0} = wings_io:get_mouse_state(),
     {X,Y} = global2local(X0, Y0),
     {B,X,Y}.
 
@@ -541,7 +541,7 @@ dispatch_event(#resize{w=W,h=H}) ->
     %% If the window has become maximized, we don't want
     %% to save the window size, but will keep the previous
     %% size.
-    case sdl_video:wm_isMaximized() of
+    case wings_io:is_maximized() of
 	false -> wings_pref:set_value(window_size, {W,H});
 	true -> ok
     end,
@@ -621,14 +621,14 @@ redraw_all() ->
     %% Remove late buffers clear due to problems with ATI cards when AA.
     gl:clear(?GL_COLOR_BUFFER_BIT bor ?GL_DEPTH_BUFFER_BIT),
     Windows = keysort(2, gb_trees:to_list(get(wm_windows))),
-    foreach(fun({Name,_}) ->
-		    dispatch_matching(fun({wm,{send_to,N,_}}) ->
-					      N =:= Name;
-					 (_) -> false
-				      end),
-		    do_dispatch(Name, redraw)
-	    end, Windows),
-    gl:swapBuffers(),
+    wings_io:foreach(fun({Name,_}) ->
+			     dispatch_matching(fun({wm,{send_to,N,_}}) ->
+						       N =:= Name;
+						  (_) -> false
+					       end),
+			     do_dispatch(Name, redraw)
+		     end, Windows),
+    wings_io:swapBuffers(),
     clean(),
     wings_io:set_cursor(get(wm_cursor)),
     event_loop().
@@ -678,7 +678,7 @@ reinit_opengl() ->
 
 init_opengl() ->
     {W,H} = get(wm_top_size),
-    set_video_mode(W, H),
+    wings_io:reset_video_mode_for_gl(W, H),
     gl:clear(?GL_COLOR_BUFFER_BIT bor ?GL_DEPTH_BUFFER_BIT),
     gl:pixelStorei(?GL_UNPACK_ALIGNMENT, 1),
     wings_io:resize(),
@@ -872,7 +872,7 @@ find_active_0(Ev) ->
     case Ev of
 	#mousebutton{x=X,y=Y} -> ok;
 	#mousemotion{x=X,y=Y} -> ok;
-	_ -> {_,X,Y} = sdl_mouse:getMouseState()
+	_ -> {_,X,Y} = wings_io:get_mouse_state()
     end,
     window_below(X, Y).
 
@@ -1018,10 +1018,6 @@ get_window_data(Name) ->
 put_window_data(Name, Data) ->
     put(wm_windows, gb_trees:update(Name, Data, get(wm_windows))).
 
-set_video_mode(W, H) ->
-    {surfacep,_} = sdl_video:setVideoMode(W, H, 0, ?SDL_OPENGL bor ?SDL_RESIZABLE),
-    ok.
-
 %%%
 %%% Button translation.
 %%%
@@ -1109,10 +1105,10 @@ message_redraw(Msg, Right) ->
 	    if 
 		MsgW+RightW < W - RMarg ->
 		    Pos = W - RightW - RMarg,
-		    wings_io:set_color(wings_pref:get_value(menu_color)),
+		    wings_io:set_color(e3d_vec:mul(wings_pref:get_value(info_line_bg),0.9)),
 		    gl:recti(Pos-Cw, 1-wings_text:height(),
 			     Pos+RightW+Cw, 3),
-		    gl:color3b(0, 0, 0),
+		    wings_io:set_color(wings_pref:get_value(info_line_text)),
 		    wings_io:text_at(Pos, Right);
 		true -> ok
 	    end
@@ -1130,7 +1126,7 @@ message_setup() ->
     wings_io:ortho_setup(none),
     {W,H} = win_size(),
     wings_io:gradient_rect(0, 0, W, H, wings_pref:get_value(info_line_bg)),
-    gl:color3b(0, 0, 0),
+    wings_io:set_color(wings_pref:get_value(info_line_text)),
     gl:translatef(10, H-5.375, 0),
     {W,H}.
 
